(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matrix imports PolyOrder
begin




(* matrices *)

types 'a vec = "'a list"
types 'a mat = "'a vec list" (* list of column-vectors *)

definition wf_vecI :: "nat \<Rightarrow> 'x vec \<Rightarrow> bool"
 where "wf_vecI n x = (length x = n)"

definition wf_mat_rcI :: "nat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" where
 "wf_mat_rcI nr nc m = (length m = nc \<and> forall (wf_vecI nr) m)"

lemma wf_mat_def: "wf_mat_rcI nr nc m = (length m = nc \<and> (\<forall> v \<in> set m. wf_vecI nr v))"
  unfolding wf_mat_rcI_def by auto

definition vec_zeroI :: "'a \<Rightarrow> nat \<Rightarrow> 'a vec" where 
 "vec_zeroI ze n = replicate n ze"

definition mat0I :: "'a \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a mat" where
  "mat0I ze nr nc = replicate nc (vec_zeroI ze nr)"

definition vec_plusI :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a vec" where 
 "vec_plusI pl v w = map (\<lambda> xy. pl (fst xy) (snd xy)) (zip v w)"


definition mat_plusI :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
 where "mat_plusI pl m1 m2 = map (\<lambda> uv. vec_plusI pl (fst uv) (snd uv)) (zip m1 m2)"

definition scalar_prodI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a" where
 "scalar_prodI ze pl ti v w = foldr (\<lambda> xy s. pl (ti (fst xy) (snd xy)) s) (zip v w) ze"

definition row :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a vec"
where "row m i \<equiv> map (\<lambda> w. w ! i) m"

definition col :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a vec"
where "col m i \<equiv> m ! i"

lemma wf_row : assumes "wf_mat_rcI nr nc m"
  and "i < nr"
  shows "wf_vecI nc (row m i)"
  using assms
  unfolding wf_vecI_def row_def wf_mat_rcI_def
  by (auto simp: wf_vecI_def) 

lemma wf_col : assumes "wf_mat_rcI nr nc m"
  and "i < nc"
  shows "wf_vecI nr (col m i)"
  using assms
  unfolding wf_vecI_def col_def wf_mat_rcI_def
  by (auto simp: wf_vecI_def) 

fun transpose :: "nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
 where "transpose nr [] = replicate nr []"
     | "transpose nr (v # m) = map (\<lambda> (vi,mi). (vi # mi)) (zip v (transpose nr m))"      

lemma wf_transpose : assumes "wf_mat_rcI nr nc m"
  shows "wf_mat_rcI nc nr (transpose nr m)"
using assms 
proof (induct m arbitrary: nc, simp add: wf_mat_rcI_def wf_vecI_def set_replicate_conv_if)
  case (Cons v m)
  from `wf_mat_rcI nr nc (v # m)` obtain ncc where nc: "nc = Suc ncc" by (cases nc, auto simp: wf_mat_rcI_def) 
  with Cons have wfRec: "wf_mat_rcI ncc nr (transpose nr m)" unfolding wf_mat_rcI_def by auto
  from `wf_mat_rcI nr nc (v # m)` have wfV: "length v = nr" unfolding wf_mat_rcI_def by (simp add:  wf_vecI_def)
  show ?case unfolding wf_mat_rcI_def
  proof (rule conjI, simp add: wfV wfRec wf_mat_rcI_def wf_vecI_def)
    show "min nr (length (transpose nr m)) = nr" using wfRec unfolding wf_mat_rcI_def by auto
  next
    show "forall (wf_vecI nc) (transpose nr (v # m))"
    proof (simp add: wf_vecI_def nc, rule ballI, simp only: split_paired_all)
      fix a b
      assume mem: "(a,b) \<in> set (zip v (transpose nr m))"
      from mem have "b \<in> set (transpose nr m)" by (rule set_zip_rightD)
      with wfRec have "length b = ncc" unfolding wf_mat_rcI_def using wf_vecI_def[of ncc] by auto
      thus "length (split op # (a,b)) = Suc ncc" by auto
    qed
  qed
qed

lemma col_transpose_is_row: 
  assumes wf: "wf_mat_rcI nr nc m"
  and i: "i < nr"
  shows "col (transpose nr m) i = row m i"
using wf 
proof (induct m arbitrary: nc , simp add: col_def row_def wf_mat_rcI_def i)
  case (Cons v m)
  from `wf_mat_rcI nr nc (v # m)` obtain ncc where nc: "nc = Suc ncc" and wf: "wf_mat_rcI nr ncc m"  by (cases nc, auto simp: wf_mat_rcI_def)
  from `wf_mat_rcI nr nc (v # m)` nc have lengths: "(\<forall> w \<in> set m. length w = nr) \<and> length v = nr \<and> length m = ncc" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
  from wf Cons have colRec: "col (transpose nr m) i = row m i" by auto
  from wf have wf_trans: "wf_mat_rcI ncc nr (transpose nr m)" by (rule wf_transpose)
  hence lengths2: "(\<forall> w \<in> set (transpose nr m). length w = ncc) \<and> length (transpose nr m) = nr" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
  show ?case proof (rule nth_equalityI, simp add: col_def row_def lengths lengths2 i, intro allI impI)
    fix j
    have simpme: "transpose nr m ! i = row m i" using colRec unfolding col_def by auto
    assume "j < length (col (transpose nr (v # m)) i)"
    hence "j < Suc ncc" by (simp add: col_def lengths2 lengths i) 
    thus "col (transpose nr (v # m)) i ! j = row (v # m) i ! j"
      by (cases j, simp add: row_def col_def i lengths lengths2, simp add: row_def col_def i lengths lengths2 simpme)
  qed
qed

lemma col_eq:
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  shows "(\<forall> i < nc. col m1 i = col m2 i) = (m1 = m2)" (is "?l = ?r")
proof
  assume ?r thus ?l by auto
next
  assume ?l show ?r 
  proof (rule nth_equalityI)
    show "length m1 = length m2" using wf1 wf2 unfolding wf_mat_rcI_def by auto
  next
    from `?l` show "\<forall> i < length m1. m1 ! i = m2 ! i" using wf1 unfolding col_def wf_mat_rcI_def by auto
  qed
qed

lemma index_eq:
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  shows "(\<forall> i < nc. \<forall> j < nr. m1 ! i ! j = m2 ! i ! j) = (m1 = m2)" (is "?l = ?r")
proof 
  assume ?r thus ?l by auto
next
  assume ?l show ?r 
  proof (simp only: col_eq[OF wf1 wf2,symmetric], unfold col_def, intro allI impI)
    fix i
    assume i: "i < nc"
    show "m1 ! i = m2 ! i"
    proof (rule nth_equalityI)      
      show "length (m1 ! i)  = length (m2 ! i)" using wf1 wf2 i unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
    next
      from `?l` i show "\<forall> j < length (m1 ! i). m1 ! i ! j = m2 ! i ! j" using wf1 wf2 unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
    qed
  qed
qed

lemma row_col : assumes "wf_mat_rcI nr nc m"  
  and "i < nr" and "j < nc"
  shows "row m i ! j = col m j ! i"
using assms unfolding wf_mat_rcI_def row_def col_def
  by auto

lemma row_eq: 
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  shows "(\<forall> i < nr. row m1 i = row m2 i) = (m1 = m2)" (is "?l = ?r")
proof 
  assume ?r thus ?l by auto
next
  assume ?l show ?r
  proof (rule nth_equalityI)
    show "length m1 = length m2" using wf1 wf2 unfolding wf_mat_rcI_def by auto
  next
    show "\<forall> i < length m1. m1 ! i = m2 ! i"
    proof (intro allI impI)
      fix i
      assume i: "i < length m1"
      show "m1 ! i = m2 ! i"
      proof (rule nth_equalityI)
	show "length (m1 ! i) = length (m2 ! i)" using wf1 wf2 i unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
      next
	show "\<forall> j < length (m1 ! i). m1 ! i ! j = m2 ! i ! j" 
	proof (intro allI impI)
	  fix j
	  assume j: "j < length (m1 ! i)"
	  from i j wf1 have i1: "i < nc" and j1: "j < nr" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
	  from `?l` j1 have "col m1 i ! j = col m2 i ! j"
	    by (simp add: row_col[OF wf1 j1 i1, symmetric] row_col[OF wf2 j1 i1, symmetric])
	  thus "m1 ! i ! j = m2 ! i ! j" unfolding col_def .
	qed
      qed
    qed
  qed
qed

lemma row_transpose_is_col:   assumes wf: "wf_mat_rcI nr nc m"
  and i: "i < nc"
  shows "row (transpose nr m) i = col m i"
proof -
  have len: "length (row (transpose nr m) i) = length (col m i)"
    using wf_transpose[OF wf]  wf i  unfolding row_def col_def wf_mat_rcI_def by (auto simp: wf_vecI_def)
  show ?thesis 
  proof (rule nth_equalityI[OF len], intro allI impI)
    fix j
    assume "j < length (row (transpose nr m) i)"
    hence j: "j < nr" using wf_transpose[OF wf] wf i unfolding row_def col_def wf_mat_rcI_def by (auto simp: wf_vecI_def)
    show "row (transpose nr m) i ! j = col m i ! j"
      by (simp only: row_col[OF wf_transpose[OF wf] i j],
	simp only: col_transpose_is_row[OF wf j],
	simp only: row_col[OF wf j i])
  qed
qed

definition matT_vec_mulI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
 where "matT_vec_mulI ze pl ti m v = map (\<lambda> w. scalar_prodI ze pl ti w v) m"

definition mat_mulI :: "'a \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat" 
where "mat_mulI ze pl ti nr m1 m2 \<equiv> map (matT_vec_mulI ze pl ti (transpose nr m1)) m2"

lemma wf_matT_vec_mulI : assumes "wf_mat_rcI nr nc m"
  shows "wf_vecI nc (matT_vec_mulI ze pl ti m v)"
  unfolding matT_vec_mulI_def
  using assms
  unfolding wf_mat_rcI_def
  by (simp add: wf_vecI_def)

lemma matT_vec_mul_to_scalar : 
  assumes "wf_mat_rcI nr nc m"
  and "wf_vecI nr v"
  and "i < nc"
  shows "matT_vec_mulI ze pl ti m v ! i = scalar_prodI ze pl ti (col m i) v"
unfolding matT_vec_mulI_def using assms unfolding wf_mat_rcI_def col_def by (auto simp: wf_vecI_def)

lemma mat_vec_mul_to_scalar : 
  assumes wf: "wf_mat_rcI nr nc m"
  and wfV: "wf_vecI nc v"
  and i: "i < nr"
  shows "matT_vec_mulI ze pl ti (transpose nr m) v ! i = scalar_prodI ze pl ti (row m i) v"
by (simp only:matT_vec_mul_to_scalar[OF wf_transpose[OF wf] wfV i],
  simp only: col_transpose_is_row[OF wf i])

lemma wf_mat_mul : assumes wf1: "wf_mat_rcI nr n m1"
  and wf2: "wf_mat_rcI n nc m2"
  shows "wf_mat_rcI nr nc (mat_mulI ze pl ti nr m1 m2)"
using assms
unfolding wf_mat_rcI_def mat_mulI_def by (auto simp: wf_matT_vec_mulI[OF wf_transpose[OF wf1]])

lemma mat_mul_to_scalar :
  assumes wf1: "wf_mat_rcI nr n m1"
  and wf2: "wf_mat_rcI n nc m2"
  and i: "i < nr"
  and j: "j < nc"
  shows "mat_mulI ze pl ti nr m1 m2 ! j ! i = scalar_prodI ze pl ti (row m1 i) (col m2 j)"
proof -
  have jlen: "j < length m2" using wf2 j unfolding wf_mat_rcI_def by auto
  have wfj: "wf_vecI n (m2 ! j)" using jlen j wf2 unfolding wf_mat_rcI_def by auto
  show ?thesis 
    unfolding mat_mulI_def
    by (simp add: jlen, simp only: mat_vec_mul_to_scalar[OF wf1 wfj i], unfold col_def, simp)
qed

lemma col_mat_mul_to_scalar :
  assumes wf1: "wf_mat_rcI nr n m1"
  and wf2: "wf_mat_rcI n nc m2"
  and j: "j < nc"
  shows "col (mat_mulI ze pl ti nr m1 m2) j = map (\<lambda> i. scalar_prodI ze pl ti (row m1 i) (col m2 j)) [0 ..< nr]" (is "col ?l j = ?r")
proof - 
  have wf12: "wf_mat_rcI nr nc ?l" by (rule wf_mat_mul[OF wf1 wf2])
  have len: "length (col ?l j) = length ?r" and nr: "length (col ?l j) = nr" using wf1 wf2 wf12 j unfolding wf_mat_rcI_def col_def by (auto simp: wf_vecI_def) 
  show ?thesis by (rule nth_equalityI[OF len], simp add: j nr, intro allI impI, unfold col_def, simp only:
    mat_mul_to_scalar[OF wf1 wf2 _ j], simp add: col_def)
qed

lemma row_mat_mul_to_scalar :
  assumes wf1: "wf_mat_rcI nr n m1"
  and wf2: "wf_mat_rcI n nc m2"
  and i: "i < nr"
  shows "row (mat_mulI ze pl ti nr m1 m2) i = map (\<lambda> j. scalar_prodI ze pl ti (row m1 i) (col m2 j)) [0 ..< nc]" (is "row ?l i = ?r")
proof - 
  have wf12: "wf_mat_rcI nr nc ?l" by (rule wf_mat_mul[OF wf1 wf2])
  hence lenL: "length ?l = nc" unfolding wf_mat_rcI_def by simp
  have len: "length (row ?l i) = length ?r" and nc: "length (row ?l i) = nc" using wf1 wf2 wf12 i unfolding wf_mat_rcI_def row_def by (auto simp: wf_vecI_def) 
  show ?thesis by (rule nth_equalityI[OF len], simp add: i nc, intro allI impI, unfold row_def, simp add: lenL, simp only: 
    mat_mul_to_scalar[OF wf1 wf2 i], simp add: row_def)
qed

lemma scalar_product_cons : 
  "scalar_prodI ze pl ti (a # as) (b # bs) = pl (ti a b) (scalar_prodI ze pl ti as bs)"
unfolding scalar_prodI_def by auto

definition vec_geI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
  where "vec_geI ge v w \<equiv> forall (\<lambda> (a,b). ge a b) (zip v w)"

definition mat_geI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_geI ge m1 m2 \<equiv> forall (\<lambda> (v,w). vec_geI ge v w) (zip m1 m2)"

definition vec_pre_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
  where "vec_pre_gtI gt v w \<equiv> any (\<lambda> (a,b). gt a b) (zip v w)"

definition mat_pre_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_pre_gtI gt m1 m2 \<equiv> any (\<lambda> (v,w). vec_pre_gtI gt v w) (zip m1 m2)"

definition mat_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_gtI ge gt m1 m2 \<equiv> mat_geI ge m1 m2 \<and> mat_pre_gtI gt m1 m2"

definition vec_max0I :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
  where "vec_max0I = map"

definition vec1I :: "'a \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a vec"
  where "vec1I ze on n i \<equiv> replicate i ze @ on # replicate (n - 1 - i) ze"

lemma vec1_index : assumes i: "i < n"
  and j: "j < n"
  shows "vec1I ze on n i ! j = (if i = j then on else ze)" (is "_ = ?r")
unfolding vec1I_def
proof -
  let ?l = "replicate i ze @ on # replicate (n - 1 - i) ze"
  have len: "length ?l > i" by auto
  have len2: "length (replicate i ze @ on # []) > i" by auto
  show "?l ! j = ?r"
  proof (cases "j = i")
    case True
    thus ?thesis by (simp add: nth_append)
  next
    case False
    show ?thesis 
    proof (cases "j < i")
      case True
      thus ?thesis by (simp add: nth_append)
    next
      case False
      with `j \<noteq> i` have gt: "j > i" by auto
      from this have "\<exists> k. j = i + Suc k" by arith
      from this obtain k where k: "j = i + Suc k" by auto
      with j i show ?thesis by (simp add: nth_append)
    qed
  qed
qed

definition mat1I :: "'a \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a mat"
  where "mat1I ze on n \<equiv> map (vec1I ze on n) [0 ..< n]"

definition mat_max0I :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
  where "mat_max0I mz = map (vec_max0I mz)"

lemma wf_vec0[simp] : "wf_vecI nr (vec_zeroI ze nr)"
  by (simp add: wf_vecI_def vec_zeroI_def)

lemma wf_mat0[simp] : "wf_mat_rcI nr nc (mat0I ze nr nc)"
  by (simp add: wf_mat_rcI_def mat0I_def, rule replicate_prop, simp)

	     
lemma wf_vec_max0: assumes "wf_vecI nr v" shows "wf_vecI nr (vec_max0I mz v)"
using assms unfolding wf_vecI_def vec_max0I_def by auto

lemma wf_mat_max0: assumes "wf_mat_rcI nr nc m" shows "wf_mat_rcI nr nc (mat_max0I mz m)"
using assms unfolding wf_mat_rcI_def mat_max0I_def
  by (auto simp: wf_vec_max0)


lemma wf_vec_plus : "\<lbrakk>wf_vecI nr u; wf_vecI nr v\<rbrakk> \<Longrightarrow> wf_vecI nr (vec_plusI pl u v)"
using assms 
unfolding vec_plusI_def wf_vecI_def
by auto

lemma wf_mat_plus : assumes "wf_mat_rcI nr nc m1" and "wf_mat_rcI nr nc m2" shows "wf_mat_rcI nr nc (mat_plusI pl m1 m2)"
using assms
unfolding wf_mat_def mat_plusI_def
proof (simp, induct nc arbitrary: m1 m2, simp)
  case (Suc nn)
  show ?case 
  proof (cases m1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons v1 mm1)
    with Suc have l1: "length mm1 = nn" by auto
    show ?thesis
    proof (cases m2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons v2 mm2)
      with Suc have l2: "length mm2 = nn" by auto
      show ?thesis by  (simp add: prems, rule conjI, rule wf_vec_plus, (simp add: prems)+, rule Suc, auto simp: prems l1 l2)
    qed
  qed
qed


lemma wf_vec1: assumes "i < nr" shows "wf_vecI nr (vec1I ze on nr i)"
unfolding wf_vecI_def vec1I_def using assms by auto

lemma wf_mat1: "wf_mat_rcI nr nr (mat1I ze on nr)"
unfolding wf_mat_rcI_def mat1I_def using wf_vec1 by auto

lemma mat0_index: assumes "i < nc" and "j < nr"
  shows "mat0I ze nr nc ! i ! j = ze"
unfolding mat0I_def vec_zeroI_def using assms by auto

lemma mat0_row: assumes "i < nr"
  shows "row (mat0I ze nr nc) i = vec_zeroI ze nc"
unfolding row_def mat0I_def vec_zeroI_def
using assms by auto

lemma mat0_col: assumes "i < nc"
  shows "col (mat0I ze nr nc) i = vec_zeroI ze nr"
unfolding mat0I_def col_def
using assms by auto


lemma vec_plus_index : 
  assumes wf1: "wf_vecI nr v1"
  and wf2: "wf_vecI nr v2"
  and i: "i < nr"
  shows "vec_plusI pl v1 v2 ! i = pl (v1 ! i)  (v2 ! i)"
using wf1 wf2 i
unfolding wf_vecI_def vec_plusI_def
proof (induct v1 arbitrary: i v2 nr, simp)
  case (Cons a v11)
  from Cons obtain b v22 where v2: "v2 = b # v22" by (cases v2, auto)
  from v2 Cons obtain nrr where nr: "nr = Suc nrr" by (force)
  from Cons show ?case
    by (cases i, simp add: v2, auto simp: v2 nr)
qed

lemma mat_plus_index : 
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  and i: "i < nc"
  and j: "j < nr"
  shows "mat_plusI pl m1 m2 ! i ! j = pl (m1 ! i ! j) (m2 ! i ! j)"
using wf1 wf2 i
unfolding mat_plusI_def wf_mat_rcI_def 
proof (simp, induct m1 arbitrary: m2 i nc, simp)
  case (Cons v1 m11)
  from Cons obtain v2 m22 where m2: "m2 = v2 # m22" by (cases m2, auto)
  from m2 Cons obtain ncc where nc: "nc = Suc ncc" by force
  show ?case
  proof (cases i, simp add: m2, rule vec_plus_index[where nr = nr], (auto simp: Cons j m2)[3])
    case (Suc ii)
    with Cons show ?thesis using m2 nc by auto
  qed
qed

lemma col_mat_plus : assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  and i: "i < nc"
  shows "col (mat_plusI pl m1 m2) i = vec_plusI pl (col m1 i) (col m2 i)"
using assms
unfolding mat_plusI_def col_def wf_mat_rcI_def
proof (induct m1 arbitrary: m2 nc i, simp)
  case (Cons v m1)
  from Cons obtain v2 m22 where m2: "m2 = v2 # m22" by (cases m2, auto)
  from m2 Cons obtain ncc where nc: "nc = Suc ncc" by force
  show ?case
  proof (cases i, simp add: m2)
    case (Suc ii)
    with Cons show ?thesis using m2 nc by auto
  qed
qed

lemma transpose_index: assumes wf: "wf_mat_rcI nr nc m"
  and i: "i < nr"
  and j: "j < nc"
  shows "transpose nr m ! i ! j = m ! j ! i"
proof -
  have "transpose nr m ! i ! j = col (transpose nr m) i ! j" unfolding col_def by simp
  also have "\<dots> = row m i ! j" using col_transpose_is_row[OF wf i] by simp
  also have "\<dots> = m ! j ! i" unfolding row_def using wf j unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
  finally show ?thesis . 
qed

lemma transpose_mat_plus: assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  shows "transpose nr (mat_plusI pl m1 m2) = mat_plusI pl (transpose nr m1) (transpose nr m2)"
proof - 
  let ?m12 = "mat_plusI pl m1 m2"
  let ?t1 = "transpose nr m1"
  let ?t2 = "transpose nr m2"
  from wf_mat_plus[OF wf1 wf2] have wf12: "wf_mat_rcI nr nc ?m12" .
  from wf_transpose[OF wf1] have wft1: "wf_mat_rcI nc nr ?t1" .
  from wf_transpose[OF wf2] have wft2: "wf_mat_rcI nc nr ?t2" .
  show ?thesis 
  proof (simp only: index_eq[OF wf_transpose[OF wf12] wf_mat_plus[OF wft1 wft2], symmetric], intro allI impI)
    fix i j
    assume i: "i < nr" and j: "j < nc"
    show "transpose nr ?m12 ! i ! j = mat_plusI pl ?t1 ?t2 ! i ! j"      
      by (simp only: transpose_index[OF wf12 i j],
	simp only: mat_plus_index[OF wft1 wft2 i j],
	simp only: mat_plus_index[OF wf1 wf2 j i],
	simp only: transpose_index[OF wf1 i j],
	simp only: transpose_index[OF wf2 i j])
  qed
qed
      

lemma row_mat_plus : assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  and i: "i < nr"
  shows "row (mat_plusI pl m1 m2) i = vec_plusI pl (row m1 i) (row m2 i)"
  by (
    simp only: col_transpose_is_row[OF wf_mat_plus[OF wf1 wf2] i, symmetric], 
    simp only: transpose_mat_plus[OF wf1 wf2],
    simp only: col_mat_plus[OF wf_transpose[OF wf1] wf_transpose[OF wf2] i],
    simp only: col_transpose_is_row[OF wf1 i],
    simp only: col_transpose_is_row[OF wf2 i])


lemma col_mat1: assumes "i < nr"
  shows "col (mat1I ze on nr) i = vec1I ze on nr i"
unfolding mat1I_def col_def using assms by auto
 
lemma transpose_mat1: "transpose nr (mat1I ze on nr) = (mat1I ze on nr)"
proof (simp only: index_eq[OF wf_transpose[OF wf_mat1] wf_mat1, symmetric], intro impI allI)
  fix i j
  assume i: "i < nr" and j: "j < nr"
  let ?I = "mat1I ze on nr"
  show "transpose nr ?I ! i ! j = ?I ! i ! j"
    by  (simp only: col_def[symmetric], 
      simp only: col_mat1[OF i],
      simp only: row_col[OF wf_transpose[OF wf_mat1] j i,symmetric],
      simp only: row_transpose_is_col[OF wf_mat1 j],
      simp only: col_mat1[OF j],
      simp only: vec1_index[OF i j],
      simp only: vec1_index[OF j i], simp)
qed

lemma row_mat1: assumes i: "i < nr"
  shows "row (mat1I ze on nr) i = vec1I ze on nr i"
by (simp only: col_transpose_is_row[OF wf_mat1 i, symmetric],
  simp only: transpose_mat1,
  simp only: col_mat1[OF i])

lemma vec_ge_to_index : assumes "wf_vecI nr v1" 
  and "wf_vecI nr v2"
  shows "vec_geI ge v1 v2 = (\<forall> i < nr. ge (v1 ! i) (v2 ! i))"
using assms
unfolding wf_vecI_def vec_geI_def 
proof (simp, induct nr arbitrary: v1 v2, simp)
  case (Suc nrr)
  from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and lw1: "length w1 = nrr" by (cases v1, auto)
  from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and lw2: "length w2 = nrr" by (cases v2, auto)
  have rec: "(\<forall> a \<in> set (zip w1 w2). split ge a) = (\<forall> i < nrr. ge (w1 ! i) (w2 ! i))"
    by (rule Suc, auto simp: Suc lw1 lw2)
  show ?case (is "?l = ?r")
  proof (rule iffI)
    assume ?r
    thus ?l using Suc v1 v2 by auto
  next
    assume ?l
    show ?r
    proof (intro allI impI)
      fix i
      assume i: "i < Suc nrr"
      show "ge (v1 ! i) (v2 ! i)"
	using `?l` v1 v2 i
	by (cases i, auto simp: rec)
    qed
  qed
qed

lemma mat_ge_to_index : assumes "wf_mat_rcI nr nc m1" 
  and "wf_mat_rcI nr nc m2"
  shows "mat_geI ge m1 m2 = (\<forall> i < nc. \<forall> j < nr. ge (m1 ! i ! j) (m2 ! i ! j))"
using assms
unfolding wf_mat_rcI_def mat_geI_def
proof (simp, induct nc arbitrary: m1 m2, simp)
  case (Suc ncc)
  from Suc obtain v1 mm1 where m1: "m1 = v1 # mm1" and lm1: "length mm1 = ncc \<and> (\<forall> a \<in> set mm1. wf_vecI nr a)" by (cases m1, auto)
  from Suc obtain v2 mm2 where m2: "m2 = v2 # mm2" and lm2: "length mm2 = ncc \<and> (\<forall> a \<in> set mm2. wf_vecI nr a)" by (cases m2, auto)
  from Suc m1 have wf1: "wf_vecI nr v1" by simp
  from Suc m2 have wf2: "wf_vecI nr v2" by simp
  have rec: "(\<forall> a \<in> set (zip mm1 mm2). split (vec_geI ge) a) = (\<forall> i < ncc. \<forall> j < nr. ge (mm1 ! i ! j) (mm2 ! i ! j))"
    by (rule Suc, auto simp: Suc lm1 lm2)
  show ?case (is "?l = ?r")
  proof (rule iffI)
    assume ?r
    thus ?l using m1 m2 lm1 lm2 rec vec_ge_to_index[OF wf1 wf2] by auto
  next
    assume ?l
    hence ge: "vec_geI ge v1 v2" and "\<forall> a \<in> set (zip mm1 mm2). split (vec_geI ge) a" using m1 m2 by auto
    with rec have ge2: " (\<forall>i<ncc. \<forall>j<nr. ge (mm1 ! i ! j) (mm2 ! i ! j))" by simp
    show ?r
    proof (rule allI, rule impI)
      fix i 
      assume i: "i < Suc ncc" 
      show "\<forall> j < nr. ge (m1 ! i ! j) (m2 ! i ! j)"
      proof (cases i, simp add: m1 m2, simp only: vec_ge_to_index[OF wf1 wf2, symmetric], rule ge)
	case (Suc ii)	
	with i have " ii < ncc" by simp
	with Suc 
	show ?thesis by (simp add: m1 m2, simp add: ge2)
      qed
    qed
  qed
qed

definition mat_arc_posI :: "('a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_arc_posI ap m \<equiv> ap (m ! 0 ! 0)"

definition vec_arc_posI :: "('a \<Rightarrow> bool) \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_arc_posI ap v \<equiv> ap (v ! 0)"


locale matrix_order = 
  fixes n :: nat (* dimension *)
  and zero :: "'a :: sshow" ("0")
  and one :: 'a ("1")
  and plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "+" 65)
  and times :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "*" 70)
  and geq :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<ge>" 21)
  and gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix ">" 21)
  and max0 :: "'a \<Rightarrow> 'a"
  and plus_single_mono :: "bool"
  and arc_pos :: "'a \<Rightarrow> bool"
  assumes n_pos: "n > 0" (* otherwise there is no 1-matrix where 1-mat > 0-mat *)
  and plus_comm: "(x :: 'a) +y = y+x"
  and plus_ass[simp]: "(x :: 'a) +(y+z) = (x+y)+z"
  and zero_left[simp]: "0+(x :: 'a) = x"
  and times_ass[simp]: "(x :: 'a) *(y*z) = (x*y)*z"
  and left_mult_zero[simp]: "0* (x :: 'a) = 0"
  and right_mult_zero[simp]: "(x :: 'a) *0 = 0"
  and distrib_left[simp]: "(x :: 'a) *(y+z) = (x*y)+(x*z)"
  and distrib_right[simp]: "((x :: 'a) +y)*z = (x*z)+(y*z)"
  and compat[trans]: "\<lbrakk>s \<ge> (t :: 'a); t > u\<rbrakk> \<Longrightarrow> s > u"
  and compat2[trans]: "\<lbrakk>s > (t :: 'a); t \<ge> u\<rbrakk> \<Longrightarrow> s > u"
  and plus_left_mono: "\<lbrakk>(x :: 'a) \<ge> y\<rbrakk> \<Longrightarrow> x + z \<ge> y + z"
  and times_left_mono: "\<lbrakk>z \<ge> 0; (x :: 'a) \<ge> y\<rbrakk> \<Longrightarrow> x * z \<ge> y * z"
  and times_right_mono: "\<lbrakk>x \<ge> 0; (y :: 'a) \<ge> z\<rbrakk> \<Longrightarrow> x * y \<ge> x * z"
  and geq_refl[simp]: "x \<ge> (x :: 'a)"
  and geq_trans[trans]: "\<lbrakk>x \<ge> y; (y :: 'a) \<ge> z\<rbrakk> \<Longrightarrow> x \<ge> z"
  and SN: "SN {(x,y) . geq y zero \<and> gt x y \<and> arc_pos y}"
  and max0_id_pos: "\<lbrakk>geq x  zero\<rbrakk> \<Longrightarrow> max0 x = x"
  and max0_pos: "geq (max0 x) zero"
  and max0_mono: "\<lbrakk>x \<ge> y\<rbrakk> \<Longrightarrow> geq (max0 x) (max0 y)"
  and max0_x: "geq (max0 x) x"
  and left_one[simp]: "1 * (x :: 'a) = x"
  and right_one[simp]: "(x :: 'a) * 1 = x"
  and one_geq_zero[simp]: "one \<ge> zero"
  and one_gt_zero[simp]: "one > zero"
  and plus_gt_left_mono: "\<lbrakk>(x :: 'a) > y; z \<ge> u; plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + u"
  and plus_gt_both_mono: "\<lbrakk>(x :: 'a) > y; z > u; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + u" 
  and times_gt_left_mono: "\<lbrakk>(x :: 'a) > y; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x * z > y * z" 
  and arc_pos_plus: "\<lbrakk>arc_pos x\<rbrakk> \<Longrightarrow> arc_pos (x + y)"
  and arc_pos_mult: "\<lbrakk>arc_pos x; arc_pos y\<rbrakk> \<Longrightarrow> arc_pos (x * y)"
  and arc_pos_max0: "arc_pos (max0 x) = arc_pos x"
  and arc_pos_one: "arc_pos one"
  and arc_pos_plus_single: "plus_single_mono \<Longrightarrow> arc_pos = (\<lambda> x. True)"
  and zero_leastI: "\<not> plus_single_mono \<Longrightarrow> x > zero" 
  and zero_leastII: "\<lbrakk>0 > x; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x = zero" 
begin
  
abbreviation wf_vec :: "'a vec \<Rightarrow> bool"
where "wf_vec \<equiv> wf_vecI n"

abbreviation wf_mat :: "'a mat \<Rightarrow> bool"
where "wf_mat \<equiv> wf_mat_rcI n n"

abbreviation vec0 :: "'a vec"
where "vec0 \<equiv> vec_zeroI zero n"

abbreviation mat0 :: "'a mat"
where "mat0 \<equiv> mat0I zero n n"

abbreviation mat1 :: "'a mat"
where "mat1 \<equiv> mat1I zero one n"

abbreviation vec_max0 :: "'a vec \<Rightarrow> 'a vec"
where "vec_max0 \<equiv> vec_max0I max0"

abbreviation mat_max0 :: "'a mat \<Rightarrow> 'a mat"
where "mat_max0 \<equiv> mat_max0I max0"

abbreviation vec_plus :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
where "vec_plus \<equiv> vec_plusI plus"

abbreviation mat_plus :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
where "mat_plus \<equiv> mat_plusI plus"

abbreviation vec_ge :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_ge \<equiv> vec_geI geq"

abbreviation vec_pre_gt :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_pre_gt \<equiv> vec_pre_gtI gt"

abbreviation vec_gt :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_gt u v \<equiv> vec_ge u v \<and> vec_pre_gt u v"

abbreviation mat_ge :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_ge \<equiv> mat_geI geq"

abbreviation mat_pre_gt :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_pre_gt \<equiv> mat_pre_gtI gt"

abbreviation mat_gt :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_gt \<equiv> mat_gtI geq gt"

abbreviation mat_gt_arc :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_gt_arc \<equiv> mat_geI gt"

abbreviation mat_arc_pos :: "'a mat \<Rightarrow> bool"
where "mat_arc_pos \<equiv> mat_arc_posI arc_pos"

abbreviation vec_arc_pos :: "'a vec \<Rightarrow> bool"
where "vec_arc_pos \<equiv> vec_arc_posI arc_pos"

abbreviation scalar_prod :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a"
where "scalar_prod \<equiv> scalar_prodI zero plus times"

abbreviation matT_vec_mul :: "'a mat \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
where "matT_vec_mul \<equiv> matT_vec_mulI zero plus times"

abbreviation mat_mul :: "nat \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
where "mat_mul \<equiv> mat_mulI zero plus times"

lemma plus_right_mono: "\<lbrakk>(y :: 'a) \<ge> z\<rbrakk> \<Longrightarrow> x + y \<ge> x + z"
  by (simp add: plus_comm[where x = x], rule plus_left_mono, auto)

lemma scalar_product_assoc: 
  assumes wfm: "wf_mat_rcI nr nc m"
  and wfr: "wf_vecI nr r"
  and wfc: "wf_vecI nc c"
  shows "scalar_prod (map (\<lambda>k. scalar_prod r (col m k)) [0..<nc]) c = scalar_prod r (map (\<lambda>k. scalar_prod (row m k) c) [0..<nr])"
using wfm wfc
unfolding col_def 
proof (induct m arbitrary: nc c)
  case Nil
  hence "nc = 0" unfolding wf_mat_rcI_def by (auto)
  with Nil show ?case 
  proof (simp add: scalar_prodI_def row_def)
    from wfr have nr: "nr = length r" unfolding wf_vecI_def by auto
    let ?term = "\<lambda> r :: 'a vec. zip r (map (\<lambda> k. zero) [0..<length r])"
    let ?fun = "\<lambda> xy. plus (times (fst xy) (snd xy))"
    have "foldr ?fun (?term r) zero = zero" 
    proof (induct r, simp)
      case (Cons d r)
      have "foldr ?fun (?term (d # r)) 0 = foldr ?fun ( (d,0) # ?term r) 0" by (simp only: map_replicate_trivial, simp)
      also have "\<dots> = 0" using Cons by simp
      finally show ?case .
    qed
    thus "0 = foldr ?fun (zip r (map (\<lambda> k. 0) [0..<nr])) 0" by (simp add: nr)
  qed
next
  case (Cons v m)
  from this obtain ncc where nc: "nc = Suc ncc" and wf: "wf_mat_rcI nr ncc m" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
  from nc `wf_vecI nc c` obtain a cc where c: "c = a # cc" and wfc: "wf_vecI ncc cc" unfolding wf_vecI_def by (cases c, auto)
  have rec: "scalar_prod (map (\<lambda> k. scalar_prod r (m ! k)) [0..<ncc]) cc = scalar_prod r (map (\<lambda> k. scalar_prod (row m k) cc) [0..<nr])"
    by (rule Cons, rule wf, rule wfc)
  have id: "map (\<lambda>k. scalar_prod r ((v # m) ! k)) [0..<Suc ncc] = scalar_prod r v # map (\<lambda> k. scalar_prod r (m ! k)) [0..<ncc]" by (induct ncc, auto)    
  from wfr have nr: "nr = length r" unfolding wf_vecI_def by auto
  with Cons have v: "length v = length r" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
  have "\<forall> i < nr. wf_vecI ncc (row m i)" by (intro allI impI, rule wf_row[OF wf], simp)
  obtain tm where tm: "tm = transpose nr m" by auto
  hence idk: "\<forall> k < length r. row m k = tm ! k" using col_transpose_is_row[OF wf] unfolding col_def by (auto simp: nr)
  hence idtm1: "map (\<lambda>k. scalar_prod (row m k) cc) [0..<length r] = map (\<lambda>k. scalar_prod (tm ! k) cc) [0..<length r]" 
    and idtm2: "map (\<lambda>k. v ! k * a + scalar_prod (row m k) cc) [0..<length r] = map (\<lambda>k. v ! k * a + scalar_prod (tm ! k) cc) [0..<length r]" by auto
  from tm wf_transpose[OF wf] have "wf_mat_rcI ncc nr tm" by simp
  with nr have "length tm = length r" and  "(\<forall> i < length r. length (tm ! i) = ncc)" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def) 
  with v have main: "scalar_prod r v * a + scalar_prod r (map (\<lambda>k. scalar_prod (tm ! k) cc) [0..<length r]) =
    scalar_prod r (map (\<lambda>k. v ! k * a + scalar_prod (tm ! k) cc) [0..<length r])" 
  proof (induct r arbitrary: v tm)
    case Nil
    thus ?case by (auto simp: scalar_prodI_def row_def)
  next
    case (Cons b r)
    from this obtain c vv where v: "v = c # vv" and vvlen: "length vv = length r" by (cases v, auto)
    from Cons obtain u mm where tm: "tm = u # mm" and mmlen: "length mm = length r"  by (cases tm, auto)
    from Cons tm have argLen: "\<forall> i < length r. length (mm ! i) = ncc" by auto
    have rec: "scalar_prod r vv * a + scalar_prod r (map (\<lambda>k. scalar_prod (mm ! k) cc) [0..<length r]) =
     scalar_prod r (map (\<lambda>k. vv ! k * a + scalar_prod (mm ! k) cc) [0..<length r])" 
      (is "?rv * a + ?recl = ?recr")
      by (rule Cons, auto simp: vvlen mmlen argLen)
    have id: "map (\<lambda>k. scalar_prod ((u # mm) ! k) cc) [0..<length (b # r)] = scalar_prod u cc # map (\<lambda>k. scalar_prod (mm ! k) cc) [0..<length r]" 
      by (simp, induct r, auto)
    have id2: "map (\<lambda>k. (c # vv) ! k * a + scalar_prod ((u # mm) ! k) cc) [0..<length (b # r)] = 
               (c * a + scalar_prod u cc) #
               map (\<lambda>k. vv ! k * a + scalar_prod (mm ! k) cc) [0..<length r]" 
      by (simp, induct r, auto)
    show ?case proof (simp only: v tm, simp only: id, simp only: id2, simp only: scalar_product_cons)
      let ?uc = "scalar_prod u cc"
      let ?bca = "(times b c) * a"
      have "(b * c + ?rv) * a + (b * ?uc + ?recl) = (?bca + ?rv * a) + (b * ?uc + ?recl)" by simp
      also have "\<dots> = (?bca + ?rv * a) + (?recl + b * ?uc)" using plus_comm[of "b * ?uc" ?recl] by simp
      also have "\<dots> = plus ?bca ((?rv * a + ?recl) + b * ?uc)" by auto
      also have "\<dots> = plus ?bca (?recr + b * ?uc)" by (simp only: rec)
      also have "\<dots> = plus ?bca (b * ?uc + ?recr)" using plus_comm[of ?recr "b * ?uc"] by simp
      also have "\<dots> = b * (c * a + ?uc) + ?recr" by simp
      finally show "(b * c + ?rv) * a + (b * ?uc + ?recl) = b * (c * a + ?uc) + ?recr" .
    qed
  qed
  show ?case 
    by (simp add: c scalar_product_cons, simp only: nc, simp only: id, simp add: scalar_product_cons, simp only: rec, simp only: nr, simp only: idtm1 idtm2, simp only: main, simp only: idtm2[symmetric], simp add: row_def scalar_product_cons)
qed

lemma mat_mul_assoc: 
  assumes wf1: "wf_mat_rcI nr n1 m1"
  and wf2: "wf_mat_rcI n1 n2 m2"
  and wf3: "wf_mat_rcI n2 nc m3"
  shows "mat_mul nr (mat_mul nr m1 m2) m3 = mat_mul nr m1 (mat_mul n1 m2 m3)" (is "?m12_3 = ?m1_23")
proof -
  let ?m12 = "mat_mul nr m1 m2"
  let ?m23 = "mat_mul n1 m2 m3"
  from wf1 wf2 have wf12: "wf_mat_rcI nr n2 ?m12" by (rule wf_mat_mul)
  from wf2 wf3 have wf23: "wf_mat_rcI n1 nc ?m23" by (rule wf_mat_mul)
  from wf1 wf23 have wf1_23: "wf_mat_rcI nr nc ?m1_23" by (rule wf_mat_mul)
  from wf12 wf3 have wf12_3: "wf_mat_rcI nr nc ?m12_3" by (rule wf_mat_mul)
  show ?thesis
  proof (simp only: col_eq[OF wf12_3 wf1_23, symmetric], unfold col_def, intro allI impI)
    fix i
    assume i: "i < nc"
    with wf1_23 wf12_3 wf3 have len: "length (?m12_3 ! i) = length (?m1_23 ! i)" and ilen: "i < length m3" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)
    show "?m12_3 ! i = ?m1_23 ! i"
    proof (rule nth_equalityI[OF len], intro allI impI)
      fix j
      assume jlen: "j < length (?m12_3 ! i)"
      with wf12_3 i have j: "j < nr" unfolding wf_mat_rcI_def by (auto simp: wf_vecI_def)      
      show "?m12_3 ! i ! j = ?m1_23 ! i ! j"
	by (simp only: mat_mul_to_scalar[OF wf12 wf3 j i],
	     simp only: mat_mul_to_scalar[OF wf1 wf23 j i], 
	     simp only: row_mat_mul_to_scalar[OF wf1 wf2 j],
	     simp only: col_mat_mul_to_scalar[OF wf2 wf3 i], 
	     simp only: scalar_product_assoc[OF wf2 wf_row[OF wf1 j] wf_col[OF wf3 i]])
    qed
  qed
qed

lemma mat_mul_assoc_n:  
  assumes wf1: "wf_mat m1"
  and wf2: "wf_mat m2"
  and wf3: "wf_mat m3"
  shows "mat_mul n (mat_mul n m1 m2) m3 = mat_mul n m1 (mat_mul n m2 m3)"
using assms
 by (rule mat_mul_assoc)



lemma vec_max0_x: "vec_ge (vec_max0 v) v"
unfolding vec_geI_def vec_max0I_def
by (auto simp: set_zip max0_x)

lemma mat_max0_x: "mat_ge (mat_max0 m) m"
unfolding mat_geI_def mat_max0I_def
by (auto simp: set_zip vec_max0_x)

lemma vec_max0_pos: assumes "wf_vecI nr v"
  shows "vec_ge (vec_max0 v) (vec_zeroI zero nr)"
using assms unfolding wf_vecI_def vec_zeroI_def vec_max0I_def vec_geI_def
by (auto simp: set_zip max0_pos)

lemma mat_max0_pos: assumes "wf_mat_rcI nr nc v"
  shows "mat_ge (mat_max0 v) (mat0I zero nr nc)"
using assms unfolding wf_mat_rcI_def mat0I_def mat_max0I_def mat_geI_def
by (auto simp: set_zip vec_max0_pos)

lemma vec_max0_id_pos: assumes "vec_ge v (vec_zeroI zero nr)" and "wf_vecI nr v"  
  shows "vec_max0 v = v"
using assms unfolding wf_vecI_def vec_zeroI_def vec_max0I_def vec_geI_def 
by (induct v arbitrary: nr, auto simp: max0_id_pos)

lemma mat_max0_id_pos: assumes "mat_ge m (mat0I zero nr nc)" and "wf_mat_rcI nr nc m"  
  shows "mat_max0 m = m"
using assms unfolding wf_mat_rcI_def mat0I_def mat_max0I_def mat_geI_def 
by (induct m arbitrary: nc, auto simp: vec_max0_id_pos)

lemma vec_max0_mono: assumes "vec_ge v w" and "wf_vecI nr v" and "wf_vecI nr w"
  shows "vec_ge (vec_max0 v) (vec_max0 w)"
using assms unfolding wf_vecI_def vec_max0I_def vec_geI_def
proof (simp, induct nr arbitrary: v w, simp)
  case (Suc nn)
  show ?case
  proof (cases v, simp)
    case (Cons a v1)
    show ?thesis
    proof (cases w, simp)
      case (Cons b w1)
      from prems have len: "(a \<ge> b) \<and> length v1 = nn \<and> length w1 = nn \<and> (\<forall> a b . (a,b) \<in> set (zip v1 w1) \<longrightarrow> (a \<ge> b))" by auto
      show ?thesis
	by (simp add: `v = a # v1` `w = b # w1`, rule conjI, rule max0_mono, simp add: len, rule Suc, auto simp: len)
    qed
  qed
qed

lemma mat_max0_mono: assumes "mat_ge v w" and "wf_mat_rcI nr nc v" and "wf_mat_rcI nr nc w"
  shows "mat_ge (mat_max0 v) (mat_max0 w)"
using assms unfolding wf_mat_rcI_def mat_max0I_def mat_geI_def
proof (simp, induct nc arbitrary: v w, simp)
  case (Suc nn)
  show ?case
  proof (cases v, simp)
    case (Cons a v1)
    show ?thesis
    proof (cases w, simp)
      case (Cons b w1)
      from prems have len: "(vec_ge a b) \<and> wf_vecI nr a \<and> wf_vecI nr b \<and> length v1 = nn \<and> length w1 = nn \<and> (\<forall> a \<in> set v1. wf_vecI nr a) \<and> (\<forall> b \<in> set w1. wf_vecI nr b) \<and> (\<forall> a b . (a,b) \<in> set (zip v1 w1) \<longrightarrow> (vec_ge a b))" by auto
      show ?thesis 
	by (simp add: `v = a # v1` `w = b # w1`, rule conjI, rule vec_max0_mono[where nr = nr], (simp add: len)+, rule Suc, auto simp: len)
    qed
  qed
qed


lemma vec_plus_comm : "vec_plus x y = vec_plus y x"
unfolding vec_plusI_def
proof (induct x arbitrary: y, simp)
  case (Cons a x)
  thus ?case 
    by (cases y, simp, auto simp: plus_comm) 
qed

lemma mat_plus_comm : "mat_plus m1 m2 = mat_plus m2 m1"
unfolding mat_plusI_def
proof (induct m1 arbitrary: m2, simp)
  case (Cons v m1)
  thus ?case
  proof (cases m2, simp)
    case (Cons w m2a)
    hence "mat_plus (v # m1) m2 = vec_plus v w # mat_plus m1 m2a" by (auto simp: mat_plusI_def)
    also have "\<dots> = vec_plus w v # mat_plus m1 m2a" using vec_plus_comm by auto
    finally show ?thesis using prems by (auto simp: mat_plusI_def)
  qed
qed

lemma vec_plus_assoc[simp]: assumes "wf_vecI nr u" and "wf_vecI nr v" and "wf_vecI nr w"
 shows "vec_plus u (vec_plus v w) = vec_plus (vec_plus u v) w"
using assms
unfolding wf_vecI_def 
proof (induct nr arbitrary: u v w, simp add: vec_plusI_def)
 case (Suc nn)
 thus ?case 
 proof (cases u, simp, cases v, simp, cases w, simp)
   fix x uu y vv z ww
   assume "u = x # uu" and "v = y # vv" and "w = z # ww"
   with Suc show "vec_plus u (vec_plus v w) = vec_plus (vec_plus u v) w" by (simp add: vec_plusI_def)
 qed
qed

lemma mat_plus_assoc[simp]: assumes "wf_mat_rcI nr nc m1" and "wf_mat_rcI nr nc m2" and "wf_mat_rcI nr nc m3"
  shows "mat_plus m1 (mat_plus m2 m3) = mat_plus (mat_plus m1 m2) m3"
using assms
proof (induct nc arbitrary: m1 m2 m3, simp add: mat_plusI_def wf_mat_rcI_def)
  case (Suc nn)
  thus ?case
  proof (cases m1, simp add: wf_mat_rcI_def, cases m2, simp add: wf_mat_rcI_def, cases m3, simp add: wf_mat_rcI_def)
    fix v1 mm1 v2 mm2 v3 mm3
    assume 
      "wf_mat_rcI nr (Suc nn) m1" and m1: "m1 = v1 # mm1" and
      "wf_mat_rcI nr (Suc nn) m2" and m2: "m2 = v2 # mm2" and
      "wf_mat_rcI nr (Suc nn) m3" and m3: "m3 = v3 # mm3"
    hence v1: "wf_vecI nr v1" and v2: "wf_vecI nr v2" and v3: "wf_vecI nr v3" by (auto simp: wf_mat_rcI_def)
    hence id: "vec_plus v1 (vec_plus v2 v3) = vec_plus (vec_plus v1 v2) v3" by (rule vec_plus_assoc)
    from m1 m2 m3 have "mat_plus m1 (mat_plus m2 m3) = vec_plus v1 (vec_plus v2 v3) # mat_plus mm1 (mat_plus mm2 mm3)" by (auto simp: mat_plusI_def)
    also have "\<dots> = vec_plus (vec_plus v1 v2) v3 # mat_plus mm1 (mat_plus mm2 mm3)" by (simp only: id)
    finally show "mat_plus m1 (mat_plus m2 m3) = mat_plus (mat_plus m1 m2) m3" using prems unfolding wf_mat_rcI_def mat_plusI_def by auto 
  qed
qed

lemma vec0_plus[simp]: assumes "wf_vecI nr u" shows "vec_plus (vec_zeroI zero nr) u = u"
using assms
unfolding wf_vecI_def vec_plusI_def vec_zeroI_def
proof (induct nr arbitrary: u, simp)
 case (Suc nn) thus ?case by (cases u, auto)
qed

lemma mat0_plus[simp]: assumes "wf_mat_rcI nr nc m" shows "mat_plus (mat0I zero nr nc) m = m"
using assms 
unfolding wf_mat_rcI_def 
proof (induct nc arbitrary: m, simp add: mat_plusI_def mat0I_def)
  case (Suc nn) 
  thus ?case proof (cases m, simp)
    case (Cons v mm)
    with Suc have wf: "wf_vecI nr v" by auto
    from Cons Suc have "mat_plus (mat0I 0 nr (Suc nn)) m = vec_plus (vec_zeroI zero nr) v # mat_plus (mat0I 0 nr nn) mm" by (auto simp: mat_plusI_def mat0I_def)
    also have "\<dots> = vec_plus (vec_zeroI zero nr) v # mm" using Suc Cons by auto
    also have "\<dots> = v # mm" by (simp only: vec0_plus wf)
    finally show ?thesis using Cons by auto
  qed
qed


lemma scalar_left_zero : "scalar_prod (vec_zeroI zero nn) v = 0"
  unfolding vec_zeroI_def scalar_prodI_def
proof (induct nn arbitrary: v, simp)
  case (Suc m)
  thus ?case by (cases v, auto)
qed

lemma scalar_right_zero : "scalar_prod v (vec_zeroI zero nn) = 0"
  unfolding vec_zeroI_def scalar_prodI_def
proof (induct v arbitrary: nn, simp)
  case (Cons a vv)
  thus ?case by (cases nn, auto)
qed



lemma mat0_mult_left: assumes wf: "wf_mat_rcI nc ncc m"
  shows "mat_mul nr (mat0I zero nr nc) m = (mat0I zero nr ncc)"
proof (simp only: index_eq[OF wf_mat_mul[OF wf_mat0 wf] wf_mat0, symmetric], intro allI impI)
  fix i j
  assume i: "i < ncc" and j: "j < nr"
  show "mat_mul nr (mat0I 0 nr nc) m ! i ! j = mat0I 0 nr ncc ! i ! j"
  by (simp only: mat_mul_to_scalar[OF wf_mat0 wf j i], 
         simp only: mat0_index[OF i j], 
         simp only: mat0_row[OF j],
         simp only: scalar_left_zero)
qed


lemma mat0_mult_right: assumes wf: "wf_mat_rcI nr nc m"
  shows "mat_mul nr m (mat0I zero nc ncc) = (mat0I zero nr ncc)"
proof (simp only: index_eq[OF wf_mat_mul[OF wf wf_mat0] wf_mat0, symmetric], intro allI impI)
  fix i j
  assume i: "i < ncc" and j: "j < nr"
  show "mat_mul nr m (mat0I 0 nc ncc) ! i ! j = mat0I 0 nr ncc ! i ! j"
    by (simp only: mat_mul_to_scalar[OF wf wf_mat0 j i],
         simp only: mat0_index[OF i j],
         simp only: mat0_col[OF i],
         simp only: scalar_right_zero)
qed

lemma scalar_vec_plus_distrib_right: 
  assumes wf1: "wf_vecI nr u"
  assumes wf2: "wf_vecI nr v"
  assumes wf3: "wf_vecI nr w"
  shows "scalar_prod u (vec_plus v w) = scalar_prod u v + scalar_prod u w"
using assms
unfolding wf_vecI_def scalar_prodI_def vec_plusI_def
proof (induct nr arbitrary: u v w, simp)
  case (Suc n)
  from Suc obtain a uu where u: "u = a # uu" by (cases u, auto)
  from Suc obtain b vv where v: "v = b # vv" by (cases v, auto)
  from Suc obtain c ww where w: "w = c # ww" by (cases w, auto)
  from Suc u v w show ?case by (auto simp: plus_comm[of _ "a*c"])
qed

lemma scalar_vec_plus_distrib_left: 
  assumes wf1: "wf_vecI nr u"
  assumes wf2: "wf_vecI nr v"
  assumes wf3: "wf_vecI nr w"
  shows "scalar_prod (vec_plus u v) w = scalar_prod u w + scalar_prod v w"
using assms
unfolding wf_vecI_def scalar_prodI_def vec_plusI_def
proof (induct nr arbitrary: u v w, simp)
  case (Suc n)
  from Suc obtain a uu where u: "u = a # uu" by (cases u, auto)
  from Suc obtain b vv where v: "v = b # vv" by (cases v, auto)
  from Suc obtain c ww where w: "w = c # ww" by (cases w, auto)
  from Suc u v w show ?case by (auto simp: plus_comm[of _ "b*c"])
qed

lemma mat_mult_plus_distrib_right: 
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nc ncc m2"
  and wf3: "wf_mat_rcI nc ncc m3"
  shows "mat_mul nr m1 (mat_plus m2 m3) = mat_plus (mat_mul nr m1 m2) (mat_mul nr m1 m3)" (is "mat_mul nr m1 ?m23 = mat_plus ?m12 ?m13")
proof -
  let ?m1_23 = "mat_mul nr m1 ?m23"
  let ?m12_13 = "mat_plus ?m12 ?m13"
  from wf_mat_plus[OF wf2 wf3] have wf23: "wf_mat_rcI nc ncc ?m23" .
  from wf_mat_mul[OF wf1 wf2] have wf12: "wf_mat_rcI nr ncc ?m12" .
  from wf_mat_mul[OF wf1 wf3] have wf13: "wf_mat_rcI nr ncc ?m13" .
  from wf_mat_mul[OF wf1 wf23] have wf1_23: "wf_mat_rcI nr ncc ?m1_23" .
  from wf_mat_plus[OF wf12 wf13] have wf12_13: "wf_mat_rcI nr ncc ?m12_13" .
  show ?thesis 
  proof (simp only: index_eq[OF wf1_23 wf12_13, symmetric], intro impI allI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    show "?m1_23 ! i ! j = ?m12_13 ! i ! j"
      by (simp only: mat_mul_to_scalar[OF wf1 wf23 j i],
           simp only: mat_plus_index[OF wf12 wf13 i j],
	   simp only: mat_mul_to_scalar[OF wf1 wf2 j i],
	   simp only: mat_mul_to_scalar[OF wf1 wf3 j i],
	   simp only: col_mat_plus[OF wf2 wf3 i],
	rule scalar_vec_plus_distrib_right[OF wf_row[OF wf1 j] wf_col[OF wf2 i] wf_col[OF wf3 i]])
  qed
qed

lemma mat_mult_plus_distrib_left: 
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  and wf3: "wf_mat_rcI nc ncc m3"
  shows "mat_mul nr (mat_plus m1 m2) m3 = mat_plus (mat_mul nr m1 m3) (mat_mul nr m2 m3)" (is "mat_mul nr ?m12 _ = mat_plus ?m13 ?m23")
proof -
  let ?m12_3 = "mat_mul nr ?m12 m3"
  let ?m13_23 = "mat_plus ?m13 ?m23"
  from wf_mat_plus[OF wf1 wf2] have wf12: "wf_mat_rcI nr nc ?m12" .
  from wf_mat_mul[OF wf1 wf3] have wf13: "wf_mat_rcI nr ncc ?m13" .
  from wf_mat_mul[OF wf2 wf3] have wf23: "wf_mat_rcI nr ncc ?m23" .
  from wf_mat_mul[OF wf12 wf3] have wf12_3: "wf_mat_rcI nr ncc ?m12_3" .
  from wf_mat_plus[OF wf13 wf23] have wf13_23: "wf_mat_rcI nr ncc ?m13_23" .
  show ?thesis 
  proof (simp only: index_eq[OF wf12_3 wf13_23, symmetric], intro impI allI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    show "?m12_3 ! i ! j = ?m13_23 ! i ! j"
      by (simp only: mat_mul_to_scalar[OF wf12 wf3 j i],
           simp only: mat_plus_index[OF wf13 wf23 i j],
	   simp only: mat_mul_to_scalar[OF wf1 wf3 j i],
	   simp only: mat_mul_to_scalar[OF wf2 wf3 j i],
	   simp only: row_mat_plus[OF wf1 wf2 j],
	   rule scalar_vec_plus_distrib_left[OF wf_row[OF wf1 j] wf_row[OF wf2 j] wf_col[OF wf3 i]])
  qed
qed


lemma scalar_left_one : assumes wf: "wf_vecI nn v"
  and i: "i < nn"
  shows "scalar_prod (vec1I zero one nn i) v = v ! i"
  using assms 
  unfolding vec1I_def wf_vecI_def 
proof (induct nn arbitrary: v i, blast)
  case (Suc n)
  from this obtain a vv where v: "v = a # vv" and lvv: "length vv = n" by (cases v, auto)
  show ?case 
  proof (cases i)
    case 0
    thus ?thesis using scalar_left_zero unfolding vec_zeroI_def by (simp add: v scalar_product_cons plus_comm[of _ zero])
  next
    case (Suc ii)
    thus ?thesis using prems lvv v by (auto simp: scalar_product_cons)
  qed
qed


lemma scalar_right_one : assumes wf: "wf_vecI nn v"
  and i: "i < nn"
  shows "scalar_prod v (vec1I zero one nn i) = v ! i"
  using assms 
  unfolding vec1I_def wf_vecI_def 
proof (induct nn arbitrary: v i, blast)
  case (Suc n)
  from this obtain a vv where v: "v = a # vv" and lvv: "length vv = n" by (cases v, auto)
  show ?case 
  proof (cases i)
    case 0
    thus ?thesis using scalar_right_zero unfolding vec_zeroI_def by (simp add: v scalar_product_cons plus_comm[of _ zero])
  next
    case (Suc ii)
    thus ?thesis using prems lvv v by (auto simp: scalar_product_cons)
  qed
qed


lemma mat1_mult_right: assumes wf: "wf_mat_rcI nr nc m"
  shows "mat_mul nr m (mat1I zero one nc) = m"
proof (simp only: index_eq[OF wf_mat_mul[OF wf wf_mat1] wf, symmetric], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "mat_mul nr m (mat1I 0 1 nc) ! i ! j = m ! i ! j"
    by (simp only: mat_mul_to_scalar[OF wf wf_mat1 j i],
    simp only: col_mat1[OF i],
    simp only: scalar_right_one[OF wf_row[OF wf j] i],
    simp only: row_col[OF wf j i],
    unfold col_def, simp)
qed


lemma mat1_mult_left: assumes wf: "wf_mat_rcI nr nc m"
  shows "mat_mul nr (mat1I zero one nr) m = m"
proof (simp only: index_eq[OF wf_mat_mul[OF wf_mat1 wf] wf, symmetric], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "mat_mul nr (mat1I 0 1 nr) m ! i ! j = m ! i ! j"
    by (simp only: mat_mul_to_scalar[OF wf_mat1 wf j i],
      simp only: row_mat1[OF j],
      simp only: scalar_left_one[OF wf_col[OF wf i] j], unfold col_def, simp)
qed


lemma vec_plus_left_mono: "\<lbrakk>vec_ge v1 v2; wf_vecI nr v1; wf_vecI nr v2; wf_vecI nr v3\<rbrakk> \<Longrightarrow> vec_ge (vec_plus v1 v3) (vec_plus v2 v3)"
unfolding vec_geI_def vec_plusI_def wf_vecI_def
proof (simp, induct nr arbitrary: v1 v2 v3, simp)
  case (Suc nn)
  show ?case 
  proof (cases v1, simp, cases v2, simp, cases v3, simp)
    fix a1 w1 a2 w2 a3 w3
    assume v1: "v1 = a1 # w1" and v2: "v2 = a2 # w2" and v3: "v3 = a3 # w3"    
    with Suc have geq: "a1 \<ge> a2" by auto
    from v1 v2 v3 Suc have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto
    show "\<forall> a \<in> set (zip (map (\<lambda> xy. fst xy + snd xy) (zip v1 v3)) 
                         (map (\<lambda> xy. fst xy + snd xy) (zip v2 v3))). split op \<ge> a"
      by (simp add: v1 v2 v3, rule conjI, rule plus_left_mono, rule geq, rule Suc, auto simp: Suc v1 v2 v3 len)
  qed
qed

lemma mat_plus_left_mono: assumes ge: "mat_ge m1 m2" and wf1: "wf_mat_rcI nr nc m1" and wf2: "wf_mat_rcI nr nc m2" and wf3: "wf_mat_rcI nr nc m3" 
  shows "mat_ge (mat_plus m1 m3) (mat_plus m2 m3)"
proof -
  from ge have "\<forall> (v,w) \<in> set (zip m1 m2). vec_ge v w" unfolding mat_geI_def by auto
  with wf1 wf2 wf3 show "mat_ge (mat_plus m1 m3) (mat_plus m2 m3)" unfolding mat_geI_def mat_plusI_def wf_mat_def
  proof (simp, induct nc arbitrary: m1 m2 m3, simp)
    case (Suc nn)
    show ?case 
    proof (cases m1, simp, cases m2, simp, cases m3, simp)
      fix v1 mm1 v2 mm2 v3 mm3
      assume v1: "m1 = v1 # mm1" and v2: "m2 = v2 # mm2" and v3: "m3 = v3 # mm3"    
      with Suc have geq: "vec_ge v1 v2" by auto
      from v1 v2 v3 Suc have len: "length mm1 = nn \<and> length mm2 = nn \<and> length mm3 = nn" by auto
      show "\<forall> a \<in> set (zip (map (\<lambda> xy. vec_plus (fst xy) (snd xy)) (zip m1 m3)) 
                      (map (\<lambda> xy. vec_plus (fst xy) (snd xy)) (zip m2 m3))). split vec_ge a"
	by (simp add: v1 v2 v3, rule conjI, rule vec_plus_left_mono[where nr = nr], rule geq,   simp add: Suc v1, simp add: Suc v2, simp add: Suc v3, rule Suc,
	  auto simp: len v1 v2 v3 Suc)
    qed
  qed
qed
      
lemma vec_ge_refl: "vec_ge v v"
  unfolding vec_geI_def 
  by (induct v, auto)

lemma mat_ge_refl: "mat_ge m m"
  unfolding mat_geI_def
  by (induct m, auto simp: vec_ge_refl)

lemma vec_ge_trans: assumes "vec_ge v1 v2" and "vec_ge v2 v3" and "wf_vecI nr v1" and "wf_vecI nr v2" and "wf_vecI nr v3" 
  shows "vec_ge v1 v3"
unfolding vec_geI_def
using assms unfolding wf_vecI_def
proof (simp, induct nr arbitrary: v1 v2 v3, simp)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by auto
      next
	case (Cons a3 w3)
	from `vec_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `vec_ge v2 v3` 
	have a12: "a1 \<ge> a2" and a23: "a2 \<ge> a3" and w12: "vec_ge w1 w2" and w23: "vec_ge w2 w3" unfolding vec_geI_def by auto
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto 
	show ?thesis 
	  by (simp add: `v1 = a1 # w1` Cons, rule conjI, rule geq_trans, rule a12, rule a23, rule Suc, rule w12, rule w23, auto simp: len w12 w23)
      qed
    qed
  qed
qed


lemma mat_ge_trans: assumes "mat_ge v1 v2" and "mat_ge v2 v3" and "wf_mat_rcI nr nc v1" and "wf_mat_rcI nr nc v2" and "wf_mat_rcI nr nc v3" 
  shows "mat_ge v1 v3"
unfolding mat_geI_def 
using assms 
proof (simp, induct nc arbitrary: v1 v2 v3, simp add: wf_mat_rcI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
      next
	case (Cons a3 w3)
	from `mat_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `mat_ge v2 v3` 
	have a12: "vec_ge a1 a2" and a23: "vec_ge a2 a3" and w12: "mat_ge w1 w2" and w23: "mat_ge w2 w3" unfolding mat_geI_def by auto
	from `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `wf_mat_rcI nr (Suc nn) v1` `wf_mat_rcI nr (Suc nn) v2` `wf_mat_rcI nr (Suc nn) v3` 
	have wf: "wf_mat_rcI nr nn w1 \<and> wf_mat_rcI nr  nn w2 \<and> wf_mat_rcI nr nn w3 \<and> wf_vecI nr a1 \<and> wf_vecI nr a2 \<and> wf_vecI nr a3" unfolding wf_mat_def by auto
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" unfolding wf_mat_def by auto 
	show ?thesis 
	  by (simp add: `v1 = a1 # w1` Cons, rule conjI, rule vec_ge_trans[where nr = nr], rule a12, rule a23, (simp add: wf)+, rule Suc, rule w12, rule w23, auto simp: wf, rule w12, rule w23)
      qed
    qed
  qed
qed


lemma vec1_ge_vec0: assumes "i < n" shows "vec_ge (vec1I zero one n i) vec0"
unfolding vec_geI_def vec_zeroI_def vec1I_def using assms
proof (auto simp: set_zip)
  fix j
  assume i: "i < n" and j: "j < n"
  show "(replicate i 0 @ 1 # replicate (n - Suc i) 0) ! j \<ge> (0 :: 'a)" (is "?repl j \<ge> (0 :: 'a)")
  proof (cases "j < i", simp add: nth_append)
    case False
    show ?thesis
    proof (cases "j - i", simp add: nth_append)
      case (Suc k)
      with False
      have repl: "?repl j = (replicate (n - Suc i) 0) ! k" by (simp add: nth_append)
      from Suc False j have "k < (n - Suc i)" by arith
      with repl show ?thesis by auto
    qed
  qed
qed

lemma mat1_ge_mat0: shows "mat_ge mat1 mat0"
unfolding mat_geI_def mat0I_def mat1I_def
by (auto simp: set_zip vec1_ge_vec0)


lemma scalar_prod_mono_left: assumes wf1: "wf_vecI nr v1"
  and wf2: "wf_vecI nr v2"
  and wf3: "wf_vecI nr v3"
  and ge1: "vec_ge v1 v2"
  and ge2: "vec_ge v3 (vec_zeroI zero nr)"
  shows "scalar_prod v1 v3 \<ge> scalar_prod v2 v3"
using assms unfolding wf_vecI_def vec_geI_def vec_zeroI_def
proof (simp)
  assume " length v1 = nr" and "length v2 = nr" and " length v3 = nr" and " \<forall>x\<in>set (zip v1 v2). split op \<ge> x" and " \<forall>x\<in>set (zip v3 (replicate nr 0)). split op \<ge> x"
  thus "scalar_prod v1 v3 \<ge> scalar_prod v2 v3"
  proof (induct nr arbitrary: v1 v2 v3, simp add: scalar_prodI_def)
    case (Suc nrr)
    from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and w1: "length w1 = nrr" by (cases v1, auto)
    from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and w2: "length w2 = nrr" by (cases v2, auto)
    from Suc obtain a3 w3 where v3: "v3 = a3 # w3" and w3: "length w3 = nrr" by (cases v3, auto)
    from Suc have rec: "scalar_prod w1 w3 \<ge> scalar_prod w2 w3" (is "geq ?l ?r")
      by (auto simp: w1 w2 w3 v1 v2 v3)
    show ?case proof (simp add: v1 v2 v3 scalar_product_cons)
      have one: "a1 * a3 \<ge> a2 * a3" using times_left_mono[of a3 a1 a2] Suc v1 v2 v3 by auto
      hence "a1 * a3 + ?l \<ge> a2 * a3 + ?l" by (rule plus_left_mono)
      also have "\<dots> \<ge> a2 * a3 + ?r" using rec by (rule plus_right_mono)
      finally show "a1 * a3 + ?l \<ge> a2 * a3 + ?r" .
    qed
  qed
qed

lemma scalar_prod_mono_right: assumes wf1: "wf_vecI nr v1"
  and wf2: "wf_vecI nr v2"
  and wf3: "wf_vecI nr v3"
  and ge1: "vec_ge v2 v3"
  and ge2: "vec_ge v1 (vec_zeroI zero nr)"
  shows "scalar_prod v1 v2 \<ge> scalar_prod v1 v3"
using assms unfolding wf_vecI_def vec_geI_def vec_zeroI_def
proof (simp)
  assume " length v1 = nr" and "length v2 = nr" and " length v3 = nr" and " \<forall>x\<in>set (zip v2 v3). split op \<ge> x" and " \<forall>x\<in>set (zip v1 (replicate nr 0)). split op \<ge> x"
  thus "scalar_prod v1 v2 \<ge> scalar_prod v1 v3"
  proof (induct nr arbitrary: v1 v2 v3, simp add: scalar_prodI_def)
    case (Suc nrr)
    from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and w1: "length w1 = nrr" by (cases v1, auto)
    from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and w2: "length w2 = nrr" by (cases v2, auto)
    from Suc obtain a3 w3 where v3: "v3 = a3 # w3" and w3: "length w3 = nrr" by (cases v3, auto)
    from Suc have rec: "scalar_prod w1 w2 \<ge> scalar_prod w1 w3" (is "geq ?l ?r")
      by (auto simp: w1 w2 w3 v1 v2 v3)
    show ?case proof (simp add: v1 v2 v3 scalar_product_cons)
      have one: "a1 * a2 \<ge> a1 * a3" using times_right_mono[of a1 a2 a3] Suc v1 v2 v3 by auto
      hence "a1 * a2 + ?l \<ge> a1 * a3 + ?l" by (rule plus_left_mono)
      also have "\<dots> \<ge> a1 * a3 + ?r" using rec by (rule plus_right_mono)
      finally show "a1 * a2 + ?l \<ge> a1 * a3 + ?r" .
    qed
  qed
qed


lemma mat_mul_left_mono:
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nr nc m2"
  and wf3: "wf_mat_rcI nc ncc m3"
  and ge1: "mat_ge m1 m2"
  and ge2: "mat_ge m3 (mat0I zero nc ncc)"
  shows "mat_ge (mat_mul nr m1 m3) (mat_mul nr m2 m3)"
proof -
  let ?m13 = "mat_mul nr m1 m3"
  let ?m23 = "mat_mul nr m2 m3"
  from wf_mat_mul[OF wf1 wf3] have wf13: "wf_mat_rcI nr ncc ?m13" .
  from wf_mat_mul[OF wf2 wf3] have wf23: "wf_mat_rcI nr ncc ?m23" .
  show ?thesis 
  proof (simp only: mat_ge_to_index[OF wf13 wf23], intro allI impI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    from ge1 have ge1a: "\<forall>i<nc. \<forall> j < nr. m1 ! i ! j \<ge>  m2 ! i ! j"
      using mat_ge_to_index[OF wf1 wf2] by simp
    from ge2 have ge2a: "\<forall>ia<nc. col m3 i ! ia \<ge> vec_zeroI 0 nc ! ia"
      using mat_ge_to_index[OF wf3 wf_mat0] i unfolding col_def mat0I_def vec_zeroI_def
      by auto      
    show "?m13 ! i ! j \<ge> ?m23 ! i ! j"
      by (simp only: mat_mul_to_scalar[OF wf1 wf3 j i],
	simp only: mat_mul_to_scalar[OF wf2 wf3 j i], 
	rule scalar_prod_mono_left[OF wf_row[OF wf1 j] wf_row[OF wf2 j] wf_col[OF wf3 i]],
	simp only: vec_ge_to_index[OF wf_row[OF wf1 j] wf_row[OF wf2 j]],
	simp add: row_col[OF wf1 j] row_col[OF wf2 j] col_def, (auto simp: ge1a j)[1],
	simp only: vec_ge_to_index[OF wf_col[OF wf3 i] wf_vec0],
	rule ge2a)
  qed
qed  


lemma mat_mul_right_mono:
  assumes wf1: "wf_mat_rcI nr nc m1"
  and wf2: "wf_mat_rcI nc ncc m2"
  and wf3: "wf_mat_rcI nc ncc m3"
  and ge1: "mat_ge m1 (mat0I zero nr nc)"
  and ge2: "mat_ge m2 m3"
  shows "mat_ge (mat_mul nr m1 m2) (mat_mul nr m1 m3)"
proof -
  let ?m12 = "mat_mul nr m1 m2"
  let ?m13 = "mat_mul nr m1 m3"
  from wf_mat_mul[OF wf1 wf2] have wf12: "wf_mat_rcI nr ncc ?m12" .
  from wf_mat_mul[OF wf1 wf3] have wf13: "wf_mat_rcI nr ncc ?m13" .
  show ?thesis 
  proof (simp only: mat_ge_to_index[OF wf12 wf13], intro allI impI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    from ge2 i have ge2a: " \<forall>ia<nc. col m2 i ! ia \<ge> col m3 i ! ia"
      using mat_ge_to_index[OF wf2 wf3] unfolding col_def by auto
    from ge1 j have ge1a: " \<forall>i<nc. m1 ! i ! j \<ge> 0" 
      using mat_ge_to_index[OF wf1 wf_mat0] unfolding mat0I_def vec_zeroI_def
      by auto
    show "?m12 ! i ! j \<ge> ?m13 ! i ! j"
      by  (simp only: mat_mul_to_scalar[OF wf1 wf2 j i],
	simp only: mat_mul_to_scalar[OF wf1 wf3 j i],
	rule scalar_prod_mono_right[OF wf_row[OF wf1 j] wf_col[OF wf2 i] wf_col[OF wf3 i]], 
	simp only: vec_ge_to_index[OF wf_col[OF wf2 i] wf_col[OF wf3 i]], rule ge2a, 
	simp only: vec_ge_to_index[OF wf_row[OF wf1 j] wf_vec0],
	simp add: row_col[OF wf1 j] vec_zeroI_def col_def, rule ge1a) 
  qed
qed

lemma mat_arc_pos_plus: assumes wf1: "wf_mat m1"
  and wf2: "wf_mat m2"
  and arc_pos: "mat_arc_pos m1"
  shows "mat_arc_pos (mat_plus m1 m2)"
proof -
  from n_pos wf1 obtain v1 mm1 where m1: "m1 = v1 # mm1" unfolding wf_mat_rcI_def by (cases m1, auto)
  from n_pos wf2 obtain v2 mm2 where m2: "m2 = v2 # mm2" unfolding wf_mat_rcI_def by (cases m2, auto)	
  from n_pos wf1 m1 obtain a1 vv1 where v1: "v1 = a1 # vv1" unfolding wf_mat_rcI_def by (cases v1, auto simp: wf_vecI_def)
  from n_pos wf2 m2 obtain a2 vv2 where v2: "v2 = a2 # vv2" unfolding wf_mat_rcI_def by (cases v2, auto simp: wf_vecI_def)
  from m1 v1 arc_pos have "arc_pos a1" unfolding mat_arc_posI_def by simp
  hence "arc_pos (plus a1 a2)" by (rule arc_pos_plus)
  with m1 v1 m2 v2 show ?thesis unfolding mat_arc_posI_def mat_plusI_def vec_plusI_def by simp
qed

lemma mat_arc_pos_mul: assumes wf1: "wf_mat m1"
  and wf2: "wf_mat m2"
  and ap1: "mat_arc_pos m1"
  and ap2: "mat_arc_pos m2"
  shows "mat_arc_pos (mat_mul n m1 m2)"
proof -
  from n_pos wf1 obtain v1 mm1 where m1: "m1 = v1 # mm1" unfolding wf_mat_rcI_def by (cases m1, auto)
  from n_pos wf2 obtain v2 mm2 where m2: "m2 = v2 # mm2" unfolding wf_mat_rcI_def by (cases m2, auto)	
  from n_pos wf1 m1 obtain a1 vv1 where v1: "v1 = a1 # vv1" unfolding wf_mat_rcI_def by (cases v1, auto simp: wf_vecI_def)
  from n_pos wf2 m2 obtain a2 vv2 where v2: "v2 = a2 # vv2" unfolding wf_mat_rcI_def by (cases v2, auto simp: wf_vecI_def)
  from m1 v1 ap1 have a1: "arc_pos a1" unfolding mat_arc_posI_def by simp
  from m2 v2 ap2 have a2: "arc_pos a2" unfolding mat_arc_posI_def by simp
  from a1 a2 have prod: "arc_pos (a1 * a2)" by (rule arc_pos_mult)
  show ?thesis unfolding mat_arc_posI_def 
    by (simp only: mat_mul_to_scalar[OF wf1 wf2 n_pos n_pos],
    simp add: m1 v1 m2 v2 col_def row_def scalar_product_cons,
    rule arc_pos_plus[OF prod])
qed

lemma mat_arc_pos_max0: assumes wf: "wf_mat m" 
  shows "mat_arc_pos (mat_max0 m) = mat_arc_pos m"
proof -
  from n_pos wf obtain v mm where m: "m = v # mm" unfolding wf_mat_rcI_def by (cases m, auto)
  from n_pos wf m obtain a vv where v: "v = a # vv" unfolding wf_mat_rcI_def by (cases v, auto simp: wf_vecI_def)
  show ?thesis unfolding mat_arc_posI_def mat_max0I_def vec_max0I_def
    by (simp add: m v, rule arc_pos_max0)
qed

lemma mat_arc_pos_mat1: "mat_arc_pos mat1"
unfolding mat1I_def mat_arc_posI_def 
using n_pos
by (auto simp: vec1I_def arc_pos_one)


(* here the part with gt starts where arctic is different from non-arctic *)
lemma vec_pre_compat: assumes "vec_ge v1 v2" and "vec_pre_gt v2 v3" and "wf_vec v1" and "wf_vec v2" and "wf_vec v3" 
  shows "vec_pre_gt v1 v3"
unfolding vec_pre_gtI_def
using assms unfolding wf_vecI_def 
proof (simp add: any_mem_conv mem_iff, induct n arbitrary: v1 v2 v3, simp add: vec_pre_gtI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by auto
      next
	case (Cons a3 w3)
	from `vec_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `vec_pre_gt v2 v3` 
	have a12: "a1 \<ge> a2" and w12: "vec_ge w1 w2" and w23: "(a2 > a3) \<or> vec_pre_gt w2 w3" unfolding vec_geI_def vec_pre_gtI_def by (cases "a2 > a3", auto)
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto 
	from w23 show ?thesis
	proof
	  assume "a2 > a3"
	  with a12 have "a1 > a3" by (rule compat)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `a1 > a3` show ?thesis by blast
	next
	  assume gt: "vec_pre_gt w2 w3"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (a > b)"
	    by (rule Suc, rule `vec_ge w1 w2`, auto simp: len w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed

lemma vec_pre_compat2: assumes "vec_pre_gt v1 v2" and "vec_ge v2 v3" and "wf_vec v1" and "wf_vec v2" and "wf_vec v3" 
  shows "vec_pre_gt v1 v3"
unfolding vec_pre_gtI_def
using assms unfolding wf_vecI_def 
proof (simp add: any_mem_conv mem_iff, induct n arbitrary: v1 v2 v3, simp add: vec_pre_gtI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by auto
      next
	case (Cons a3 w3)
	from `vec_pre_gt v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `vec_ge v2 v3` 
	have a12: "a2 \<ge> a3" and w12: "vec_ge w2 w3" and w23: "(a1 > a2) \<or> vec_pre_gt w1 w2" unfolding vec_geI_def vec_pre_gtI_def by (cases "a1 > a2", auto)
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto 
	from w23 show ?thesis
	proof
	  assume "a1 > a2"
	  from this and a12 have "a1 > a3" by (rule compat2)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `a1 > a3` show ?thesis by blast
	next
	  assume gt: "vec_pre_gt w1 w2"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (a > b)"
	    by (rule Suc, rule gt, rule w12, auto simp: len w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed

lemma mat_pre_compat: assumes "mat_ge v1 v2" and "mat_pre_gt v2 v3" and "wf_mat_rcI n nc v1" and "wf_mat_rcI n nc v2" and "wf_mat_rcI n nc v3" 
  shows "mat_pre_gt v1 v3"
unfolding mat_pre_gtI_def
using assms 
proof (simp add: any_mem_conv mem_iff, induct nc arbitrary: v1 v2 v3, simp add: mat_pre_gtI_def wf_mat_rcI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
      next
	case (Cons a3 w3)
	from `mat_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `mat_pre_gt v2 v3` `wf_mat_rcI n (Suc nn) v1` `wf_mat_rcI n (Suc nn) v2` `wf_mat_rcI n (Suc nn) v3`
	have a12: "vec_ge a1 a2" and w12: "mat_ge w1 w2" and w23: "vec_pre_gt a2 a3 \<or> mat_pre_gt w2 w3" 
	  and wf: "wf_mat_rcI n nn w1 \<and> wf_mat_rcI n nn w2 \<and> wf_mat_rcI n nn w3 \<and> wf_vec a1 \<and> wf_vec a2 \<and> wf_vec a3" unfolding mat_geI_def mat_pre_gtI_def wf_mat_rcI_def by (cases "vec_pre_gt a2 a3", auto)
	from w23 show ?thesis
	proof
	  assume gt: "vec_pre_gt a2 a3"
	  have "vec_pre_gt a1 a3" by (rule vec_pre_compat, rule a12, auto simp: gt wf)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `vec_pre_gt a1 a3` show ?thesis by blast
	next
	  assume gt: "mat_pre_gt w2 w3"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (vec_pre_gt a b)"
	    by (rule Suc, rule `mat_ge w1 w2`, auto simp: wf w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed

lemma mat_pre_compat2: assumes "mat_pre_gt v1 v2" and "mat_ge v2 v3" and "wf_mat_rcI n nc v1" and "wf_mat_rcI n nc v2" and "wf_mat_rcI n nc v3" 
  shows "mat_pre_gt v1 v3"
unfolding mat_pre_gtI_def
using assms 
proof (simp add: any_mem_conv mem_iff, induct nc arbitrary: v1 v2 v3, simp add: mat_pre_gtI_def wf_mat_rcI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
      next
	case (Cons a3 w3)
	from `mat_ge v2 v3` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `mat_pre_gt v1 v2` `wf_mat_rcI n (Suc nn) v1` `wf_mat_rcI n (Suc nn) v2` `wf_mat_rcI n (Suc nn) v3`
	have a12: "vec_ge a2 a3" and w12: "mat_ge w2 w3" and w23: "vec_pre_gt a1 a2 \<or> mat_pre_gt w1 w2" 
	  and wf: "wf_mat_rcI n nn w1 \<and> wf_mat_rcI n nn w2 \<and> wf_mat_rcI n nn w3 \<and> wf_vec a1 \<and> wf_vec a2 \<and> wf_vec a3" unfolding mat_geI_def mat_pre_gtI_def wf_mat_rcI_def by (cases "vec_pre_gt a1 a2", auto)
	from w23 show ?thesis
	proof
	  assume gt: "vec_pre_gt a1 a2"
	  have "vec_pre_gt a1 a3" by (rule vec_pre_compat2, rule gt, rule a12, auto simp: gt wf)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `vec_pre_gt a1 a3` show ?thesis by blast
	next
	  assume gt: "mat_pre_gt w1 w2"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (vec_pre_gt a b)"
	    by (rule Suc, rule gt, rule `mat_ge w2 w3`, auto simp: wf w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed


lemma mat_gt_compat: assumes ge: "mat_ge v1 v2" and gt: "mat_gt v2 v3" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" 
  shows "mat_gt v1 v3"
proof -
  from gt have ge1: "mat_ge v2 v3" and gt1: "mat_pre_gt v2 v3" unfolding mat_gtI_def by auto
  show ?thesis unfolding mat_gtI_def
    by (rule conjI, rule mat_ge_trans, rule ge, rule ge1, rule wf1, auto simp: wf2 wf3, rule mat_pre_compat, rule ge, rule gt1, rule wf1, auto simp: wf2 wf3)
qed  

lemma mat_gt_compat2: assumes gt: "mat_gt v1 v2" and ge: "mat_ge v2 v3" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" 
  shows "mat_gt v1 v3"
proof -
  from gt have ge1: "mat_ge v1 v2" and gt1: "mat_pre_gt v1 v2" unfolding mat_gtI_def by auto
  show ?thesis unfolding mat_gtI_def
    by (rule conjI, rule mat_ge_trans, rule ge1, rule ge, rule wf1, auto simp: wf2 wf3, rule mat_pre_compat2, rule gt1, rule ge, rule wf1, auto simp: wf2 wf3)
qed


lemma vec_plus_pre_gt_left_mono: assumes gt: "vec_pre_gt v1 v2" and ge: "vec_ge v3 v4" and wf1: "wf_vec v1" and wf2: "wf_vec v2" and wf3: "wf_vec v3" and wf4: "wf_vec v4"
  and single: "plus_single_mono"
  shows "vec_pre_gt (vec_plus v1 v3) (vec_plus v2 v4)"
proof - 
  from gt ge wf1 wf2 wf3 wf4
  show "vec_pre_gt (vec_plus v1 v3) (vec_plus v2 v4)" unfolding vec_pre_gtI_def wf_vecI_def vec_plusI_def vec_geI_def
  proof (induct n arbitrary: v1 v2 v3 v4, simp)
    case (Suc nn)
    hence "length v1 = Suc nn \<and> length v2 = Suc nn \<and> length v3 = Suc nn \<and> length v4 = Suc nn" by auto
    thus ?case
    proof (cases v1, simp, cases v2, simp, cases v3, simp, cases v4, simp)
      fix a1 w1 a2 w2 a3 w3 a4 w4
      assume v1: "v1 = a1 # w1" and v2: "v2 = a2 # w2" and v3: "v3 = a3 # w3" and v4: "v4 = a4 # w4"
      with Suc have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn \<and> length w4 = nn" by auto
      let ?pl = "\<lambda> u1 u2. map (\<lambda> xy. plus (fst xy) (snd xy)) (zip u1 u2)"
      show "any (\<lambda>(a, b). a > b) (zip (?pl v1 v3) (?pl v2 v4))"
      proof (cases "a1 > a2")
	case True
	from Suc v1 v2 v3 v4 have ge: "a3 \<ge> a4" by auto
	have "a1 + a3 > a2 + a4" by (rule plus_gt_left_mono, auto simp: True ge single)
	with v1 v2 v3 v4 show ?thesis by auto
      next
	case False
	with Suc v1 v2 v3 v4 have tp: "any (\<lambda>(a,b). a > b) (zip w1 w2)" and ta: "forall (\<lambda>(a,b). a \<ge> b) (zip w3 w4)" by auto
	have "any (\<lambda>(a, b). a > b) (zip (?pl w1 w3) (?pl w2 w4))" 
	  by (rule Suc, rule tp, rule ta, auto simp: len)
	with v1 v2 v3 v4 False show ?thesis by auto
      qed
    qed
  qed
qed	
    


lemma mat_plus_pre_gt_left_mono: assumes gt: "mat_pre_gt v1 v2" and ge: "mat_ge v3 v4" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" and wf4: "wf_mat_rcI n nc v4"
  and single: "plus_single_mono"
  shows "mat_pre_gt (mat_plus v1 v3) (mat_plus v2 v4)"
proof -
  from gt ge wf1 wf2 wf3 wf4
  show "mat_pre_gt (mat_plus v1 v3) (mat_plus v2 v4)" unfolding mat_pre_gtI_def wf_mat_rcI_def mat_plusI_def mat_geI_def
  proof (induct nc arbitrary: v1 v2 v3 v4, simp)
    case (Suc nn)
    hence "length v1 = Suc nn \<and> length v2 = Suc nn \<and> length v3 = Suc nn \<and> length v4 = Suc nn" by auto
    thus ?case
    proof (cases v1, simp, cases v2, simp, cases v3, simp, cases v4, simp)
      fix a1 w1 a2 w2 a3 w3 a4 w4
      assume v1: "v1 = a1 # w1" and v2: "v2 = a2 # w2" and v3: "v3 = a3 # w3" and v4: "v4 = a4 # w4"
      with Suc have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn \<and> length w4 = nn" by auto
      from Suc v1 v2 v3 v4 have wf: "wf_vec a1 \<and> wf_vec a2 \<and> wf_vec a3 \<and> wf_vec a4 \<and> 
	forall wf_vec w1 \<and> forall wf_vec w2 \<and> forall wf_vec w3 \<and> forall wf_vec w4 \<and>
	vec_ge a3 a4 \<and> forall (\<lambda> (a,b). vec_ge a b) (zip w3 w4)" by auto
      let ?pl = "\<lambda> u1 u2. map (\<lambda> xy. vec_plus (fst xy) (snd xy)) (zip u1 u2)"
      show "any (\<lambda>(a, b). vec_pre_gt a b) (zip (?pl v1 v3) (?pl v2 v4))"
      proof (cases "vec_pre_gt a1 a2")
	case True
	from Suc v1 v2 v3 v4 have ge: "vec_ge a3 a4" by auto
	have "vec_pre_gt (vec_plus a1 a3) (vec_plus a2 a4)" by (rule vec_plus_pre_gt_left_mono, auto simp: True single wf)
	with v1 v2 v3 v4 show ?thesis by auto
      next
	case False
	with Suc v1 v2 v3 v4 have tp: "any (\<lambda>(a,b). vec_pre_gt a b) (zip w1 w2)" by auto
	have "any (\<lambda>(a, b). vec_pre_gt a b) (zip (?pl w1 w3) (?pl w2 w4))" 
	  by (rule Suc, rule tp, auto simp only: len wf)
	with v1 v2 v3 v4 False show ?thesis by auto
      qed
    qed
  qed
qed	

lemma mat_plus_gt_left_mono: assumes gt: "mat_gt v1 v2" and ge: "mat_ge v3 v4" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" and wf4: "wf_mat_rcI n nc v4"
  and single: "plus_single_mono"
  shows "mat_gt (mat_plus v1 v3) (mat_plus v2 v4)"
proof -
  from gt have ge1: "mat_ge v1 v2" and gt1: "mat_pre_gt v1 v2" unfolding mat_gtI_def by auto
  show ?thesis unfolding mat_gtI_def 
    by (rule conjI, rule mat_ge_trans[where ?v2.0 = "mat_plus v2 v3" and nc = nc and nr = n], rule mat_plus_left_mono[where nc = nc and nr = n], auto simp: ge1 ge wf1 wf2 wf3 wf4 wf_mat_plus, simp add: mat_plus_comm[where ?m1.0 = v2],  rule mat_plus_left_mono[where nc = nc and nr = n], rule ge, auto simp: wf1 wf2 wf3 wf4,
    rule mat_plus_pre_gt_left_mono[where nc = nc], rule gt1, rule ge, auto simp: wf1 wf2 wf3 wf4 single)
qed


abbreviation vec_ns :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool" ("(_ \<succeq>v _)" [51,51] 50) 
 where "v \<succeq>v w \<equiv> (wf_vec v \<and> wf_vec w \<and> vec_ge v w)"

abbreviation vec_s :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool" ("(_ \<succ>v _)" [51,51] 50)
 where "v \<succ>v w \<equiv> (wf_vec v \<and> wf_vec w \<and> vec_ge w vec0 \<and> vec_gt v w)"

lemma vec_sn : assumes psm: plus_single_mono
  shows "SN {(v,w) . v \<succ>v w}"
unfolding SN_ars_def SN_elt_def 
proof clarify
 fix e f 
 assume ass: "\<forall> i. (f i, f (Suc i)) \<in> {(v,w). v \<succ>v w}"
 hence ns_vec: "\<And> i. (f i \<succeq>v f (Suc i))" by auto
 hence len_n: "\<And> i. length (f i) = n" by (auto simp: wf_vecI_def)
 with ns_vec have ns: "\<And> i j. j < n \<Longrightarrow> f i ! j \<ge> f (Suc i) ! j" by (auto simp: vec_geI_def set_zip)
 have vec0: "\<And> v. wf_vec v \<Longrightarrow> vec_ge v vec0 = (\<forall> i < n. v ! i \<ge> 0)" 
 proof -
   fix v
   assume "wf_vec v" hence n: "n = length v" unfolding wf_vecI_def by auto
   show "vec_ge v vec0 = (\<forall> i < n. v ! i \<ge> 0)"
     unfolding vec_geI_def vec_zeroI_def 
     by (simp only: n, simp add: all_set_conv_all_nth)
 qed      
 from ass len_n have "\<And> i. \<exists> j < n. ((wf_vec (f (Suc i)) \<and> vec_ge (f (Suc i)) vec0)) \<and> (f i ! j > f (Suc i) ! j)" unfolding vec_pre_gtI_def by (auto simp: any_mem_conv mem_iff set_zip wf_vecI_def)
 hence s: "\<And> i. \<exists> j < n. ((f (Suc i)) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)" using vec0 by blast
 show False
 proof (cases "\<exists> j < n. \<forall> m . \<exists> i . i \<ge> m \<and> (f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)")
   case False
   with False have "\<forall> j < n. \<exists> m . \<forall> i . i \<ge> m \<longrightarrow> \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))" by blast
   hence "\<exists> m. \<forall> j < n. \<forall> i. i \<ge> m \<longrightarrow> \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))"
   proof (induct n, simp)
     case (Suc nn)
     from Suc obtain m1 where res1: "\<forall> j < nn. \<forall> i \<ge> m1 . \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))" by auto
     from Suc obtain m2 where res2: "\<forall> i \<ge> m2. \<not> ((f (Suc i) ! nn \<ge> 0) \<and> (f i ! nn > f (Suc i) ! nn))" by auto
     let ?m = "max m1 m2"
     have "\<forall> j < Suc nn. \<forall> i \<ge> ?m. \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))" 
     proof (rule allI)
	fix j
	show "j < Suc nn \<longrightarrow> (\<forall> i \<ge> ?m . \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)))"
	  using res1 res2 by (cases "j = nn", auto)
     qed
     thus ?case ..
   qed
   with s show False by blast      
 next
   case True
   from this obtain j where jlen: "j < n" and jinf: "\<forall> m. \<exists> i \<ge> m. (f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)" by auto   
   obtain g where gdef: "g = (\<lambda> i. f i ! j)" by auto
   let ?S = "{(x,y). (y \<ge> 0) \<and> ((x :: 'a) > y)}"
   let ?NS = "{(x,y). (x :: 'a) \<ge> y}"
   from SN have SNstart: "SN ?S (g 0)" unfolding SN_ars_def using arc_pos_plus_single[OF psm] by simp
   from ns jlen gdef have steps: "\<forall> i. (g i, g (Suc i)) \<in> ?NS \<union> ?S" by simp
   from compat have "?S O ?NS \<subseteq> ?S" by auto
   from steps this SNstart have "\<exists> m. \<forall> i \<ge> m. (g i, g (Suc i)) \<in> ?NS - ?S" by (rule non_strict_ending)
   with gdef jinf show False by blast
 qed
qed

abbreviation mat_ns :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" ("(_ \<succeq>_ _)" [51,51,51] 50) 
 where "v \<succeq>nc w \<equiv> (wf_mat_rcI n nc v \<and> wf_mat_rcI n nc w \<and> mat_ge v w)"

abbreviation mat_s :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" ("(_ \<succ>_ _)" [51,51,51] 50)
 where "v \<succ>nc w \<equiv> (wf_mat_rcI n nc v \<and> wf_mat_rcI n nc w \<and> mat_ge w (mat0I 0 n nc) \<and> mat_gt v w)"

lemma mat_sn : assumes psm: plus_single_mono
  shows "SN {(v,w) . v \<succ>nc w}"
unfolding SN_ars_def SN_elt_def 
proof clarify
 fix e f 
 assume ass: "\<forall> i. (f i, f (Suc i)) \<in> {(v,w). v \<succ>nc w}"
 hence ns_vec: "\<And> i. (f i \<succeq>nc f (Suc i))" unfolding mat_gtI_def by auto
 hence len_n: "\<And> i. length (f i) = nc" by (auto simp: wf_mat_rcI_def)
 have vec0: "\<And> v. wf_mat_rcI n nc v \<Longrightarrow> mat_ge v (mat0I 0 n nc) = (\<forall> i < nc. vec_ge (v ! i) vec0)" 
 proof -
   fix v :: "'a mat"
   assume "wf_mat_rcI n nc v" hence n: "nc = length v" unfolding wf_mat_rcI_def by auto
   show "mat_ge v (mat0I 0 n nc) = (\<forall> i < nc. vec_ge (v ! i) vec0)"
     unfolding mat_geI_def mat0I_def 
     by (simp only: n, simp add: all_set_conv_all_nth)
 qed   
 have ns: "\<And> i j. j < nc \<Longrightarrow> wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f i ! j) (f (Suc i) ! j)" 
 proof -
   fix i j
   assume j: "j < nc"
   have len: "length (f i) = nc" using len_n by auto
   have len2: "length (f (Suc i)) = nc" using len_n by auto
   have vec: "(f i \<succeq>nc f (Suc i))" using ns_vec by auto
   with j have "vec_ge (f i ! j) (f (Suc i) ! j)" unfolding mat_geI_def 
     by (simp add: all_set_conv_all_nth len len2)
   with len vec j show "wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f i ! j) (f (Suc i) ! j)" 
     unfolding wf_mat_rcI_def by auto
 qed   
 have s: "\<And> i. \<exists> j < nc. (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))"
 proof -
   fix i
   have len: "length (f i) = nc" using len_n by auto
   have len2: "length (f (Suc i)) = nc" using len_n by auto
   from ass have stri: "f i \<succ>nc  f (Suc i)" by auto
   hence "mat_gt (f i) (f (Suc i))" by auto
   from this obtain j where j: "j < nc" and gt: "vec_gt (f i ! j) (f (Suc i) ! j)" unfolding mat_gtI_def mat_geI_def mat_pre_gtI_def
     by (auto simp: any_mem_conv mem_iff set_zip all_set_conv_all_nth len len2)
   from stri j have ze: "vec_ge (f (Suc i) ! j) vec0" by (auto simp: vec0)
   with gt stri j have "wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j)" 
     unfolding wf_mat_rcI_def by auto
   with j show "\<exists> j < nc. (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by blast
 qed   
 show False
 proof (cases "\<exists> j < nc. \<forall> m . \<exists> i . i \<ge> m \<and> wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j)")
   case False
   with False have "\<forall> j < nc. \<exists> m . \<forall> i . i \<ge> m \<longrightarrow> \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by blast
   hence "\<exists> m. \<forall> j < nc. \<forall> i. i \<ge> m \<longrightarrow> \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))"
   proof (induct nc, simp)
     case (Suc nn)
     from Suc obtain m1 where res1: "\<forall> j < nn. \<forall> i \<ge> m1 . \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by auto
     from Suc obtain m2 where res2: "\<forall> i \<ge> m2. \<not> (wf_vec (f i ! nn) \<and> wf_vec (f (Suc i) ! nn) \<and> vec_ge (f (Suc i) ! nn) vec0 \<and> vec_gt (f i ! nn) (f (Suc i) ! nn))" by auto
     let ?m = "max m1 m2"
     have "\<forall> j < Suc nn. \<forall> i \<ge> ?m. \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))"
     proof (rule allI)
	fix j
	show "j < Suc nn \<longrightarrow> (\<forall> i \<ge> ?m . \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j)))"
	  using res1 res2 by (cases "j = nn", auto)
     qed
     thus ?case by blast
   qed
   with s show False by blast      
 next
   case True
   from this obtain j where jlen: "j < nc" and jinf: "\<forall> m. \<exists> i \<ge> m. (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by auto   
   obtain g where gdef: "g = (\<lambda> i. f i ! j)" by auto
   let ?S = "{(x,y). wf_vec x \<and> wf_vec y \<and> vec_ge y vec0 \<and> vec_gt x y}"
   let ?NS = "{(x,y). wf_vec x \<and> wf_vec y \<and> vec_ge x y}"
   from vec_sn[OF psm] have SNstart: "SN ?S (g 0)" unfolding SN_ars_def by simp
   from ns jlen gdef have steps: "\<forall> i. (g i, g (Suc i)) \<in> ?NS \<union> ?S" by simp
   have "?S O ?NS \<subseteq> ?S" 
   proof (clarify)
     fix a b x y z
     assume x: "wf_vec x" and y: "wf_vec y" and z: "wf_vec z" and xy: "vec_ge x y" and yzg: "vec_pre_gt y z" and yz: "vec_ge y z"
     show "vec_ge x z \<and> vec_pre_gt x z"
       by (rule conjI, rule vec_ge_trans[where nr = n], rule xy, rule yz, auto simp: x y z, rule vec_pre_compat, rule xy, rule yzg, auto simp: x y z)
   qed     
   from steps this SNstart have "\<exists> m. \<forall> i \<ge> m. (g i, g (Suc i)) \<in> ?NS - ?S" by (rule non_strict_ending)
   with gdef jinf show False by blast
 qed
qed


lemma vec1_pre_gt_vec0: assumes "i < n" shows "vec_pre_gt (vec1I zero one n i) vec0"
unfolding vec_pre_gtI_def vec_zeroI_def vec1I_def using assms
proof (auto simp: set_zip any_mem_conv mem_iff)
  let ?one = "\<lambda> j. (replicate i 0 @ 1 # replicate (n - Suc i) (0 :: 'a)) ! j"
  let ?zero = "\<lambda> i. (replicate n (0 :: 'a)) ! i"
  have gt: "?one i > ?zero i" using assms by (simp add: nth_append)
  with assms show "\<exists> a b. (\<exists> j. a = ?one j \<and> b = ?zero j \<and> j < n) \<and> (a > b)" by blast 
qed


lemma mat1_pre_gt_mat0: shows "mat_pre_gt mat1 mat0"
proof -
  show ?thesis unfolding mat_pre_gtI_def mat0I_def mat1I_def
  proof (auto simp: any_mem_conv mem_iff set_zip)
    let ?one = "\<lambda> i. map (vec1I (0 :: 'a) 1 n) [0..<n] ! i"
    let ?zero = "\<lambda> i. (replicate n vec0) ! i"
    have gt: "vec_pre_gt (?one 0) (?zero 0)" using n_pos by (simp add: vec1_pre_gt_vec0)
    with n_pos show "\<exists> a b. (\<exists> j. a = ?one j \<and> b = ?zero j \<and> j < n) \<and> (vec_pre_gt a b)" by blast 
  qed
qed

lemma mat1_gt_mat0: shows "mat_gt mat1 mat0"
unfolding mat_gtI_def using mat1_ge_mat0 mat1_pre_gt_mat0 by auto


lemma mono_poly_order_carrier: assumes mono: "plus_single_mono" shows 
  "poly_order_carrier wf_mat mat0 mat_plus (mat_mul n) mat_ge mat_gt plus_single_mono mat1 mat_max0 mat_arc_pos"
  by (rule poly_order_carrier.intro, rule poly_carrier.intro, 
    rule wf_mat0, 
    rule wf_mat_plus, simp+, 
    rule mat_plus_comm, 
    rule mat_plus_assoc, simp+,
    rule wf_mat_mul, simp+,
    rule mat_mul_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_compat, simp+,
    rule mat_gt_compat2, simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mul_left_mono, simp+,
    rule mat_mul_right_mono, simp+,
    rule mat_ge_refl,
    rule mat_ge_trans, simp+,
    rule mat_plus_gt_left_mono, simp+,
    (simp add: mono)+,
    rule poly_order_carrier_axioms.intro,
    simp only: mat_arc_posI_def, simp add: arc_pos_plus_single[OF mono], rule mat_sn[OF mono],
    rule mat_max0_id_pos, simp+,
    rule mat_max0_pos, simp+,
    rule mat_max0_mono, simp+,
    rule mat_max0_x,
    rule wf_mat_max0, simp,
    rule wf_mat1,
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat1_gt_mat0,
    rule mat_arc_pos_plus, simp+,
    rule mat_arc_pos_mul, simp+,
    rule mat_arc_pos_max0, simp+,
    rule mat_arc_pos_mat1)


lemma mat_gt_arc_compat: assumes ge: "mat_ge m1 m2" and gt: "mat_gt_arc m2 m3" and wf1: "wf_mat_rcI nr nc m1" and wf2: "wf_mat_rcI nr nc m2" and wf3: "wf_mat_rcI nr nc m3" 
  shows "mat_gt_arc m1 m3"
proof (simp only: mat_ge_to_index[OF wf1 wf3], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  from ge i j have one: "m1 ! i ! j \<ge> m2 ! i ! j" using mat_ge_to_index[OF wf1 wf2] by auto
  from gt i j have two: "m2 ! i ! j > m3 ! i ! j" using mat_ge_to_index[OF wf2 wf3] by auto
  from one two show "m1 ! i ! j > m3 ! i ! j" by (rule compat)
qed

lemma mat_gt_arc_compat2: assumes gt: "mat_gt_arc m1 m2" and ge: "mat_ge m2 m3" and wf1: "wf_mat_rcI nr nc m1" and wf2: "wf_mat_rcI nr nc m2" and wf3: "wf_mat_rcI nr nc m3" 
  shows "mat_gt_arc m1 m3"
proof (simp only: mat_ge_to_index[OF wf1 wf3], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  from gt i j have one: "m1 ! i ! j > m2 ! i ! j" using mat_ge_to_index[OF wf1 wf2] by auto
  from ge i j have two: "m2 ! i ! j \<ge> m3 ! i ! j" using mat_ge_to_index[OF wf2 wf3] by auto
  from one two show "m1 ! i ! j > m3 ! i ! j" by (rule compat2)
qed

lemma mat_gt_arc_plus_mono: assumes gt1: "mat_gt_arc x y"
  and gt2: "mat_gt_arc z u"
  and wfx: "wf_mat_rcI nr nc x"
  and wfy: "wf_mat_rcI nr nc y"
  and wfz: "wf_mat_rcI nr nc z"
  and wfu: "wf_mat_rcI nr nc u"
  and psm: "\<not> plus_single_mono"
  shows "mat_gt_arc (mat_plus x z) (mat_plus y u)"
proof -
  let ?xz = "mat_plus x z"
  let ?yu = "mat_plus y u"
  from wf_mat_plus[OF wfx wfz] have wfxz: "wf_mat_rcI nr nc ?xz" .
  from wf_mat_plus[OF wfy wfu] have wfyu: "wf_mat_rcI nr nc ?yu" .
  show ?thesis 
  proof (simp only: mat_ge_to_index[OF wfxz wfyu], intro allI impI)
    fix i j
    assume i: "i < nc" and j: "j < nr"
    show "mat_plus x z ! i ! j > mat_plus y u ! i ! j"
    proof (
	simp only: mat_plus_index[OF wfx wfz i j],
	simp only: mat_plus_index[OF wfy wfu i j],
        rule plus_gt_both_mono[OF _ _ psm])
      show "x ! i ! j > y ! i ! j" using gt1 i j mat_ge_to_index[OF wfx wfy] by auto
    next
      show "z ! i ! j > u ! i ! j" using gt2 i j mat_ge_to_index[OF wfz wfu] by auto
    qed
  qed
qed

lemma scalar_prod_left_mono: assumes wf1: "wf_vecI nr v1"
  and wf2: "wf_vecI nr v2"
  and wf3: "wf_vecI nr v3"
  and gt1: "vec_geI gt v1 v2"
  and psm: "\<not> plus_single_mono"
  shows "scalar_prod v1 v3 > scalar_prod v2 v3"
proof -
  from vec_ge_to_index[OF wf1 wf2] gt1 have g1: "\<forall> i < nr. gt (v1 ! i) (v2 ! i)" by auto
  from g1 wf1 wf2 wf3 show ?thesis unfolding wf_vecI_def
  proof (induct nr arbitrary: v1 v2 v3, simp add: scalar_prodI_def zero_leastI[OF psm])
    case (Suc nrr)
    from Suc obtain a1 w1 where v1: "v1 = a1 # w1" and w1: "length w1 = nrr" by (cases v1, auto)
    from Suc obtain a2 w2 where v2: "v2 = a2 # w2" and w2: "length w2 = nrr" by (cases v2, auto)
    from Suc obtain a3 w3 where v3: "v3 = a3 # w3" and w3: "length w3 = nrr" by (cases v3, auto)
    from Suc v1 v2 have a12: "a1 > a2" and w12: "\<forall> i < nrr. w1 ! i > w2 ! i" by auto
    have rec: "scalar_prod w1 w3 > scalar_prod w2 w3" 
      by (rule Suc, auto simp: w1 w2 w3 w12)
    have a: "a1 * a3 > a2 * a3" by (rule times_gt_left_mono[OF a12 psm])
    show ?case 
      by (simp add: v1 v2 v3 scalar_product_cons, rule plus_gt_both_mono[OF a rec psm])
  qed
qed

lemma mat_gt_arc_mul_mono: assumes gt1: "mat_gt_arc x y"
  and wfx: "wf_mat_rcI nr nc x"
  and wfy: "wf_mat_rcI nr nc y"
  and wfz: "wf_mat_rcI nc ncc z"
  and psm: "\<not> plus_single_mono"
  shows "mat_gt_arc (mat_mul nr x z) (mat_mul nr y z)"
proof -
  let ?xz = "mat_mul nr x z"
  let ?yz = "mat_mul nr y z"
  from wf_mat_mul[OF wfx wfz] have wfxz: "wf_mat_rcI nr ncc ?xz" .
  from wf_mat_mul[OF wfy wfz] have wfyz: "wf_mat_rcI nr ncc ?yz" .
  show ?thesis 
  proof (simp only: mat_ge_to_index[OF wfxz wfyz], intro allI impI)
    fix i j
    assume i: "i < ncc" and j: "j < nr"
    have wfxj : "wf_vecI nc (row x j)" using wf_row[OF wfx j] .
    have wfyj : "wf_vecI nc (row y j)" using wf_row[OF wfy j] .
    have wfzi : "wf_vecI nc (col z i)" using wf_col[OF wfz i] .
    show "?xz ! i ! j > ?yz ! i ! j"
    proof (
        simp only: mat_mul_to_scalar[OF wfx wfz j i],
	simp only: mat_mul_to_scalar[OF wfy wfz j i],
	rule scalar_prod_left_mono[OF wfxj wfyj wfzi _  psm],
	simp only: vec_ge_to_index[OF wfxj wfyj],
	intro allI impI
      )
      fix k
      assume k: "k < nc"
      from gt1 mat_ge_to_index[OF wfx wfy, of gt] j k
      show "row x j ! k > row y j ! k"
	by (
	  simp only: row_col[OF wfx j k],
	  simp only: row_col[OF wfy j k],
	  unfold col_def, auto)
    qed
  qed
qed

lemma mat0_leastI: assumes wf: "wf_mat_rcI nr nc m"
  and psm: "\<not> plus_single_mono"
  shows "mat_gt_arc m (mat0I zero nr nc)"
proof (simp only: mat_ge_to_index[OF wf wf_mat0], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "m ! i ! j > mat0I 0 nr nc ! i ! j"
    by (simp only: mat0_index[OF i j] zero_leastI[OF psm])
qed

lemma mat0_leastII: 
  assumes gt: "mat_gt_arc (mat0I zero nr nc) m"
  and wf: "wf_mat_rcI nr nc m"
  and psm: "\<not> plus_single_mono"
  shows "m = mat0I zero nr nc"
proof (simp only: index_eq[OF wf wf_mat0, symmetric], intro allI impI)
  fix i j
  assume i: "i < nc" and j: "j < nr"
  show "m ! i ! j = mat0I 0 nr nc ! i ! j"
  proof (simp only: mat0_index[OF i j], rule zero_leastII[OF _ psm])
    show "0 > m ! i ! j" using i j gt mat_ge_to_index[OF wf_mat0 wf, of gt zero] mat0_index[OF i j, of zero] by force
  qed
qed

lemma mat_gt_arc_SN: 
  "SN {(x, y). wf_mat x \<and> wf_mat y \<and> mat_ge y mat0 \<and> mat_arc_pos y \<and> mat_gt_arc x y}" (is "SN ?rel")
proof (rule ccontr)
  assume "\<not> SN ?rel"
  from this obtain f and x where "f (0 :: nat) = x" and steps: "\<forall> i. (f i, f (Suc i)) \<in> ?rel" unfolding SN_ars_def SN_elt_def by blast
  hence pos: "\<forall> i. arc_pos (f (Suc i) ! 0 ! 0)" unfolding mat_arc_posI_def by blast
  have gt: "\<forall> i. geq (f (Suc i) ! 0 ! 0) zero \<and> (f i ! 0 ! 0 > f (Suc i) ! 0 ! 0)"
  proof
    fix i
    from steps have wf1: "wf_mat (f i)" and wf2: "wf_mat (f (Suc i))" and gt: "mat_gt_arc (f i) (f (Suc i))" and ge: "mat_ge (f (Suc i)) mat0" by auto
    show "(geq (f (Suc i) ! 0 ! 0) zero) \<and> (f i ! 0 ! 0 > f (Suc i) ! 0 ! 0)" using mat_ge_to_index[OF wf1 wf2, of gt] mat_ge_to_index[OF wf2 wf_mat0, of geq] mat0_index[OF n_pos n_pos, of zero] gt ge n_pos by force
  qed
  from pos gt SN show False unfolding SN_ars_def SN_elt_def by force
qed

lemma mat1_gt_arc_mat0: assumes psm: "\<not> plus_single_mono" shows "mat_gt_arc mat1 mat0"
proof (simp only: mat_ge_to_index[OF wf_mat1 wf_mat0], intro allI impI)
  fix i j
  assume i: "i < n" and j: "j < n"
  show "mat1 ! i ! j > mat0 ! i ! j"
    by (simp only: mat0_index[OF i j],
      rule zero_leastI[OF psm])
qed

lemma non_mono_poly_order_carrier: assumes mono: "\<not> plus_single_mono" shows 
  "poly_order_carrier wf_mat mat0 mat_plus (mat_mul n) mat_ge mat_gt_arc plus_single_mono mat1 mat_max0 mat_arc_pos"
by (rule poly_order_carrier.intro, rule poly_carrier.intro, 
    rule wf_mat0, 
    rule wf_mat_plus, simp+, 
    rule mat_plus_comm, 
    rule mat_plus_assoc, simp+,
    rule wf_mat_mul, simp+,
    rule mat_mul_assoc_n[symmetric], simp+,
    rule mat0_mult_left, simp, 
    rule mat0_mult_right, simp,
    rule mat_mult_plus_distrib_right, simp+,
    rule mat_mult_plus_distrib_left, simp+,
    rule mat_gt_arc_compat, simp+,
    rule mat_gt_arc_compat2, simp+,
    rule mat_plus_left_mono, simp+,
    rule mat_mul_left_mono, simp+,
    rule mat_mul_right_mono, simp+,
    rule mat_ge_refl,
    rule mat_ge_trans, simp+,
    simp add: mono,
    rule mat_gt_arc_plus_mono, simp+,
    rule mat_gt_arc_mul_mono, simp+,
    rule mat0_leastI, simp+,
    rule mat0_leastII, simp+,
    rule poly_order_carrier_axioms.intro,
    rule mat_gt_arc_SN,
    rule mat_max0_id_pos, simp+,
    rule mat_max0_pos, simp+,
    rule mat_max0_mono, simp+,
    rule mat_max0_x,
    rule wf_mat_max0, simp, 
    rule wf_mat1,
    rule mat1_mult_left, simp,
    rule mat1_mult_right, simp,
    rule mat1_ge_mat0,
    rule mat1_gt_arc_mat0[OF mono],
    rule mat_arc_pos_plus, simp+,
    rule mat_arc_pos_mul, simp+,
    rule mat_arc_pos_max0, simp+,
    rule mat_arc_pos_mat1)
end 

end
