{-# OPTIONS_GHC -fglasgow-exts #-}

module Loop where {

import qualified ErrorMonad;
import qualified List;
import qualified Term;
import qualified Nat;
import qualified Product_Type;
import qualified Position;
import qualified Show;
import qualified Eithera;
import qualified Option;

rewrites ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => (Term.Term a b, Term.Term a b) ->
                         Term.Ctxt a b ->
                           Term.Sub a b ->
                             (Term.Term a b, Term.Term a b) ->
                               [(Term.Term a b, Term.Term a b)] ->
                                 Eithera.Eithera [Char] ();
rewrites (s, t) c sigma (l, r) ra =
  ErrorMonad.checkB
    (s == Term.ctxt_apply c (Term.sub_apply l sigma) &&
      (t == Term.ctxt_apply c (Term.sub_apply r sigma) &&
        List.member (l, r) ra))
    (List.append
      (List.append (List.append ['T', 'e', 'r', 'm', ' '] (Term.sshow_term s))
        [' ', 'd', 'o', 'e', 's', ' ', 'n', 'o', 't', ' ', 'r', 'e', 'w', 'r',
          'i', 't', 'e', ' ', 't', 'o', ' ', 't', 'e', 'r', 'm', ' '])
      (Term.sshow_term t));

subst_safe ::
  forall a b.
    (Eq a,
      Eq b) => [a] ->
                 [(Term.Term b a, Term.Term b a)] ->
                   Term.Sub b a -> Maybe (Term.Sub b a);
subst_safe vs [] sigma = Just sigma;
subst_safe vs ((Term.Var x, t) : ts) sigma =
  (if List.member (Term.Var x, t) ts then Loop.subst_safe vs ts sigma
    else (if List.member x vs && not (Term.sub_apply (Term.Var x) sigma == t)
           then Nothing
           else (case Loop.subst_safe (x : vs) ts sigma of {
                  Nothing -> Nothing;
                  Just a ->
                    let {
                      (Term.Sub s) = a;
                    } in Just (Term.Sub (\ y -> (if x == y then t else s y)));
                })));

decomp :: forall a b. [a] -> [b] -> Maybe [(a, b)];
decomp [] [] = Just [];
decomp (uu : uv) [] = Nothing;
decomp [] (uw : ux) = Nothing;
decomp (s : ss) (t : ts) =
  (case Loop.decomp ss ts of {
    Nothing -> Nothing;
    Just pairs -> Just ((s, t) : pairs);
  });

decompose_one ::
  forall a b.
    (Eq a) => Term.Term a b ->
                Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose_one (Term.Var x) t = Just [(Term.Var x, t)];
decompose_one (Term.Fun v va) (Term.Var y) = Nothing;
decompose_one (Term.Fun f ts) (Term.Fun g ss) =
  (if f == g then Loop.decomp ts ss else Nothing);

decompose_n ::
  forall a b.
    (Eq a) => Nat.Nat ->
                Term.Term a b ->
                  Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose_n Nat.Zero_nat s t = Loop.decompose_one s t;
decompose_n (Nat.Suc n) s t =
  (case Loop.decompose_n n s t of {
    Nothing -> Nothing;
    Just a ->
      List.foldla
        (\ aa (b @ (sa, ta)) ->
          (case aa of {
            Nothing -> Nothing;
            Just ys ->
              (case Loop.decompose_one sa ta of {
                Nothing -> Nothing;
                Just xs -> Just (List.append xs ys);
              });
          }))
        (Just []) a;
  });

decompose ::
  forall a b.
    (Eq a) => Term.Term a b ->
                Term.Term a b -> Maybe [(Term.Term a b, Term.Term a b)];
decompose s t = Loop.decompose_n (Term.sz s) s t;

decompose_list ::
  forall a b.
    (Eq a) => [(Term.Term a b, Term.Term a b)] ->
                Maybe [(Term.Term a b, Term.Term a b)];
decompose_list ts =
  List.foldla
    (\ a (b @ (s, t)) ->
      (case a of {
        Nothing -> Nothing;
        Just ls ->
          (case Loop.decompose s t of {
            Nothing -> Nothing;
            Just ys -> Just (List.append ls ys);
          });
      }))
    (Just []) ts;

match_all ::
  forall a b.
    (Eq a, Eq b) => [(Term.Term a b, Term.Term a b)] -> Maybe (Term.Sub a b);
match_all ts =
  (case Loop.decompose_list ts of {
    Nothing -> Nothing;
    Just ls -> Loop.subst_safe [] ls (Term.Sub Term.Var);
  });

getpr ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Term.Term a b ->
                         Term.Term a b ->
                           Position.Pos ->
                             (Term.Term a b, Term.Term a b) ->
                               Maybe (Term.Term a b,
                                       (Term.Ctxt a b,
 (Term.Sub a b, (Term.Term a b, Term.Term a b))));
getpr s t p (l, r) =
  (case Loop.match_all [(l, Term.subt_at s p), (r, Term.subt_at t p)] of {
    Nothing -> Nothing;
    Just sigma ->
      (if ErrorMonad.isValue
            (Loop.rewrites (s, t) (Term.ctxt_of_pos_term p s) sigma (l, r)
              [(l, r)])
        then Just (s, (Term.ctxt_of_pos_term p s, (sigma, (l, r))))
        else Nothing);
  });

getp ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Term.Term a b ->
                         Term.Term a b ->
                           Position.Pos ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Maybe (Term.Term a b,
                                       (Term.Ctxt a b,
 (Term.Sub a b, (Term.Term a b, Term.Term a b))));
getp s t p [] = Nothing;
getp s t p (r : rs) =
  (case Loop.getpr s t p r of {
    Nothing -> Loop.getp s t p rs;
    Just a -> Just a;
  });

find_get_p ::
  forall f v.
    (Eq f, Show.Sshow f, Eq v,
      Show.Sshow v) => Term.Term f v ->
                         Term.Term f v ->
                           [Position.Pos] ->
                             [(Term.Term f v, Term.Term f v)] ->
                               Maybe (Term.Term f v,
                                       (Term.Ctxt f v,
 (Term.Sub f v, (Term.Term f v, Term.Term f v))));
find_get_p s t [] r = Nothing;
find_get_p s t (p : ps) r =
  (case Loop.getp s t p r of {
    Nothing -> Loop.find_get_p s t ps r;
    Just a -> Just a;
  });

get ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => Term.Term a b ->
                         Term.Term a b ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Eithera.Eithera [Char]
                               (Term.Term a b,
                                 (Term.Ctxt a b,
                                   (Term.Sub a b,
                                     (Term.Term a b, Term.Term a b))));
get s t r =
  (case Loop.find_get_p s t (Term.poss_list s) r of {
    Nothing ->
      ErrorMonad.errora
        (List.append
          (List.append
            (List.append
              (List.append
                ['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'f', 'i',
                  'n', 'd', ' ', 'r', 'e', 'd', 'u', 'c', 't', 'i', 'o', 'n',
                  ' ', 'b', 'e', 't', 'w', 'e', 'e', 'n', ' ', 't', 'e', 'r',
                  'm', ' ']
                (Term.sshow_term s))
              [' ', 'a', 'n', 'd', ' ', 't', 'e', 'r', 'm', ' '])
            (Term.sshow_term t))
          ['.']);
    Just a -> ErrorMonad.returna a;
  });

get_list ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b)] ->
                         Term.Term a b ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Eithera.Eithera [Char]
                               [(Term.Term a b,
                                  (Term.Ctxt a b,
                                    (Term.Sub a b,
                                      (Term.Term a b, Term.Term a b))))];
get_list [] u r =
  ErrorMonad.errora
    ['L', 'o', 'o', 'p', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n', 'o',
      'n', '-', 'e', 'm', 'p', 't', 'y'];
get_list [s] u r =
  ErrorMonad.bind (Loop.get s u r) (\ y -> ErrorMonad.returna [y]);
get_list (s : t : ts) u r =
  ErrorMonad.bind (Loop.get s t r)
    (\ d ->
      ErrorMonad.bind (Loop.get_list (t : ts) u r)
        (\ y -> ErrorMonad.returna (d : y)));

get_loop ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b)] ->
                         Term.Ctxt a b ->
                           Term.Sub a b ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Eithera.Eithera [Char]
                                 [(Term.Term a b,
                                    (Term.Ctxt a b,
                                      (Term.Sub a b,
(Term.Term a b, Term.Term a b))))];
get_loop [] uu uv uw =
  ErrorMonad.errora
    ['L', 'o', 'o', 'p', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'n', 'o',
      'n', '-', 'e', 'm', 'p', 't', 'y'];
get_loop (t : ts) d tau r =
  Loop.get_list (t : ts) (Term.ctxt_apply d (Term.sub_apply t tau)) r;

rewrites_to ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b,
                          (Term.Ctxt a b,
                            (Term.Sub a b, (Term.Term a b, Term.Term a b))))]
                         -> Term.Term a b ->
                              [(Term.Term a b, Term.Term a b)] ->
                                Eithera.Eithera [Char] ();
rewrites_to [(s, (c, (sigma, rule)))] t r = Loop.rewrites (s, t) c sigma rule r;
rewrites_to ((s, (c, (sigma, rule))) : (t, (d, (tau, rule'))) : xs) u r =
  ErrorMonad.bind (Loop.rewrites (s, t) c sigma rule r)
    (\ x -> Loop.rewrites_to ((t, (d, (tau, rule'))) : xs) u r);

check_loop_d ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b,
                          (Term.Ctxt a b,
                            (Term.Sub a b, (Term.Term a b, Term.Term a b))))]
                         -> Term.Ctxt a b ->
                              Term.Sub a b ->
                                [(Term.Term a b, Term.Term a b)] ->
                                  Eithera.Eithera [Char] ();
check_loop_d [] uu uv uw =
  ErrorMonad.errora
    ['T', 'h', 'e', 'r', 'e', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'a',
      't', ' ', 'l', 'e', 'a', 's', 't', ' ', 'o', 'n', 'e', ' ', 'r', 'e', 'w',
      'r', 'i', 't', 'e', ' ', 's', 't', 'e', 'p'];
check_loop_d (v : va) d tau r =
  Loop.rewrites_to (v : va)
    (Term.ctxt_apply d (Term.sub_apply (fst (List.hd (v : va))) tau)) r;

check_loop ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => ([(Term.Term a b)], (Term.Ctxt a b, Term.Sub a b)) ->
                         [(Term.Term a b, Term.Term a b)] ->
                           Eithera.Eithera [Char] ();
check_loop (ts, (c, sigma)) r =
  ErrorMonad.bind (Loop.get_loop ts c sigma r)
    (\ ls -> Loop.check_loop_d ls c sigma r);

checkLoop ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b, Term.Term a b)] ->
                         ([(Term.Term a b)],
                           (Term.Ctxt a b, [(Term.Term a b, Term.Term a b)]))
                           -> Eithera.Eithera [[Char]] ();
checkLoop r (ts, (c, sub)) =
  ErrorMonad.catcha
    (case Loop.subst_safe [] sub (Term.Sub Term.Var) of {
      Nothing ->
        ErrorMonad.errora
          ['s', 'u', 'b', 's', 't', 'i', 't', 'u', 't', 'i', 'o', 'n', ' ', 'n',
            'o', 't', ' ', 'u', 'n', 'i', 'q', 'u', 'e'];
      Just sigma -> Loop.check_loop (ts, (c, sigma)) r;
    })
    (\ x ->
      ErrorMonad.errora
        [['e', 'r', 'r', 'o', 'r', ' ', 'i', 'n', ' ', 'c', 'e', 'r', 't', 'i',
           'f', 'i', 'c', 'a', 't', 'i', 'o', 'n', ' ', 'o', 'f', ' ', 'l', 'o',
           'o', 'p'],
          x]);

}
