{-# OPTIONS_GHC -fglasgow-exts #-}

module SCT where {

import qualified Util;
import qualified Product_Type;
import qualified List;
import qualified DpTransImpl;
import qualified Term;
import qualified UsableRules;
import qualified PolyOrder;
import qualified RedPairImpl;
import qualified Show;
import qualified Orderings;
import qualified Integer;
import qualified Nat;
import qualified Matrix;
import qualified Eithera;
import qualified HOL;
import qualified Edg;
import qualified ErrorMonad;
import qualified Option;
import qualified Trs;

data Scg a b = Null | Scg a a [(b, b)] [(b, b)];

comp :: forall ap. (Eq ap) => [(ap, ap)] -> [(ap, ap)] -> [(ap, ap)];
comp es es' =
  List.remdups
    (Util.flat_map
      (\ (a @ (x, y)) ->
        Util.flat_map (\ (aa @ (y', z)) -> (if y == y' then [(x, z)] else []))
          es')
      es);

eq_scg ::
  forall pp ap. (Eq pp, Eq ap) => SCT.Scg pp ap -> SCT.Scg pp ap -> Bool;
eq_scg (SCT.Scg a b c d) SCT.Null = False;
eq_scg SCT.Null (SCT.Scg a b c d) = False;
eq_scg (SCT.Scg pp1 pp2 list1 list2) (SCT.Scg pp1' pp2' list1' list2') =
  pp1 == pp1' && (pp2 == pp2' && (list1 == list1' && list2 == list2'));
eq_scg SCT.Null SCT.Null = True;

instance (Eq a, Eq b) => Eq (SCT.Scg a b) where {
  a == b = SCT.eq_scg a b;
};

combine :: forall a b. (Eq b) => SCT.Scg a b -> SCT.Scg a b -> SCT.Scg a b;
combine (SCT.Scg p q str wk) (SCT.Scg p' q' str' wk') =
  SCT.Scg p q (List.remdups (List.append str str'))
    (List.remdups (List.append wk wk'));
combine SCT.Null s = SCT.Null;
combine (SCT.Scg v va vb vc) SCT.Null = SCT.Null;

subsumes :: forall a b. (Eq a, Eq b) => SCT.Scg a b -> SCT.Scg a b -> Bool;
subsumes (SCT.Scg p q str wk) (SCT.Scg p' q' str' wk') =
  p == p' &&
    (q == q' &&
      (List.list_all (\ x -> List.member x str') str &&
        List.list_all (\ x -> List.member x (List.append str' wk')) wk));
subsumes g SCT.Null = True;
subsumes SCT.Null (SCT.Scg v va vb vc) = False;

subtract_list :: forall a. (Eq a) => [a] -> [a] -> [a];
subtract_list xs ys = List.filtera (\ x -> not (List.member x ys)) xs;

scg_comp ::
  forall a b.
    (Eq b) => (a -> a -> Bool) -> SCT.Scg a b -> SCT.Scg a b -> SCT.Scg a b;
scg_comp conn (SCT.Scg p q str wk) (SCT.Scg p' q' str' wk') =
  (if not (conn q p') then SCT.Null
    else let {
           strs =
             List.remdups
               (List.append (SCT.comp str str')
                 (List.append (SCT.comp str wk') (SCT.comp wk str')));
           a = SCT.subtract_list (List.remdups (SCT.comp wk wk')) strs;
         } in SCT.Scg p q' strs a);
scg_comp conn SCT.Null g = SCT.Null;
scg_comp conn (SCT.Scg v va vb vc) SCT.Null = SCT.Null;

in_situ :: forall pp ap. (Eq ap) => SCT.Scg pp ap -> Bool;
in_situ SCT.Null = True;
in_situ (SCT.Scg p q str wk) = List.list_ex (\ (a @ (aa, b)) -> aa == b) str;

sagiv :: forall a b. (Eq a, Eq b) => (a -> a -> Bool) -> SCT.Scg a b -> Bool;
sagiv conn g =
  (if SCT.in_situ g then True
    else let {
           gg = SCT.scg_comp conn g g;
         } in (if SCT.subsumes gg g then False
                else SCT.sagiv conn (SCT.combine g gg)));

insert_scg ::
  forall pp ap.
    (Eq pp,
      Eq ap) => SCT.Scg pp ap -> [(SCT.Scg pp ap)] -> Maybe [(SCT.Scg pp ap)];
insert_scg g gs =
  (if g == SCT.Null then Nothing
    else (if List.list_ex (\ g' -> SCT.subsumes g' g) gs then Nothing
           else Just (g : List.filtera (\ g' -> not (SCT.subsumes g g')) gs)));

mk_trancl ::
  forall a b.
    (Eq a,
      Eq b) => (a -> a -> Bool) ->
                 [(SCT.Scg a b)] -> [(SCT.Scg a b)] -> [(SCT.Scg a b)];
mk_trancl conn [] s = s;
mk_trancl conn (g : gs) s =
  (case SCT.insert_scg g s of {
    Nothing -> SCT.mk_trancl conn gs s;
    Just s' ->
      SCT.mk_trancl conn
        (List.append (List.mapa (SCT.scg_comp conn g) s')
          (List.append (List.mapa (\ f -> SCT.scg_comp conn f g) s') gs))
        s';
  });

check_SCT ::
  forall pp ap.
    (Eq pp, Eq ap) => (pp -> pp -> Bool) -> [(SCT.Scg pp ap)] -> Bool;
check_SCT conn gs =
  let {
    a = SCT.mk_trancl conn gs [];
  } in List.list_all (SCT.sagiv conn) a;

check_supt ::
  forall f v.
    (Eq f, Show.Sshow f, Eq v,
      Show.Sshow v) => (Term.Term f v, Term.Term f v) ->
                         Eithera.Eithera [[Char]] ();
check_supt lr =
  ErrorMonad.catcha
    (ErrorMonad.checkB (DpTransImpl.checkSupterm (fst lr) (snd lr))
      [['t', 'h', 'e', ' ', 'r', 'h', 's', ' ', 'i', 's', ' ', 'n', 'o', 't',
         ' ', 'a', ' ', 'p', 'r', 'o', 'p', 'e', 'r', ' ', 's', 'u', 'b', 't',
         'e', 'r', 'm', ' ', 'o', 'f', ' ', 't', 'h', 'e', ' ', 'l', 'h', 's']])
    (\ x ->
      ErrorMonad.errora
        (List.concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' '],
             Term.sshow_term (fst lr), [' ', '-', '>', ' '],
             Term.sshow_term (snd lr)]
          : x));

check_supteq ::
  forall f v.
    (Eq f, Show.Sshow f, Eq v,
      Show.Sshow v) => (Term.Term f v, Term.Term f v) ->
                         Eithera.Eithera [[Char]] ();
check_supteq lr =
  ErrorMonad.catcha
    (ErrorMonad.checkB
      (fst lr == snd lr || DpTransImpl.checkSupterm (fst lr) (snd lr))
      [['t', 'h', 'e', ' ', 'r', 'h', 's', ' ', 'i', 's', ' ', 'n', 'o', 't',
         ' ', 'a', ' ', 's', 'u', 'b', 't', 'e', 'r', 'm', ' ', 'o', 'f', ' ',
         't', 'h', 'e', ' ', 'l', 'h', 's']])
    (\ x ->
      ErrorMonad.errora
        (List.concata
           [['C', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'o', 'r', 'i',
              'e', 'n', 't', ' ', 'r', 'u', 'l', 'e', ' '],
             Term.sshow_term (fst lr), [' ', '-', '>', ' '],
             Term.sshow_term (snd lr)]
          : x));

check_sct_ur_af_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => ((Term.Term a b, Term.Term a b) ->
                         Eithera.Eithera [[Char]] ())
                         -> ((Term.Term a b, Term.Term a b) ->
                              Eithera.Eithera [[Char]] ())
                              -> [[Char]] ->
                                   [(Term.Term a b, Term.Term a b)] ->
                                     [((Term.Term a b, Term.Term a b),
([(Nat.Nat, Nat.Nat)], [(Nat.Nat, Nat.Nat)]))]
                                       -> [(Term.Term a b, Term.Term a b)] ->
    Maybe (a -> [(Term.Term a b, Term.Term a b)]) ->
      Maybe (a -> [(Term.Term a b, Term.Term a b)]) ->
        [(Term.Term a b, Term.Term a b)] ->
          (a -> Nat.Nat -> Nat.Nat -> Bool) -> Eithera.Eithera [[Char]] ();
check_sct_ur_af_proc checkSS checkNSS representation p gs r rm rrm ur pia =
  let {
    edgeCheck = Edg.checkEdgRM rm rrm p True;
  } in ErrorMonad.catcha
         (case rm of {
           Nothing ->
             ErrorMonad.errora
               [['T', 'R', 'S', ' ', 'h', 'a', 's', ' ', 'v', 'a', 'r', 'i',
                  'a', 'b', 'l', 'e', 's', ' ', 'a', 's', ' ', 'l', 'e', 'f',
                  't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', 's']];
           Just rma ->
             ErrorMonad.bind
               (ErrorMonad.bind
                 (ErrorMonad.bind
                   (ErrorMonad.bind
                     (ErrorMonad.catcha
                       (ErrorMonad.catcha
                         (ErrorMonad.checkAllMI
                           (\ lr ->
                             UsableRules.checkUrClosedTRM_af rma ur pia
                               (snd lr))
                           ur)
                         (\ x -> ErrorMonad.errora (snd x)))
                       (\ x ->
                         ErrorMonad.errora
                           [['e', 'r', 'r', 'o', 'r', ' ', 'w', 'h', 'e', 'n',
                              ' ', 'c', 'h', 'e', 'c', 'k', 'i', 'n', 'g', ' ',
                              'c', 'l', 'o', 's', 'u', 'r', 'e', ' ', 'p', 'r',
                              'o', 'p', 'e', 'r', 't', 'i', 'e', 's', ' ', 'o',
                              'f', ' ', 'r', 'h', 's', ' ', 'o', 'f', ' ', 'u',
                              's', 'a', 'b', 'l', 'e', ' ', 'r', 'u', 'l', 'e',
                              's'],
                             x]))
                     (\ x ->
                       ErrorMonad.catcha
                         (ErrorMonad.checkSubset p (List.mapa fst gs))
                         (\ xa ->
                           ErrorMonad.errora
                             [List.append
                                ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ',
                                  'n', 'o', ' ', 's', 'i', 'z', 'e', '-', 'c',
                                  'h', 'a', 'n', 'g', 'e', ' ', 'g', 'r', 'a',
                                  'p', 'h', ' ', 'f', 'o', 'r', ' ', 'D', 'P',
                                  ' ']
                                (Trs.showRule xa)])))
                   (\ x ->
                     ErrorMonad.catcha
                       (ErrorMonad.checkAllMI
                         (\ (a @ (aa, b)) ->
                           let {
                             (s, t) = aa;
                           } in (\ (ab @ (stri, nstri)) ->
                                  ErrorMonad.catcha
                                    (ErrorMonad.bind
                                      (ErrorMonad.bind
(ErrorMonad.bind
  (ErrorMonad.checkB
    (not (Term.is_var s) &&
      (not (Term.is_var t) && List.nulla (rma (Option.the (Term.root t)))))
    [['r', 'o', 'o', 't', 's', ' ', 'o', 'f', ' ', 'D', 'P', ' ', 'm', 'u', 's',
       't', ' ', 'b', 'e', ' ', 'c', 'o', 'n', 's', 't', 'r', 'u', 'c', 't',
       'o', 'r', 's']])
  (\ xa ->
    ErrorMonad.catcha
      (ErrorMonad.checkAllMI
        (\ i ->
          ErrorMonad.checkB (Nat.less_nat i (List.size_list (Term.args s)))
            [List.append
               ['l', 'e', 'f', 't', '-', 'i', 'n', 'd', 'e', 'x', ' ', 't', 'o',
                 ' ', 'l', 'a', 'r', 'g', 'e']
               (Show.sshow_nat i)])
        (List.remdups (List.mapa fst (List.append stri nstri))))
      (\ xb -> ErrorMonad.errora (snd xb))))
(\ xa ->
  ErrorMonad.catcha
    (ErrorMonad.checkAllMI
      (\ j ->
        ErrorMonad.checkB
          (Nat.less_nat j (List.size_list (Term.args t)) &&
            ErrorMonad.isValue
              (UsableRules.checkUrClosedTRM_af rma ur pia
                (List.nth (Term.args t) j)))
          [List.append
             ['r', 'i', 'g', 'h', 't', '-', 'i', 'n', 'd', 'e', 'x', ' ', 't',
               'o', ' ', 'l', 'a', 'r', 'g', 'e', ' ', 'o', 'r', ' ', 'a', 'r',
               'g', 'u', 'm', 'e', 'n', 't', ' ', 'v', 'i', 'o', 'l', 'a', 't',
               'e', 's', ' ', 'u', 's', 'a', 'b', 'l', 'e', '-', 'r', 'u', 'l',
               'e', 's', ' ', 'c', 'o', 'n', 'd', 'i', 't', 'i', 'o', 'n']
             (Show.sshow_nat j)])
      (List.remdups (List.mapa snd (List.append stri nstri))))
    (\ xb -> ErrorMonad.errora (snd xb))))
                                      (\ xa ->
let {
  ss = Term.args s;
  ts = Term.args t;
} in ErrorMonad.bind
       (ErrorMonad.catcha
         (ErrorMonad.checkAllMI
           (\ (ac @ (i, j)) ->
             ErrorMonad.checkB
               (ErrorMonad.isValue (checkSS (List.nth ss i, List.nth ts j)))
               [List.append
                  (List.append
                    (List.append
                      ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't',
                        'h', ' ', 'e', 'd', 'g', 'e', ' ']
                      (Show.sshow_nat i))
                    [' ', '-', 'S', '-', '>', ' '])
                  (Show.sshow_nat j)])
           stri)
         (\ xb -> ErrorMonad.errora (snd xb)))
       (\ xb ->
         ErrorMonad.catcha
           (ErrorMonad.checkAllMI
             (\ (ac @ (i, j)) ->
               ErrorMonad.checkB
                 (ErrorMonad.isValue (checkNSS (List.nth ss i, List.nth ts j)))
                 [List.append
                    (List.append
                      (List.append
                        ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't',
                          'h', ' ', 'e', 'd', 'g', 'e', ' ']
                        (Show.sshow_nat i))
                      [' ', '-', 'N', 'S', '-', '>', ' '])
                    (Show.sshow_nat j)])
             nstri)
           (\ xc -> ErrorMonad.errora (snd xc)))))
                                    (\ xa ->
                                      ErrorMonad.errora
(List.append
   ['p', 'r', 'o', 'b', 'l', 'e', 'm', 's', ' ', 'w', 'i', 't', 'h', ' ', 'D',
     'P', ' ']
   (Trs.showRule (s, t))
  : xa)))
                             b)
                         gs)
                       (\ xa -> ErrorMonad.errora (snd xa))))
                 (\ x ->
                   ErrorMonad.catcha
                     (ErrorMonad.catcha (ErrorMonad.checkAllMI checkNSS ur)
                       (\ xa -> ErrorMonad.errora (snd xa)))
                     (\ xa ->
                       ErrorMonad.errora
                         (['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h',
                            'e', 'n', ' ', 'o', 'r', 'i', 'e', 'n', 't', 'i',
                            'n', 'g', ' ', 'u', 's', 'a', 'b', 'l', 'e', ' ',
                            'r', 'u', 'l', 'e', 's']
                           : xa))))
               (\ x ->
                 ErrorMonad.checkB
                   (SCT.check_SCT (\ st uv -> edgeCheck (st, uv))
                     (List.mapa
                       (\ (a @ (st, aa)) ->
                         let {
                           (ab, b) = aa;
                         } in SCT.Scg st st ab b)
                       gs))
                   [['s', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ',
                      'a', 'n', 'a', 'l', 'y', 's', 'i', 's', ' ', 'f', 'a',
                      'i', 'l', 'e', 'd']]);
         })
         (\ x ->
           ErrorMonad.errora
             (['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
                'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e', '-', 'c',
                'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 't', 'h', 'e', ' ', 'f',
                'o', 'l', 'l', 'o', 'w', 'i', 'n', 'g', ' ']
               : List.append representation
                   (['f', 'o', 'r', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l',
                      'o', 'w', 'i', 'n', 'g', ' ', 'r', 'e', 'a', 's', 'o',
                      'n']
                     : x)));

create_sct_ur_af_proc_comp ::
  forall a b c.
    (Eq a, Show.Sshow a, Eq b, Show.Sshow b, Eq c,
      Show.Sshow c) => (a -> Bool,
                         (a, (a, (a -> a -> a,
                                   (a -> a -> a,
                                     (a -> a,
                                       (a -> a -> Bool,
 (a -> a -> Bool, (Bool, ([Char], a -> Bool))))))))))
                         -> [((b, Nat.Nat), (a, [a]))] ->
                              [(Term.Term b c, Term.Term b c)] ->
                                [(Term.Term b c, Term.Term b c)] ->
                                  Maybe (b -> [(Term.Term b c, Term.Term b c)])
                                    -> Maybe (b ->
       [(Term.Term b c, Term.Term b c)])
 -> [(Term.Term b c, Term.Term b c)] ->
      [((Term.Term b c, Term.Term b c),
         ([(Nat.Nat, Nat.Nat)], [(Nat.Nat, Nat.Nat)]))]
        -> Eithera.Eithera [[Char]] ();
create_sct_ur_af_proc_comp c i p r rm rrm ur graphs =
  let {
    pia = PolyOrder.to_lpoly_inter (PolyOrder.get_comp_1 c) i;
  } in ErrorMonad.bind
         (PolyOrder.check_lpoly_coeffs (PolyOrder.get_comp_wf_elem c)
           (PolyOrder.get_comp_apos c)
           (\ x -> PolyOrder.get_comp_ge c x (PolyOrder.get_comp_0 c)) i)
         (\ x ->
           SCT.check_sct_ur_af_proc
             (PolyOrder.check_polo_sI (PolyOrder.get_comp_mono c)
               (PolyOrder.get_comp_0 c) (PolyOrder.get_comp_1 c)
               (PolyOrder.get_comp_plus c) (PolyOrder.get_comp_times c)
               (PolyOrder.get_comp_max0 c) (PolyOrder.get_comp_ge c)
               (PolyOrder.get_comp_gt c) pia)
             (PolyOrder.check_polo_nsI (PolyOrder.get_comp_0 c)
               (PolyOrder.get_comp_1 c) (PolyOrder.get_comp_plus c)
               (PolyOrder.get_comp_times c) (PolyOrder.get_comp_max0 c)
               (PolyOrder.get_comp_ge c) pia)
             (List.append
                (List.append
                  ['p', 'o', 'l', 'y', 'n', 'o', 'm', 'i', 'a', 'l', ' ', 'o',
                    'r', 'd', 'e', 'r', ' ', '(']
                  (PolyOrder.get_comp_description c))
                [')']
               : PolyOrder.create_lpoly_repr c i)
             p graphs r rm rrm ur
             (PolyOrder.create_afI (PolyOrder.get_comp_0 c) i));

create_sct_ur_af_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => RedPairImpl.Redpair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Maybe (a -> [(Term.Term a b, Term.Term a b)]) ->
                               Maybe (a -> [(Term.Term a b, Term.Term a b)]) ->
                                 [(Term.Term a b, Term.Term a b)] ->
                                   [((Term.Term a b, Term.Term a b),
                                      ([(Nat.Nat, Nat.Nat)],
[(Nat.Nat, Nat.Nat)]))]
                                     -> Eithera.Eithera [[Char]] ();
create_sct_ur_af_proc (RedPairImpl.Int_carrier ia) =
  SCT.create_sct_ur_af_proc_comp
    ((\ x -> True),
      (0, (1, ((\ a b -> a + b),
                ((\ a b -> a * b),
                  (Orderings.maxa 0,
                    ((\ x y -> y <= x),
                      ((\ x y -> y < x),
                        (True, (RedPairImpl.int_name, (\ x -> True)))))))))))
    ia;
create_sct_ur_af_proc (RedPairImpl.Nat_carrier i) =
  SCT.create_sct_ur_af_proc_comp
    ((\ x -> True),
      (Nat.Zero_nat,
        (Nat.one_nat,
          (Nat.plus_nat,
            (Nat.times_nat,
              (id, ((\ x y -> Nat.less_eq_nat y x),
                     ((\ x y -> Nat.less_nat y x),
                       (True, (RedPairImpl.nat_name, (\ x -> True)))))))))))
    i;
create_sct_ur_af_proc (RedPairImpl.Arctic_carrier ida) =
  SCT.create_sct_ur_af_proc_comp
    ((\ x -> True),
      (RedPairImpl.MinInfty,
        (RedPairImpl.Num_arc 0,
          (RedPairImpl.plus_arctic,
            (RedPairImpl.mult_arctic,
              (id, (RedPairImpl.geq_arctic,
                     (RedPairImpl.gt_arctic,
                       (False,
                         (RedPairImpl.arctic_name,
                           RedPairImpl.pos_arctic))))))))))
    ida;
create_sct_ur_af_proc (RedPairImpl.Int_mat_carrier n ib) =
  SCT.create_sct_ur_af_proc_comp
    (Matrix.wf_mat_rcI (Nat.Suc n) (Nat.Suc n),
      (Matrix.mat0I 0 (Nat.Suc n) (Nat.Suc n),
        (Matrix.mat1I 0 1 (Nat.Suc n),
          (Matrix.mat_plusI (\ a b -> a + b),
            (Matrix.mat_mulI 0 (\ a b -> a + b) (\ a b -> a * b) (Nat.Suc n),
              (Matrix.mat_max0I (Orderings.maxa 0),
                (Matrix.mat_geI (\ x y -> y <= x),
                  (Matrix.mat_gtI (\ x y -> y <= x) (\ x y -> y < x),
                    (True,
                      (RedPairImpl.int_mat_name,
                        Matrix.mat_arc_posI (\ x -> True)))))))))))
    ib;
create_sct_ur_af_proc (RedPairImpl.Arc_mat_carrier n ic) =
  SCT.create_sct_ur_af_proc_comp
    (Matrix.wf_mat_rcI (Nat.Suc n) (Nat.Suc n),
      (Matrix.mat0I RedPairImpl.MinInfty (Nat.Suc n) (Nat.Suc n),
        (Matrix.mat1I RedPairImpl.MinInfty (RedPairImpl.Num_arc 0) (Nat.Suc n),
          (Matrix.mat_plusI RedPairImpl.plus_arctic,
            (Matrix.mat_mulI RedPairImpl.MinInfty RedPairImpl.plus_arctic
               RedPairImpl.mult_arctic (Nat.Suc n),
              (Matrix.mat_max0I id,
                (Matrix.mat_geI RedPairImpl.geq_arctic,
                  (Matrix.mat_geI RedPairImpl.gt_arctic,
                    (False,
                      (RedPairImpl.arc_mat_name,
                        Matrix.mat_arc_posI RedPairImpl.pos_arctic))))))))))
    ic;

check_sct_subterm_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => [(Term.Term a b, Term.Term a b)] ->
                         [((Term.Term a b, Term.Term a b),
                            ([(Nat.Nat, Nat.Nat)], [(Nat.Nat, Nat.Nat)]))]
                           -> [(Term.Term a b, Term.Term a b)] ->
                                Maybe (a -> [(Term.Term a b, Term.Term a b)]) ->
                                  Maybe (a -> [(Term.Term a b, Term.Term a b)])
                                    -> Eithera.Eithera [[Char]] ();
check_sct_subterm_proc p gs r rm rrm =
  let {
    edge_check = Edg.checkEdgRM rm rrm p True;
  } in ErrorMonad.catcha
         (case rm of {
           Nothing ->
             ErrorMonad.errora
               [['T', 'R', 'S', ' ', 'h', 'a', 's', ' ', 'v', 'a', 'r', 'i',
                  'a', 'b', 'l', 'e', 's', ' ', 'a', 's', ' ', 'l', 'e', 'f',
                  't', '-', 'h', 'a', 'n', 'd', ' ', 's', 'i', 'd', 'e', 's']];
           Just rma ->
             ErrorMonad.bind
               (ErrorMonad.bind
                 (ErrorMonad.catcha
                   (ErrorMonad.checkSubset p (List.mapa fst gs))
                   (\ x ->
                     ErrorMonad.errora
                       [List.append
                          ['t', 'h', 'e', 'r', 'e', ' ', 'i', 's', ' ', 'n',
                            'o', ' ', 's', 'i', 'z', 'e', '-', 'c', 'h', 'a',
                            'n', 'g', 'e', ' ', 'g', 'r', 'a', 'p', 'h', ' ',
                            'f', 'o', 'r', ' ', 'D', 'P', ' ']
                          (Trs.showRule x)]))
                 (\ x ->
                   ErrorMonad.catcha
                     (ErrorMonad.checkAllMI
                       (\ (a @ (aa, b)) ->
                         let {
                           (s, t) = aa;
                         } in (\ (ab @ (stri, nstri)) ->
                                ErrorMonad.catcha
                                  (ErrorMonad.bind
                                    (ErrorMonad.checkB
                                      (not (Term.is_var s) &&
(not (Term.is_var t) && List.nulla (rma (Option.the (Term.root t)))))
                                      [['r', 'o', 'o', 't', 's', ' ', 'o', 'f',
 ' ', 'D', 'P', ' ', 'm', 'u', 's', 't', ' ', 'b', 'e', ' ', 'c', 'o', 'n', 's',
 't', 'r', 'u', 'c', 't', 'o', 'r', 's']])
                                    (\ xa ->
                                      let {
ss = Term.args s;
ts = Term.args t;
n = List.size_list ss;
m = List.size_list ts;
                                      } in
ErrorMonad.bind
  (ErrorMonad.catcha
    (ErrorMonad.checkAllMI
      (\ (ac @ (i, j)) ->
        ErrorMonad.checkB
          (Nat.less_nat i n &&
            (Nat.less_nat j m &&
              ErrorMonad.isValue
                (SCT.check_supt (List.nth ss i, List.nth ts j))))
          [List.append
             (List.append
               (List.append
                 ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h',
                   ' ', 'e', 'd', 'g', 'e', ' ']
                 (Show.sshow_nat i))
               [' ', '|', '>', ' '])
             (Show.sshow_nat j)])
      stri)
    (\ xb -> ErrorMonad.errora (snd xb)))
  (\ xb ->
    ErrorMonad.catcha
      (ErrorMonad.checkAllMI
        (\ (ac @ (i, j)) ->
          ErrorMonad.checkB
            (Nat.less_nat i n &&
              (Nat.less_nat j m &&
                ErrorMonad.isValue
                  (SCT.check_supteq (List.nth ss i, List.nth ts j))))
            [List.append
               (List.append
                 (List.append
                   ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'i', 't', 'h',
                     ' ', 'e', 'd', 'g', 'e', ' ']
                   (Show.sshow_nat i))
                 [' ', '|', '>', '=', ' '])
               (Show.sshow_nat j)])
        nstri)
      (\ xc -> ErrorMonad.errora (snd xc)))))
                                  (\ xa ->
                                    ErrorMonad.errora
                                      (List.append
 ['p', 'r', 'o', 'b', 'l', 'e', 'm', 's', ' ', 'w', 'i', 't', 'h', ' ', 'D',
   'P', ' ']
 (Trs.showRule (s, t))
: xa)))
                           b)
                       gs)
                     (\ xa -> ErrorMonad.errora (snd xa))))
               (\ x ->
                 ErrorMonad.checkB
                   (SCT.check_SCT (\ st uv -> edge_check (st, uv))
                     (List.mapa
                       (\ (a @ (st, aa)) ->
                         let {
                           (ab, b) = aa;
                         } in SCT.Scg st st ab b)
                       gs))
                   [['s', 'i', 'z', 'e', '-', 'c', 'h', 'a', 'n', 'g', 'e', ' ',
                      'a', 'n', 'a', 'l', 'y', 's', 'i', 's', ' ', 'f', 'a',
                      'i', 'l', 'e', 'd']]);
         })
         (\ x ->
           ErrorMonad.errora
             (['c', 'o', 'u', 'l', 'd', ' ', 'n', 'o', 't', ' ', 'a', 'p', 'p',
                'l', 'y', ' ', 't', 'h', 'e', ' ', 's', 'i', 'z', 'e', '-', 'c',
                'h', 'a', 'n', 'g', 'e', ' ', 'p', 'r', 'o', 'c', 'e', 's', 's',
                'o', 'r', ' ', 'w', 'i', 't', 'h', ' ', 's', 'u', 'b', 't', 'e',
                'r', 'm', '-', 'r', 'e', 'l', 'a', 't', 'i', 'o', 'n', ' ', 'f',
                'o', 'r', ' ', 't', 'h', 'e', ' ', 'f', 'o', 'l', 'l', 'o', 'w',
                'i', 'n', 'g', ' ', 'r', 'e', 'a', 's', 'o', 'n']
               : x));

}
