structure Ceta = 
struct

fun id x = x;

type 'a eq = {eq : 'a -> 'a -> bool};
fun eq (A_:'a eq) = #eq A_;

datatype nat = Zero_nat | Suc of nat;

datatype ('a, 'b) scg = Null | Scg of 'a * 'a * ('b * 'b) list * ('b * 'b) list;

datatype 'a shp = Sharp of 'a | Plain of 'a;

datatype xml =
  Xml of
    char list * (char list * char list) list * xml list * (char list) option;

fun leta s f = f s;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
fun less_eq (A_:'a ord) = #less_eq A_;
fun less (A_:'a ord) = #less A_;

fun nat_aux i n =
  (if IntInf.<= (i, (0 : IntInf.int)) then n
    else nat_aux (IntInf.- (i, (1 : IntInf.int))) (Suc n));

fun nat i = nat_aux i Zero_nat;

fun hd (x :: xs) = x;

fun tl [] = []
  | tl (x :: xs) = xs;

datatype ('a, 'b) term = Var of 'b | Fun of 'a * ('a, 'b) term list;

datatype ('a, 'b) sub = Sub of ('b -> ('a, 'b) term);

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun plus_nat (Suc m) n = plus_nat m (Suc n)
  | plus_nat Zero_nat n = n;

val one_nat : nat = Suc Zero_nat

fun sz (Var x) = one_nat
  | sz (Fun (f, ts)) =
    plus_nat one_nat (foldl (fn a => fn x => plus_nat a (sz x)) Zero_nat ts);

datatype 'a poly = PVar of nat | PNum of 'a | PSum of 'a poly list |
  PMult of 'a poly list;

fun fst (a, b) = a;

fun nth (x :: xs) (Suc n) = nth xs n
  | nth (x :: xs) Zero_nat = x;

fun eqop A_ a b = eq A_ a b;

datatype ('a, 'b) hterm = HHole | HFun of 'a * ('a, 'b) hterm list;

fun option_case f1 f2 (SOME a) = f2 a
  | option_case f1 f2 NONE = f1;

fun mergeHTerms A_ [] [] = SOME []
  | mergeHTerms A_ [] (uu :: uv) = NONE
  | mergeHTerms A_ (uw :: ux) [] = NONE
  | mergeHTerms A_ (HHole :: hsl) (h :: hsr) =
    (case mergeHTerms A_ hsl hsr of NONE => NONE | SOME res => SOME (h :: res))
  | mergeHTerms A_ (HFun (f, hs) :: hsl) (HHole :: hsr) =
    (case mergeHTerms A_ hsl hsr of NONE => NONE
       | SOME res => SOME (HFun (f, hs) :: res))
  | mergeHTerms A_ (HFun (f, fs) :: hsl) (HFun (g, gs) :: hsr) =
    (if eqop A_ f g
      then (case mergeHTerms A_ fs gs of NONE => NONE
              | SOME mhs =>
                (case mergeHTerms A_ hsl hsr of NONE => NONE
                   | SOME res => SOME (HFun (f, mhs) :: res)))
      else NONE);

fun merge A_ hl hr =
  (case mergeHTerms A_ [hl] [hr] of NONE => NONE
     | SOME res => SOME (nth res Zero_nat));

fun mergeX A_ B_ x h [] = SOME []
  | mergeX A_ B_ x h ((ht, y) :: ps) =
    (if eqop A_ x y
      then (case merge B_ h ht of NONE => NONE | SOME m => mergeX A_ B_ x m ps)
      else (case mergeX A_ B_ x h ps of NONE => NONE
              | SOME pairs => SOME ((ht, y) :: pairs)));

fun mergeAll A_ B_ [] = true
  | mergeAll A_ B_ ((h, x) :: ps) =
    (case mergeX B_ A_ x h ps of NONE => false | SOME a => mergeAll A_ B_ a);

fun append [] ys = ys
  | append (x :: xs) ys = x :: append xs ys;

fun decomp [] [] = SOME []
  | decomp (uu :: uv) [] = NONE
  | decomp [] (uw :: ux) = NONE
  | decomp (s :: ss) (t :: ts) =
    (case decomp ss ts of NONE => NONE | SOME pairs => SOME ((s, t) :: pairs));

fun gcMatchList A_ [] = SOME []
  | gcMatchList A_ ((HHole, uu) :: ps) = gcMatchList A_ ps
  | gcMatchList A_ ((HFun (f, hs), Fun (g, ss)) :: ps) =
    (if eqop A_ f g
      then (case decomp hs ss of NONE => NONE
              | SOME pairs => gcMatchList A_ (append pairs ps))
      else NONE)
  | gcMatchList A_ ((HFun (f, hs), Var x) :: ps) =
    (case gcMatchList A_ ps of NONE => NONE
       | SOME pairs => SOME ((HFun (f, hs), x) :: pairs));

fun gcMatchI A_ B_ h t =
  (case gcMatchList A_ [(h, t)] of NONE => false | SOME a => mergeAll A_ B_ a);

fun any uu [] = false
  | any p (x :: xs) = p x orelse any p xs;

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun capI A_ B_ uu (Var uv) = HHole
  | capI A_ B_ r (Fun (f, ts)) =
    let
      val h = HFun (f, map (capI A_ B_ r) ts);
    in
      (if any (fn ra => gcMatchI A_ B_ h (fst ra)) r then HHole else h)
    end;

fun rev xs = foldl (fn xsa => fn x => x :: xsa) [] xs;

fun less_nat m (Suc n) = less_eq_nat m n
  | less_nat n Zero_nat = false
and less_eq_nat (Suc m) n = less_nat m n
  | less_eq_nat Zero_nat n = true;

fun upt i j = (if less_nat i j then i :: upt (Suc i) j else []);

fun list_case f1 f2 (a :: lista) = f2 a lista
  | list_case f1 f2 [] = f1;

fun zip xs [] = []
  | zip xs (y :: ys) = (case xs of [] => [] | z :: zs => (z, y) :: zip zs ys);

datatype ('a, 'b) either = Left of 'a | Right of 'b;

fun return a = Right a;

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

fun concat [] = []
  | concat (x :: xs) = append x (concat xs);

fun rev_append [] ys = ys
  | rev_append (x :: xs) ys = rev_append xs (x :: ys);

fun flat_map_aux f [] ys = rev ys
  | flat_map_aux f (x :: xs) ys = flat_map_aux f xs (rev_append (f x) ys);

fun flat_map f xs = flat_map_aux f xs [];

val sEmpty : char list = []

type 'a sshow = {sshow : 'a -> char list};
fun sshow (A_:'a sshow) = #sshow A_;

fun sshow_term C_ D_ (Var x) = sshow D_ x
  | sshow_term C_ D_ (Fun (f, ts)) =
    append (sshow C_ f)
      ((case ts of [] => sEmpty
         | s :: ss =>
           append [#"("]
             (append
               (append (sshow_term C_ D_ s)
                 (flat_map (fn x => concat [[#","], sshow_term C_ D_ x]) ss))
               [#")"])));

fun error a = Left a;

fun size_list [] = Zero_nat
  | size_list (a :: lista) = plus_nat (size_list lista) (Suc Zero_nat);

datatype pos = Empty | PCons of nat * pos;

fun minus_nat (Suc m) (Suc n) = minus_nat m n
  | minus_nat Zero_nat n = Zero_nat
  | minus_nat m Zero_nat = m;

fun split f (a, b) = f a b;

fun foldr f [] a = a
  | foldr f (x :: xs) a = f x (foldr f xs a);

fun poss_list (Var x) = [Empty]
  | poss_list (Fun (f, ts)) =
    let
      val a =
        foldr (fn t => fn a as (i, ps) =>
                (minus_nat i one_nat,
                  append ps (map (fn aa => PCons (i, aa)) (poss_list t))))
          ts (minus_nat (size_list ts) one_nat, []);
      val (i, aa) = a;
    in
      Empty :: aa
    end;

fun nat_case f1 f2 (Suc nat) = f2 nat
  | nat_case f1 f2 Zero_nat = f1;

fun drop n [] = []
  | drop n (x :: xs) = (case n of Zero_nat => x :: xs | Suc m => drop m xs);

fun take n [] = []
  | take n (x :: xs) = (case n of Zero_nat => [] | Suc m => x :: take m xs);

datatype ('a, 'b) ctxt = Hole |
  More of 'a * ('a, 'b) term list * ('a, 'b) ctxt * ('a, 'b) term list;

fun ctxt_of_pos_term Empty t = Hole
  | ctxt_of_pos_term (PCons (i, ps)) (Fun (f, ts)) =
    More (f, take i ts, ctxt_of_pos_term ps (nth ts i), drop (Suc i) ts);

fun eqa A_ B_ (x1, y1) (x2, y2) = eqop A_ x1 x2 andalso eqop B_ y1 y2;

fun eqb A_ B_ = {eq = eqa A_ B_} : ('a * 'b) eq;

fun member A_ x [] = false
  | member A_ x (y :: ys) = eqop A_ x y orelse member A_ x ys;

fun sub_apply (Var x) (Sub sigma) = sigma x
  | sub_apply (Fun (f, ss)) sigmaa =
    Fun (f, map (fn t => sub_apply t sigmaa) ss);

fun ctxt_apply Hole s = s
  | ctxt_apply (More (f, ss1, c, ss2)) s =
    Fun (f, append ss1 (ctxt_apply c s :: ss2));

fun null [] = true
  | null (x :: xs) = false;

fun list_all2 p (x :: xs) (y :: ys) = p x y andalso list_all2 p xs ys
  | list_all2 p xs [] = null xs
  | list_all2 p [] ys = null ys;

fun eq_term C_ D_ (Fun (f, ts)) (Fun (g, us)) =
  eq C_ f g andalso list_all2 (eq_term C_ D_) ts us
  | eq_term C_ D_ (Var x) (Var y) = eq D_ x y
  | eq_term C_ D_ (Fun (b, c)) (Var a) = false
  | eq_term C_ D_ (Var a) (Fun (b, c)) = false;

fun eq_terma A_ B_ = {eq = eq_term A_ B_} : ('a, 'b) term eq;

fun checkB b e = (if b then return () else error e);

fun rewrites (A1_, A2_) (B1_, B2_) (s, t) c sigma (l, r) ra =
  checkB
    (eqop (eq_terma A1_ B1_) s (ctxt_apply c (sub_apply l sigma)) andalso
      (eqop (eq_terma A1_ B1_) t (ctxt_apply c (sub_apply r sigma)) andalso
        member (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) (l, r) ra))
    (append
      (append (append [#"T", #"e", #"r", #"m", #" "] (sshow_term A2_ B2_ s))
        [#" ", #"d", #"o", #"e", #"s", #" ", #"n", #"o", #"t", #" ", #"r", #"e",
          #"w", #"r", #"i", #"t", #"e", #" ", #"t", #"o", #" ", #"t", #"e",
          #"r", #"m", #" "])
      (sshow_term A2_ B2_ t));

fun either_case f1 f2 (Right b) = f2 b
  | either_case f1 f2 (Left a) = f1 a;

fun isValue p = (case p of Left l => false | Right r => true);

fun subt_at s Empty = s
  | subt_at (Fun (f, ss)) (PCons (i, p)) = subt_at (nth ss i) p;

fun sub_case f1 (Sub funa) = f1 funa;

fun subst_safe A_ B_ vs [] sigma = SOME sigma
  | subst_safe A_ B_ vs ((Var x, t) :: ts) sigma =
    (if member (eqb (eq_terma B_ A_) (eq_terma B_ A_)) (Var x, t) ts
      then subst_safe A_ B_ vs ts sigma
      else (if member A_ x vs andalso
                 not (eqop (eq_terma B_ A_) (sub_apply (Var x) sigma) t)
             then NONE
             else (case subst_safe A_ B_ (x :: vs) ts sigma of NONE => NONE
                     | SOME a =>
                       let
                         val Sub s = a;
                       in
                         SOME (Sub (fn y => (if eqop A_ x y then t else s y)))
                       end)));

fun decompa [] [] = SOME []
  | decompa (uu :: uv) [] = NONE
  | decompa [] (uw :: ux) = NONE
  | decompa (s :: ss) (t :: ts) =
    (case decompa ss ts of NONE => NONE | SOME pairs => SOME ((s, t) :: pairs));

fun decompose_one A_ (Var x) t = SOME [(Var x, t)]
  | decompose_one A_ (Fun (v, va)) (Var y) = NONE
  | decompose_one A_ (Fun (f, ts)) (Fun (g, ss)) =
    (if eqop A_ f g then decompa ts ss else NONE);

fun decompose_n A_ Zero_nat s t = decompose_one A_ s t
  | decompose_n A_ (Suc n) s t =
    (case decompose_n A_ n s t of NONE => NONE
       | SOME a =>
         foldl (fn aa => fn b as (sa, ta) =>
                 (case aa of NONE => NONE
                    | SOME ys =>
                      (case decompose_one A_ sa ta of NONE => NONE
                         | SOME xs => SOME (append xs ys))))
           (SOME []) a);

fun decompose A_ s t = decompose_n A_ (sz s) s t;

fun decompose_list A_ ts =
  foldl (fn a => fn b as (s, t) =>
          (case a of NONE => NONE
             | SOME ls =>
               (case decompose A_ s t of NONE => NONE
                  | SOME ys => SOME (append ls ys))))
    (SOME []) ts;

fun match_all A_ B_ ts =
  (case decompose_list A_ ts of NONE => NONE
     | SOME ls => subst_safe B_ A_ [] ls (Sub Var));

fun getpr (A1_, A2_) (B1_, B2_) s t p (l, r) =
  (case match_all A1_ B1_ [(l, subt_at s p), (r, subt_at t p)] of NONE => NONE
     | SOME sigma =>
       (if isValue
             (rewrites (A1_, A2_) (B1_, B2_) (s, t) (ctxt_of_pos_term p s) sigma
               (l, r) [(l, r)])
         then SOME (s, (ctxt_of_pos_term p s, (sigma, (l, r)))) else NONE));

fun getp (A1_, A2_) (B1_, B2_) s t p [] = NONE
  | getp (A1_, A2_) (B1_, B2_) s t p (r :: rs) =
    (case getpr (A1_, A2_) (B1_, B2_) s t p r
       of NONE => getp (A1_, A2_) (B1_, B2_) s t p rs | SOME a => SOME a);

fun find_get_p (F1_, F2_) (V1_, V2_) s t [] r = NONE
  | find_get_p (F1_, F2_) (V1_, V2_) s t (p :: ps) r =
    (case getp (F1_, F2_) (V1_, V2_) s t p r
       of NONE => find_get_p (F1_, F2_) (V1_, V2_) s t ps r | SOME a => SOME a);

fun get (A1_, A2_) (B1_, B2_) s t r =
  (case find_get_p (A1_, A2_) (B1_, B2_) s t (poss_list s) r
     of NONE =>
       error (append
               (append
                 (append
                   (append
                     [#"C", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t",
                       #" ", #"f", #"i", #"n", #"d", #" ", #"r", #"e", #"d",
                       #"u", #"c", #"t", #"i", #"o", #"n", #" ", #"b", #"e",
                       #"t", #"w", #"e", #"e", #"n", #" ", #"t", #"e", #"r",
                       #"m", #" "]
                     (sshow_term A2_ B2_ s))
                   [#" ", #"a", #"n", #"d", #" ", #"t", #"e", #"r", #"m", #" "])
                 (sshow_term A2_ B2_ t))
               [#"."])
     | SOME a => return a);

fun prod_case f1 (a, b) = f1 a b;

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if member A_ x xs then remdups A_ xs else x :: remdups A_ xs);

fun comp Ap_ es es' =
  remdups (eqb Ap_ Ap_)
    (flat_map
      (fn a as (x, y) =>
        flat_map (fn aa as (y', z) => (if eqop Ap_ y y' then [(x, z)] else []))
          es')
      es);

val tabC : char = #"\^I"

val tab : char list = [tabC]

val newlineC : char = #"\^J"

val newline : char list = [newlineC]

val wspace : char list = append newline (append tab [#" "])

val eq_char = {eq = (fn a => fn b => ((a : char) = b))} : char eq;

fun trim (c :: s) = (if member eq_char c wspace then trim s else c :: s)
  | trim [] = [];

fun capRM2 A_ B_ uu (Var uv) = HHole
  | capRM2 A_ B_ rm (Fun (f, ts)) =
    let
      val h = HFun (f, map (capRM2 A_ B_ rm) ts);
    in
      (if any (fn r => gcMatchI A_ B_ h (fst r)) (rm f) then HHole else h)
    end;

fun capRM A_ B_ r =
  (case r of NONE => (fn t => HHole) | SOME a => capRM2 A_ B_ a);

val eq_int = {eq = (fn a => fn b => ((a : IntInf.int) = b))} : IntInf.int eq;

fun eq_nat (Suc a) Zero_nat = false
  | eq_nat Zero_nat (Suc a) = false
  | eq_nat (Suc nat) (Suc nat') = eq_nat nat nat'
  | eq_nat Zero_nat Zero_nat = true;

val eq_nata = {eq = eq_nat} : nat eq;

fun eq_list A_ (a :: b) [] = false
  | eq_list A_ [] (a :: b) = false
  | eq_list A_ (a :: lista) (a' :: list') =
    eq A_ a a' andalso eq_list A_ lista list'
  | eq_list A_ [] [] = true;

fun eq_scg Pp_ Ap_ (Scg (a, b, c, d)) Null = false
  | eq_scg Pp_ Ap_ Null (Scg (a, b, c, d)) = false
  | eq_scg Pp_ Ap_ (Scg (pp1, pp2, list1, list2))
    (Scg (pp1', pp2', list1', list2')) =
    eq Pp_ pp1 pp1' andalso
      (eq Pp_ pp2 pp2' andalso
        (eq_list (eqb Ap_ Ap_) list1 list1' andalso
          eq_list (eqb Ap_ Ap_) list2 list2'))
  | eq_scg Pp_ Ap_ Null Null = true;

fun eq_scga A_ B_ = {eq = eq_scg A_ B_} : ('a, 'b) scg eq;

fun combine B_ (Scg (p, q, str, wk)) (Scg (p', q', str', wk')) =
  Scg (p, q, remdups (eqb B_ B_) (append str str'),
        remdups (eqb B_ B_) (append wk wk'))
  | combine B_ Null s = Null
  | combine B_ (Scg (v, va, vb, vc)) Null = Null;

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

fun subsumes A_ B_ (Scg (p, q, str, wk)) (Scg (p', q', str', wk')) =
  eqop A_ p p' andalso
    (eqop A_ q q' andalso
      (list_all (fn x => member (eqb B_ B_) x str') str andalso
        list_all (fn x => member (eqb B_ B_) x (append str' wk')) wk))
  | subsumes A_ B_ g Null = true
  | subsumes A_ B_ Null (Scg (v, va, vb, vc)) = false;

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun subtract_list A_ xs ys = filter (fn x => not (member A_ x ys)) xs;

fun scg_comp B_ conn (Scg (p, q, str, wk)) (Scg (p', q', str', wk')) =
  (if not (conn q p') then Null
    else let
           val strs =
             remdups (eqb B_ B_)
               (append (comp B_ str str')
                 (append (comp B_ str wk') (comp B_ wk str')));
           val a =
             subtract_list (eqb B_ B_) (remdups (eqb B_ B_) (comp B_ wk wk'))
               strs;
         in
           Scg (p, q', strs, a)
         end)
  | scg_comp B_ conn Null g = Null
  | scg_comp B_ conn (Scg (v, va, vb, vc)) Null = Null;

fun list_ex p [] = false
  | list_ex p (x :: xs) = p x orelse list_ex p xs;

fun in_situ Ap_ Null = true
  | in_situ Ap_ (Scg (p, q, str, wk)) =
    list_ex (fn a as (aa, b) => eqop Ap_ aa b) str;

fun sagiv A_ B_ conn g =
  (if in_situ B_ g then true
    else let
           val gg = scg_comp B_ conn g g;
         in
           (if subsumes A_ B_ gg g then false
             else sagiv A_ B_ conn (combine B_ g gg))
         end);

fun empty x = false;

fun args (Var uu) = []
  | args (Fun (uv, ts)) = ts;

fun root (Var x) = NONE
  | root (Fun (f, ts)) = SOME f;

fun eq_shp F_ (Plain b) (Sharp a) = false
  | eq_shp F_ (Sharp a) (Plain b) = false
  | eq_shp F_ (Plain f) (Plain f') = eq F_ f f'
  | eq_shp F_ (Sharp f) (Sharp f') = eq F_ f f';

fun eq_shpa A_ = {eq = eq_shp A_} : 'a shp eq;

datatype domain = Natural | Arctic | Int_mat of nat | Arc_mat of nat;

fun eq_option A_ (SOME a) NONE = false
  | eq_option A_ NONE (SOME a) = false
  | eq_option A_ (SOME a) (SOME a') = eq A_ a a'
  | eq_option A_ NONE NONE = true;

fun eq_lista A_ = {eq = eq_list A_} : ('a list) eq;

fun eq_xml (Xml (name1, attrs1, children1, text1))
  (Xml (name2, attrs2, children2, text2)) =
  eq_list eq_char name1 name2 andalso
    (eq_list (eqb (eq_lista eq_char) (eq_lista eq_char)) attrs1 attrs2 andalso
      (list_all2 eq_xml children1 children2 andalso
        eq_option (eq_lista eq_char) text1 text2));

val eq_xmla = {eq = eq_xml} : xml eq;

datatype xmldoc = Xmldoc of char list * xml;

val ord_int =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int ord;

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

fun the (SOME x) = x;

fun plaina (Var x) = Var x
  | plaina (Fun (f, ss)) = Fun (Plain f, map plaina ss);

fun sharpa (Var x) = Var x
  | sharpa (Fun (f, ss)) = Fun (Sharp f, map plaina ss);

val dquote : char list = [#"\""]

datatype 'a spL = SimpProj of ('a * nat) list;

fun times_nat Zero_nat n = Zero_nat
  | times_nat (Suc m) n = plus_nat n (times_nat m n);

fun digit2nat x =
  (if eqop eq_char x #"0" then return Zero_nat
    else (if eqop eq_char x #"1" then return one_nat
           else (if eqop eq_char x #"2" then return (nat (2 : IntInf.int))
                  else (if eqop eq_char x #"3"
                         then return (nat (3 : IntInf.int))
                         else (if eqop eq_char x #"4"
                                then return (nat (4 : IntInf.int))
                                else (if eqop eq_char x #"5"
                                       then return (nat (5 : IntInf.int))
                                       else (if eqop eq_char x #"6"
      then return (nat (6 : IntInf.int))
      else (if eqop eq_char x #"7" then return (nat (7 : IntInf.int))
             else (if eqop eq_char x #"8" then return (nat (8 : IntInf.int))
                    else (if eqop eq_char x #"9"
                           then return (nat (9 : IntInf.int))
                           else error (append [x]
[#" ", #"i", #"s", #" ", #"n", #"o", #"t", #" ", #"a", #" ", #"d", #"i", #"g",
  #"i", #"t"])))))))))));

fun bind p vq = (case p of Left a => Left a | Right a => vq a);

fun string2natHelp n [] = return n
  | string2natHelp n (d :: s) =
    bind (digit2nat d)
      (fn m =>
        string2natHelp (plus_nat (times_nat (nat (10 : IntInf.int)) n) m) s);

fun string2nat s =
  (if null s
    then error [#"c", #"a", #"n", #"n", #"o", #"t", #" ", #"c", #"o", #"n",
                 #"v", #"e", #"r", #"t", #" ", #"e", #"m", #"p", #"t", #"y",
                 #" ", #"s", #"t", #"r", #"i", #"n", #"g", #" ", #"i", #"n",
                 #"t", #"o", #" ", #"n", #"u", #"m", #"b", #"e", #"r"]
    else string2natHelp Zero_nat s);

fun replicate Zero_nat x = []
  | replicate (Suc n) x = x :: replicate n x;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun showAtt (a, v) = append a (append [#"="] (append dquote (append v dquote)));

fun showAtts asa =
  foldr (fn a => fn s => append [#" "] (append (showAtt a) s)) asa [];

fun showXMLindent ind i (Xml (n, a, c, t)) =
  concat
    [ind, [#"<"], n, showAtts a,
      (if null c andalso is_none t then [#"/", #">"]
        else concat
               [[#">"],
                 (if not (null c)
                   then append
                          (flat_map
                            (showXMLindent (append ind (replicate i #" ")) i) c)
                          ind
                   else []),
                 (if is_none t then [] else the t), [#"<", #"/"], n, [#">"]])];

fun sshow_xml xml = trim (showXMLindent newline (nat (2 : IntInf.int)) xml);

fun xml2text reqName (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso
          (null cs andalso (not (is_none text) andalso not (null (the text)))))
    then return (the text)
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"e", #"x", #"t", #"r", #"a", #"c", #"t", #" ", #"t", #"e",
                    #"x", #"t", #" ", #"f", #"o", #"r", #" "],
                   reqName, [#" ", #"f", #"r", #"o", #"m", #" ", #" "], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml2nat reqName x = bind (xml2text reqName x) string2nat;

fun xml3elements reqName p1 p2 p3 combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso
          (eqop eq_nata (size_list cs) (nat (3 : IntInf.int)) andalso
            is_none text))
    then bind (p1 (nth cs Zero_nat))
           (fn a =>
             bind (p2 (nth cs one_nat))
               (fn b =>
                 bind (p3 (nth cs (nat (2 : IntInf.int))))
                   (return o combi a b)))
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun mapMonad uu [] = return []
  | mapMonad f (xml :: rest) =
    (case f xml of Left a => error a
       | Right a =>
         (case mapMonad f rest of Left aa => error aa
            | Right asa => return (a :: asa)));

fun xmlManyElements reqName p combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso is_none text)
    then bind (mapMonad p cs) (return o combi)
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml2sp xml2name =
  xmlManyElements
    [#"a", #"r", #"g", #"u", #"m", #"e", #"n", #"t", #"F", #"i", #"l", #"t",
      #"e", #"r"]
    (xml3elements
      [#"a", #"r", #"g", #"u", #"m", #"e", #"n", #"t", #"F", #"i", #"l", #"t",
        #"e", #"r", #"E", #"n", #"t", #"r", #"y"]
      xml2name (xml2nat [#"a", #"r", #"i", #"t", #"y"])
      (xml2nat [#"c", #"o", #"l", #"l", #"a", #"p", #"s", #"i", #"n", #"g"])
      (fn f => fn a => fn n => (f, minus_nat n one_nat)))
    SimpProj;

fun fun_upd A_ f a b = (fn x => (if eqop A_ x a then b else f x));

fun implies p q = not p orelse q;

fun sshowa C_ D_ (x, y) =
  concat [sshow C_ x, [#" ", #"-", #">", #" "], sshow D_ y];

fun is_var (Var uu) = true
  | is_var (Fun (uv, uw)) = false;

fun showRule A_ B_ (l, r) =
  append (append (sshow_term A_ B_ l) [#" ", #"-", #">", #" "])
    (sshow_term A_ B_ r);

fun showRules A_ B_ r = map (showRule A_ B_) r;

fun showDPP A_ B_ p r =
  [#"d", #"e", #"p", #"e", #"n", #"d", #"e", #"n", #"c", #"y", #" ", #"p", #"a",
    #"i", #"r", #"s", #":", #" "]
    :: append (showRules A_ B_ p)
         ([] ::
           [#"r", #"e", #"w", #"r", #"i", #"t", #"e", #" ", #"s", #"y", #"s",
             #"t", #"e", #"m", #":"]
             :: append (showRules A_ B_ r) [[]]);

fun showTRS A_ B_ r = append (showRules A_ B_ r) [[]];

fun forall p [] = true
  | forall p (x :: xs) = p x andalso forall p xs;

fun lookup A_ uu [] = NONE
  | lookup A_ a ((a', b) :: xs) =
    (if eqop A_ a a' then SOME b else lookup A_ a xs);

fun snd (a, b) = b;

fun parseError s allowed =
  error (append
          [#"e", #"x", #"p", #"e", #"c", #"t", #"e", #"d", #" ", #"s", #"y",
            #"m", #"b", #"o", #"l", #"s", #" "]
          (append allowed
            (append
              [#" ", #"b", #"u", #"t", #" ", #"f", #"o", #"u", #"n", #"d", #" "]
              (take (nat (30 : IntInf.int)) s))));

fun consume2 [] allowed s =
  parseError s (flat_map (fn sa => append sa [#","]) allowed)
  | consume2 (t :: asa) allowed s =
    (if eqop (eq_lista eq_char) (map snd (zip t s)) t
      then return (t, trim (drop (size_list t) s)) else consume2 asa allowed s);

fun consume allowed s = consume2 allowed allowed s;

val letters : char list =
  [#"a", #"b", #"c", #"d", #"e", #"f", #"g", #"h", #"i", #"j", #"k", #"l", #"m",
    #"n", #"o", #"p", #"q", #"r", #"s", #"t", #"u", #"v", #"w", #"x", #"y",
    #"z", #"A", #"B", #"C", #"D", #"E", #"F", #"G", #"H", #"I", #"J", #"K",
    #"L", #"M", #"N", #"O", #"P", #"Q", #"R", #"S", #"T", #"U", #"V", #"W",
    #"X", #"Y", #"Z", #"_", #"0", #"1", #"2", #"3", #"4", #"5", #"6", #"7",
    #"8", #"9", #"&", #";", #":"]

fun nat2int Zero_nat = (0 : IntInf.int)
  | nat2int (Suc n) = IntInf.+ ((1 : IntInf.int), nat2int n);

fun string2int s =
  (case s
     of [] =>
       error [#"c", #"a", #"n", #"n", #"o", #"t", #" ", #"c", #"o", #"n", #"v",
               #"e", #"r", #"t", #" ", #"e", #"m", #"p", #"t", #"y", #" ", #"s",
               #"t", #"r", #"i", #"n", #"g", #" ", #"i", #"n", #"t", #"o", #" ",
               #"n", #"u", #"m", #"b", #"e", #"r"]
     | h :: t =>
       let
         val a = (if eqop eq_char h #"-" then (false, t) else (true, s));
         val (pos, ss) = a;
       in
         bind (string2nat ss)
           (fn n =>
             (if pos orelse eqop eq_nata n Zero_nat then return (nat2int n)
               else return (IntInf.- ((0 : IntInf.int), nat2int n))))
       end);

fun xml2int reqName x = bind (xml2text reqName x) string2int;

fun list2position [] = Empty
  | list2position (n :: ns) = PCons (n, list2position ns);

fun xmlChange p f x = bind (p x) (fn a => return (f a));

fun xml2position x =
  xmlChange (xml2nat [#"p", #"o", #"s", #"i", #"t", #"i", #"o", #"n"])
    (fn n => minus_nat n one_nat) x;

fun xml2pos x =
  xmlManyElements
    [#"p", #"o", #"s", #"i", #"t", #"i", #"o", #"n", #"I", #"n", #"T", #"e",
      #"r", #"m"]
    xml2position list2position x;

fun string2bool s =
  (if eqop (eq_lista eq_char) s [#"t", #"r", #"u", #"e"] then return true
    else (if eqop (eq_lista eq_char) s [#"f", #"a", #"l", #"s", #"e"]
           then return false
           else error (append
                        (append
                          [#"c", #"a", #"n", #"n", #"o", #"t", #" ", #"c", #"o",
                            #"n", #"v", #"e", #"r", #"t", #" "]
                          s)
                        [#" ", #"i", #"n", #"t", #"o", #" ", #"B", #"o", #"o",
                          #"l", #"e", #"a", #"n"])));

fun xml2bool reqName x = bind (xml2text reqName x) string2bool;

fun xml1element reqName p1 combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso
          (eqop eq_nata (size_list cs) one_nat andalso is_none text))
    then bind (p1 (nth cs Zero_nat)) (return o combi)
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml1CmanyElements reqName p1 p2 combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso (is_none text andalso not (null cs)))
    then bind (p1 (hd cs))
           (fn a => bind (mapMonad (p2 a) (tl cs)) (return o combi a))
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml1manyElements reqName p1 p2 = xml1CmanyElements reqName p1 (fn a => p2);

fun xml2var x = bind (xml2text [#"v", #"a", #"r"] x) (return o Var);

fun xmlGetName (Xml (name, uu, uv, uw)) = name;

fun xmlOptions ps x =
  (case lookup (eq_lista eq_char) (xmlGetName x) ps
     of NONE =>
       error (concat
               [[#"e", #"x", #"p", #"e", #"c", #"t", #"e", #"d", #" ", #"o",
                  #"n", #"e", #" ", #"o", #"f", #":", #" "],
                 flat_map (fn p => append (fst p) [#" "]) ps, newline,
                 [#"b", #"u", #"t", #" ", #"f", #"o", #"u", #"n", #"d"],
                 newline, sshow_xml x])
     | SOME p => p x);

fun xml2term xml2name x =
  xmlOptions
    [([#"v", #"a", #"r"], xml2var),
      ([#"f", #"u", #"n", #"a", #"p", #"p"],
        xml1manyElements [#"f", #"u", #"n", #"a", #"p", #"p"] xml2name
          (xml1element [#"a", #"r", #"g"] (xml2term xml2name) id)
          (fn a => fn b => Fun (a, b)))]
    x;

fun xml2elements reqName p1 p2 combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso
          (eqop eq_nata (size_list cs) (nat (2 : IntInf.int)) andalso
            is_none text))
    then bind (p1 (nth cs Zero_nat))
           (fn a => bind (p2 (nth cs one_nat)) (return o combi a))
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml2rule xml2name =
  xml2elements [#"r", #"u", #"l", #"e"]
    (xml1element [#"l", #"h", #"s"] (xml2term xml2name) id)
    (xml1element [#"r", #"h", #"s"] (xml2term xml2name) id)
    (fn a => fn b => (a, b));

fun xml2scg xml2name =
  xml1manyElements
    [#"s", #"i", #"z", #"e", #"C", #"h", #"a", #"n", #"g", #"e", #"G", #"r",
      #"a", #"p", #"h"]
    (xml2rule xml2name)
    (xml3elements [#"e", #"d", #"g", #"e"] xml2position
      (xml2bool [#"s", #"t", #"r", #"i", #"c", #"t"]) xml2position
      (fn p => fn s => fn q => (s, (p, q))))
    (fn lr => fn edges =>
      (lr, (map snd (filter fst edges),
             map snd (filter (fn x => not (fst x)) edges))));

fun lefta (Left l) = l;

fun vec_zeroI ze n = replicate n ze;

fun mat0I ze nr nc = replicate nc (vec_zeroI ze nr);

fun vec1I ze on n i =
  append (replicate i ze)
    (on :: replicate (minus_nat (minus_nat n one_nat) i) ze);

fun mat1I ze on n = map (vec1I ze on n) (upt Zero_nat n);

fun sshow_shp B_ (Sharp s) = append (sshow B_ s) [#"#"]
  | sshow_shp B_ (Plain s) = sshow B_ s;

fun sshow_shpa A_ = {sshow = sshow_shp A_} : 'a shp sshow;

fun splitterFun allowed (c :: s) =
  (if allowed c
    then let
           val a = splitterFun allowed s;
           val (l, aa) = a;
         in
           (c :: l, aa)
         end
    else ([], c :: s))
  | splitterFun uu [] = ([], []);

fun parseText s =
  let
    val a = splitterFun (fn c => not (eqop (eq_lista eq_char) [c] [#"<"])) s;
    val (l, r) = a;
  in
    (if null (trim l) then return (NONE, r)
      else return (SOME (rev (trim (rev (trim l)))), r))
  end;

fun parseAttribute s =
  bind (consume [dquote] s)
    (fn a as (uu, s1) =>
      let
        val aa =
          splitterFun (fn c => not (eqop (eq_lista eq_char) [c] dquote)) s1;
        val (n, s2) = aa;
      in
        bind (consume [dquote] s2) (fn ab as (uua, s3) => return (n, s3))
      end);

fun splitter allowed s = splitterFun (fn c => member eq_char c allowed) s;

fun parseName s =
  let
    val a = splitter letters s;
    val (l, r) = a;
  in
    (if null l then parseError s letters else return (l, trim r))
  end;

fun parseAttributes [] = return ([], [])
  | parseAttributes (c :: s) =
    (if member eq_char c [#"/", #">"] then return ([], c :: s)
      else bind (parseName (c :: s))
             (fn a as (n, s1) =>
               bind (consume [[#"="]] s1)
                 (fn aa as (uu, s2) =>
                   bind (parseAttribute s2)
                     (fn ab as (ac, s3) =>
                       bind (parseAttributes s3)
                         (fn b as (nas, s4) => return ((n, ac) :: nas, s4))))));

fun parseXMLList s =
  (if null s orelse
        eqop (eq_lista eq_char) (take (nat (2 : IntInf.int)) s) [#"<", #"/"]
    then return ([], s)
    else bind (consume [[#"<"]] s)
           (fn a as (uu, s1) =>
             bind (parseName s1)
               (fn aa as (n, s2) =>
                 bind (parseAttributes s2)
                   (fn ab as (nas, s3) =>
                     bind (consume [[#"/", #">"], [#">"]] s3)
                       (fn ac as (clo, s4) =>
                         (if eqop (eq_lista eq_char) clo [#"/", #">"]
                           then bind (parseXMLList s4)
                                  (fn ad as (li, s5) =>
                                    return (Xml (n, nas, [], NONE) :: li, s5))
                           else bind (parseText s4)
                                  (fn ad as (t, s5) =>
                                    bind (parseXMLList s5)
                                      (fn ae as (chi, s6) =>
bind (consume [append [#"<", #"/"] (append n [#">"])] s6)
  (fn af as (uua, s7) =>
    bind (parseXMLList s7)
      (fn ag as (li, s8) => return (Xml (n, nas, chi, t) :: li, s8)))))))))));

fun parseXML s =
  bind (consume [[#"<"]] s)
    (fn a as (uu, s1) =>
      bind (parseName s1)
        (fn aa as (n, s2) =>
          bind (parseAttributes s2)
            (fn ab as (nas, s3) =>
              bind (consume [[#"/", #">"], [#">"]] s3)
                (fn ac as (clo, s4) =>
                  (if eqop (eq_lista eq_char) clo [#"/", #">"]
                    then return (Xml (n, nas, [], NONE), trim s4)
                    else bind (parseText s4)
                           (fn ad as (t, s5) =>
                             bind (parseXMLList s5)
                               (fn ae as (chi, s6) =>
                                 bind (consume
[append [#"<", #"/"] (append n [#">"])] s6)
                                   (fn af as (uua, s7) =>
                                     return (Xml (n, nas, chi, t), s7)))))))));

fun scanUpto enda [] =
  error (append
          [#"d", #"i", #"d", #" ", #"n", #"o", #"t", #" ", #"f", #"i", #"n",
            #"d", #" ", #"r", #"e", #"q", #"u", #"i", #"r", #"e", #"d", #" ",
            #"t", #"o", #"k", #"e", #"n", #" "]
          enda)
  | scanUpto enda (c :: s) =
    (if eqop (eq_lista eq_char) (map snd (zip enda (c :: s))) enda
      then return
             (take (size_list enda) (c :: s),
               trim (drop (size_list enda) (c :: s)))
      else bind (scanUpto enda s)
             (fn a as (res, rest) => return (c :: res, rest)));

fun xml2poly xml2coeff x =
  xml1element [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l"]
    (xmlOptions
      [([#"c", #"o", #"e", #"f", #"f", #"i", #"c", #"i", #"e", #"n", #"t"],
         xml1element
           [#"c", #"o", #"e", #"f", #"f", #"i", #"c", #"i", #"e", #"n", #"t"]
           xml2coeff PNum),
        ([#"v", #"a", #"r", #"i", #"a", #"b", #"l", #"e"],
          xmlChange (xml2nat [#"v", #"a", #"r", #"i", #"a", #"b", #"l", #"e"])
            (fn n => PVar (minus_nat n one_nat))),
        ([#"s", #"u", #"m"],
          xmlManyElements [#"s", #"u", #"m"] (xml2poly xml2coeff) PSum),
        ([#"p", #"r", #"o", #"d", #"u", #"c", #"t"],
          xmlManyElements [#"p", #"r", #"o", #"d", #"u", #"c", #"t"]
            (xml2poly xml2coeff) PMult)])
    id x;

fun xml_term C_ D_ (Var x) = Xml ([#"v", #"a", #"r"], [], [], SOME (sshow D_ x))
  | xml_term C_ D_ (Fun (f, ts)) =
    Xml ([#"f", #"u", #"n", #"a", #"p", #"p"], [],
          Xml ([#"n", #"a", #"m", #"e"], [], [], SOME (sshow C_ f)) ::
            map (fn t => Xml ([#"a", #"r", #"g"], [], [xml_term C_ D_ t], NONE))
              ts,
          NONE);

fun isLeft (Left uu) = true
  | isLeft (Right uv) = false;

fun righta (Right r) = r;

fun get_list (A1_, A2_) (B1_, B2_) [] u r =
  error [#"L", #"o", #"o", #"p", #" ", #"m", #"u", #"s", #"t", #" ", #"b", #"e",
          #" ", #"n", #"o", #"n", #"-", #"e", #"m", #"p", #"t", #"y"]
  | get_list (A1_, A2_) (B1_, B2_) [s] u r =
    bind (get (A1_, A2_) (B1_, B2_) s u r) (fn y => return [y])
  | get_list (A1_, A2_) (B1_, B2_) (s :: t :: ts) u r =
    bind (get (A1_, A2_) (B1_, B2_) s t r)
      (fn d =>
        bind (get_list (A1_, A2_) (B1_, B2_) (t :: ts) u r)
          (fn y => return (d :: y)));

fun get_loop (A1_, A2_) (B1_, B2_) [] uu uv uw =
  error [#"L", #"o", #"o", #"p", #" ", #"m", #"u", #"s", #"t", #" ", #"b", #"e",
          #" ", #"n", #"o", #"n", #"-", #"e", #"m", #"p", #"t", #"y"]
  | get_loop (A1_, A2_) (B1_, B2_) (t :: ts) d tau r =
    get_list (A1_, A2_) (B1_, B2_) (t :: ts) (ctxt_apply d (sub_apply t tau)) r;

fun maxa less_eq a b = (if less_eq a b then b else a);

fun max A_ a = maxa (less_eq A_) a;

fun mina less_eq a b = (if less_eq a b then a else b);

fun min A_ a = mina (less_eq A_) a;

fun insert_scg Pp_ Ap_ g gs =
  (if eqop (eq_scga Pp_ Ap_) g Null then NONE
    else (if list_ex (fn g' => subsumes Pp_ Ap_ g' g) gs then NONE
           else SOME (g :: filter (fn g' => not (subsumes Pp_ Ap_ g g')) gs)));

fun mk_trancl A_ B_ conn [] s = s
  | mk_trancl A_ B_ conn (g :: gs) s =
    (case insert_scg A_ B_ g s of NONE => mk_trancl A_ B_ conn gs s
       | SOME s' =>
         mk_trancl A_ B_ conn
           (append (map (scg_comp B_ conn g) s')
             (append (map (fn f => scg_comp B_ conn f g) s') gs))
           s');

fun check_SCT Pp_ Ap_ conn gs =
  let
    val a = mk_trancl Pp_ Ap_ conn gs [];
  in
    list_all (sagiv Pp_ Ap_ conn) a
  end;

fun showList xs = flat_map (fn x => append x newline) xs;

fun divmod m n =
  (if eqop eq_nata n Zero_nat orelse less_nat m n then (Zero_nat, m)
    else let
           val a = divmod (minus_nat m n) n;
           val (q, aa) = a;
         in
           (Suc q, aa)
         end);

fun mod_nat m n = snd (divmod m n);

fun div_nat m n = fst (divmod m n);

fun digit2string n =
  (if eqop eq_nata n Zero_nat then [#"0"]
    else (if eqop eq_nata n one_nat then [#"1"]
           else (if eqop eq_nata n (nat (2 : IntInf.int)) then [#"2"]
                  else (if eqop eq_nata n (nat (3 : IntInf.int)) then [#"3"]
                         else (if eqop eq_nata n (nat (4 : IntInf.int))
                                then [#"4"]
                                else (if eqop eq_nata n (nat (5 : IntInf.int))
                                       then [#"5"]
                                       else (if eqop eq_nata n
          (nat (6 : IntInf.int))
      then [#"6"]
      else (if eqop eq_nata n (nat (7 : IntInf.int)) then [#"7"]
             else (if eqop eq_nata n (nat (8 : IntInf.int)) then [#"8"]
                    else (if eqop eq_nata n (nat (9 : IntInf.int)) then [#"9"]
                           else [#"n", #"o", #"t", #" ", #"a", #" ", #"d", #"i",
                                  #"g", #"i", #"t"]))))))))));

fun sshow_nat n =
  (if less_nat n (nat (10 : IntInf.int)) then digit2string n
    else append (sshow_nat (div_nat n (nat (10 : IntInf.int))))
           (digit2string (mod_nat n (nat (10 : IntInf.int)))));

fun sshow_int i =
  (if IntInf.< (i, (0 : IntInf.int))
    then append [#"-"] (sshow_nat (nat (IntInf.- ((0 : IntInf.int), i))))
    else sshow_nat (nat i));

val sshow_inta = {sshow = sshow_int} : IntInf.int sshow;

val sshow_nata = {sshow = sshow_nat} : nat sshow;

fun defs_list A_ [] = []
  | defs_list A_ ((Fun (f, uu), uv) :: ts) =
    let
      val fs = defs_list A_ ts;
    in
      (if member A_ f fs then fs else f :: fs)
    end
  | defs_list A_ ((Var uw, ux) :: ts) = defs_list A_ ts;

fun fitLength uu Zero_nat uv = []
  | fitLength a (Suc v) [] = replicate (Suc v) a
  | fitLength a (Suc n) (b :: asa) = b :: fitLength a n asa;

fun sum_lpoly uu [] ys = ys
  | sum_lpoly uv (v :: va) [] = v :: va
  | sum_lpoly pl (x :: xs) (y :: ys) = pl x y :: sum_lpoly pl xs ys;

fun isRight (Left uu) = false
  | isRight (Right uv) = true;

fun xml2label xml2name xml =
  bind (xmlManyElements [#"l", #"a", #"b", #"e", #"l"]
         (xmlOptions
           [([#"n", #"u", #"m", #"b", #"e", #"r"],
              xmlChange (xml2nat [#"n", #"u", #"m", #"b", #"e", #"r"]) Left),
             ([#"n", #"a", #"m", #"e"], xmlChange xml2name Right),
             ([#"s", #"h", #"a", #"r", #"p"], xmlChange xml2name Right),
             ([#"l", #"a", #"b", #"e", #"l", #"e", #"d", #"S", #"y", #"m", #"b",
                #"o", #"l"],
               xmlChange xml2name Right)])
         id xml)
    (fn xs =>
      (if forall isLeft xs then return (Left (map lefta xs))
        else (if forall isRight xs then return (Right (map righta xs))
               else error [#"m", #"i", #"x", #"e", #"d", #" ", #"l", #"a", #"b",
                            #"e", #"l", #"s"])));

fun xml2rules xml2name =
  xmlManyElements [#"r", #"u", #"l", #"e", #"s"] (xml2rule xml2name) id;

fun xml_rules A_ C_ rls =
  Xml ([#"r", #"u", #"l", #"e", #"s"], [],
        map (fn a as (l, r) =>
              Xml ([#"r", #"u", #"l", #"e"], [],
                    [Xml ([#"l", #"h", #"s"], [], [xml_term A_ C_ l], NONE),
                      Xml ([#"r", #"h", #"s"], [], [xml_term A_ C_ r], NONE)],
                    NONE))
          rls,
        NONE);

fun checkEdgRM A_ B_ rm rrm p precomputeEdges =
  (if null p
    then (case rrm
            of NONE =>
              (fn a as (aa, b) =>
                let
                  val (s, t) = aa;
                in
                  (fn ab as (u, v) => gcMatchI A_ B_ (capRM A_ B_ rm t) u)
                end
                  b)
            | SOME a =>
              (fn aa as (ab, b) =>
                let
                  val (s, t) = ab;
                in
                  (fn ac as (u, v) =>
                    gcMatchI A_ B_ (capRM A_ B_ rm t) u andalso
                      gcMatchI A_ B_ (capRM A_ B_ rrm u) t)
                end
                  b))
    else let
           val forward =
             map (fn t => (t, capRM A_ B_ rm t))
               (remdups (eq_terma A_ B_) (map snd p));
         in
           (case rrm
              of NONE =>
                (if precomputeEdges
                  then let
                         val edges =
                           flat_map
                             (fn a as (s, t) =>
                               map (fn aa as (u, v) =>
                                     ((t, u),
                                       gcMatchI A_ B_
 (the (lookup (eq_terma A_ B_) t forward)) u))
                                 p)
                             p;
                       in
                         (fn a as (aa, b) =>
                           let
                             val (s, t) = aa;
                           in
                             (fn ab as (u, v) =>
                               (case lookup
                                       (eqb (eq_terma A_ B_) (eq_terma A_ B_))
                                       (t, u) edges
                                  of NONE => gcMatchI A_ B_ (capRM A_ B_ rm t) u
                                  | SOME ba => ba))
                           end
                             b)
                       end
                  else (fn a as (aa, b) =>
                         let
                           val (s, t) = aa;
                         in
                           (fn ab as (u, v) =>
                             gcMatchI A_ B_
                               ((case lookup (eq_terma A_ B_) t forward
                                  of NONE => capRM A_ B_ rm t | SOME ct => ct))
                               u)
                         end
                           b))
              | SOME x =>
                let
                  val backward =
                    map (fn u => (u, capRM A_ B_ rrm u))
                      (remdups (eq_terma A_ B_) (map fst p));
                in
                  (if precomputeEdges
                    then let
                           val edges =
                             flat_map
                               (fn a as (s, t) =>
                                 map (fn aa as (u, v) =>
                                       ((t, u),
 gcMatchI A_ B_ (the (lookup (eq_terma A_ B_) t forward)) u andalso
   gcMatchI A_ B_ (the (lookup (eq_terma A_ B_) u backward)) t))
                                   p)
                               p;
                         in
                           (fn a as (aa, b) =>
                             let
                               val (s, t) = aa;
                             in
                               (fn ab as (u, v) =>
                                 (case lookup
 (eqb (eq_terma A_ B_) (eq_terma A_ B_)) (t, u) edges
                                    of NONE =>
                                      gcMatchI A_ B_ (capRM A_ B_ rm t) u
andalso gcMatchI A_ B_ (capRM A_ B_ rrm u) t
                                    | SOME ba => ba))
                             end
                               b)
                         end
                    else (fn a as (aa, b) =>
                           let
                             val (s, t) = aa;
                           in
                             (fn ab as (u, v) =>
                               gcMatchI A_ B_
                                 ((case lookup (eq_terma A_ B_) t forward
                                    of NONE => capRM A_ B_ rm t
                                    | SOME ct => ct))
                                 u
                                 andalso
                                 gcMatchI A_ B_
                                   ((case lookup (eq_terma A_ B_) u backward
                                      of NONE => capRM A_ B_ rrm u
                                      | SOME cu => cu))
                                   t)
                           end
                             b))
                end)
         end);

fun partition p [] = ([], [])
  | partition p (x :: xs) =
    let
      val a = partition p xs;
      val (yes, no) = a;
    in
      (if p x then (x :: yes, no) else (yes, x :: no))
    end;

fun removeAll A_ x [] = []
  | removeAll A_ x (y :: xs) =
    (if eqop A_ x y then removeAll A_ x xs else y :: removeAll A_ x xs);

fun rewrites_to (A1_, A2_) (B1_, B2_) [(s, (c, (sigma, rule)))] t r =
  rewrites (A1_, A2_) (B1_, B2_) (s, t) c sigma rule r
  | rewrites_to (A1_, A2_) (B1_, B2_)
    ((s, (c, (sigma, rule))) :: (t, (d, (tau, rule'))) :: xs) u r =
    bind (rewrites (A1_, A2_) (B1_, B2_) (s, t) c sigma rule r)
      (fn x =>
        rewrites_to (A1_, A2_) (B1_, B2_) ((t, (d, (tau, rule'))) :: xs) u r);

fun check_loop_d (A1_, A2_) (B1_, B2_) [] uu uv uw =
  error [#"T", #"h", #"e", #"r", #"e", #" ", #"m", #"u", #"s", #"t", #" ", #"b",
          #"e", #" ", #"a", #"t", #" ", #"l", #"e", #"a", #"s", #"t", #" ",
          #"o", #"n", #"e", #" ", #"r", #"e", #"w", #"r", #"i", #"t", #"e",
          #" ", #"s", #"t", #"e", #"p"]
  | check_loop_d (A1_, A2_) (B1_, B2_) (v :: va) d tau r =
    rewrites_to (A1_, A2_) (B1_, B2_) (v :: va)
      (ctxt_apply d (sub_apply (fst (hd (v :: va))) tau)) r;

fun check_loop (A1_, A2_) (B1_, B2_) (ts, (c, sigma)) r =
  bind (get_loop (A1_, A2_) (B1_, B2_) ts c sigma r)
    (fn ls => check_loop_d (A1_, A2_) (B1_, B2_) ls c sigma r);

fun catch p eq = (case p of Left a => eq a | Right a => Right a);

fun checkLoop (A1_, A2_) (B1_, B2_) r (ts, (c, sub)) =
  catch ((case subst_safe B1_ A1_ [] sub (Sub Var)
           of NONE =>
             error [#"s", #"u", #"b", #"s", #"t", #"i", #"t", #"u", #"t", #"i",
                     #"o", #"n", #" ", #"n", #"o", #"t", #" ", #"u", #"n", #"i",
                     #"q", #"u", #"e"]
           | SOME sigma => check_loop (A1_, A2_) (B1_, B2_) (ts, (c, sigma)) r))
    (fn x =>
      error [[#"e", #"r", #"r", #"o", #"r", #" ", #"i", #"n", #" ", #"c", #"e",
               #"r", #"t", #"i", #"f", #"i", #"c", #"a", #"t", #"i", #"o", #"n",
               #" ", #"o", #"f", #" ", #"l", #"o", #"o", #"p"],
              x]);

fun vec_geI ge v w = forall (fn a as (aa, b) => ge aa b) (zip v w);

fun mat_geI ge m1 m2 = forall (fn a as (aa, b) => vec_geI ge aa b) (zip m1 m2);

fun vec_pre_gtI gt v w = any (fn a as (aa, b) => gt aa b) (zip v w);

fun mat_pre_gtI gt m1 m2 =
  any (fn a as (aa, b) => vec_pre_gtI gt aa b) (zip m1 m2);

fun mat_gtI ge gt m1 m2 = mat_geI ge m1 m2 andalso mat_pre_gtI gt m1 m2;

fun wf_vecI n x = eqop eq_nata (size_list x) n;

fun checkSupterm A_ B_ (Var x) uu = false
  | checkSupterm A_ B_ (Fun (f, ts)) s =
    member (eq_terma A_ B_) s ts orelse any (fn t => checkSupterm A_ B_ t s) ts;

fun check_supt (F1_, F2_) (V1_, V2_) lr =
  catch (checkB (checkSupterm F1_ V1_ (fst lr) (snd lr))
          [[#"t", #"h", #"e", #" ", #"r", #"h", #"s", #" ", #"i", #"s", #" ",
             #"n", #"o", #"t", #" ", #"a", #" ", #"p", #"r", #"o", #"p", #"e",
             #"r", #" ", #"s", #"u", #"b", #"t", #"e", #"r", #"m", #" ", #"o",
             #"f", #" ", #"t", #"h", #"e", #" ", #"l", #"h", #"s"]])
    (fn x =>
      error (concat
               [[#"C", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                  #"o", #"r", #"i", #"e", #"n", #"t", #" ", #"r", #"u", #"l",
                  #"e", #" "],
                 sshow_term F2_ V2_ (fst lr), [#" ", #"-", #">", #" "],
                 sshow_term F2_ V2_ (snd lr)]
              :: x));

fun sshow_char x = [x];

val sshow_chara = {sshow = sshow_char} : char sshow;

fun sshow_list A_ xs = flat_map (sshow A_) xs;

fun sshow_lista A_ = {sshow = sshow_list A_} : ('a list) sshow;

fun sshow_terma A_ B_ = {sshow = sshow_term A_ B_} : ('a, 'b) term sshow;

fun transpose nr [] = replicate nr []
  | transpose nr (v :: m) =
    map (fn a as (aa, b) => aa :: b) (zip v (transpose nr m));

fun scalar_prodI ze pl ti v w =
  foldr (fn xy => pl (ti (fst xy) (snd xy))) (zip v w) ze;

fun matT_vec_mulI ze pl ti m v = map (fn w => scalar_prodI ze pl ti w v) m;

fun mat_mulI ze pl ti nr m1 m2 =
  map (matT_vec_mulI ze pl ti (transpose nr m1)) m2;

fun list_union A_ [] ys = ys
  | list_union A_ (x :: xs) ys =
    let
      val zs = list_union A_ xs ys;
    in
      (if member A_ x zs then zs else x :: zs)
    end;

fun domain_case f1 f2 f3 f4 (Arc_mat nat) = f4 nat
  | domain_case f1 f2 f3 f4 (Int_mat nat) = f3 nat
  | domain_case f1 f2 f3 f4 Arctic = f2
  | domain_case f1 f2 f3 f4 Natural = f1;

datatype ('a, 'b) lab = Lab of ('a, 'b) lab * 'b |
  FunLab of ('a, 'b) lab * ('a, 'b) lab list | UnLab of 'a;

fun id_labeling uu cs = Left cs;

fun stripXMLHeader s =
  (if eqop (eq_lista eq_char) (take (nat (2 : IntInf.int)) (trim s))
        [#"<", #"?"]
    then bind (scanUpto [#"?", #">"] s)
           (fn a as (header1, s1) =>
             bind (stripXMLHeader s1)
               (fn aa as (header2, s2) =>
                 return (append header1 header2, trim s2)))
    else return ([], trim s));

fun debug i t x = x;

fun parseXMLDoc s =
  debug [#"0"]
    [#"p", #"a", #"r", #"s", #"i", #"n", #"g", #" ", #"t", #"x", #"t", #" ",
      #"t", #"o", #" ", #"x", #"m", #"l"]
    (bind (stripXMLHeader s)
      (fn a as (header, rest) =>
        bind (parseXML rest)
          (fn aa as (xml, enda) =>
            (if null enda then return (Xmldoc (header, xml))
              else error (append
                           [#"t", #"h", #"e", #"r", #"e", #" ", #"i", #"s",
                             #" ", #"t", #"e", #"x", #"t", #" ", #"a", #"f",
                             #"t", #"e", #"r", #" ", #"r", #"o", #"o", #"t",
                             #" ", #"e", #"l", #"e", #"m", #"e", #"n", #"t",
                             #":", #" "]
                           (append newline enda))))));

fun xml4elements reqName p1 p2 p3 p4 combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso
          (eqop eq_nata (size_list cs) (nat (4 : IntInf.int)) andalso
            is_none text))
    then bind (p1 (nth cs Zero_nat))
           (fn a =>
             bind (p2 (nth cs one_nat))
               (fn b =>
                 bind (p3 (nth cs (nat (2 : IntInf.int))))
                   (fn c =>
                     bind (p4 (nth cs (nat (3 : IntInf.int))))
                       (return o combi a b c))))
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml0elements reqName combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso (null cs andalso is_none text))
    then return combi
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml2context xml2name x =
  xmlOptions
    [([#"b", #"o", #"x"], xml0elements [#"b", #"o", #"x"] Hole),
      ([#"f", #"u", #"n", #"C", #"o", #"n", #"t", #"e", #"x", #"t"],
        xml4elements
          [#"f", #"u", #"n", #"C", #"o", #"n", #"t", #"e", #"x", #"t"] xml2name
          (xmlManyElements [#"b", #"e", #"f", #"o", #"r", #"e"]
            (xml2term xml2name) id)
          (xml2context xml2name)
          (xmlManyElements [#"a", #"f", #"t", #"e", #"r"] (xml2term xml2name)
            id)
          (fn a => fn b => fn c => fn d => More (a, b, c, d)))]
    x;

datatype arithFun = Arg of nat | Const of nat | Sum of arithFun list |
  Max of arithFun list | Min of arithFun list | Prod of arithFun list;

datatype ('a, 'b) sl_inter =
  SL_Inter of nat * ((('a, 'b) lab * nat) * arithFun) list;

datatype arctic = MinInfty | Num_arc of IntInf.int;

datatype 'a redpair = Nat_carrier of (('a * nat) * (nat * nat list)) list |
  Int_carrier of (('a * nat) * (IntInf.int * IntInf.int list)) list |
  Arctic_carrier of (('a * nat) * (arctic * arctic list)) list |
  Int_mat_carrier of
    nat * (('a * nat) *
            ((IntInf.int list) list * ((IntInf.int list) list) list)) list
  | Arc_mat_carrier of
      nat * (('a * nat) *
              ((arctic list) list * ((arctic list) list) list)) list;

datatype ('a, 'b, 'c) dPProof = PisEmpty |
  RedPairProc of
    ('a, 'b) lab redpair *
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
      ('a, 'b, 'c) dPProof
  | RedPairUrProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | DepGraphProc of
      (('a, 'b, 'c) dPProof option *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list) list
  | MonoRedPairProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | MonoRedPairUrProc of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | UnlabProc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SemlabProc of
      ('a, 'b) sl_inter *
        (('a, 'b) lab -> nat list -> ('b, (('a, 'b) lab list)) either) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | RootlabProc of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SpscProc of
      ('a, 'b) lab spL *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) dPProof
  | SizeChangeSubTerm of
      (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
        ((nat * nat) list * (nat * nat) list)) list
  | SizeChangeRedPair of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        (((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) *
          ((nat * nat) list * (nat * nat) list)) list;

fun xml2arith_fun xml =
  xmlOptions
    [([#"n", #"a", #"t", #"u", #"r", #"a", #"l"],
       xmlChange (xml2nat [#"n", #"a", #"t", #"u", #"r", #"a", #"l"]) Const),
      ([#"v", #"a", #"r", #"i", #"a", #"b", #"l", #"e"],
        xmlChange (xml2nat [#"v", #"a", #"r", #"i", #"a", #"b", #"l", #"e"])
          (fn n => Arg (minus_nat n one_nat))),
      ([#"s", #"u", #"m"],
        xmlManyElements [#"s", #"u", #"m"] xml2arith_fun Sum),
      ([#"p", #"r", #"o", #"d", #"u", #"c", #"t"],
        xmlManyElements [#"p", #"r", #"o", #"d", #"u", #"c", #"t"] xml2arith_fun
          Prod),
      ([#"m", #"i", #"n", #"i", #"m", #"u", #"m"],
        xmlManyElements [#"m", #"i", #"n", #"i", #"m", #"u", #"m"] xml2arith_fun
          Min),
      ([#"m", #"a", #"x", #"i", #"m", #"u", #"m"],
        xmlManyElements [#"m", #"a", #"x", #"i", #"m", #"u", #"m"] xml2arith_fun
          Max)]
    xml;

fun xml2sl_inter xml2name =
  xml1element [#"m", #"o", #"d", #"e", #"l"]
    (xmlOptions
      [([#"f", #"i", #"n", #"i", #"t", #"e", #"M", #"o", #"d", #"e", #"l"],
         xml1manyElements
           [#"f", #"i", #"n", #"i", #"t", #"e", #"M", #"o", #"d", #"e", #"l"]
           (xmlChange
             (xml2nat
               [#"c", #"a", #"r", #"r", #"i", #"e", #"r", #"S", #"i", #"z",
                 #"e"])
             (fn n => minus_nat n one_nat))
           (xml3elements [#"i", #"n", #"t", #"e", #"r", #"p", #"r", #"e", #"t"]
             xml2name (xml2nat [#"a", #"r", #"i", #"t", #"y"]) xml2arith_fun
             (fn f => fn a => (fn b => ((f, a), b))))
           (fn n => fn xs => SOME (SL_Inter (n, xs)))),
        ([#"r", #"o", #"o", #"t", #"L", #"a", #"b", #"e", #"l", #"i", #"n",
           #"g"],
          xml0elements
            [#"r", #"o", #"o", #"t", #"L", #"a", #"b", #"e", #"l", #"i", #"n",
              #"g"]
            NONE)])
    id;

fun xml5elements reqName p1 p2 p3 p4 p5 combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso
          (eqop eq_nata (size_list cs) (nat (5 : IntInf.int)) andalso
            is_none text))
    then bind (p1 (nth cs Zero_nat))
           (fn a =>
             bind (p2 (nth cs one_nat))
               (fn b =>
                 bind (p3 (nth cs (nat (2 : IntInf.int))))
                   (fn c =>
                     bind (p4 (nth cs (nat (3 : IntInf.int))))
                       (fn d =>
                         bind (p5 (nth cs (nat (4 : IntInf.int))))
                           (return o combi a b c d)))))
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

val arc_mat_name : char list =
  [#"m", #"a", #"t", #"r", #"i", #"c", #"e", #"s", #" ", #"o", #"v", #"e", #"r",
    #" ", #"t", #"h", #"e", #" ", #"a", #"r", #"c", #"t", #"i", #"c", #" ",
    #"s", #"e", #"m", #"i", #"-", #"r", #"i", #"n", #"g"]

fun mat_arc_posI ap m = ap (nth (nth m Zero_nat) Zero_nat);

val int_mat_name : char list =
  [#"m", #"a", #"t", #"r", #"i", #"c", #"e", #"s", #" ", #"o", #"v", #"e", #"r",
    #" ", #"t", #"h", #"e", #" ", #"n", #"a", #"t", #"u", #"r", #"a", #"l",
    #"s", #" ", #"w", #"i", #"t", #"h", #" ", #"n", #"e", #"g", #"a", #"t",
    #"i", #"v", #"e", #" ", #"c", #"o", #"n", #"s", #"t", #"a", #"n", #"t",
    #"s"]

fun vec_max0I a = map a;

fun mat_max0I mz = map (vec_max0I mz);

fun vec_plusI pl v w = map (fn xy => pl (fst xy) (snd xy)) (zip v w);

fun mat_plusI pl m1 m2 =
  map (fn uv => vec_plusI pl (fst uv) (snd uv)) (zip m1 m2);

fun wf_mat_rcI nr nc m =
  eqop eq_nata (size_list m) nc andalso forall (wf_vecI nr) m;

fun xml2vec_coeff xml2coeff =
  xmlManyElements [#"v", #"e", #"c", #"t", #"o", #"r"]
    (xml1element
      [#"c", #"o", #"e", #"f", #"f", #"i", #"c", #"i", #"e", #"n", #"t"]
      xml2coeff id)
    id;

fun xml2mat_coeff xml2coeff =
  xmlManyElements [#"m", #"a", #"t", #"r", #"i", #"x"] (xml2vec_coeff xml2coeff)
    id;

fun pos_arctic MinInfty = false
  | pos_arctic (Num_arc n) = IntInf.<= ((0 : IntInf.int), n);

val arctic_name : char list =
  [#"a", #"r", #"c", #"t", #"i", #"c", #" ", #"s", #"e", #"m", #"i", #"r", #"i",
    #"n", #"g"]

fun gt_arctic uu MinInfty = true
  | gt_arctic MinInfty (Num_arc uv) = false
  | gt_arctic (Num_arc x) (Num_arc y) = IntInf.< (y, x);

fun geq_arctic xa MinInfty = true
  | geq_arctic MinInfty (Num_arc uu) = false
  | geq_arctic (Num_arc x) (Num_arc y) = IntInf.<= (y, x);

fun mult_arctic MinInfty ya = MinInfty
  | mult_arctic (Num_arc v) MinInfty = MinInfty
  | mult_arctic (Num_arc x) (Num_arc y) = Num_arc (IntInf.+ (x, y));

fun plus_arctic MinInfty ya = ya
  | plus_arctic (Num_arc v) MinInfty = Num_arc v
  | plus_arctic (Num_arc x) (Num_arc y) = Num_arc (max ord_int x y);

fun eq_arctic (Num_arc a) MinInfty = false
  | eq_arctic MinInfty (Num_arc a) = false
  | eq_arctic (Num_arc int) (Num_arc int') = ((int : IntInf.int) = int')
  | eq_arctic MinInfty MinInfty = true;

val eq_arctica = {eq = eq_arctic} : arctic eq;

fun xml2arctic_coeff x =
  xmlOptions
    [([#"i", #"n", #"t", #"e", #"g", #"e", #"r"],
       xmlChange (xml2int [#"i", #"n", #"t", #"e", #"g", #"e", #"r"]) Num_arc),
      ([#"m", #"i", #"n", #"u", #"s", #"I", #"n", #"f", #"i", #"n", #"i", #"t",
         #"y"],
        xml0elements
          [#"m", #"i", #"n", #"u", #"s", #"I", #"n", #"f", #"i", #"n", #"i",
            #"t", #"y"]
          MinInfty)]
    x;

fun get_comp_times
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = ti;

fun get_comp_plus
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = pl;

fun get_comp_1 (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) =
  on;

val int_name : char list =
  [#"n", #"a", #"t", #"u", #"r", #"a", #"l", #"s", #" ", #"w", #"i", #"t", #"h",
    #" ", #"n", #"e", #"g", #"a", #"t", #"i", #"v", #"e", #" ", #"c", #"o",
    #"n", #"s", #"t", #"a", #"n", #"t", #"s"]

fun get_comp_0 (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) =
  ze;

fun poly_to_lpoly A_ uu uv uw ux (PNum i) = return (i, [])
  | poly_to_lpoly A_ ze on pl ti (PVar x) =
    return (ze, append (replicate x ze) [on])
  | poly_to_lpoly A_ ze on pl ti (PSum [p]) = poly_to_lpoly A_ ze on pl ti p
  | poly_to_lpoly A_ ze on pl ti (PSum (p :: q :: ps)) =
    bind (poly_to_lpoly A_ ze on pl ti p)
      (fn a as (cp, ncp) =>
        bind (poly_to_lpoly A_ ze on pl ti (PSum (q :: ps)))
          (fn aa as (cq, ncq) => return (pl cp cq, sum_lpoly pl ncp ncq)))
  | poly_to_lpoly A_ ze on pl ti (PMult [p]) = poly_to_lpoly A_ ze on pl ti p
  | poly_to_lpoly A_ ze on pl ti (PMult (p :: q :: ps)) =
    bind (poly_to_lpoly A_ ze on pl ti p)
      (fn a as (cp, ncp) =>
        bind (poly_to_lpoly A_ ze on pl ti (PMult (q :: ps)))
          (fn aa as (cq, ncq) =>
            (if forall (eqop A_ ze) ncp then return (ti cp cq, map (ti cp) ncq)
              else (if forall (eqop A_ ze) ncq
                     then return (ti cp cq, map (fn x => ti x cq) ncp)
                     else error [#"n", #"o", #"n", #"-", #"l", #"i", #"n", #"e",
                                  #"a", #"r", #" ", #"p", #"o", #"l", #"y",
                                  #"n", #"o", #"m", #"i", #"a", #"l", #"s",
                                  #" ", #"n", #"o", #"t", #" ", #"s", #"u",
                                  #"p", #"p", #"o", #"r", #"t", #"e", #"d"]))));

fun xml2int_coeff x = xml2int [#"i", #"n", #"t", #"e", #"g", #"e", #"r"] x;

fun xml2redpair xml2name =
  xmlOptions
    [([#"r", #"e", #"d", #"P", #"a", #"i", #"r"],
       xml1element [#"r", #"e", #"d", #"P", #"a", #"i", #"r"]
         (xml1CmanyElements
           [#"i", #"n", #"t", #"e", #"r", #"p", #"r", #"e", #"t", #"a", #"t",
             #"i", #"o", #"n"]
           (xml1element [#"t", #"y", #"p", #"e"]
             (xml2elements
               [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l"]
               (xml1element [#"d", #"o", #"m", #"a", #"i", #"n"]
                 (xmlOptions
                   [([#"n", #"a", #"t", #"u", #"r", #"a", #"l", #"s"],
                      xml0elements
                        [#"n", #"a", #"t", #"u", #"r", #"a", #"l", #"s"]
                        Natural),
                     ([#"a", #"r", #"c", #"t", #"i", #"c"],
                       xml0elements [#"a", #"r", #"c", #"t", #"i", #"c"]
                         Arctic),
                     ([#"a", #"r", #"c", #"t", #"i", #"c", #"B", #"e", #"l",
                        #"o", #"w", #"Z", #"e", #"r", #"o"],
                       xml0elements
                         [#"a", #"r", #"c", #"t", #"i", #"c", #"B", #"e", #"l",
                           #"o", #"w", #"Z", #"e", #"r", #"o"]
                         Arctic),
                     ([#"m", #"a", #"t", #"r", #"i", #"c", #"e", #"s"],
                       xml3elements
                         [#"m", #"a", #"t", #"r", #"i", #"c", #"e", #"s"]
                         (xml2nat
                           [#"d", #"i", #"m", #"e", #"n", #"s", #"i", #"o",
                             #"n"])
                         (xml2nat
                           [#"s", #"t", #"r", #"i", #"c", #"t", #"D", #"i",
                             #"m", #"e", #"n", #"s", #"i", #"o", #"n"])
                         (xml1element [#"d", #"o", #"m", #"a", #"i", #"n"]
                           (xmlOptions
                             [([#"n", #"a", #"t", #"u", #"r", #"a", #"l", #"s"],
                                xml0elements
                                  [#"n", #"a", #"t", #"u", #"r", #"a", #"l",
                                    #"s"]
                                  true),
                               ([#"a", #"r", #"c", #"t", #"i", #"c"],
                                 xml0elements
                                   [#"a", #"r", #"c", #"t", #"i", #"c"] false),
                               ([#"a", #"r", #"c", #"t", #"i", #"c", #"B", #"e",
                                  #"l", #"o", #"w", #"Z", #"e", #"r", #"o"],
                                 xml0elements
                                   [#"a", #"r", #"c", #"t", #"i", #"c", #"B",
                                     #"e", #"l", #"o", #"w", #"Z", #"e", #"r",
                                     #"o"]
                                   false)])
                           id)
                         (fn di => fn sd => fn doa =>
                           (if doa then Int_mat di else Arc_mat di)))])
                 id)
               (xml2nat [#"d", #"e", #"g", #"r", #"e", #"e"])
               (fn typea => fn uu => typea))
             id)
           (fn a =>
             (case a
                of Natural =>
                  xml3elements
                    [#"i", #"n", #"t", #"e", #"r", #"p", #"r", #"e", #"t"]
                    xml2name (xml2nat [#"a", #"r", #"i", #"t", #"y"])
                    (fn x =>
                      bind (xml2poly xml2int_coeff x)
                        (poly_to_lpoly eq_int
                          (get_comp_0
                            ((fn xa => true),
                              ((0 : IntInf.int),
                                ((1 : IntInf.int),
                                  ((fn aa => fn b => IntInf.+ (aa, b)),
                                    ((fn aa => fn b => IntInf.* (aa, b)),
                                      (max ord_int (0 : IntInf.int),
((fn xa => fn y => IntInf.<= (y, xa)),
  ((fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_name, (fn xa => true))))))))))))
                          (get_comp_1
                            ((fn xa => true),
                              ((0 : IntInf.int),
                                ((1 : IntInf.int),
                                  ((fn aa => fn b => IntInf.+ (aa, b)),
                                    ((fn aa => fn b => IntInf.* (aa, b)),
                                      (max ord_int (0 : IntInf.int),
((fn xa => fn y => IntInf.<= (y, xa)),
  ((fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_name, (fn xa => true))))))))))))
                          (get_comp_plus
                            ((fn xa => true),
                              ((0 : IntInf.int),
                                ((1 : IntInf.int),
                                  ((fn aa => fn b => IntInf.+ (aa, b)),
                                    ((fn aa => fn b => IntInf.* (aa, b)),
                                      (max ord_int (0 : IntInf.int),
((fn xa => fn y => IntInf.<= (y, xa)),
  ((fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_name, (fn xa => true))))))))))))
                          (get_comp_times
                            ((fn xa => true),
                              ((0 : IntInf.int),
                                ((1 : IntInf.int),
                                  ((fn aa => fn b => IntInf.+ (aa, b)),
                                    ((fn aa => fn b => IntInf.* (aa, b)),
                                      (max ord_int (0 : IntInf.int),
((fn xa => fn y => IntInf.<= (y, xa)),
  ((fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_name, (fn xa => true))))))))))))))
                    (fn f => fn aa => fn poly =>
                      Left (Left ((f, aa),
                                   (fst poly,
                                     fitLength
                                       (get_comp_0
 ((fn x => true),
   ((0 : IntInf.int),
     ((1 : IntInf.int),
       ((fn ab => fn b => IntInf.+ (ab, b)),
         ((fn ab => fn b => IntInf.* (ab, b)),
           (max ord_int (0 : IntInf.int),
             ((fn x => fn y => IntInf.<= (y, x)),
               ((fn x => fn y => IntInf.< (y, x)),
                 (true, (int_name, (fn x => true))))))))))))
                                       aa (snd poly)))))
                | Arctic =>
                  xml3elements
                    [#"i", #"n", #"t", #"e", #"r", #"p", #"r", #"e", #"t"]
                    xml2name (xml2nat [#"a", #"r", #"i", #"t", #"y"])
                    (fn x =>
                      bind (xml2poly xml2arctic_coeff x)
                        (poly_to_lpoly eq_arctica
                          (get_comp_0
                            ((fn xa => true),
                              (MinInfty,
                                (Num_arc (0 : IntInf.int),
                                  (plus_arctic,
                                    (mult_arctic,
                                      (id,
(geq_arctic, (gt_arctic, (false, (arctic_name, pos_arctic)))))))))))
                          (get_comp_1
                            ((fn xa => true),
                              (MinInfty,
                                (Num_arc (0 : IntInf.int),
                                  (plus_arctic,
                                    (mult_arctic,
                                      (id,
(geq_arctic, (gt_arctic, (false, (arctic_name, pos_arctic)))))))))))
                          (get_comp_plus
                            ((fn xa => true),
                              (MinInfty,
                                (Num_arc (0 : IntInf.int),
                                  (plus_arctic,
                                    (mult_arctic,
                                      (id,
(geq_arctic, (gt_arctic, (false, (arctic_name, pos_arctic)))))))))))
                          (get_comp_times
                            ((fn xa => true),
                              (MinInfty,
                                (Num_arc (0 : IntInf.int),
                                  (plus_arctic,
                                    (mult_arctic,
                                      (id,
(geq_arctic, (gt_arctic, (false, (arctic_name, pos_arctic)))))))))))))
                    (fn f => fn aa => fn poly =>
                      Left (Right
                             ((f, aa),
                               (fst poly,
                                 fitLength
                                   (get_comp_0
                                     ((fn x => true),
                                       (MinInfty,
 (Num_arc (0 : IntInf.int),
   (plus_arctic,
     (mult_arctic,
       (id, (geq_arctic, (gt_arctic, (false, (arctic_name, pos_arctic)))))))))))
                                   aa (snd poly)))))
                | Int_mat n =>
                  xml3elements
                    [#"i", #"n", #"t", #"e", #"r", #"p", #"r", #"e", #"t"]
                    xml2name (xml2nat [#"a", #"r", #"i", #"t", #"y"])
                    (fn x =>
                      bind (xml2poly (xml2mat_coeff xml2int_coeff) x)
                        (poly_to_lpoly (eq_lista (eq_lista eq_int))
                          (get_comp_0
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I (0 : IntInf.int)
                                 (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I (0 : IntInf.int) (1 : IntInf.int)
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI
                                     (fn aa => fn b => IntInf.+ (aa, b)),
                                    (mat_mulI (0 : IntInf.int)
                                       (fn aa => fn b => IntInf.+ (aa, b))
                                       (fn aa => fn b => IntInf.* (aa, b))
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I (max ord_int (0 : IntInf.int)),
(mat_geI (fn xa => fn y => IntInf.<= (y, xa)),
  (mat_gtI (fn xa => fn y => IntInf.<= (y, xa))
     (fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_mat_name, mat_arc_posI (fn xa => true))))))))))))
                          (get_comp_1
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I (0 : IntInf.int)
                                 (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I (0 : IntInf.int) (1 : IntInf.int)
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI
                                     (fn aa => fn b => IntInf.+ (aa, b)),
                                    (mat_mulI (0 : IntInf.int)
                                       (fn aa => fn b => IntInf.+ (aa, b))
                                       (fn aa => fn b => IntInf.* (aa, b))
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I (max ord_int (0 : IntInf.int)),
(mat_geI (fn xa => fn y => IntInf.<= (y, xa)),
  (mat_gtI (fn xa => fn y => IntInf.<= (y, xa))
     (fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_mat_name, mat_arc_posI (fn xa => true))))))))))))
                          (get_comp_plus
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I (0 : IntInf.int)
                                 (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I (0 : IntInf.int) (1 : IntInf.int)
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI
                                     (fn aa => fn b => IntInf.+ (aa, b)),
                                    (mat_mulI (0 : IntInf.int)
                                       (fn aa => fn b => IntInf.+ (aa, b))
                                       (fn aa => fn b => IntInf.* (aa, b))
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I (max ord_int (0 : IntInf.int)),
(mat_geI (fn xa => fn y => IntInf.<= (y, xa)),
  (mat_gtI (fn xa => fn y => IntInf.<= (y, xa))
     (fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_mat_name, mat_arc_posI (fn xa => true))))))))))))
                          (get_comp_times
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I (0 : IntInf.int)
                                 (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I (0 : IntInf.int) (1 : IntInf.int)
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI
                                     (fn aa => fn b => IntInf.+ (aa, b)),
                                    (mat_mulI (0 : IntInf.int)
                                       (fn aa => fn b => IntInf.+ (aa, b))
                                       (fn aa => fn b => IntInf.* (aa, b))
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I (max ord_int (0 : IntInf.int)),
(mat_geI (fn xa => fn y => IntInf.<= (y, xa)),
  (mat_gtI (fn xa => fn y => IntInf.<= (y, xa))
     (fn xa => fn y => IntInf.< (y, xa)),
    (true, (int_mat_name, mat_arc_posI (fn xa => true))))))))))))))
                    (fn f => fn aa => fn poly =>
                      Right (Left ((f, aa),
                                    (fst poly,
                                      fitLength
(get_comp_0
  (wf_mat_rcI (Suc (minus_nat n one_nat)) (Suc (minus_nat n one_nat)),
    (mat0I (0 : IntInf.int) (Suc (minus_nat n one_nat))
       (Suc (minus_nat n one_nat)),
      (mat1I (0 : IntInf.int) (1 : IntInf.int) (Suc (minus_nat n one_nat)),
        (mat_plusI (fn ab => fn b => IntInf.+ (ab, b)),
          (mat_mulI (0 : IntInf.int) (fn ab => fn b => IntInf.+ (ab, b))
             (fn ab => fn b => IntInf.* (ab, b)) (Suc (minus_nat n one_nat)),
            (mat_max0I (max ord_int (0 : IntInf.int)),
              (mat_geI (fn x => fn y => IntInf.<= (y, x)),
                (mat_gtI (fn x => fn y => IntInf.<= (y, x))
                   (fn x => fn y => IntInf.< (y, x)),
                  (true, (int_mat_name, mat_arc_posI (fn x => true))))))))))))
aa (snd poly)))))
                | Arc_mat n =>
                  xml3elements
                    [#"i", #"n", #"t", #"e", #"r", #"p", #"r", #"e", #"t"]
                    xml2name (xml2nat [#"a", #"r", #"i", #"t", #"y"])
                    (fn x =>
                      bind (xml2poly (xml2mat_coeff xml2arctic_coeff) x)
                        (poly_to_lpoly (eq_lista (eq_lista eq_arctica))
                          (get_comp_0
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I MinInfty (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I MinInfty (Num_arc (0 : IntInf.int))
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI plus_arctic,
                                    (mat_mulI MinInfty plus_arctic mult_arctic
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I id,
(mat_geI geq_arctic,
  (mat_geI gt_arctic, (false, (arc_mat_name, mat_arc_posI pos_arctic)))))))))))
                          (get_comp_1
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I MinInfty (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I MinInfty (Num_arc (0 : IntInf.int))
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI plus_arctic,
                                    (mat_mulI MinInfty plus_arctic mult_arctic
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I id,
(mat_geI geq_arctic,
  (mat_geI gt_arctic, (false, (arc_mat_name, mat_arc_posI pos_arctic)))))))))))
                          (get_comp_plus
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I MinInfty (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I MinInfty (Num_arc (0 : IntInf.int))
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI plus_arctic,
                                    (mat_mulI MinInfty plus_arctic mult_arctic
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I id,
(mat_geI geq_arctic,
  (mat_geI gt_arctic, (false, (arc_mat_name, mat_arc_posI pos_arctic)))))))))))
                          (get_comp_times
                            (wf_mat_rcI (Suc (minus_nat n one_nat))
                               (Suc (minus_nat n one_nat)),
                              (mat0I MinInfty (Suc (minus_nat n one_nat))
                                 (Suc (minus_nat n one_nat)),
                                (mat1I MinInfty (Num_arc (0 : IntInf.int))
                                   (Suc (minus_nat n one_nat)),
                                  (mat_plusI plus_arctic,
                                    (mat_mulI MinInfty plus_arctic mult_arctic
                                       (Suc (minus_nat n one_nat)),
                                      (mat_max0I id,
(mat_geI geq_arctic,
  (mat_geI gt_arctic,
    (false, (arc_mat_name, mat_arc_posI pos_arctic)))))))))))))
                    (fn f => fn aa => fn poly =>
                      Right (Right
                              ((f, aa),
                                (fst poly,
                                  fitLength
                                    (get_comp_0
                                      (wf_mat_rcI (Suc (minus_nat n one_nat))
 (Suc (minus_nat n one_nat)),
(mat0I MinInfty (Suc (minus_nat n one_nat)) (Suc (minus_nat n one_nat)),
  (mat1I MinInfty (Num_arc (0 : IntInf.int)) (Suc (minus_nat n one_nat)),
    (mat_plusI plus_arctic,
      (mat_mulI MinInfty plus_arctic mult_arctic (Suc (minus_nat n one_nat)),
        (mat_max0I id,
          (mat_geI geq_arctic,
            (mat_geI gt_arctic,
              (false, (arc_mat_name, mat_arc_posI pos_arctic)))))))))))
                                    aa (snd poly)))))))
           (fn typea => fn pi =>
             (case typea of Natural => Int_carrier (map (lefta o lefta) pi)
                | Arctic => Arctic_carrier (map (righta o lefta) pi)
                | Int_mat n =>
                  Int_mat_carrier (minus_nat n one_nat, map (lefta o righta) pi)
                | Arc_mat n =>
                  Arc_mat_carrier
                    (minus_nat n one_nat, map (righta o righta) pi))))
         id)];

fun xml2ndChoice reqName p1 choiceName p2 combi (Xml (name, atts, cs, text)) =
  (if eqop (eq_lista eq_char) name reqName andalso
        (null atts andalso
          (less_eq_nat (nat (2 : IntInf.int)) (size_list cs) andalso
            is_none text))
    then bind (p1 (nth cs Zero_nat))
           (fn a =>
             bind (xml2bool choiceName (nth cs one_nat))
               (fn choice =>
                 (if choice
                   then bind (p2 (nth cs (minus_nat (size_list cs) one_nat)))
                          (fn b => return (combi a (SOME b)))
                   else return (combi a NONE))))
    else error (concat
                 [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ", #"e", #"l",
                    #"e", #"m", #"e", #"n", #"t", #" ", #"(", #"e", #"x", #"p",
                    #"e", #"c", #"t", #"e", #"d", #" "],
                   reqName, [#")"], newline,
                   sshow_xml (Xml (name, atts, cs, text))]));

fun xml2dpProof ld xml2name x =
  xmlOptions
    [([#"p", #"I", #"s", #"E", #"m", #"p", #"t", #"y"],
       xml0elements [#"p", #"I", #"s", #"E", #"m", #"p", #"t", #"y"] PisEmpty),
      ([#"d", #"e", #"p", #"G", #"r", #"a", #"p", #"h", #"P", #"r", #"o", #"c"],
        xmlManyElements
          [#"d", #"e", #"p", #"G", #"r", #"a", #"p", #"h", #"P", #"r", #"o",
            #"c"]
          (xml2ndChoice [#"c", #"o", #"m", #"p", #"o", #"n", #"e", #"n", #"t"]
            (xml1element [#"d", #"p", #"s"] (xml2rules (xml2name ld)) id)
            [#"r", #"e", #"a", #"l", #"S", #"c", #"c"] (xml2dpProof ld xml2name)
            (fn dps => fn prfOpt => (prfOpt, dps)))
          DepGraphProc),
      ([#"r", #"e", #"d", #"P", #"a", #"i", #"r", #"P", #"r", #"o", #"c"],
        xml3elements
          [#"r", #"e", #"d", #"P", #"a", #"i", #"r", #"P", #"r", #"o", #"c"]
          (xml2redpair (xml2name ld))
          (xml1element [#"d", #"p", #"s"] (xml2rules (xml2name ld)) id)
          (xml2dpProof ld xml2name)
          (fn a => fn b => fn c => RedPairProc (a, b, c))),
      ([#"r", #"e", #"d", #"P", #"a", #"i", #"r", #"U", #"r", #"P", #"r", #"o",
         #"c"],
        xml4elements
          [#"r", #"e", #"d", #"P", #"a", #"i", #"r", #"U", #"r", #"P", #"r",
            #"o", #"c"]
          (xml2redpair (xml2name ld))
          (xml1element [#"d", #"p", #"s"] (xml2rules (xml2name ld)) id)
          (xml1element
            [#"u", #"s", #"a", #"b", #"l", #"e", #"R", #"u", #"l", #"e", #"s"]
            (xml2rules (xml2name ld)) id)
          (xml2dpProof ld xml2name)
          (fn a => fn b => fn c => fn d => RedPairUrProc (a, b, c, d))),
      ([#"m", #"o", #"n", #"o", #"R", #"e", #"d", #"P", #"a", #"i", #"r", #"P",
         #"r", #"o", #"c"],
        xml4elements
          [#"m", #"o", #"n", #"o", #"R", #"e", #"d", #"P", #"a", #"i", #"r",
            #"P", #"r", #"o", #"c"]
          (xml2redpair (xml2name ld))
          (xml1element [#"d", #"p", #"s"] (xml2rules (xml2name ld)) id)
          (xml1element [#"t", #"r", #"s"] (xml2rules (xml2name ld)) id)
          (xml2dpProof ld xml2name)
          (fn a => fn b => fn c => fn d => MonoRedPairProc (a, b, c, d))),
      ([#"m", #"o", #"n", #"o", #"R", #"e", #"d", #"P", #"a", #"i", #"r", #"U",
         #"r", #"P", #"r", #"o", #"c"],
        xml5elements
          [#"m", #"o", #"n", #"o", #"R", #"e", #"d", #"P", #"a", #"i", #"r",
            #"U", #"r", #"P", #"r", #"o", #"c"]
          (xml2redpair (xml2name ld))
          (xml1element [#"d", #"p", #"s"] (xml2rules (xml2name ld)) id)
          (xml1element [#"t", #"r", #"s"] (xml2rules (xml2name ld)) id)
          (xml1element
            [#"u", #"s", #"a", #"b", #"l", #"e", #"R", #"u", #"l", #"e", #"s"]
            (xml2rules (xml2name ld)) id)
          (xml2dpProof ld xml2name)
          (fn a => fn b => fn c => fn d => fn e =>
            MonoRedPairUrProc (a, b, c, d, e))),
      ([#"s", #"e", #"m", #"l", #"a", #"b", #"P", #"r", #"o", #"c"],
        xml4elements
          [#"s", #"e", #"m", #"l", #"a", #"b", #"P", #"r", #"o", #"c"]
          (xml2sl_inter (xml2name ld))
          (xml1element [#"d", #"p", #"s"] (xml2rules (xml2name (Suc ld))) id)
          (xml1element [#"t", #"r", #"s"] (xml2rules (xml2name (Suc ld))) id)
          (xml2dpProof (Suc ld) xml2name)
          (fn a =>
            (case a of NONE => (fn aa => fn b => fn c => RootlabProc (aa, b, c))
               | SOME i =>
                 (fn aa => fn b => fn c =>
                   SemlabProc (i, id_labeling, aa, b, c))))),
      ([#"u", #"n", #"l", #"a", #"b", #"P", #"r", #"o", #"c"],
        xml3elements [#"u", #"n", #"l", #"a", #"b", #"P", #"r", #"o", #"c"]
          (xml1element [#"d", #"p", #"s"]
            (xml2rules (xml2name (minus_nat ld one_nat))) id)
          (xml1element [#"t", #"r", #"s"]
            (xml2rules (xml2name (minus_nat ld one_nat))) id)
          (xml2dpProof (minus_nat ld one_nat) xml2name)
          (fn a => fn b => fn c => UnlabProc (a, b, c))),
      ([#"s", #"u", #"b", #"t", #"e", #"r", #"m", #"P", #"r", #"o", #"c"],
        xml3elements
          [#"s", #"u", #"b", #"t", #"e", #"r", #"m", #"P", #"r", #"o", #"c"]
          (xml2sp (xml2name ld))
          (xml1element [#"d", #"p", #"s"] (xml2rules (xml2name ld)) id)
          (xml2dpProof ld xml2name)
          (fn a => fn b => fn c => SpscProc (a, b, c))),
      ([#"s", #"i", #"z", #"e", #"C", #"h", #"a", #"n", #"g", #"e", #"C", #"r",
         #"i", #"t", #"e", #"r", #"i", #"o", #"n"],
        xml1manyElements
          [#"s", #"i", #"z", #"e", #"C", #"h", #"a", #"n", #"g", #"e", #"C",
            #"r", #"i", #"t", #"e", #"r", #"i", #"o", #"n"]
          (xmlOptions
            [([#"s", #"u", #"b", #"t", #"e", #"r", #"m", #"C", #"r", #"i", #"t",
                #"e", #"r", #"i", #"o", #"n"],
               xml0elements
                 [#"s", #"u", #"b", #"t", #"e", #"r", #"m", #"C", #"r", #"i",
                   #"t", #"e", #"r", #"i", #"o", #"n"]
                 NONE),
              ([#"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n", #"P",
                 #"a", #"i", #"r"],
                xml2elements
                  [#"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n", #"P",
                    #"a", #"i", #"r"]
                  (xml2redpair (xml2name ld))
                  (xml1element
                    [#"u", #"s", #"a", #"b", #"l", #"e", #"R", #"u", #"l", #"e",
                      #"s"]
                    (xml2rules (xml2name ld)) id)
                  (fn redp => fn ur => SOME (redp, ur)))])
          (xml2scg (xml2name ld))
          (fn a =>
            (case a of NONE => SizeChangeSubTerm
               | SOME redp_ur =>
                 (fn aa => SizeChangeRedPair (fst redp_ur, snd redp_ur, aa)))))]
    x;

fun foldR f [] b = b
  | foldR f (x :: xs) b = f x xs (foldR f xs b);

fun mul_pvarsI A_ uu uv a [] = []
  | mul_pvarsI A_ ze ti a ((x, b) :: vas) =
    let
      val p = ti a b;
      val res = mul_pvarsI A_ ze ti a vas;
    in
      (if eqop A_ p ze then res else (x, p) :: res)
    end;

datatype ('a, 'b) l_poly = LPoly of 'b * ('a * 'b) list;

fun mul_lpolyI A_ ze ti a (LPoly (b, vas)) =
  LPoly (ti a b, mul_pvarsI A_ ze ti a vas);

fun list_sumI ze pl [] = ze
  | list_sumI ze pl (x :: xs) = pl x (list_sumI ze pl xs);

fun add_varI A_ B_ uu uv x a [] = [(x, a)]
  | add_varI A_ B_ ze pl x a ((y, b) :: vas) =
    (if eqop B_ x y
      then let
             val s = pl a b;
           in
             (if eqop A_ s ze then vas else (x, s) :: vas)
           end
      else (y, b) :: add_varI A_ B_ ze pl x a vas);

fun sum_pvarsI A_ B_ uu uv [] vbs = vbs
  | sum_pvarsI A_ B_ ze pl ((x, a) :: vas) vbs =
    (if eqop A_ a ze then sum_pvarsI A_ B_ ze pl vas vbs
      else sum_pvarsI A_ B_ ze pl vas (add_varI A_ B_ ze pl x a vbs));

fun sum_lpolyI A_ B_ ze pl (LPoly (a, vas)) (LPoly (b, vbs)) =
  LPoly (pl a b, sum_pvarsI A_ B_ ze pl vas vbs);

fun l_poly_case f1 (LPoly (a, lista)) = f1 a lista;

fun pleftI A_ C_ ze on pl ti mz pi (Var x) = LPoly (ze, [(x, on)])
  | pleftI A_ C_ ze on pl ti mz pi (Fun (f, ts)) =
    let
      val a = pi f (size_list ts);
      val (c, asa) = a;
      val LPoly (d, aa) =
        sum_lpolyI A_ C_ ze pl (LPoly (c, []))
          (list_sumI (LPoly (ze, [])) (sum_lpolyI A_ C_ ze pl)
            (map (fn at =>
                   mul_lpolyI A_ ze ti (fst at)
                     (pleftI A_ C_ ze on pl ti mz pi (snd at)))
              (zip asa ts)));
    in
      (case aa of [] => LPoly (mz d, [])
         | ab :: lista => LPoly (d, ab :: lista))
    end;

fun check_supteq (F1_, F2_) (V1_, V2_) lr =
  catch (checkB
          (eqop (eq_terma F1_ V1_) (fst lr) (snd lr) orelse
            checkSupterm F1_ V1_ (fst lr) (snd lr))
          [[#"t", #"h", #"e", #" ", #"r", #"h", #"s", #" ", #"i", #"s", #" ",
             #"n", #"o", #"t", #" ", #"a", #" ", #"s", #"u", #"b", #"t", #"e",
             #"r", #"m", #" ", #"o", #"f", #" ", #"t", #"h", #"e", #" ", #"l",
             #"h", #"s"]])
    (fn x =>
      error (concat
               [[#"C", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                  #"o", #"r", #"i", #"e", #"n", #"t", #" ", #"r", #"u", #"l",
                  #"e", #" "],
                 sshow_term F2_ V2_ (fst lr), [#" ", #"-", #">", #" "],
                 sshow_term F2_ V2_ (snd lr)]
              :: x));

fun xml2trsInput xml2name =
  xml1element [#"i", #"n", #"p", #"u", #"t"]
    (xml1element [#"t", #"r", #"s", #"I", #"n", #"p", #"u", #"t"]
      (xml1element [#"t", #"r", #"s"] (xml2rules xml2name) id) id)
    id;

datatype ('a, 'b, 'c) tRSProof =
  DPTrans of
    ((('a, 'b) lab shp, 'c) term * (('a, 'b) lab shp, 'c) term) list *
      (('a, 'b) lab shp, 'b, 'c) dPProof
  | RuleRemoval of
      ('a, 'b) lab redpair *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Unlab of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Semlab of
      ('a, 'b) sl_inter *
        (('a, 'b) lab -> nat list -> ('b, (('a, 'b) lab list)) either) *
        ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | Rootlab of
      ((('a, 'b) lab, 'c) term * (('a, 'b) lab, 'c) term) list *
        ('a, 'b, 'c) tRSProof
  | RisEmpty;

fun xml2trsProof xml2name xml2sname xml2lsname x =
  xmlOptions
    [([#"r", #"I", #"s", #"E", #"m", #"p", #"t", #"y"],
       xml0elements [#"r", #"I", #"s", #"E", #"m", #"p", #"t", #"y"] RisEmpty),
      ([#"d", #"p", #"T", #"r", #"a", #"n", #"s"],
        xml3elements [#"d", #"p", #"T", #"r", #"a", #"n", #"s"]
          (xml1element [#"d", #"p", #"s"] (xml2rules xml2sname) id)
          (xml2bool
            [#"m", #"a", #"r", #"k", #"e", #"d", #"S", #"y", #"m", #"b", #"o",
              #"l", #"s"])
          (xml2dpProof Zero_nat xml2lsname)
          (fn dps => fn uu => (fn a => DPTrans (dps, a)))),
      ([#"r", #"u", #"l", #"e", #"R", #"e", #"m", #"o", #"v", #"a", #"l"],
        xml3elements
          [#"r", #"u", #"l", #"e", #"R", #"e", #"m", #"o", #"v", #"a", #"l"]
          (xml2redpair xml2name)
          (xml1element [#"t", #"r", #"s"] (xml2rules xml2name) id)
          (xml2trsProof xml2name xml2sname xml2lsname)
          (fn a => fn b => fn c => RuleRemoval (a, b, c))),
      ([#"u", #"n", #"l", #"a", #"b"],
        xml2elements [#"u", #"n", #"l", #"a", #"b"]
          (xml1element [#"t", #"r", #"s"] (xml2rules xml2name) id)
          (xml2trsProof xml2name xml2sname xml2lsname)
          (fn a => fn b => Unlab (a, b))),
      ([#"s", #"e", #"m", #"l", #"a", #"b"],
        xml3elements [#"s", #"e", #"m", #"l", #"a", #"b"]
          (xml2sl_inter xml2name)
          (xml1element [#"t", #"r", #"s"] (xml2rules xml2name) id)
          (xml2trsProof xml2name xml2sname xml2lsname)
          (fn a =>
            (case a of NONE => (fn aa => fn b => Rootlab (aa, b))
               | SOME i => (fn aa => fn b => Semlab (i, id_labeling, aa, b)))))]
    x;

fun checkF f g x = (if f x then return () else error (g x));

fun checkAllMI f xs =
  foldr (fn x => fn iq =>
          bind (catch (f x) (fn xa => error (x, xa))) (fn xa => iq))
    xs (return ());

fun check_edges ss g c d =
  catch (checkAllMI
          (fn n =>
            catch (checkAllMI
                    (fn m =>
                      checkB (not (g (m, n)))
                        (concat
                          [[#"e", #"d", #"g", #"e", #" ", #"f", #"r", #"o",
                             #"m", #" "],
                            ss m, [#" ", #"t", #"o", #" "], ss n]))
                    d)
              (fn x => error (snd x)))
          c)
    (fn x => error (snd x));

fun prightI A_ C_ ze on pl ti mz pi (Var x) = LPoly (ze, [(x, on)])
  | prightI A_ C_ ze on pl ti mz pi (Fun (f, ts)) =
    let
      val a = pi f (size_list ts);
      val (c, asa) = a;
      val LPoly (d, aa) =
        sum_lpolyI A_ C_ ze pl (LPoly (c, []))
          (list_sumI (LPoly (ze, [])) (sum_lpolyI A_ C_ ze pl)
            (map (fn at =>
                   mul_lpolyI A_ ze ti (fst at)
                     (prightI A_ C_ ze on pl ti mz pi (snd at)))
              (zip asa ts)));
    in
      LPoly (mz d, aa)
    end;

datatype ('a, 'b) tRSDisproof =
  Loop of
    (('a, 'b) term list *
      (('a, 'b) ctxt * (('a, 'b) term * ('a, 'b) term) list))
  | NotWellFormed;

fun funs_term_list' (Var uu) = []
  | funs_term_list' (Fun (f, ts)) = f :: flat_map funs_term_list' ts;

fun funs_rule_list' r =
  append (funs_term_list' (fst r)) (funs_term_list' (snd r));

fun funs_trs_list F_ trs = remdups F_ (flat_map funs_rule_list' trs);

fun xml2plainName x =
  xmlOptions
    [([#"n", #"a", #"m", #"e"],
       xmlChange (xml2text [#"n", #"a", #"m", #"e"]) UnLab),
      ([#"l", #"a", #"b", #"e", #"l", #"e", #"d", #"S", #"y", #"m", #"b", #"o",
         #"l"],
        xml2elements
          [#"l", #"a", #"b", #"e", #"l", #"e", #"d", #"S", #"y", #"m", #"b",
            #"o", #"l"]
          xml2plainName (xml2label xml2plainName)
          (fn f => fn a =>
            (case a of Left aa => Lab (f, aa) | Right aa => FunLab (f, aa))))]
    x;

fun xml2sharpName x =
  (if eqop (eq_lista eq_char) (xmlGetName x) [#"s", #"h", #"a", #"r", #"p"]
    then xml1element [#"s", #"h", #"a", #"r", #"p"] xml2plainName Sharp x
    else xmlChange xml2plainName Plain x);

fun length_unique A_ [] = Zero_nat
  | length_unique A_ (x :: xs) =
    (if member A_ x xs then length_unique A_ xs else Suc (length_unique A_ xs));

datatype certificationResult = Certified of bool | Error of char list;

fun vars_term_list' (Var x) = [x]
  | vars_term_list' (Fun (uu, ts)) = flat_map vars_term_list' ts;

fun vars_term_list B_ t = remdups B_ (vars_term_list' t);

fun enum_vectors c [] = [[]]
  | enum_vectors c (x :: xs) =
    let
      val a = enum_vectors c xs;
    in
      flat_map (fn vec => map (fn ca => (x, ca) :: vec) c) a
    end;

fun assign_from_vector A_ vec x = the (lookup A_ x vec);

fun eq_lab A_ B_ (FunLab (fa, ls)) (FunLab (g, ks)) =
  eq_lab A_ B_ fa g andalso list_all2 (eq_lab A_ B_) ls ks
  | eq_lab A_ B_ (UnLab c) (FunLab (a, ba)) = false
  | eq_lab A_ B_ (UnLab c) (Lab (a, b)) = false
  | eq_lab A_ B_ (FunLab (ca, d)) (Lab (a, b)) = false
  | eq_lab A_ B_ (FunLab (a, ba)) (UnLab c) = false
  | eq_lab A_ B_ (Lab (a, b)) (UnLab c) = false
  | eq_lab A_ B_ (Lab (a, b)) (FunLab (ca, d)) = false
  | eq_lab A_ B_ (UnLab f) (UnLab f') = eq A_ f f'
  | eq_lab A_ B_ (Lab (lab, l)) (Lab (lab', l')) =
    eq_lab A_ B_ lab lab' andalso eq B_ l l';

fun eq_laba A_ B_ = {eq = eq_lab A_ B_} : ('a, 'b) lab eq;

fun sshow_lab C_ D_ (UnLab f) = sshow C_ f
  | sshow_lab C_ D_ (Lab (fa, l)) =
    append (append (append (sshow_lab C_ D_ fa) [#"["]) (sshow D_ l)) [#"]"]
  | sshow_lab C_ D_ (FunLab (fa, la)) =
    append
      (append (append (sshow_lab C_ D_ fa) [#"["])
        (flat_map (sshow_lab C_ D_) la))
      [#"]"];

fun sshow_laba A_ B_ = {sshow = sshow_lab A_ B_} : ('a, 'b) lab sshow;

fun label f (Left l) = Lab (f, l)
  | label f (Right la) = FunLab (f, la);

fun eval_lab i l alpha (Var x) = (alpha x, Var x)
  | eval_lab i l alpha (Fun (f, ts)) =
    let
      val clts = map (eval_lab i l alpha) ts;
      val cs = map fst clts;
      val c = i f cs;
      val lts = map snd clts;
      val la = l f cs;
    in
      (c, Fun (label f la, lts))
    end;

fun check_sl_rule_ass (A1_, A2_) (B1_, B2_) (C1_, C2_) (D1_, D2_) mc i la lR
  alpha (l, r) =
  let
    val cl_ll = eval_lab i la alpha l;
    val cr_lr = eval_lab i la alpha r;
  in
    bind (checkB (implies mc (eqop C1_ (fst cl_ll) (fst cr_lr)))
           (append
             (append
               (append
                 (append
                   (append [#"r", #"u", #"l", #"e", #" "]
                     (showRule (sshow_laba A2_ B2_) D2_ (l, r)))
                   [#" ", #"v", #"i", #"o", #"l", #"a", #"t", #"e", #"s", #" ",
                     #"m", #"o", #"d", #"e", #"l", #" ", #"c", #"o", #"n", #"d",
                     #"i", #"t", #"i", #"o", #"n", #",", #" ", #"[", #"l", #"h",
                     #"s", #"]", #" ", #"=", #" "])
                 (sshow C2_ (fst cl_ll)))
               [#",", #" ", #"[", #"r", #"h", #"s", #"]", #" ", #"=", #" "])
             (sshow C2_ (fst cr_lr))))
      (fn x =>
        checkB
          (member
            (eqb (eq_terma (eq_laba A1_ B1_) D1_)
              (eq_terma (eq_laba A1_ B1_) D1_))
            (snd cl_ll, snd cr_lr) lR)
          (append
            (append
              [#"l", #"a", #"b", #"e", #"l", #"e", #"d", #" ", #"r", #"u", #"l",
                #"e", #" "]
              (showRule (sshow_laba A2_ B2_) D2_ (snd cl_ll, snd cr_lr)))
            [#" ", #"m", #"i", #"s", #"s", #"i", #"n", #"g"]))
  end;

fun check_sl_rule (A1_, A2_) (B1_, B2_) (C1_, C2_) (D1_, D2_) i la c mc lR
  (l, r) =
  catch (checkAllMI
          (fn alpha =>
            check_sl_rule_ass (A1_, A2_) (B1_, B2_) (C1_, C2_) (D1_, D2_) mc i
              la lR alpha (l, r))
          (map (assign_from_vector D1_)
            (enum_vectors c
              (list_union D1_ (vars_term_list D1_ l) (vars_term_list D1_ r)))))
    (fn x => error (snd x));

fun check_sl_model_lab_trs (F1_, F2_) (L1_, L2_) (C1_, C2_) (V1_, V2_) i l c lR
  r = catch (checkAllMI
              (check_sl_rule (F1_, F2_) (L1_, L2_) (C1_, C2_) (V1_, V2_) i l c
                true lR)
              r)
        (fn x => error (snd x));

fun rl_fun uu cs = Right cs;

fun rl_inter_list A_ fs d g uu = (if member A_ g fs then g else d);

fun check_sl_lab_trs (F1_, F2_) (L1_, L2_) (C1_, C2_) (V1_, V2_) i l c lP p =
  catch (checkAllMI
          (check_sl_rule (F1_, F2_) (L1_, L2_) (C1_, C2_) (V1_, V2_) i l c false
            lP)
          p)
    (fn x => error (snd x));

fun get_lab (Lab (f, l)) = Left l
  | get_lab (FunLab (f, la)) = Right la;

fun unlab (Lab (f, l)) = f
  | unlab (FunLab (f, la)) = f
  | unlab (UnLab v) = UnLab v;

fun map_term fg (Var x) = Var x
  | map_term fg (Fun (f, ts)) = Fun (fg f, map (map_term fg) ts);

fun check_sl_only_lab_and_decr (B1_, B2_) (A1_, A2_) (C1_, C2_) gr r =
  (fn xs =>
    catch (checkAllMI
            (fn a as (ll, lr) =>
              let
                val l = map_term unlab ll;
                val ra = map_term unlab lr;
              in
                checkB
                  (if eqop (eq_terma (eq_laba A1_ B1_) C1_) l ra
                    then not (is_var ll) andalso
                           (not (is_var lr) andalso
                             gr (get_lab (the (root ll)),
                                  get_lab (the (root lr))))
                           orelse
                           eqop (eq_terma (eq_laba A1_ B1_) C1_) ll lr andalso
                             member
                               (eqb (eq_terma (eq_laba A1_ B1_) C1_)
                                 (eq_terma (eq_laba A1_ B1_) C1_))
                               (l, ra) r
                    else member
                           (eqb (eq_terma (eq_laba A1_ B1_) C1_)
                             (eq_terma (eq_laba A1_ B1_) C1_))
                           (l, ra) r)
                  (append
                    (append
                      [#"l", #"a", #"b", #"e", #"l", #"e", #"d", #" ", #"r",
                        #"u", #"l", #"e", #" "]
                      (showRule (sshow_laba A2_ B2_) C2_ (l, ra)))
                    [#" ", #"i", #"s", #" ", #"n", #"e", #"i", #"t", #"h", #"e",
                      #"r", #" ", #"d", #"e", #"c", #"r", #"e", #"a", #"s",
                      #"i", #"n", #"g", #" ", #"n", #"o", #"r", #" ", #"a",
                      #" ", #"l", #"a", #"b", #"e", #"l", #"e", #"d", #" ",
                      #"v", #"e", #"r", #"s", #"i", #"o", #"n", #" ", #"o",
                      #"f", #" ", #"o", #"r", #"i", #"g", #"i", #"n", #"a",
                      #"l", #" ", #"r", #"u", #"l", #"e"])
              end)
            xs)
      (fn x => error (snd x)));

fun check_sem_lab_tt (F1_, F2_) (L1_, L2_) (V1_, V2_) check_lab check_model_lab
  gr i l r lR =
  catch (bind (check_model_lab lR r)
          (fn x =>
            check_sl_only_lab_and_decr (L1_, L2_) (F1_, F2_) (V1_, V2_) gr r
              lR))
    (fn x =>
      error [[#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"i", #"t",
               #"h", #" ", #"l", #"a", #"b", #"e", #"l", #"e", #"d", #" ", #"T",
               #"R", #"S", #":"],
              x]);

fun rl_tt (F1_, F2_) (L1_, L2_) r lR =
  let
    val fs = funs_trs_list (eq_laba F1_ L1_) r;
  in
    bind (checkB (not (null fs))
           [[#"r", #"o", #"o", #"t", #"-", #"l", #"a", #"b", #"e", #"l", #"i",
              #"n", #"g", #" ", #"r", #"e", #"q", #"u", #"i", #"r", #"e", #"s",
              #" ", #"a", #"t", #" ", #"l", #"e", #"a", #"s", #"t", #" ", #"o",
              #"n", #"e", #" ", #"f", #"u", #"n", #"c", #"t", #"i", #"o", #"n",
              #" ", #"s", #"y", #"m", #"b", #"o", #"l", #" ", #"i", #"n", #" ",
              #"t", #"h", #"e", #" ", #"s", #"i", #"g", #"n", #"a", #"t", #"u",
              #"r", #"e"]])
      (fn x =>
        let
          val d = hd fs;
        in
          check_sem_lab_tt (F1_, F2_) (L1_, L2_)
            (eq_lista eq_char, sshow_lista sshow_chara)
            (check_sl_lab_trs (F1_, F2_) (L1_, L2_)
              (eq_laba F1_ L1_, sshow_laba F2_ L2_)
              (eq_lista eq_char, sshow_lista sshow_chara)
              (rl_inter_list (eq_laba F1_ L1_) fs d) rl_fun fs)
            (check_sl_model_lab_trs (F1_, F2_) (L1_, L2_)
              (eq_laba F1_ L1_, sshow_laba F2_ L2_)
              (eq_lista eq_char, sshow_lista sshow_chara)
              (rl_inter_list (eq_laba F1_ L1_) fs d) rl_fun fs)
            empty (rl_inter_list (eq_laba F1_ L1_) fs d) rl_fun r lR
        end)
  end;

fun indent xs = newline :: map (append [#" ", #" "]) xs;

val terminationGoal : xml =
  Xml ([#"g", #"o", #"a", #"l"], [],
        [Xml ([#"t", #"e", #"r", #"m", #"i", #"n", #"a", #"t", #"i", #"o",
                #"n"],
               [], [], NONE)],
        NONE)

fun xml2substitution xml2name =
  xmlManyElements
    [#"s", #"u", #"b", #"s", #"t", #"i", #"t", #"u", #"t", #"i", #"o", #"n"]
    (xml2elements [#"s", #"u", #"b", #"s", #"t", #"E", #"n", #"t", #"r", #"y"]
      xml2var (xml2term xml2name) (fn a => fn b => (a, b)))
    id;

fun xmlMany2Elements reqName p1 p2 p3 combi (Xml (name, atts, cs, text)) =
  let
    val ds = rev cs;
  in
    (if eqop (eq_lista eq_char) name reqName andalso
          (null atts andalso
            (is_none text andalso
              less_eq_nat (nat (2 : IntInf.int)) (size_list cs)))
      then bind (mapMonad p1 (rev (tl (tl ds))))
             (fn a =>
               bind (p2 (nth ds one_nat))
                 (fn b => bind (p3 (nth ds Zero_nat)) (return o combi a b)))
      else error (concat
                   [[#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                      #"t", #"r", #"a", #"n", #"s", #"f", #"o", #"r", #"m",
                      #" ", #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l",
                      #"o", #"w", #"i", #"n", #"g", #" ", #"x", #"m", #"l",
                      #" ", #"e", #"l", #"e", #"m", #"e", #"n", #"t", #" ",
                      #"(", #"e", #"x", #"p", #"e", #"c", #"t", #"e", #"d",
                      #" "],
                     reqName, [#")"], newline,
                     sshow_xml (Xml (name, atts, cs, text))]))
  end;

fun xml2trsDisproof xml2name =
  xmlOptions
    [([#"v", #"a", #"r", #"i", #"a", #"b", #"l", #"e", #"C", #"o", #"n", #"d",
        #"i", #"t", #"i", #"o", #"n"],
       xml0elements
         [#"v", #"a", #"r", #"i", #"a", #"b", #"l", #"e", #"C", #"o", #"n",
           #"d", #"i", #"t", #"i", #"o", #"n"]
         NotWellFormed),
      ([#"l", #"o", #"o", #"p"],
        xmlMany2Elements [#"l", #"o", #"o", #"p"]
          (xml3elements [#"r", #"e", #"d", #"e", #"x"] (xml2term xml2name)
            xml2pos (xml2rule xml2name) (fn t => fn p => fn r => t))
          (xml2substitution xml2name) (xml2context xml2name)
          (fn ts => fn sigma => fn c => Loop (ts, (c, sigma))))];

val nonTerminationGoal : xml =
  Xml ([#"g", #"o", #"a", #"l"], [],
        [Xml ([#"n", #"o", #"n", #"T", #"e", #"r", #"m", #"i", #"n", #"a", #"t",
                #"i", #"o", #"n"],
               [], [], NONE)],
        NONE)

datatype ('a, 'b, 'c) certificationProblem =
  CertificationProblem of
    bool * (('a, 'c) term * ('a, 'c) term) list *
      (('a, 'b, 'c) tRSProof, ('a, 'c) tRSDisproof) either;

fun xml2certProblem xml2uname xml2name xml2sname xml2lsname
  (Xml (name, uu, cs, text)) =
  (if eqop (eq_lista eq_char) name
        [#"c", #"e", #"r", #"t", #"i", #"f", #"i", #"c", #"a", #"t", #"i", #"o",
          #"n", #"P", #"r", #"o", #"b", #"l", #"e", #"m"]
        andalso
        (is_none text andalso less_eq_nat (nat (3 : IntInf.int)) (size_list cs))
    then bind (xml2trsInput xml2uname (nth cs one_nat))
           (fn trs =>
             (if eqop eq_xmla (nth cs Zero_nat) terminationGoal
               then bind (xml1element [#"p", #"r", #"o", #"o", #"f"]
                           (xml2trsProof xml2name xml2sname xml2lsname) id
                           (nth cs (nat (2 : IntInf.int))))
                      (fn p =>
                        return (CertificationProblem (true, trs, Left p)))
               else (if eqop eq_xmla (nth cs Zero_nat) nonTerminationGoal
                      then bind (xml1element [#"p", #"r", #"o", #"o", #"f"]
                                  (xml2trsDisproof xml2uname) id
                                  (nth cs (nat (2 : IntInf.int))))
                             (fn p =>
                               return
                                 (CertificationProblem (false, trs, Right p)))
                      else error [#"u", #"n", #"k", #"n", #"o", #"w", #"n",
                                   #" ", #"g", #"o", #"a", #"l", #" ", #"t",
                                   #"y", #"p", #"e", #" ", #"i", #"n", #" ",
                                   #"c", #"e", #"r", #"t", #"i", #"f", #"i",
                                   #"c", #"a", #"t", #"i", #"o", #"n", #" ",
                                   #"p", #"r", #"o", #"b", #"l", #"e", #"m"])))
    else error [#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"p", #"a", #"r", #"s", #"e", #" ", #"r", #"o", #"o", #"t",
                 #" ", #"e", #"l", #"e", #"m", #"e", #"n", #"t"]);

fun computeDPsOfTerm A_ B_ d l lshp (Var uu) = []
  | computeDPsOfTerm A_ B_ d l lshp (Fun (f, ts)) =
    append
      (if member A_ f d andalso not (checkSupterm A_ B_ l (Fun (f, ts)))
        then [(lshp, sharpa (Fun (f, ts)))] else [])
      (foldr
        (fn t =>
          list_union (eqb (eq_terma (eq_shpa A_) B_) (eq_terma (eq_shpa A_) B_))
            (computeDPsOfTerm A_ B_ d l lshp t))
        ts []);

fun computeDPsOfRule A_ B_ d (Fun (f, ss), r) =
  computeDPsOfTerm A_ B_ d (Fun (f, ss)) (sharpa (Fun (f, ss))) r
  | computeDPsOfRule A_ B_ d (Var x, r) = [];

fun computeDPs A_ B_ r =
  let
    val d = defs_list A_ r;
  in
    foldr (fn lr =>
            list_union
              (eqb (eq_terma (eq_shpa A_) B_) (eq_terma (eq_shpa A_) B_))
              (computeDPsOfRule A_ B_ d lr))
      r []
  end;

fun checkSubset A_ xs ys =
  catch (checkAllMI (checkF (fn x => member A_ x ys) id) xs)
    (fn x => error (snd x));

fun checkWfTRS F_ (V1_, V2_) r =
  catch (bind (catch
                (checkAllMI
                  (fn lr =>
                    checkB (not (is_var (fst lr)))
                      (concat
                        [[#"t", #"h", #"e", #" ", #"l", #"e", #"f", #"t", #" ",
                           #"s", #"i", #"d", #"e", #" ", #"o", #"f", #" ", #"r",
                           #"u", #"l", #"e", #" "],
                          showRule F_ V2_ lr,
                          [#" ", #"i", #"s", #" ", #"a", #" ", #"v", #"a", #"r",
                            #"i", #"a", #"b", #"l", #"e"]]))
                  r)
                (fn x => error (snd x)))
          (fn x =>
            catch (checkAllMI
                    (fn lr =>
                      catch (checkSubset V1_ (vars_term_list V1_ (snd lr))
                              (vars_term_list V1_ (fst lr)))
                        (fn xa =>
                          error (concat
                                  [[#"t", #"h", #"e", #" ", #"v", #"a", #"r",
                                     #"i", #"a", #"b", #"l", #"e", #" "],
                                    sshow V2_ xa,
                                    [#" ", #"o", #"c", #"c", #"u", #"r", #"s",
                                      #" ", #"o", #"n", #"l", #"y", #" ", #"i",
                                      #"n", #" ", #"t", #"h", #"e", #" ", #"r",
                                      #"i", #"g", #"h", #"t", #"-", #"h", #"a",
                                      #"n", #"d", #" ", #"s", #"i", #"d", #"e",
                                      #" ", #"o", #"f", #" ", #"r", #"u", #"l",
                                      #"e", #" "],
                                    showRule F_ V2_ lr])))
                    r)
              (fn xa => error (snd xa))))
    (fn x =>
      error [[#"t", #"h", #"e", #" ", #"T", #"R", #"S", #" ", #"i", #"s", #" ",
               #"n", #"o", #"t", #" ", #"w", #"e", #"l", #"l", #"-", #"f", #"o",
               #"r", #"m", #"e", #"d"],
              x]);

fun checkDPs (F1_, F2_) (V1_, V2_) r p =
  catch (bind (checkWfTRS F2_ (V1_, V2_) r)
          (fn x =>
            catch (checkSubset
                    (eqb (eq_terma (eq_shpa F1_) V1_)
                      (eq_terma (eq_shpa F1_) V1_))
                    (computeDPs F1_ V1_ r) p)
              (fn xa =>
                error [concat
                         [[#"t", #"h", #"e", #" ", #"D", #"P", #" "],
                           showRule (sshow_shpa F2_) V2_ xa,
                           [#" ", #"d", #"o", #"e", #"s", #" ", #"n", #"o",
                             #"t", #" ", #"a", #"p", #"p", #"e", #"a", #"r",
                             #" ", #"i", #"n", #" ", #"t", #"h", #"e", #" ",
                             #"D", #"P", #" ", #"p", #"r", #"o", #"b", #"l",
                             #"e", #"m"]]])))
    (fn x =>
      error ([#"t", #"h", #"e", #" ", #"D", #"P", #"-", #"t", #"r", #"a", #"n",
               #"s", #"f", #"o", #"r", #"m", #"a", #"t", #"i", #"o", #"n", #" ",
               #"i", #"s", #" ", #"n", #"o", #"t", #" ", #"a", #"p", #"p", #"l",
               #"i", #"e", #"d", #" ", #"c", #"o", #"r", #"r", #"e", #"c", #"t",
               #"l", #"y", #"."]
              :: x));

fun create_afI A_ B_ ze [] f n i = true
  | create_afI A_ B_ ze (((g, m), (c, coeffs)) :: rest) f n i =
    (if eqop B_ f g andalso eqop eq_nata m n
      then less_nat i (size_list coeffs) andalso not (eqop A_ (nth coeffs i) ze)
      else create_afI A_ B_ ze rest f n i);

val nat_name : char list = [#"n", #"a", #"t", #"u", #"r", #"a", #"l", #"s"]

fun check_sem_lab_proc (F1_, F2_) (L1_, L2_) (V1_, V2_) check_lab
  check_model_lab gr i l p r lP lR =
  bind (bind (catch (check_lab lP p)
               (fn x =>
                 error [[#"n", #"o", #"t", #" ", #"a", #"l", #"l", #" ", #"l",
                          #"a", #"b", #"e", #"l", #"e", #"d", #" ", #"D", #"P",
                          #"s", #" ", #"p", #"r", #"e", #"s", #"e", #"n", #"t",
                          #":"],
                         x]))
         (fn x =>
           catch (check_model_lab lR r)
             (fn xa =>
               error [[#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w",
                        #"i", #"t", #"h", #" ", #"l", #"a", #"b", #"e", #"l",
                        #"e", #"d", #" ", #"T", #"R", #"S", #":"],
                       xa])))
    (fn x =>
      catch (check_sl_only_lab_and_decr (L1_, L2_) (F1_, F2_) (V1_, V2_) gr r
              lR)
        (fn xa => error [xa]));

fun list_diff A_ xs [] = xs
  | list_diff A_ xs (y :: ys) = list_diff A_ (removeAll A_ y xs) ys;

fun rl_proc (F1_, F2_) (L1_, L2_) p r lP lR =
  let
    val fs =
      list_union (eq_laba F1_ L1_) (funs_trs_list (eq_laba F1_ L1_) r)
        (list_diff (eq_laba F1_ L1_) (funs_trs_list (eq_laba F1_ L1_) p)
          (defs_list (eq_laba F1_ L1_) p));
  in
    bind (checkB (not (null fs))
           [[#"r", #"o", #"o", #"t", #"-", #"l", #"a", #"b", #"e", #"l", #"i",
              #"n", #"g", #" ", #"r", #"e", #"q", #"u", #"i", #"r", #"e", #"s",
              #" ", #"a", #"t", #" ", #"l", #"e", #"a", #"s", #"t", #" ", #"o",
              #"n", #"e", #" ", #"f", #"u", #"n", #"c", #"t", #"i", #"o", #"n",
              #" ", #"s", #"y", #"m", #"b", #"o", #"l", #" ", #"i", #"n", #" ",
              #"t", #"h", #"e", #" ", #"s", #"i", #"g", #"n", #"a", #"t", #"u",
              #"r", #"e"]])
      (fn x =>
        let
          val d = hd fs;
        in
          check_sem_lab_proc (F1_, F2_) (L1_, L2_)
            (eq_lista eq_char, sshow_lista sshow_chara)
            (check_sl_lab_trs (F1_, F2_) (L1_, L2_)
              (eq_laba F1_ L1_, sshow_laba F2_ L2_)
              (eq_lista eq_char, sshow_lista sshow_chara)
              (rl_inter_list (eq_laba F1_ L1_) fs d) rl_fun fs)
            (check_sl_model_lab_trs (F1_, F2_) (L1_, L2_)
              (eq_laba F1_ L1_, sshow_laba F2_ L2_)
              (eq_lista eq_char, sshow_lista sshow_chara)
              (rl_inter_list (eq_laba F1_ L1_) fs d) rl_fun fs)
            empty (rl_inter_list (eq_laba F1_ L1_) fs d) rl_fun p r lP lR
        end)
  end;

fun xml2labSharpName Zero_nat = xmlChange xml2sharpName UnLab
  | xml2labSharpName (Suc n) =
    xml2elements
      [#"l", #"a", #"b", #"e", #"l", #"e", #"d", #"S", #"y", #"m", #"b", #"o",
        #"l"]
      (xml2labSharpName n) (xml2label (xml2labSharpName n))
      (fn f => fn a =>
        (case a of Left aa => Lab (f, aa) | Right aa => FunLab (f, aa)));

fun xmldoc2certProblem (Xmldoc (header, xml)) =
  debug [#"0"]
    [#"p", #"a", #"r", #"s", #"i", #"n", #"g", #" ", #"x", #"m", #"l", #" ",
      #"t", #"o", #" ", #"c", #"e", #"r", #"t", #"i", #"f", #"i", #"c", #"a",
      #"t", #"i", #"o", #"n", #" ", #"p", #"r", #"o", #"b", #"l", #"e", #"m"]
    (xml2certProblem (xml2text [#"n", #"a", #"m", #"e"]) xml2plainName
      xml2sharpName xml2labSharpName xml);

fun parseCertProblem s = bind (parseXMLDoc s) xmldoc2certProblem;

fun computeID r = map (fn lr => (plaina (fst lr), plaina (snd lr))) r;

fun get_comp_ge (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))
  = ge;

fun get_comp_gt (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap))))))))))
  = gt;

fun show_pvars (A1_, A2_) B_ on [] = []
  | show_pvars (A1_, A2_) B_ on ((x, c) :: vas) =
    append (if eqop A1_ on c then [] else sshow A2_ c) (sshow B_ x) ::
      show_pvars (A1_, A2_) B_ on vas;

fun show_lpoly (A1_, A2_) B_ ze on (LPoly (c, cs)) =
  (case show_pvars (A1_, A2_) B_ on cs of [] => sshow A2_ c
     | s :: ss =>
       append
         (append
           (if eqop A1_ c ze then []
             else append (sshow A2_ c) [#" ", #"+", #" "])
           s)
         (flat_map (append [#" ", #"+", #" "]) ss));

fun lookup_rest A_ x [] = NONE
  | lookup_rest A_ x ((y, c) :: ycs) =
    (if eqop A_ x y then SOME (c, ycs)
      else (case lookup_rest A_ x ycs of NONE => NONE
              | SOME a =>
                let
                  val (d, yccs) = a;
                in
                  SOME (d, (y, c) :: yccs)
                end));

fun search_len t s n =
  (if eqop (eq_lista eq_char) (take n t) s then SOME ([], drop n t)
    else (case t of [] => NONE
            | h :: tt =>
              (case search_len tt s n of NONE => NONE
                 | SOME a =>
                   let
                     val (bef, aft) = a;
                   in
                     SOME (h :: bef, aft)
                   end)));

fun reverse_trs r = map (fn lr => (snd lr, fst lr)) r;

fun check_pvarsI B_ ze ge vas [] =
  catch (checkAllMI (fn va => checkB (ge (snd va) ze) (fst va)) vas)
    (fn x => error (snd x))
  | check_pvarsI B_ ze ge vas ((x, b) :: vbs) =
    let
      val a =
        (case lookup_rest B_ x vas of NONE => (ze, vas)
           | SOME a => let
                         val (aa, ba) = a;
                       in
                         (aa, ba)
                       end);
      val (a', vas') = a;
    in
      bind (checkB (ge a' b) x) (fn xa => check_pvarsI B_ ze ge vas' vbs)
    end;

fun check_lpoly_sI (A1_, A2_) (B1_, B2_) mode ze on ge gt (LPoly (a, vas))
  (LPoly (b, vbs)) =
  catch (bind (checkB (gt a b)
                [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"h",
                  #"e", #"n", #" ", #"c", #"o", #"m", #"p", #"a", #"r", #"i",
                  #"n", #"g", #" ", #"c", #"o", #"n", #"s", #"t", #"a", #"n",
                  #"t", #" ", #"p", #"a", #"r", #"t"])
          (fn x =>
            catch (check_pvarsI B1_ ze (if mode then ge else gt) vas vbs)
              (fn xa =>
                error (append
                        [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w",
                          #"h", #"e", #"n", #" ", #"c", #"o", #"m", #"p", #"a",
                          #"r", #"i", #"n", #"g", #" ", #"c", #"o", #"e", #"f",
                          #"f", #"i", #"c", #"i", #"e", #"n", #"t", #"s", #" ",
                          #"o", #"f", #" ", #"v", #"a", #"r", #"i", #"a", #"b",
                          #"l", #"e", #" "]
                        (sshow B2_ xa)))))
    (fn x =>
      error [append
               (append
                 (append
                   [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"h",
                     #"e", #"n", #" ", #"c", #"o", #"m", #"p", #"a", #"r", #"i",
                     #"n", #"g", #" "]
                   (show_lpoly (A1_, A2_) B2_ ze on (LPoly (a, vas))))
                 [#" ", #">", #" "])
               (show_lpoly (A1_, A2_) B2_ ze on (LPoly (b, vbs))),
              x]);

fun check_polo_sI (A1_, A2_) B_ (C1_, C2_) mode ze on pl ti mz ge gt pi (s, t) =
  let
    val left = pleftI A1_ C1_ ze on pl ti mz pi s;
    val right = prightI A1_ C1_ ze on pl ti mz pi t;
  in
    catch (check_lpoly_sI (A1_, A2_) (C1_, C2_) mode ze on ge gt left right)
      (fn x =>
        error (append
                 (append
                   (append
                     [#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t",
                       #" ", #"e", #"n", #"s", #"u", #"r", #"e", #" "]
                     (sshow_term B_ C2_ s))
                   [#" ", #">", #" "])
                 (sshow_term B_ C2_ t)
                :: x))
  end;

fun get_comp_apos
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = ap;

fun get_comp_max0
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = mz;

fun get_comp_mono
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = mo;

fun sshow_arctic (Num_arc x) = sshow_int x
  | sshow_arctic MinInfty = [#"-", #"i", #"n", #"f"];

val sshow_arctica = {sshow = sshow_arctic} : arctic sshow;

fun fresh_strings_list name offset used n =
  take n
    (filter
      (fn s =>
        not (member (eq_lista eq_char) s (remdups (eq_lista eq_char) used)))
      (map (fn i => append name (sshow_nat (plus_nat i offset)))
        (upt Zero_nat (plus_nat n (length_unique (eq_lista eq_char) used)))));

fun checkAllCombinations c xs =
  foldR (fn x => fn ys => fn res =>
          catch (checkAllMI
                  (fn y =>
                    bind (catch (c x y) (fn xa => error ((x, y), xa)))
                      (fn xa => res))
                  ys)
            (fn xa => error (snd xa)))
    xs (return ());

fun check_graph_decomp ss g rcs =
  bind (catch (checkAllCombinations (check_edges ss g) (map snd rcs))
         (fn x => error (snd x)))
    (fn x =>
      catch (checkAllMI (fn c => check_edges ss g c c)
              (map snd (filter (fn rc => not (fst rc)) rcs)))
        (fn xa => error (snd xa)));

fun check_lpoly_nsI (A1_, A2_) (B1_, B2_) ze on ge (LPoly (a, vas))
  (LPoly (b, vbs)) =
  catch (bind (checkB (ge a b)
                [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"h",
                  #"e", #"n", #" ", #"c", #"o", #"m", #"p", #"a", #"r", #"i",
                  #"n", #"g", #" ", #"c", #"o", #"n", #"s", #"t", #"a", #"n",
                  #"t", #" ", #"p", #"a", #"r", #"t", #"s"])
          (fn x =>
            catch (check_pvarsI B1_ ze ge vas vbs)
              (fn xa =>
                error (append
                        [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w",
                          #"h", #"e", #"n", #" ", #"c", #"o", #"m", #"p", #"a",
                          #"r", #"i", #"n", #"g", #" ", #"c", #"o", #"e", #"f",
                          #"f", #"i", #"c", #"i", #"e", #"n", #"t", #"s", #" ",
                          #"o", #"f", #" ", #"v", #"a", #"r", #"i", #"a", #"b",
                          #"l", #"e", #" "]
                        (sshow B2_ xa)))))
    (fn x =>
      error [append
               (append
                 (append
                   [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"h",
                     #"e", #"n", #" ", #"c", #"o", #"m", #"p", #"a", #"r", #"i",
                     #"n", #"g", #" "]
                   (show_lpoly (A1_, A2_) B2_ ze on (LPoly (a, vas))))
                 [#" ", #">", #"=", #" "])
               (show_lpoly (A1_, A2_) B2_ ze on (LPoly (b, vbs))),
              x]);

fun check_polo_nsI (A1_, A2_) B_ (C1_, C2_) ze on pl ti mz ge pi (s, t) =
  let
    val left = pleftI A1_ C1_ ze on pl ti mz pi s;
    val right = prightI A1_ C1_ ze on pl ti mz pi t;
  in
    catch (check_lpoly_nsI (A1_, A2_) (C1_, C2_) ze on ge left right)
      (fn x =>
        error (append
                 (append
                   (append
                     [#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t",
                       #" ", #"e", #"n", #"s", #"u", #"r", #"e", #" "]
                     (sshow_term B_ C2_ s))
                   [#" ", #">", #"=", #" "])
                 (sshow_term B_ C2_ t)
                :: x))
  end;

fun to_lpoly_inter B_ on i f n =
  (case lookup (eqb B_ eq_nata) (f, n) i of NONE => (on, replicate n on)
     | SOME cas => cas);

fun checkRedPairProc A_ B_ checkSS checkNSS name representation p rP r =
  let
    val a = partition (fn foo => isValue (checkSS foo)) p;
    val (sp, nSp) = a;
  in
    catch (bind (bind (catch
                        (catch (checkAllMI checkNSS r) (fn x => error (snd x)))
                        (fn x =>
                          error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                   #" ", #"w", #"h", #"e", #"n", #" ", #"o",
                                   #"r", #"i", #"e", #"n", #"t", #"i", #"n",
                                   #"g", #" ", #"T", #"R", #"S"]
                                  :: x)))
                  (fn x =>
                    catch (catch (checkAllMI checkNSS nSp)
                            (fn xa => error (snd xa)))
                      (fn xa =>
                        error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ",
                                 #"w", #"h", #"e", #"n", #" ", #"o", #"r", #"i",
                                 #"e", #"n", #"t", #"i", #"n", #"g", #" ", #"D",
                                 #"P", #"s"]
                                :: xa))))
            (fn x =>
              catch (checkSubset (eqb (eq_terma A_ B_) (eq_terma A_ B_)) nSp rP)
                (fn xa =>
                  error ((case checkSS xa of Left err => err
                           | Right ok =>
                             [[#"i", #"n", #"t", #"e", #"r", #"n", #"a", #"l",
                                #" ", #"e", #"r", #"r", #"o", #"r"],
                               [#"p", #"l", #"e", #"a", #"s", #"e", #" ", #"c",
                                 #"o", #"n", #"t", #"a", #"c", #"t", #" ", #"o",
                                 #"n", #"e", #" ", #"o", #"f", #" ", #"t", #"h",
                                 #"e", #" ", #"C", #"e", #"T", #"A", #" ", #"m",
                                 #"e", #"m", #"b", #"e", #"r", #"s"],
                               [#"h", #"t", #"t", #"p", #":", #"/", #"/", #"c",
                                 #"l", #"-", #"i", #"n", #"f", #"o", #"r", #"m",
                                 #"a", #"t", #"i", #"k", #".", #"u", #"i", #"b",
                                 #"k", #".", #"a", #"c", #".", #"a", #"t", #"/",
                                 #"s", #"o", #"f", #"t", #"w", #"a", #"r", #"e",
                                 #"/", #"c", #"e", #"t", #"a"],
                               [#"c", #"e", #"t", #"a", #"@", #"i", #"n", #"f",
                                 #"o", #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                 #"u", #"i", #"b", #"k", #".", #"a", #"c", #".",
                                 #"a", #"t"]])))))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n", #" ",
                 #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o", #"c", #"e",
                 #"s", #"s", #"o", #"r", #" ", #"w", #"i", #"t", #"h", #" ",
                 #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                 #"i", #"n", #"g"]
                :: append representation ([] :: x)))
  end;

fun matchCapRMBelow A_ B_ l rm (Fun (f, ts)) =
  gcMatchI A_ B_ (HFun (f, map (capRM A_ B_ rm) ts)) l;

fun checkUrClosedTRM_af (A1_, A2_) (B1_, B2_) uu uv uw (Var x) = return ()
  | checkUrClosedTRM_af (A1_, A2_) (B1_, B2_) rm ur pi (Fun (f, ts)) =
    let
      val n = size_list ts;
    in
      bind (catch
             (checkAllMI
               (fn i =>
                 (if pi f n i
                   then checkUrClosedTRM_af (A1_, A2_) (B1_, B2_) rm ur pi
                          (nth ts i)
                   else return ()))
               (upt Zero_nat n))
             (fn x => error (snd x)))
        (fn x =>
          catch (checkAllMI
                  (fn lr =>
                    checkB
                      (member (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) lr ur
                        orelse
                        not (matchCapRMBelow A1_ B1_ (fst lr) (SOME rm)
                              (Fun (f, ts))))
                      (append
                        (append
                          (append
                            (append
                              [#"d", #"u", #"e", #" ", #"t", #"o", #" ", #"t",
                                #"h", #"e", #" ", #"s", #"u", #"b", #"t", #"e",
                                #"r", #"m", #" "]
                              (sshow_term A2_ B2_ (Fun (f, ts))))
                            [#" ", #"o", #"f", #" ", #"s", #"o", #"m", #"e",
                              #" ", #"u", #"s", #"a", #"b", #"l", #"e", #" ",
                              #"r", #"h", #"s", #",", #" ", #"r", #"u", #"l",
                              #"e", #" "])
                          (showRule A2_ B2_ lr))
                        [#" ", #"s", #"h", #"o", #"u", #"l", #"d", #" ", #"b",
                          #"e", #" ", #"u", #"s", #"a", #"b", #"l", #"e",
                          #"."]))
                  (rm f))
            (fn xa => error (snd xa)))
    end;

fun check_sct_ur_af_proc (A1_, A2_) (B1_, B2_) checkSS checkNSS representation p
  gs r rm rrm ur pi =
  let
    val edgeCheck = checkEdgRM A1_ B1_ rm rrm p true;
  in
    catch ((case rm
             of NONE =>
               error [[#"T", #"R", #"S", #" ", #"h", #"a", #"s", #" ", #"v",
                        #"a", #"r", #"i", #"a", #"b", #"l", #"e", #"s", #" ",
                        #"a", #"s", #" ", #"l", #"e", #"f", #"t", #"-", #"h",
                        #"a", #"n", #"d", #" ", #"s", #"i", #"d", #"e", #"s"]]
             | SOME rma =>
               bind (bind (bind (bind (catch
(catch
  (checkAllMI
    (fn lr => checkUrClosedTRM_af (A1_, A2_) (B1_, B2_) rma ur pi (snd lr)) ur)
  (fn x => error (snd x)))
(fn x =>
  error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e", #"n", #" ",
           #"c", #"h", #"e", #"c", #"k", #"i", #"n", #"g", #" ", #"c", #"l",
           #"o", #"s", #"u", #"r", #"e", #" ", #"p", #"r", #"o", #"p", #"e",
           #"r", #"t", #"i", #"e", #"s", #" ", #"o", #"f", #" ", #"r", #"h",
           #"s", #" ", #"o", #"f", #" ", #"u", #"s", #"a", #"b", #"l", #"e",
           #" ", #"r", #"u", #"l", #"e", #"s"],
          x]))
                                  (fn x =>
                                    catch (checkSubset
    (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) p (map fst gs))
                                      (fn xa =>
error [append
         [#"t", #"h", #"e", #"r", #"e", #" ", #"i", #"s", #" ", #"n", #"o",
           #" ", #"s", #"i", #"z", #"e", #"-", #"c", #"h", #"a", #"n", #"g",
           #"e", #" ", #"g", #"r", #"a", #"p", #"h", #" ", #"f", #"o", #"r",
           #" ", #"D", #"P", #" "]
         (showRule A2_ B2_ xa)])))
                            (fn x =>
                              catch (checkAllMI
                                      (fn a as (aa, b) =>
let
  val (s, t) = aa;
in
  (fn ab as (stri, nstri) =>
    catch (bind (bind (bind (checkB
                              (not (is_var s) andalso
                                (not (is_var t) andalso
                                  null (rma (the (root t)))))
                              [[#"r", #"o", #"o", #"t", #"s", #" ", #"o", #"f",
                                 #" ", #"D", #"P", #" ", #"m", #"u", #"s", #"t",
                                 #" ", #"b", #"e", #" ", #"c", #"o", #"n", #"s",
                                 #"t", #"r", #"u", #"c", #"t", #"o", #"r",
                                 #"s"]])
                        (fn xa =>
                          catch (checkAllMI
                                  (fn i =>
                                    checkB (less_nat i (size_list (args s)))
                                      [append
 [#"l", #"e", #"f", #"t", #"-", #"i", #"n", #"d", #"e", #"x", #" ", #"t", #"o",
   #" ", #"l", #"a", #"r", #"g", #"e"]
 (sshow_nat i)])
                                  (remdups eq_nata
                                    (map fst (append stri nstri))))
                            (fn xb => error (snd xb))))
                  (fn xa =>
                    catch (checkAllMI
                            (fn j =>
                              checkB
                                (less_nat j (size_list (args t)) andalso
                                  isValue
                                    (checkUrClosedTRM_af (A1_, A2_) (B1_, B2_)
                                      rma ur pi (nth (args t) j)))
                                [append
                                   [#"r", #"i", #"g", #"h", #"t", #"-", #"i",
                                     #"n", #"d", #"e", #"x", #" ", #"t", #"o",
                                     #" ", #"l", #"a", #"r", #"g", #"e", #" ",
                                     #"o", #"r", #" ", #"a", #"r", #"g", #"u",
                                     #"m", #"e", #"n", #"t", #" ", #"v", #"i",
                                     #"o", #"l", #"a", #"t", #"e", #"s", #" ",
                                     #"u", #"s", #"a", #"b", #"l", #"e", #"-",
                                     #"r", #"u", #"l", #"e", #"s", #" ", #"c",
                                     #"o", #"n", #"d", #"i", #"t", #"i", #"o",
                                     #"n"]
                                   (sshow_nat j)])
                            (remdups eq_nata (map snd (append stri nstri))))
                      (fn xb => error (snd xb))))
            (fn xa =>
              let
                val ss = args s;
                val ts = args t;
              in
                bind (catch
                       (checkAllMI
                         (fn ac as (i, j) =>
                           checkB (isValue (checkSS (nth ss i, nth ts j)))
                             [append
                                (append
                                  (append
                                    [#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                      #" ", #"w", #"i", #"t", #"h", #" ", #"e",
                                      #"d", #"g", #"e", #" "]
                                    (sshow_nat i))
                                  [#" ", #"-", #"S", #"-", #">", #" "])
                                (sshow_nat j)])
                         stri)
                       (fn xb => error (snd xb)))
                  (fn xb =>
                    catch (checkAllMI
                            (fn ac as (i, j) =>
                              checkB (isValue (checkNSS (nth ss i, nth ts j)))
                                [append
                                   (append
                                     (append
                                       [#"p", #"r", #"o", #"b", #"l", #"e",
 #"m", #" ", #"w", #"i", #"t", #"h", #" ", #"e", #"d", #"g", #"e", #" "]
                                       (sshow_nat i))
                                     [#" ", #"-", #"N", #"S", #"-", #">", #" "])
                                   (sshow_nat j)])
                            nstri)
                      (fn xc => error (snd xc)))
              end))
      (fn xa =>
        error (append
                 [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #"s", #" ", #"w",
                   #"i", #"t", #"h", #" ", #"D", #"P", #" "]
                 (showRule A2_ B2_ (s, t))
                :: xa)))
end
  b)
                                      gs)
                                (fn xa => error (snd xa))))
                      (fn x =>
                        catch (catch (checkAllMI checkNSS ur)
                                (fn xa => error (snd xa)))
                          (fn xa =>
                            error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                     #" ", #"w", #"h", #"e", #"n", #" ", #"o",
                                     #"r", #"i", #"e", #"n", #"t", #"i", #"n",
                                     #"g", #" ", #"u", #"s", #"a", #"b", #"l",
                                     #"e", #" ", #"r", #"u", #"l", #"e", #"s"]
                                    :: xa))))
                 (fn x =>
                   checkB
                     (check_SCT (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_))
                       eq_nata (fn st => fn uv => edgeCheck (st, uv))
                       (map (fn a as (st, aa) =>
                              let
                                val (ab, b) = aa;
                              in
                                Scg (st, st, ab, b)
                              end)
                         gs))
                     [[#"s", #"i", #"z", #"e", #"-", #"c", #"h", #"a", #"n",
                        #"g", #"e", #" ", #"a", #"n", #"a", #"l", #"y", #"s",
                        #"i", #"s", #" ", #"f", #"a", #"i", #"l", #"e",
                        #"d"]])))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"s", #"i", #"z", #"e", #"-", #"c", #"h", #"a", #"n", #"g",
                 #"e", #" ", #"p", #"r", #"o", #"c", #"e", #"s", #"s", #"o",
                 #"r", #" ", #"w", #"i", #"t", #"h", #" ", #"t", #"h", #"e",
                 #" ", #"f", #"o", #"l", #"l", #"o", #"w", #"i", #"n", #"g",
                 #" "]
                :: append representation
                     ([#"f", #"o", #"r", #" ", #"t", #"h", #"e", #" ", #"f",
                        #"o", #"l", #"l", #"o", #"w", #"i", #"n", #"g", #" ",
                        #"r", #"e", #"a", #"s", #"o", #"n"]
                       :: x)))
  end;

fun checkNotWellFormed F_ (V1_, V2_) r =
  checkB (not (isValue (checkWfTRS F_ (V1_, V2_) r)))
    [[#"T", #"h", #"e", #" ", #"T", #"R", #"S", #" ", #"i", #"s", #" ", #"w",
       #"e", #"l", #"l", #" ", #"f", #"o", #"r", #"m", #"e", #"d"]];

fun checkTRSDisproof (A1_, A2_) (B1_, B2_) i r (Loop l) =
  debug i [#"L", #"o", #"o", #"p"]
    (catch (checkLoop (A1_, A2_) (B1_, B2_) r l)
      (fn x =>
        error (append i
                 [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"i", #"n",
                   #" ", #"l", #"o", #"o", #"p", #" ", #"p", #"r", #"o", #"o",
                   #"f"]
                :: x)))
  | checkTRSDisproof (A1_, A2_) (B1_, B2_) i r NotWellFormed =
    debug i
      [#"n", #"o", #"t", #" ", #"w", #"e", #"l", #"l", #"-", #"f", #"o", #"r",
        #"m", #"e", #"d"]
      (catch (checkNotWellFormed A2_ (B1_, B2_) r)
        (fn x =>
          error (append i
                   [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"i", #"n",
                     #" ", #"n", #"o", #"t", #"-", #"w", #"e", #"l", #"l", #"-",
                     #"f", #"o", #"r", #"m", #"e", #"d", #" ", #"p", #"r", #"o",
                     #"o", #"f"]
                  :: x)));

fun init_lab f = UnLab f;

fun map_rule fg lr = (map_term fg (fst lr), map_term fg (snd lr));

fun sshow_arithFun (Arg i) = append [#"x", #"_"] (sshow_nat (Suc i))
  | sshow_arithFun (Const n) = sshow_nat n
  | sshow_arithFun (Sum (f :: fs)) =
    append (sshow_arithFun f)
      (flat_map (fn g => append [#" ", #"+", #" "] (sshow_arithFun g)) fs)
  | sshow_arithFun (Prod (f :: fs)) =
    append
      (append (append [#"("] (sshow_arithFun f))
        (flat_map (fn g => append [#" ", #"*", #" "] (sshow_arithFun g)) fs))
      [#")"]
  | sshow_arithFun (Min (f :: fs)) =
    append
      (append (append [#"m", #"i", #"n", #"("] (sshow_arithFun f))
        (flat_map (fn g => append [#","] (sshow_arithFun g)) fs))
      [#")"]
  | sshow_arithFun (Max (f :: fs)) =
    append
      (append (append [#"m", #"a", #"x", #"("] (sshow_arithFun f))
        (flat_map (fn g => append [#","] (sshow_arithFun g)) fs))
      [#")"];

fun sshow_sl_inter C_ D_ (SL_Inter (n, fnas)) =
  showList
    (append
       (append
         [#"c", #"a", #"r", #"r", #"i", #"e", #"r", #" ", #"{", #"0", #",",
           #".", #".", #".", #","]
         (sshow_nat n))
       [#"}"]
      :: map (fn a as (aa, b) =>
               let
                 val (f, na) = aa;
               in
                 (fn ab =>
                   append
                     (append
                       (append
                         (append (append [#"["] (sshow_lab C_ D_ f)) [#"/"])
                         (sshow_nat na))
                       [#"]", #" ", #"=", #" "])
                     (sshow_arithFun ab))
               end
                 b)
           fnas);

fun take_default def [] uu = def
  | take_default uv (x :: xs) Zero_nat = x
  | take_default def (uw :: xs) (Suc i) = take_default def xs i;

fun eval_arithFun_unbound nats (Arg i) = take_default Zero_nat nats i
  | eval_arithFun_unbound nats (Const n) = n
  | eval_arithFun_unbound nats (Sum []) = Zero_nat
  | eval_arithFun_unbound nats (Sum (f :: fs)) =
    plus_nat (eval_arithFun_unbound nats f)
      (eval_arithFun_unbound nats (Sum fs))
  | eval_arithFun_unbound nats (Prod []) = one_nat
  | eval_arithFun_unbound nats (Prod (f :: fs)) =
    times_nat (eval_arithFun_unbound nats f)
      (eval_arithFun_unbound nats (Prod fs))
  | eval_arithFun_unbound nats (Max [f]) = eval_arithFun_unbound nats f
  | eval_arithFun_unbound nats (Max (f :: v :: va)) =
    max ord_nat (eval_arithFun_unbound nats f)
      (eval_arithFun_unbound nats (Max (v :: va)))
  | eval_arithFun_unbound nats (Min [f]) = eval_arithFun_unbound nats f
  | eval_arithFun_unbound nats (Min (f :: v :: va)) =
    min ord_nat (eval_arithFun_unbound nats f)
      (eval_arithFun_unbound nats (Min (v :: va)));

fun eval_arithFun c f nats = mod_nat (eval_arithFun_unbound nats f) c;

fun sl_inter_to_inter A_ B_ (SL_Inter (c, ls)) fl cs =
  (case lookup (eqb (eq_laba A_ B_) eq_nata) (fl, size_list cs) ls
     of NONE => Zero_nat | SOME f => eval_arithFun (Suc c) f cs);

fun get_largest_element (SL_Inter (n, uu)) = n;

fun create_sem_lab_tt (F1_, F2_) (L1_, L2_) (V1_, V2_) sli l r lr =
  let
    val c = get_largest_element sli;
    val ca = upt Zero_nat (Suc c);
    val i = sl_inter_to_inter F1_ L1_ sli;
  in
    catch (check_sem_lab_tt (F1_, F2_) (L1_, L2_) (V1_, V2_)
            (check_sl_lab_trs (F1_, F2_) (L1_, L2_) (eq_nata, sshow_nata)
              (V1_, V2_) i l ca)
            (check_sl_model_lab_trs (F1_, F2_) (L1_, L2_) (eq_nata, sshow_nata)
              (V1_, V2_) i l ca)
            (fn x => false) i l r lr)
      (fn x => error (sshow_sl_inter F2_ L2_ sli :: x))
  end;

fun map_trsI fg r = map (map_rule fg) r;

fun unlab_of_trsI F_ L_ V_ r =
  remdups (eqb (eq_terma (eq_laba F_ L_) V_) (eq_terma (eq_laba F_ L_) V_))
    (map_trsI unlab
      (filter
        (fn a as (l, ra) =>
          eqop (eq_terma (eq_laba F_ L_) V_) l ra orelse
            not (eqop (eq_terma (eq_laba F_ L_) V_) (map_term unlab l)
                  (map_term unlab ra)))
        r));

fun unlab_trs_tt (F1_, F2_) (L1_, L2_) (V1_, V2_) r rR =
  let
    val u = unlab_of_trsI F1_ L1_ V1_ r;
  in
    catch (bind (checkSubset
                  (eqb (eq_terma (eq_laba F1_ L1_) V1_)
                    (eq_terma (eq_laba F1_ L1_) V1_))
                  rR u)
            (fn x =>
              checkSubset
                (eqb (eq_terma (eq_laba F1_ L1_) V1_)
                  (eq_terma (eq_laba F1_ L1_) V1_))
                u rR))
      (fn x =>
        error (append (showRules (sshow_laba F2_ L2_) V2_ rR)
                (append
                  [[#" ", #" ", #"d", #"o", #"e", #"s", #" ", #"n", #"o", #"t",
                     #" ", #"c", #"o", #"r", #"r", #"e", #"s", #"p", #"o", #"n",
                     #"d", #" ", #"t", #"o", #" ", #"u", #"n", #"l", #"a", #"b",
                     #"e", #"l", #"e", #"d", #" ", #"T", #"R", #"S", #" ",
                     #" "]]
                  (showRules (sshow_laba F2_ L2_) V2_ u))))
  end;

fun get_comp_description
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = de;

fun create_lpoly_repr (A1_, A2_) (B1_, B2_) c i =
  let
    val pi = to_lpoly_inter B1_ (get_comp_1 c) i;
  in
    append
      [append
         [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l", #" ",
           #"i", #"n", #"t", #"e", #"r", #"p", #"r", #"e", #"t", #"r", #"a",
           #"t", #"i", #"o", #"n", #" ", #"o", #"v", #"e", #"r", #" "]
         (get_comp_description c)]
      (map (fn a as (f, n) =>
             let
               val t =
                 Fun (f, map Var
                           (fresh_strings_list [#"x", #"_"] one_nat [] n));
             in
               append
                 (append
                   (append [#"P", #"o", #"l", #"("]
                     (sshow_term B2_ (sshow_lista sshow_chara) t))
                   [#")", #" ", #"=", #" "])
                 (show_lpoly (A1_, A2_) (sshow_lista sshow_chara) (get_comp_0 c)
                   (get_comp_1 c)
                   (pleftI A1_ (eq_lista eq_char) (get_comp_0 c) (get_comp_1 c)
                     (get_comp_plus c) (get_comp_times c) (get_comp_max0 c) pi
                     t))
             end)
        (remdups (eqb B1_ eq_nata) (map fst i)))
  end;

fun check_poly_monoI A_ B_ ze ge cm =
  (fn xs =>
    catch (checkAllMI
            (fn a as (aa, b) =>
              let
                val (f, n) = aa;
              in
                (fn ab as (c, cs) =>
                  catch (bind (bind (checkB (ge c ze)
                                      [append
 (append
   (append
     [#"c", #"o", #"n", #"s", #"t", #"a", #"n", #"t", #" ", #"p", #"a", #"r",
       #"t", #" "]
     (sshow A_ c))
   [#" ", #"m", #"u", #"s", #"t", #" ", #"b", #"e", #" ", #"a", #"t", #" ",
     #"l", #"e", #"a", #"s", #"t", #" "])
 (sshow A_ ze)])
                                (fn x =>
                                  checkB (less_eq_nat n (size_list cs))
                                    [[#"t", #"h", #"e", #" ", #"l", #"a", #"s",
                                       #"t", #" ", #"a", #"r", #"g", #"u", #"m",
                                       #"e", #"n", #"t", #" ", #"i", #"s", #" ",
                                       #"i", #"g", #"n", #"o", #"r", #"e",
                                       #"d"]]))
                          (fn x =>
                            catch (checkAllMI
                                    (fn d =>
                                      checkB (cm d)
[append
   (append
     [#"c", #"o", #"e", #"f", #"f", #"i", #"c", #"i", #"e", #"n", #"t", #" "]
     (sshow A_ d))
   [#" ", #"i", #"s", #" ", #"n", #"o", #"t", #" ", #"a", #"l", #"l", #"o",
     #"w", #"e", #"d"]])
                                    cs)
                              (fn xa => error (snd xa))))
                    (fn x =>
                      error (append
                               (append
                                 (append
                                   [#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                     #" ", #"w", #"i", #"t", #"h", #" ", #"m",
                                     #"o", #"n", #"o", #"t", #"o", #"n", #"i",
                                     #"c", #"i", #"t", #"y", #" ", #"d", #"u",
                                     #"e", #" ", #"t", #"o", #" ", #"i", #"n",
                                     #"t", #"e", #"r", #"p", #"r", #"e", #"t",
                                     #"a", #"t", #"i", #"o", #"n", #" ", #"o",
                                     #"f", #" "]
                                   (sshow B_ f))
                                 [#"/"])
                               (sshow_nat n)
                              :: x)))
              end
                b)
            xs)
      (fn x => error (snd x)));

fun checkRuleRemovalProc A_ B_ checkMono checkSS checkNSS name representation r
  rR =
  let
    val a = partition (fn foo => isValue (checkSS foo)) r;
    val (sp, nSp) = a;
  in
    catch (bind (bind (bind (catch checkMono
                              (fn x =>
                                error ([#"p", #"r", #"o", #"b", #"l", #"e",
 #"m", #" ", #"w", #"i", #"t", #"h", #" ", #"m", #"o", #"n", #"o", #"t", #"o",
 #"n", #"i", #"c", #"i", #"t", #"y", #" ", #"o", #"f", #" ", #"s", #"t", #"r",
 #"i", #"c", #"t", #" ", #"o", #"r", #"d", #"e", #"r"]
:: x)))
                        (fn x =>
                          catch (catch (checkAllMI checkNSS r)
                                  (fn xa => error (snd xa)))
                            (fn xa =>
                              error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                       #" ", #"w", #"h", #"e", #"n", #" ", #"o",
                                       #"r", #"i", #"e", #"n", #"t", #"i", #"n",
                                       #"g", #" ", #"r", #"u", #"l", #"e", #"s"]
                                      :: xa))))
                  (fn x =>
                    catch (checkSubset (eqb (eq_terma A_ B_) (eq_terma A_ B_))
                            nSp rR)
                      (fn xa =>
                        error ((case checkSS xa of Left err => err
                                 | Right ok =>
                                   [[#"i", #"n", #"t", #"e", #"r", #"n", #"a",
                                      #"l", #" ", #"e", #"r", #"r", #"o", #"r"],
                                     [#"p", #"l", #"e", #"a", #"s", #"e", #" ",
                                       #"c", #"o", #"n", #"t", #"a", #"c", #"t",
                                       #" ", #"o", #"n", #"e", #" ", #"o", #"f",
                                       #" ", #"t", #"h", #"e", #" ", #"C", #"e",
                                       #"T", #"A", #" ", #"m", #"e", #"m", #"b",
                                       #"e", #"r", #"s"],
                                     [#"h", #"t", #"t", #"p", #":", #"/", #"/",
                                       #"c", #"l", #"-", #"i", #"n", #"f", #"o",
                                       #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                       #"u", #"i", #"b", #"k", #".", #"a", #"c",
                                       #".", #"a", #"t", #"/", #"s", #"o", #"f",
                                       #"t", #"w", #"a", #"r", #"e", #"/", #"c",
                                       #"e", #"t", #"a"],
                                     [#"c", #"e", #"t", #"a", #"@", #"i", #"n",
                                       #"f", #"o", #"r", #"m", #"a", #"t", #"i",
                                       #"k", #".", #"u", #"i", #"b", #"k", #".",
                                       #"a", #"c", #".", #"a", #"t"]])))))
            (fn x =>
              catch (checkSubset (eqb (eq_terma A_ B_) (eq_terma A_ B_)) rR r)
                (fn xa =>
                  error ((case checkSS xa of Left err => err
                           | Right ok =>
                             [[#"i", #"n", #"t", #"e", #"r", #"n", #"a", #"l",
                                #" ", #"e", #"r", #"r", #"o", #"r"],
                               [#"p", #"l", #"e", #"a", #"s", #"e", #" ", #"c",
                                 #"o", #"n", #"t", #"a", #"c", #"t", #" ", #"o",
                                 #"n", #"e", #" ", #"o", #"f", #" ", #"t", #"h",
                                 #"e", #" ", #"C", #"e", #"T", #"A", #" ", #"m",
                                 #"e", #"m", #"b", #"e", #"r", #"s"],
                               [#"h", #"t", #"t", #"p", #":", #"/", #"/", #"c",
                                 #"l", #"-", #"i", #"n", #"f", #"o", #"r", #"m",
                                 #"a", #"t", #"i", #"k", #".", #"u", #"i", #"b",
                                 #"k", #".", #"a", #"c", #".", #"a", #"t", #"/",
                                 #"s", #"o", #"f", #"t", #"w", #"a", #"r", #"e",
                                 #"/", #"c", #"e", #"t", #"a"],
                               [#"c", #"e", #"t", #"a", #"@", #"i", #"n", #"f",
                                 #"o", #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                 #"u", #"i", #"b", #"k", #".", #"a", #"c", #".",
                                 #"a", #"t"]])))))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"r", #"u", #"l", #"e", #" ", #"r", #"e", #"m", #"o", #"v",
                 #"a", #"l", #" ", #"p", #"r", #"o", #"c", #"e", #"s", #"s",
                 #"o", #"r", #" ", #"w", #"i", #"t", #"h", #" ", #"t", #"h",
                 #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w", #"i", #"n",
                 #"g"]
                :: append representation ([] :: x)))
  end;

fun get_comp_wf_elem
  (wfe, (ze, (on, (pl, (ti, (mz, (ge, (gt, (mo, (de, ap)))))))))) = wfe;

fun check_lpoly_coeffs A_ B_ wfe apos ch i =
  catch (checkAllMI
          (fn a as (aa, b) =>
            let
              val (f, n) = aa;
            in
              (fn ab as (c, cs) =>
                catch (bind (bind (checkB (wfe c)
                                    [append
                                       (append
 [#"c", #"o", #"n", #"s", #"t", #"a", #"n", #"t", #" ", #"p", #"a", #"r", #"t",
   #" "]
 (sshow A_ c))
                                       [#" ", #"i", #"s", #" ", #"n", #"o",
 #"t", #" ", #"w", #"e", #"l", #"l", #"-", #"f", #"o", #"r", #"m", #"e", #"d"]])
                              (fn x =>
                                checkB (apos c orelse any apos (take n cs))
                                  [[#"c", #"o", #"u", #"l", #"d", #" ", #"n",
                                     #"o", #"t", #" ", #"f", #"i", #"n", #"d",
                                     #" ", #"p", #"o", #"s", #"i", #"t", #"i",
                                     #"v", #"e", #" ", #"e", #"n", #"t", #"r",
                                     #"y", #" ", #"w", #"h", #"i", #"c", #"h",
                                     #" ", #"i", #"s", #" ", #"r", #"e", #"q",
                                     #"u", #"i", #"r", #"e", #"d", #" ", #"f",
                                     #"o", #"r", #" ", #"a", #"r", #"c", #"t",
                                     #"i", #"c", #" ", #"i", #"n", #"t", #"e",
                                     #"r", #"p", #"r", #"e", #"t", #"a", #"t",
                                     #"i", #"o", #"n", #"s"]]))
                        (fn x =>
                          catch (checkAllMI
                                  (fn ac =>
                                    checkB (ch ac andalso wfe ac)
                                      [append
 (append
   [#"c", #"o", #"e", #"f", #"f", #"i", #"c", #"i", #"e", #"n", #"t", #" "]
   (sshow A_ ac))
 [#" ", #"i", #"s", #" ", #"n", #"o", #"t", #" ", #"a", #"l", #"l", #"o", #"w",
   #"e", #"d"]])
                                  cs)
                            (fn xa => error (snd xa))))
                  (fn x =>
                    error (append
                             (append
                               (append
                                 [#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                   #" ", #"w", #"i", #"t", #"h", #" ", #"i",
                                   #"n", #"t", #"e", #"r", #"p", #"r", #"e",
                                   #"t", #"a", #"t", #"i", #"o", #"n", #" ",
                                   #"o", #"f", #" "]
                                 (sshow B_ f))
                               [#"/"])
                             (sshow_nat n)
                            :: x)))
            end
              b)
          i)
    (fn x => error (snd x));

fun create_rule_removal_proc_comp (A1_, A2_) (B1_, B2_) (C1_, C2_) c cm i r rR =
  let
    val pi = to_lpoly_inter B1_ (get_comp_1 c) i;
  in
    bind (check_lpoly_coeffs A2_ B2_ (get_comp_wf_elem c) (get_comp_apos c)
           (fn x => get_comp_ge c x (get_comp_0 c)) i)
      (fn x =>
        checkRuleRemovalProc B1_ C1_
          (check_poly_monoI A2_ B2_ (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (A1_, A2_) B2_ (C1_, C2_) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (A1_, A2_) B2_ (C1_, C2_) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (append
            (append
              [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l", #" ",
                #"o", #"r", #"d", #"e", #"r", #" ", #"("]
              (get_comp_description c))
            [#")"])
          (create_lpoly_repr (A1_, A2_) (B1_, B2_) c i) r rR)
  end;

fun create_rule_removal_proc (A1_, A2_) (B1_, B2_) (Int_carrier ia) =
  create_rule_removal_proc_comp (eq_int, sshow_inta) (A1_, A2_) (B1_, B2_)
    ((fn x => true),
      ((0 : IntInf.int),
        ((1 : IntInf.int),
          ((fn a => fn b => IntInf.+ (a, b)),
            ((fn a => fn b => IntInf.* (a, b)),
              (max ord_int (0 : IntInf.int),
                ((fn x => fn y => IntInf.<= (y, x)),
                  ((fn x => fn y => IntInf.< (y, x)),
                    (true, (int_name, (fn x => true)))))))))))
    (fn a => IntInf.< ((0 : IntInf.int), a)) ia
  | create_rule_removal_proc (A1_, A2_) (B1_, B2_) (Nat_carrier i) =
    create_rule_removal_proc_comp (eq_nata, sshow_nata) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (Zero_nat,
          (one_nat,
            (plus_nat,
              (times_nat,
                (id, ((fn x => fn y => less_eq_nat y x),
                       ((fn x => fn y => less_nat y x),
                         (true, (nat_name, (fn x => true)))))))))))
      (fn x => not (eqop eq_nata x Zero_nat)) i
  | create_rule_removal_proc (A1_, A2_) (B1_, B2_) (Arctic_carrier id) =
    (fn uu => fn uua =>
      error [[#"a", #"r", #"c", #"t", #"i", #"c", #" ", #"n", #"o", #"t", #" ",
               #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ", #"f",
               #"o", #"r", #" ", #"T", #"R", #"S", #"s"]])
  | create_rule_removal_proc (A1_, A2_) (B1_, B2_) (Arc_mat_carrier (n, ic)) =
    (fn uu => fn uua =>
      error [[#"a", #"r", #"c", #"t", #"i", #"c", #" ", #"n", #"o", #"t", #" ",
               #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ", #"f",
               #"o", #"r", #" ", #"T", #"R", #"S", #"s"]])
  | create_rule_removal_proc (A1_, A2_) (B1_, B2_) (Int_mat_carrier (n, ib)) =
    (fn uu => fn uua =>
      error [[#"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ", #"m",
               #"a", #"t", #"r", #"i", #"x", #"-", #"o", #"r", #"d", #"e", #"r",
               #"s", #" ", #"n", #"o", #"t", #" ", #"y", #"e", #"t", #" ", #"s",
               #"u", #"p", #"p", #"o", #"r", #"t", #"e", #"d"]]);

fun computeRuleMap A_ [] = SOME (fn f => [])
  | computeRuleMap A_ ((Fun (f, ts), r) :: rules) =
    (case computeRuleMap A_ rules of NONE => NONE
       | SOME rm => SOME (fun_upd A_ rm f ((Fun (f, ts), r) :: rm f)))
  | computeRuleMap A_ ((Var uu, uv) :: rules) = NONE;

fun create_sct_ur_af_proc_comp (A1_, A2_) (B1_, B2_) (C1_, C2_) c i p r rm rrm
  ur graphs =
  let
    val pi = to_lpoly_inter B1_ (get_comp_1 c) i;
  in
    bind (check_lpoly_coeffs A2_ B2_ (get_comp_wf_elem c) (get_comp_apos c)
           (fn x => get_comp_ge c x (get_comp_0 c)) i)
      (fn x =>
        check_sct_ur_af_proc (B1_, B2_) (C1_, C2_)
          (check_polo_sI (A1_, A2_) B2_ (C1_, C2_) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (A1_, A2_) B2_ (C1_, C2_) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (append
             (append
               [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l",
                 #" ", #"o", #"r", #"d", #"e", #"r", #" ", #"("]
               (get_comp_description c))
             [#")"]
            :: create_lpoly_repr (A1_, A2_) (B1_, B2_) c i)
          p graphs r rm rrm ur (create_afI A1_ B1_ (get_comp_0 c) i))
  end;

fun create_sct_ur_af_proc (A1_, A2_) (B1_, B2_) (Int_carrier ia) =
  create_sct_ur_af_proc_comp (eq_int, sshow_inta) (A1_, A2_) (B1_, B2_)
    ((fn x => true),
      ((0 : IntInf.int),
        ((1 : IntInf.int),
          ((fn a => fn b => IntInf.+ (a, b)),
            ((fn a => fn b => IntInf.* (a, b)),
              (max ord_int (0 : IntInf.int),
                ((fn x => fn y => IntInf.<= (y, x)),
                  ((fn x => fn y => IntInf.< (y, x)),
                    (true, (int_name, (fn x => true)))))))))))
    ia
  | create_sct_ur_af_proc (A1_, A2_) (B1_, B2_) (Nat_carrier i) =
    create_sct_ur_af_proc_comp (eq_nata, sshow_nata) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (Zero_nat,
          (one_nat,
            (plus_nat,
              (times_nat,
                (id, ((fn x => fn y => less_eq_nat y x),
                       ((fn x => fn y => less_nat y x),
                         (true, (nat_name, (fn x => true)))))))))))
      i
  | create_sct_ur_af_proc (A1_, A2_) (B1_, B2_) (Arctic_carrier ida) =
    create_sct_ur_af_proc_comp (eq_arctica, sshow_arctica) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (MinInfty,
          (Num_arc (0 : IntInf.int),
            (plus_arctic,
              (mult_arctic,
                (id, (geq_arctic,
                       (gt_arctic, (false, (arctic_name, pos_arctic))))))))))
      ida
  | create_sct_ur_af_proc (A1_, A2_) (B1_, B2_) (Int_mat_carrier (n, ib)) =
    create_sct_ur_af_proc_comp
      (eq_lista (eq_lista eq_int), sshow_lista (sshow_lista sshow_inta))
      (A1_, A2_) (B1_, B2_)
      (wf_mat_rcI (Suc n) (Suc n),
        (mat0I (0 : IntInf.int) (Suc n) (Suc n),
          (mat1I (0 : IntInf.int) (1 : IntInf.int) (Suc n),
            (mat_plusI (fn a => fn b => IntInf.+ (a, b)),
              (mat_mulI (0 : IntInf.int) (fn a => fn b => IntInf.+ (a, b))
                 (fn a => fn b => IntInf.* (a, b)) (Suc n),
                (mat_max0I (max ord_int (0 : IntInf.int)),
                  (mat_geI (fn x => fn y => IntInf.<= (y, x)),
                    (mat_gtI (fn x => fn y => IntInf.<= (y, x))
                       (fn x => fn y => IntInf.< (y, x)),
                      (true,
                        (int_mat_name, mat_arc_posI (fn x => true)))))))))))
      ib
  | create_sct_ur_af_proc (A1_, A2_) (B1_, B2_) (Arc_mat_carrier (n, ic)) =
    create_sct_ur_af_proc_comp
      (eq_lista (eq_lista eq_arctica), sshow_lista (sshow_lista sshow_arctica))
      (A1_, A2_) (B1_, B2_)
      (wf_mat_rcI (Suc n) (Suc n),
        (mat0I MinInfty (Suc n) (Suc n),
          (mat1I MinInfty (Num_arc (0 : IntInf.int)) (Suc n),
            (mat_plusI plus_arctic,
              (mat_mulI MinInfty plus_arctic mult_arctic (Suc n),
                (mat_max0I id,
                  (mat_geI geq_arctic,
                    (mat_geI gt_arctic,
                      (false, (arc_mat_name, mat_arc_posI pos_arctic))))))))))
      ic;

fun check_sct_subterm_proc (A1_, A2_) (B1_, B2_) p gs r rm rrm =
  let
    val edge_check = checkEdgRM A1_ B1_ rm rrm p true;
  in
    catch ((case rm
             of NONE =>
               error [[#"T", #"R", #"S", #" ", #"h", #"a", #"s", #" ", #"v",
                        #"a", #"r", #"i", #"a", #"b", #"l", #"e", #"s", #" ",
                        #"a", #"s", #" ", #"l", #"e", #"f", #"t", #"-", #"h",
                        #"a", #"n", #"d", #" ", #"s", #"i", #"d", #"e", #"s"]]
             | SOME rma =>
               bind (bind (catch
                            (checkSubset
                              (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) p
                              (map fst gs))
                            (fn x =>
                              error [append
                                       [#"t", #"h", #"e", #"r", #"e", #" ",
 #"i", #"s", #" ", #"n", #"o", #" ", #"s", #"i", #"z", #"e", #"-", #"c", #"h",
 #"a", #"n", #"g", #"e", #" ", #"g", #"r", #"a", #"p", #"h", #" ", #"f", #"o",
 #"r", #" ", #"D", #"P", #" "]
                                       (showRule A2_ B2_ x)]))
                      (fn x =>
                        catch (checkAllMI
                                (fn a as (aa, b) =>
                                  let
                                    val (s, t) = aa;
                                  in
                                    (fn ab as (stri, nstri) =>
                                      catch
(bind (checkB
        (not (is_var s) andalso
          (not (is_var t) andalso null (rma (the (root t)))))
        [[#"r", #"o", #"o", #"t", #"s", #" ", #"o", #"f", #" ", #"D", #"P",
           #" ", #"m", #"u", #"s", #"t", #" ", #"b", #"e", #" ", #"c", #"o",
           #"n", #"s", #"t", #"r", #"u", #"c", #"t", #"o", #"r", #"s"]])
  (fn xa =>
    let
      val ss = args s;
      val ts = args t;
      val n = size_list ss;
      val m = size_list ts;
    in
      bind (catch
             (checkAllMI
               (fn ac as (i, j) =>
                 checkB
                   (less_nat i n andalso
                     (less_nat j m andalso
                       isValue
                         (check_supt (A1_, A2_) (B1_, B2_)
                           (nth ss i, nth ts j))))
                   [append
                      (append
                        (append
                          [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w",
                            #"i", #"t", #"h", #" ", #"e", #"d", #"g", #"e",
                            #" "]
                          (sshow_nat i))
                        [#" ", #"|", #">", #" "])
                      (sshow_nat j)])
               stri)
             (fn xb => error (snd xb)))
        (fn xb =>
          catch (checkAllMI
                  (fn ac as (i, j) =>
                    checkB
                      (less_nat i n andalso
                        (less_nat j m andalso
                          isValue
                            (check_supteq (A1_, A2_) (B1_, B2_)
                              (nth ss i, nth ts j))))
                      [append
                         (append
                           (append
                             [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ",
                               #"w", #"i", #"t", #"h", #" ", #"e", #"d", #"g",
                               #"e", #" "]
                             (sshow_nat i))
                           [#" ", #"|", #">", #"=", #" "])
                         (sshow_nat j)])
                  nstri)
            (fn xc => error (snd xc)))
    end))
(fn xa =>
  error (append
           [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #"s", #" ", #"w", #"i",
             #"t", #"h", #" ", #"D", #"P", #" "]
           (showRule A2_ B2_ (s, t))
          :: xa)))
                                  end
                                    b)
                                gs)
                          (fn xa => error (snd xa))))
                 (fn x =>
                   checkB
                     (check_SCT (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_))
                       eq_nata (fn st => fn uv => edge_check (st, uv))
                       (map (fn a as (st, aa) =>
                              let
                                val (ab, b) = aa;
                              in
                                Scg (st, st, ab, b)
                              end)
                         gs))
                     [[#"s", #"i", #"z", #"e", #"-", #"c", #"h", #"a", #"n",
                        #"g", #"e", #" ", #"a", #"n", #"a", #"l", #"y", #"s",
                        #"i", #"s", #" ", #"f", #"a", #"i", #"l", #"e",
                        #"d"]])))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"s", #"i", #"z", #"e", #"-", #"c", #"h", #"a", #"n", #"g",
                 #"e", #" ", #"p", #"r", #"o", #"c", #"e", #"s", #"s", #"o",
                 #"r", #" ", #"w", #"i", #"t", #"h", #" ", #"s", #"u", #"b",
                 #"t", #"e", #"r", #"m", #"-", #"r", #"e", #"l", #"a", #"t",
                 #"i", #"o", #"n", #" ", #"f", #"o", #"r", #" ", #"t", #"h",
                 #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w", #"i", #"n",
                 #"g", #" ", #"r", #"e", #"a", #"s", #"o", #"n"]
                :: x))
  end;

fun create_sp_representation A_ (SimpProj []) = []
  | create_sp_representation A_ (SimpProj ((f, n) :: sp)) =
    concat
      [[#"p", #"i", #"("], sshow A_ f, [#")", #" ", #"=", #" "], sshow_nat n]
      :: create_sp_representation A_ (SimpProj sp);

fun check_for_var F_ V_ rule =
  catch (checkB (not (is_var (fst rule)))
          [[#"t", #"h", #"e", #" ", #"l", #"h", #"s", #" ", #"i", #"s", #" ",
             #"a", #" ", #"v", #"a", #"r", #"i", #"a", #"b", #"l", #"e"]])
    (fn x =>
      error (concat
               [sshow_term F_ V_ (fst rule),
                 [#" ", #"i", #"s", #" ", #"a", #" ", #"v", #"a", #"r", #"i",
                   #"a", #"b", #"l", #"e"]]
              :: x));

fun lift2term p (Var x) = Var x
  | lift2term p (Fun (f, ss)) = nth ss (p f);

fun check_supteq_proj (F1_, F2_) (V1_, V2_) p lr =
  catch (checkB
          (eqop (eq_terma F1_ V1_) (lift2term p (fst lr)) (lift2term p (snd lr))
            orelse
            checkSupterm F1_ V1_ (lift2term p (fst lr)) (lift2term p (snd lr)))
          [[#"t", #"h", #"e", #" ", #"p", #"r", #"o", #"j", #"e", #"c", #"t",
             #"e", #"d", #" ", #"r", #"h", #"s", #" ", #"i", #"s", #" ", #"n",
             #"o", #"t", #" ", #"a", #" ", #"s", #"u", #"b", #"t", #"e", #"r",
             #"m", #" ", #"o", #"f", #" ", #"t", #"h", #"e", #" ", #"p", #"r",
             #"o", #"j", #"e", #"c", #"t", #"e", #"d", #" ", #"l", #"h", #"s"]])
    (fn x =>
      error (concat
               [[#"C", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                  #"o", #"r", #"i", #"e", #"n", #"t", #" ", #"r", #"u", #"l",
                  #"e", #" "],
                 sshow_term F2_ V2_ (fst lr), [#" ", #"-", #">", #" "],
                 sshow_term F2_ V2_ (snd lr)]
              :: concat
                   [[#"i", #".", #"e", #".", #",", #" "],
                     sshow_term F2_ V2_ (lift2term p (fst lr)),
                     [#" ", #"|", #">", #"=", #" "],
                     sshow_term F2_ V2_ (lift2term p (snd lr))]
                   :: x));

fun check_projection (A1_, A2_) B_ ds p (Fun (f, ss), Fun (g, ts)) =
  bind (bind (bind (bind (bind (checkB (less_nat Zero_nat (size_list ss))
                                 [[#"l", #"h", #"s", #" ", #"i", #"s", #" ",
                                    #"a", #" ", #"c", #"o", #"n", #"s", #"t",
                                    #"a", #"n", #"t"]])
                           (fn x =>
                             checkB (not (member A1_ f ds))
                               [concat
                                  [sshow A2_ f,
                                    [#" ", #"i", #"s", #" ", #"d", #"e", #"f",
                                      #"i", #"n", #"e", #"d", #" ", #"i", #"n",
                                      #" ", #"R"]]]))
                     (fn x =>
                       checkB (less_nat (p f) (size_list ss))
                         [concat
                            [sshow_nat (p f),
                              [#" ", #"i", #"s", #" ", #"n", #"o", #"t", #" ",
                                #"a", #"n", #" ", #"a", #"r", #"g", #"u", #"m",
                                #"e", #"n", #"t", #" ", #"p", #"o", #"s", #"i",
                                #"t", #"i", #"o", #"n", #" ", #"o", #"f", #" "],
                              sshow_term A2_ B_ (Fun (f, ss))]]))
               (fn x =>
                 checkB (less_nat Zero_nat (size_list ts))
                   [[#"r", #"h", #"s", #" ", #"i", #"s", #" ", #"a", #" ", #"c",
                      #"o", #"n", #"s", #"t", #"a", #"n", #"t"]]))
         (fn x =>
           checkB (not (member A1_ g ds))
             [concat
                [sshow A2_ g,
                  [#" ", #"i", #"s", #" ", #"d", #"e", #"f", #"i", #"n", #"e",
                    #"d", #" ", #"i", #"n", #" ", #"R"]]]))
    (fn x =>
      checkB (less_nat (p g) (size_list ts))
        [concat
           [sshow_nat (p g),
             [#" ", #"i", #"s", #" ", #"n", #"o", #"t", #" ", #"a", #"n", #" ",
               #"a", #"r", #"g", #"u", #"m", #"e", #"n", #"t", #" ", #"p", #"o",
               #"s", #"i", #"t", #"i", #"o", #"n", #" ", #"o", #"f", #" "],
             sshow_term A2_ B_ (Fun (g, ts))]])
  | check_projection (A1_, A2_) B_ ds p (Var v, t) =
    error [[#"c", #"a", #"n", #"n", #"o", #"t", #" ", #"p", #"r", #"o", #"j",
             #"e", #"c", #"t", #" ", #"a", #" ", #"v", #"a", #"r", #"i", #"a",
             #"b", #"l", #"e"]]
  | check_projection (A1_, A2_) B_ ds p (s, Var v) =
    error [[#"c", #"a", #"n", #"n", #"o", #"t", #" ", #"p", #"r", #"o", #"j",
             #"e", #"c", #"t", #" ", #"a", #" ", #"v", #"a", #"r", #"i", #"a",
             #"b", #"l", #"e"]];

fun check_supt_proj (F1_, F2_) (V1_, V2_) p lr =
  catch (checkB
          (checkSupterm F1_ V1_ (lift2term p (fst lr)) (lift2term p (snd lr)))
          [[#"t", #"h", #"e", #" ", #"p", #"r", #"o", #"j", #"e", #"c", #"t",
             #"e", #"d", #" ", #"r", #"h", #"s", #" ", #"i", #"s", #" ", #"n",
             #"o", #"t", #" ", #"a", #" ", #"p", #"r", #"o", #"p", #"e", #"r",
             #" ", #"s", #"u", #"b", #"t", #"e", #"r", #"m", #" ", #"o", #"f",
             #" ", #"t", #"h", #"e", #" ", #"p", #"r", #"o", #"j", #"e", #"c",
             #"t", #"e", #"d", #" ", #"l", #"h", #"s"]])
    (fn x =>
      error (concat
               [[#"C", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                  #"o", #"r", #"i", #"e", #"n", #"t", #" ", #"r", #"u", #"l",
                  #"e", #" "],
                 sshow_term F2_ V2_ (fst lr), [#" ", #"-", #">", #" "],
                 sshow_term F2_ V2_ (snd lr)]
              :: concat
                   [[#"i", #".", #"e", #".", #",", #" "],
                     sshow_term F2_ V2_ (lift2term p (fst lr)),
                     [#" ", #"|", #">", #" "],
                     sshow_term F2_ V2_ (lift2term p (snd lr))]
                   :: x));

fun check_spsc_proc (A1_, A2_) (B1_, B2_) proj rep p rP r =
  let
    val ds = defs_list A1_ r;
    val a =
      partition
        (fn ra => isValue (check_supt_proj (A1_, A2_) (B1_, B2_) proj ra)) p;
    val (sp, nSp) = a;
  in
    catch (bind (bind (bind (catch
                              (catch
                                (checkAllMI
                                  (check_projection (A1_, A2_) B2_ ds proj) p)
                                (fn x => error (snd x)))
                              (fn x =>
                                error ([#"t", #"h", #"e", #" ", #"s", #"u",
 #"p", #"p", #"l", #"i", #"e", #"d", #" ", #"p", #"r", #"o", #"j", #"e", #"c",
 #"t", #"i", #"o", #"n", #" ", #"i", #"s", #" ", #"m", #"a", #"l", #"f", #"o",
 #"r", #"m", #"e", #"d"]
:: x)))
                        (fn x =>
                          catch (catch
                                  (checkAllMI
                                    (check_supteq_proj (A1_, A2_) (B1_, B2_)
                                      proj)
                                    nSp)
                                  (fn xa => error (snd xa)))
                            (fn xa =>
                              error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                       #" ", #"w", #"h", #"e", #"n", #" ", #"o",
                                       #"r", #"i", #"e", #"n", #"t", #"i", #"n",
                                       #"g", #" ", #"D", #"P", #"s"]
                                      :: xa))))
                  (fn x =>
                    catch (catch (checkAllMI (check_for_var A2_ B2_) r)
                            (fn xa => error (snd xa)))
                      (fn xa =>
                        error ([#"o", #"n", #"e", #" ", #"o", #"f", #" ", #"t",
                                 #"h", #"e", #" ", #"l", #"h", #"s", #"s", #" ",
                                 #"o", #"f", #" ", #"R", #" ", #"i", #"s", #" ",
                                 #"a", #" ", #"v", #"a", #"r", #"i", #"a", #"b",
                                 #"l", #"e"]
                                :: xa))))
            (fn x =>
              catch (checkSubset (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) nSp
                      rP)
                (fn xa =>
                  error ((case check_supt_proj (A1_, A2_) (B1_, B2_) proj xa
                           of Left err => err
                           | Right ok =>
                             [[#"i", #"n", #"t", #"e", #"r", #"n", #"a", #"l",
                                #" ", #"e", #"r", #"r", #"o", #"r"],
                               [#"p", #"l", #"e", #"a", #"s", #"e", #" ", #"c",
                                 #"o", #"n", #"t", #"a", #"c", #"t", #" ", #"o",
                                 #"n", #"e", #" ", #"o", #"f", #" ", #"t", #"h",
                                 #"e", #" ", #"C", #"e", #"T", #"A", #" ", #"m",
                                 #"e", #"m", #"b", #"e", #"r", #"s"],
                               [#"h", #"t", #"t", #"p", #":", #"/", #"/", #"c",
                                 #"l", #"-", #"i", #"n", #"f", #"o", #"r", #"m",
                                 #"a", #"t", #"i", #"k", #".", #"u", #"i", #"b",
                                 #"k", #".", #"a", #"c", #".", #"a", #"t", #"/",
                                 #"s", #"o", #"f", #"t", #"w", #"a", #"r", #"e",
                                 #"/", #"c", #"e", #"t", #"a"],
                               [#"c", #"e", #"t", #"a", #"@", #"i", #"n", #"f",
                                 #"o", #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                 #"u", #"i", #"b", #"k", #".", #"a", #"c", #".",
                                 #"a", #"t"]])))))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"s", #"u", #"b", #"t", #"e", #"r", #"m", #" ", #"c", #"r",
                 #"i", #"t", #"e", #"r", #"i", #"o", #"n", #" ", #"p", #"r",
                 #"o", #"c", #"e", #"s", #"s", #"o", #"r", #" ", #"w", #"i",
                 #"t", #"h", #" ", #"t", #"h", #"e", #" ", #"f", #"o", #"l",
                 #"l", #"o", #"w", #"i", #"n", #"g", #" ", #"s", #"i", #"m",
                 #"p", #"l", #"e", #" ", #"p", #"r", #"o", #"j", #"e", #"c",
                 #"t", #"i", #"o", #"n"]
                :: append rep ([] :: x)))
  end;

fun create_sp A_ (SimpProj sp) f =
  (case lookup A_ f sp of NONE => Zero_nat | SOME n => n);

fun create_spsc_proc (A1_, A2_) (B1_, B2_) sp p rP r =
  let
    val pi = create_sp A1_ sp;
  in
    check_spsc_proc (A1_, A2_) (B1_, B2_) pi (create_sp_representation A2_ sp) p
      rP r
  end;

fun create_sem_lab_proc (F1_, F2_) (L1_, L2_) (V1_, V2_) sli l r1 r2 r3 r4 =
  let
    val c = get_largest_element sli;
    val ca = upt Zero_nat (Suc c);
    val i = sl_inter_to_inter F1_ L1_ sli;
  in
    catch (check_sem_lab_proc (F1_, F2_) (L1_, L2_) (V1_, V2_)
            (check_sl_lab_trs (F1_, F2_) (L1_, L2_) (eq_nata, sshow_nata)
              (V1_, V2_) i l ca)
            (check_sl_model_lab_trs (F1_, F2_) (L1_, L2_) (eq_nata, sshow_nata)
              (V1_, V2_) i l ca)
            (fn x => false) i l r1 r2 r3 r4)
      (fn x => error (sshow_sl_inter F2_ L2_ sli :: x))
  end;

fun check_unlab_proc (F1_, F2_) (L1_, L2_) (V1_, V2_) p r ulP ulR =
  let
    val ulr = unlab_of_trsI F1_ L1_ V1_ r;
  in
    catch (bind (bind (checkSubset
                        (eqb (eq_terma (eq_laba F1_ L1_) V1_)
                          (eq_terma (eq_laba F1_ L1_) V1_))
                        (map_trsI unlab p) ulP)
                  (fn x =>
                    checkSubset
                      (eqb (eq_terma (eq_laba F1_ L1_) V1_)
                        (eq_terma (eq_laba F1_ L1_) V1_))
                      ulr ulR))
            (fn x =>
              checkSubset
                (eqb (eq_terma (eq_laba F1_ L1_) V1_)
                  (eq_terma (eq_laba F1_ L1_) V1_))
                ulR ulr))
      (fn x =>
        error [append
                 [#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"i",
                   #"t", #"h", #" ", #"r", #"u", #"l", #"e", #" "]
                 (showRule (sshow_laba F2_ L2_) V2_ x)])
  end;

fun checkUrClosedTRM (A1_, A2_) (B1_, B2_) uu uv (Var x) = return ()
  | checkUrClosedTRM (A1_, A2_) (B1_, B2_) rm ur (Fun (f, ts)) =
    bind (catch (checkAllMI (checkUrClosedTRM (A1_, A2_) (B1_, B2_) rm ur) ts)
           (fn x => error (snd x)))
      (fn x =>
        catch (checkAllMI
                (fn lr =>
                  checkB
                    (member (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) lr ur
                      orelse
                      not (matchCapRMBelow A1_ B1_ (fst lr) (SOME rm)
                            (Fun (f, ts))))
                    (append
                      (append
                        (append
                          (append
                            [#"d", #"u", #"e", #" ", #"t", #"o", #" ", #"t",
                              #"h", #"e", #" ", #"s", #"u", #"b", #"t", #"e",
                              #"r", #"m", #" "]
                            (sshow_term A2_ B2_ (Fun (f, ts))))
                          [#" ", #"o", #"f", #" ", #"s", #"o", #"m", #"e", #" ",
                            #"u", #"s", #"a", #"b", #"l", #"e", #" ", #"r",
                            #"h", #"s", #",", #" ", #"r", #"u", #"l", #"e",
                            #" "])
                        (showRule A2_ B2_ lr))
                      [#" ", #"s", #"h", #"o", #"u", #"l", #"d", #" ", #"b",
                        #"e", #" ", #"u", #"s", #"a", #"b", #"l", #"e", #"."]))
                (rm f))
          (fn xa => error (snd xa)));

fun matchCapBelowI A_ B_ l r (Fun (f, ts)) =
  gcMatchI A_ B_ (HFun (f, map (capI A_ B_ r) ts)) l;

fun checkUrClosedT (A1_, A2_) (B1_, B2_) uu uv (Var x) = return ()
  | checkUrClosedT (A1_, A2_) (B1_, B2_) r ur (Fun (f, ts)) =
    bind (catch (checkAllMI (checkUrClosedT (A1_, A2_) (B1_, B2_) r ur) ts)
           (fn x => error (snd x)))
      (fn x =>
        catch (checkAllMI
                (fn lr =>
                  checkB
                    (member (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) lr ur
                      orelse
                      not (matchCapBelowI A1_ B1_ (fst lr) r (Fun (f, ts))))
                    (append
                      (append
                        (append
                          (append
                            [#"d", #"u", #"e", #" ", #"t", #"o", #" ", #"t",
                              #"h", #"e", #" ", #"s", #"u", #"b", #"t", #"e",
                              #"r", #"m", #" "]
                            (sshow_term A2_ B2_ (Fun (f, ts))))
                          [#" ", #"o", #"f", #" ", #"s", #"o", #"m", #"e", #" ",
                            #"u", #"s", #"a", #"b", #"l", #"e", #" ", #"r",
                            #"h", #"s", #",", #" ", #"r", #"u", #"l", #"e",
                            #" "])
                        (showRule A2_ B2_ lr))
                      [#" ", #"s", #"h", #"o", #"u", #"l", #"d", #" ", #"b",
                        #"e", #" ", #"u", #"s", #"a", #"b", #"l", #"e", #"."]))
                r)
          (fn xa => error (snd xa)));

fun checkUrPClosed (A1_, A2_) (B1_, B2_) r ur p =
  bind (catch
         (catch
           (checkAllMI
             (fn lr => checkUrClosedT (A1_, A2_) (B1_, B2_) r ur (snd lr)) ur)
           (fn x => error (snd x)))
         (fn x =>
           error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e", #"n",
                    #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n", #"g", #" ",
                    #"c", #"l", #"o", #"s", #"u", #"r", #"e", #" ", #"p", #"r",
                    #"o", #"p", #"e", #"r", #"t", #"i", #"e", #"s", #" ", #"o",
                    #"f", #" ", #"r", #"h", #"s", #" ", #"o", #"f", #" ", #"u",
                    #"s", #"a", #"b", #"l", #"e", #" ", #"r", #"u", #"l", #"e",
                    #"s"],
                   x]))
    (fn x =>
      catch (catch
              (checkAllMI
                (fn st => checkUrClosedT (A1_, A2_) (B1_, B2_) r ur (snd st)) p)
              (fn xa => error (snd xa)))
        (fn xa =>
          error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e", #"n",
                   #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n", #"g", #" ",
                   #"c", #"l", #"o", #"s", #"u", #"r", #"e", #" ", #"p", #"r",
                   #"o", #"p", #"e", #"r", #"t", #"i", #"e", #"s", #" ", #"o",
                   #"f", #" ", #"r", #"h", #"s", #" ", #"o", #"f", #" ", #"D",
                   #"P", #"s"],
                  xa]));

fun checkUrPClosedRM (A1_, A2_) (B1_, B2_) r rmo ur p =
  (case rmo of NONE => checkUrPClosed (A1_, A2_) (B1_, B2_) r ur p
     | SOME rm =>
       bind (catch
              (catch
                (checkAllMI
                  (fn lr =>
                    checkUrClosedTRM (A1_, A2_) (B1_, B2_) rm ur (snd lr))
                  ur)
                (fn x => error (snd x)))
              (fn x =>
                error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e",
                         #"n", #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n",
                         #"g", #" ", #"c", #"l", #"o", #"s", #"u", #"r", #"e",
                         #" ", #"p", #"r", #"o", #"p", #"e", #"r", #"t", #"i",
                         #"e", #"s", #" ", #"o", #"f", #" ", #"r", #"h", #"s",
                         #" ", #"o", #"f", #" ", #"u", #"s", #"a", #"b", #"l",
                         #"e", #" ", #"r", #"u", #"l", #"e", #"s"],
                        x]))
         (fn x =>
           catch (catch
                   (checkAllMI
                     (fn st =>
                       checkUrClosedTRM (A1_, A2_) (B1_, B2_) rm ur (snd st))
                     p)
                   (fn xa => error (snd xa)))
             (fn xa =>
               error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e",
                        #"n", #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n",
                        #"g", #" ", #"c", #"l", #"o", #"s", #"u", #"r", #"e",
                        #" ", #"p", #"r", #"o", #"p", #"e", #"r", #"t", #"i",
                        #"e", #"s", #" ", #"o", #"f", #" ", #"r", #"h", #"s",
                        #" ", #"o", #"f", #" ", #"D", #"P", #"s"],
                       xa])));

fun checkMonoRedPairUrProc (A1_, A2_) (B1_, B2_) checkMono checkSS checkNSS name
  representation p rP r rm rR ur =
  let
    val a = partition (fn foo => isValue (checkSS foo)) p;
    val (sp, nSp) = a;
    val aa = partition (fn foo => isValue (checkSS foo)) ur;
    val (sr, nSr) = aa;
  in
    catch (bind (bind (bind (bind (bind (bind
  (catch checkMono
    (fn x =>
      error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"i", #"t",
               #"h", #" ", #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c",
               #"i", #"t", #"y", #" ", #"o", #"f", #" ", #"s", #"t", #"r", #"i",
               #"c", #"t", #" ", #"o", #"r", #"d", #"e", #"r"]
              :: x)))
  (fn x =>
    catch (checkUrPClosedRM (A1_, A2_) (B1_, B2_) r rm ur p)
      (fn xa =>
        error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"i",
                 #"t", #"h", #" ", #"c", #"l", #"o", #"s", #"u", #"r", #"e",
                 #" ", #"p", #"r", #"o", #"p", #"e", #"r", #"t", #"i", #"e",
                 #"s", #" ", #"o", #"f", #" ", #"u", #"s", #"a", #"b", #"l",
                 #"e", #" ", #"r", #"u", #"l", #"e", #"s"]
                :: xa))))
                                    (fn x =>
                                      catch
(catch (checkAllMI checkNSS nSr) (fn xa => error (snd xa)))
(fn xa =>
  error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"h", #"e",
           #"n", #" ", #"o", #"r", #"i", #"e", #"n", #"t", #"i", #"n", #"g",
           #" ", #"u", #"s", #"a", #"b", #"l", #"e", #" ", #"r", #"u", #"l",
           #"e", #"s"]
          :: xa))))
                              (fn x =>
                                catch (catch (checkAllMI checkNSS nSp)
(fn xa => error (snd xa)))
                                  (fn xa =>
                                    error ([#"p", #"r", #"o", #"b", #"l", #"e",
     #"m", #" ", #"w", #"h", #"e", #"n", #" ", #"o", #"r", #"i", #"e", #"n",
     #"t", #"i", #"n", #"g", #" ", #"D", #"P", #"s"]
    :: xa))))
                        (fn x =>
                          catch (checkSubset
                                  (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) rR
                                  r)
                            (fn xa =>
                              error [[#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                       #":", #" ", #"a", #"d", #"d", #"i", #"t",
                                       #"i", #"o", #"n", #"a", #"l", #" ", #"r",
                                       #"u", #"l", #"e", #"(", #"s", #")", #" ",
                                       #"t", #"o", #" ", #"T", #"R", #"S", #" ",
                                       #"a", #"d", #"d", #"e", #"d"]])))
                  (fn x =>
                    catch (checkSubset
                            (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) nSr rR)
                      (fn xa =>
                        error ((case checkSS xa of Left err => err
                                 | Right ok =>
                                   [[#"i", #"n", #"t", #"e", #"r", #"n", #"a",
                                      #"l", #" ", #"e", #"r", #"r", #"o", #"r"],
                                     [#"p", #"l", #"e", #"a", #"s", #"e", #" ",
                                       #"c", #"o", #"n", #"t", #"a", #"c", #"t",
                                       #" ", #"o", #"n", #"e", #" ", #"o", #"f",
                                       #" ", #"t", #"h", #"e", #" ", #"C", #"e",
                                       #"T", #"A", #" ", #"m", #"e", #"m", #"b",
                                       #"e", #"r", #"s"],
                                     [#"h", #"t", #"t", #"p", #":", #"/", #"/",
                                       #"c", #"l", #"-", #"i", #"n", #"f", #"o",
                                       #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                       #"u", #"i", #"b", #"k", #".", #"a", #"c",
                                       #".", #"a", #"t", #"/", #"s", #"o", #"f",
                                       #"t", #"w", #"a", #"r", #"e", #"/", #"c",
                                       #"e", #"t", #"a"],
                                     [#"c", #"e", #"t", #"a", #"@", #"i", #"n",
                                       #"f", #"o", #"r", #"m", #"a", #"t", #"i",
                                       #"k", #".", #"u", #"i", #"b", #"k", #".",
                                       #"a", #"c", #".", #"a", #"t"]])))))
            (fn x =>
              catch (checkSubset (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) nSp
                      rP)
                (fn xa =>
                  error ((case checkSS xa of Left err => err
                           | Right ok =>
                             [[#"i", #"n", #"t", #"e", #"r", #"n", #"a", #"l",
                                #" ", #"e", #"r", #"r", #"o", #"r"],
                               [#"p", #"l", #"e", #"a", #"s", #"e", #" ", #"c",
                                 #"o", #"n", #"t", #"a", #"c", #"t", #" ", #"o",
                                 #"n", #"e", #" ", #"o", #"f", #" ", #"t", #"h",
                                 #"e", #" ", #"C", #"e", #"T", #"A", #" ", #"m",
                                 #"e", #"m", #"b", #"e", #"r", #"s"],
                               [#"h", #"t", #"t", #"p", #":", #"/", #"/", #"c",
                                 #"l", #"-", #"i", #"n", #"f", #"o", #"r", #"m",
                                 #"a", #"t", #"i", #"k", #".", #"u", #"i", #"b",
                                 #"k", #".", #"a", #"c", #".", #"a", #"t", #"/",
                                 #"s", #"o", #"f", #"t", #"w", #"a", #"r", #"e",
                                 #"/", #"c", #"e", #"t", #"a"],
                               [#"c", #"e", #"t", #"a", #"@", #"i", #"n", #"f",
                                 #"o", #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                 #"u", #"i", #"b", #"k", #".", #"a", #"c", #".",
                                 #"a", #"t"]])))))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"(", #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c",
                 #")", #" ", #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o",
                 #"n", #" ", #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o",
                 #"c", #"e", #"s", #"s", #"o", #"r", #" ", #"w", #"i", #"t",
                 #"h", #" ", #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l",
                 #"o", #"w", #"i", #"n", #"g"]
                :: append representation
                     ([] ::
                       [#"a", #"n", #"d", #" ", #"t", #"h", #"e", #" ", #"f",
                         #"o", #"l", #"l", #"o", #"w", #"i", #"n", #"g", #" ",
                         #"u", #"s", #"a", #"b", #"l", #"e", #" ", #"r", #"u",
                         #"l", #"e", #"s"]
                         :: append (showRules A2_ B2_ ur) ([] :: x))))
  end;

fun create_mono_redpair_ur_proc_comp (A1_, A2_) (B1_, B2_) (C1_, C2_) c cm i p
  rP r rm rR ur =
  let
    val pi = to_lpoly_inter B1_ (get_comp_1 c) i;
  in
    bind (check_lpoly_coeffs A2_ B2_ (get_comp_wf_elem c) (get_comp_apos c)
           (fn x => get_comp_ge c x (get_comp_0 c)) i)
      (fn x =>
        checkMonoRedPairUrProc (B1_, B2_) (C1_, C2_)
          (check_poly_monoI A2_ B2_ (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (A1_, A2_) B2_ (C1_, C2_) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (A1_, A2_) B2_ (C1_, C2_) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (append
            (append
              [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l", #" ",
                #"o", #"r", #"d", #"e", #"r", #" ", #"("]
              (get_comp_description c))
            [#")"])
          (create_lpoly_repr (A1_, A2_) (B1_, B2_) c i) p rP r rm rR ur)
  end;

fun create_mono_redpair_ur_proc (A1_, A2_) (B1_, B2_) (Int_carrier ia) =
  create_mono_redpair_ur_proc_comp (eq_int, sshow_inta) (A1_, A2_) (B1_, B2_)
    ((fn x => true),
      ((0 : IntInf.int),
        ((1 : IntInf.int),
          ((fn a => fn b => IntInf.+ (a, b)),
            ((fn a => fn b => IntInf.* (a, b)),
              (max ord_int (0 : IntInf.int),
                ((fn x => fn y => IntInf.<= (y, x)),
                  ((fn x => fn y => IntInf.< (y, x)),
                    (true, (int_name, (fn x => true)))))))))))
    (fn a => IntInf.< ((0 : IntInf.int), a)) ia
  | create_mono_redpair_ur_proc (A1_, A2_) (B1_, B2_) (Nat_carrier i) =
    create_mono_redpair_ur_proc_comp (eq_nata, sshow_nata) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (Zero_nat,
          (one_nat,
            (plus_nat,
              (times_nat,
                (id, ((fn x => fn y => less_eq_nat y x),
                       ((fn x => fn y => less_nat y x),
                         (true, (nat_name, (fn x => true)))))))))))
      (fn x => not (eqop eq_nata x Zero_nat)) i
  | create_mono_redpair_ur_proc (A1_, A2_) (B1_, B2_) (Arctic_carrier id) =
    (fn uu => fn uua => fn uub => fn uuc => fn uud => fn uue =>
      error [[#"a", #"r", #"c", #"t", #"i", #"c", #" ", #"n", #"o", #"t", #" ",
               #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ", #"f",
               #"o", #"r", #" ", #"T", #"R", #"S", #"s"]])
  | create_mono_redpair_ur_proc (A1_, A2_) (B1_, B2_) (Arc_mat_carrier (n, ic))
    = (fn uu => fn uua => fn uub => fn uuc => fn uud => fn uue =>
        error [[#"a", #"r", #"c", #"t", #"i", #"c", #" ", #"n", #"o", #"t",
                 #" ", #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c",
                 #" ", #"f", #"o", #"r", #" ", #"T", #"R", #"S", #"s"]])
  | create_mono_redpair_ur_proc (A1_, A2_) (B1_, B2_) (Int_mat_carrier (n, ib))
    = (fn uu => fn uua => fn uub => fn uuc => fn uud => fn uue =>
        error [[#"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ",
                 #"m", #"a", #"t", #"r", #"i", #"x", #"-", #"o", #"r", #"d",
                 #"e", #"r", #"s", #" ", #"n", #"o", #"t", #" ", #"y", #"e",
                 #"t", #" ", #"s", #"u", #"p", #"p", #"o", #"r", #"t", #"e",
                 #"d"]]);

fun checkMonoRedPairProc A_ B_ checkMono checkSS checkNSS name representation p
  rP r rR =
  let
    val a = partition (fn foo => isValue (checkSS foo)) p;
    val (sp, nSp) = a;
    val aa = partition (fn foo => isValue (checkSS foo)) r;
    val (sr, nSr) = aa;
  in
    catch (bind (bind (bind (bind (bind (catch checkMono
  (fn x =>
    error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"i", #"t",
             #"h", #" ", #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c",
             #"i", #"t", #"y", #" ", #"o", #"f", #" ", #"s", #"t", #"r", #"i",
             #"c", #"t", #" ", #"o", #"r", #"d", #"e", #"r"]
            :: x)))
                                    (fn x =>
                                      catch
(catch (checkAllMI checkNSS nSr) (fn xa => error (snd xa)))
(fn xa =>
  error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"w", #"h", #"e",
           #"n", #" ", #"o", #"r", #"i", #"e", #"n", #"t", #"i", #"n", #"g",
           #" ", #"T", #"R", #"S"]
          :: xa))))
                              (fn x =>
                                catch (catch (checkAllMI checkNSS nSp)
(fn xa => error (snd xa)))
                                  (fn xa =>
                                    error ([#"p", #"r", #"o", #"b", #"l", #"e",
     #"m", #" ", #"w", #"h", #"e", #"n", #" ", #"o", #"r", #"i", #"e", #"n",
     #"t", #"i", #"n", #"g", #" ", #"D", #"P", #"s"]
    :: xa))))
                        (fn x =>
                          catch (checkSubset
                                  (eqb (eq_terma A_ B_) (eq_terma A_ B_)) rR r)
                            (fn xa =>
                              error [[#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                       #":", #" ", #"a", #"d", #"d", #"i", #"t",
                                       #"i", #"o", #"n", #"a", #"l", #" "]])))
                  (fn x =>
                    catch (checkSubset (eqb (eq_terma A_ B_) (eq_terma A_ B_))
                            nSr rR)
                      (fn xa =>
                        error ((case checkSS xa of Left err => err
                                 | Right ok =>
                                   [[#"i", #"n", #"t", #"e", #"r", #"n", #"a",
                                      #"l", #" ", #"e", #"r", #"r", #"o", #"r"],
                                     [#"p", #"l", #"e", #"a", #"s", #"e", #" ",
                                       #"c", #"o", #"n", #"t", #"a", #"c", #"t",
                                       #" ", #"o", #"n", #"e", #" ", #"o", #"f",
                                       #" ", #"t", #"h", #"e", #" ", #"C", #"e",
                                       #"T", #"A", #" ", #"m", #"e", #"m", #"b",
                                       #"e", #"r", #"s"],
                                     [#"h", #"t", #"t", #"p", #":", #"/", #"/",
                                       #"c", #"l", #"-", #"i", #"n", #"f", #"o",
                                       #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                       #"u", #"i", #"b", #"k", #".", #"a", #"c",
                                       #".", #"a", #"t", #"/", #"s", #"o", #"f",
                                       #"t", #"w", #"a", #"r", #"e", #"/", #"c",
                                       #"e", #"t", #"a"],
                                     [#"c", #"e", #"t", #"a", #"@", #"i", #"n",
                                       #"f", #"o", #"r", #"m", #"a", #"t", #"i",
                                       #"k", #".", #"u", #"i", #"b", #"k", #".",
                                       #"a", #"c", #".", #"a", #"t"]])))))
            (fn x =>
              catch (checkSubset (eqb (eq_terma A_ B_) (eq_terma A_ B_)) nSp rP)
                (fn xa =>
                  error ((case checkSS xa of Left err => err
                           | Right ok =>
                             [[#"i", #"n", #"t", #"e", #"r", #"n", #"a", #"l",
                                #" ", #"e", #"r", #"r", #"o", #"r"],
                               [#"p", #"l", #"e", #"a", #"s", #"e", #" ", #"c",
                                 #"o", #"n", #"t", #"a", #"c", #"t", #" ", #"o",
                                 #"n", #"e", #" ", #"o", #"f", #" ", #"t", #"h",
                                 #"e", #" ", #"C", #"e", #"T", #"A", #" ", #"m",
                                 #"e", #"m", #"b", #"e", #"r", #"s"],
                               [#"h", #"t", #"t", #"p", #":", #"/", #"/", #"c",
                                 #"l", #"-", #"i", #"n", #"f", #"o", #"r", #"m",
                                 #"a", #"t", #"i", #"k", #".", #"u", #"i", #"b",
                                 #"k", #".", #"a", #"c", #".", #"a", #"t", #"/",
                                 #"s", #"o", #"f", #"t", #"w", #"a", #"r", #"e",
                                 #"/", #"c", #"e", #"t", #"a"],
                               [#"c", #"e", #"t", #"a", #"@", #"i", #"n", #"f",
                                 #"o", #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                 #"u", #"i", #"b", #"k", #".", #"a", #"c", #".",
                                 #"a", #"t"]])))))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n", #" ",
                 #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o", #"c", #"e",
                 #"s", #"s", #"o", #"r", #" ", #"w", #"i", #"t", #"h", #" ",
                 #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                 #"i", #"n", #"g"]
                :: append representation ([] :: x)))
  end;

fun create_mono_redpair_proc_comp (A1_, A2_) (B1_, B2_) (C1_, C2_) c cm i p rP r
  rR =
  let
    val pi = to_lpoly_inter B1_ (get_comp_1 c) i;
  in
    bind (check_lpoly_coeffs A2_ B2_ (get_comp_wf_elem c) (get_comp_apos c)
           (fn x => get_comp_ge c x (get_comp_0 c)) i)
      (fn x =>
        checkMonoRedPairProc B1_ C1_
          (check_poly_monoI A2_ B2_ (get_comp_0 c) (get_comp_ge c) cm i)
          (check_polo_sI (A1_, A2_) B2_ (C1_, C2_) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (A1_, A2_) B2_ (C1_, C2_) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (append
            (append
              [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l", #" ",
                #"o", #"r", #"d", #"e", #"r", #" ", #"("]
              (get_comp_description c))
            [#")"])
          (create_lpoly_repr (A1_, A2_) (B1_, B2_) c i) p rP r rR)
  end;

fun create_mono_redpair_proc (A1_, A2_) (B1_, B2_) (Int_carrier ia) =
  create_mono_redpair_proc_comp (eq_int, sshow_inta) (A1_, A2_) (B1_, B2_)
    ((fn x => true),
      ((0 : IntInf.int),
        ((1 : IntInf.int),
          ((fn a => fn b => IntInf.+ (a, b)),
            ((fn a => fn b => IntInf.* (a, b)),
              (max ord_int (0 : IntInf.int),
                ((fn x => fn y => IntInf.<= (y, x)),
                  ((fn x => fn y => IntInf.< (y, x)),
                    (true, (int_name, (fn x => true)))))))))))
    (fn a => IntInf.< ((0 : IntInf.int), a)) ia
  | create_mono_redpair_proc (A1_, A2_) (B1_, B2_) (Nat_carrier i) =
    create_mono_redpair_proc_comp (eq_nata, sshow_nata) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (Zero_nat,
          (one_nat,
            (plus_nat,
              (times_nat,
                (id, ((fn x => fn y => less_eq_nat y x),
                       ((fn x => fn y => less_nat y x),
                         (true, (nat_name, (fn x => true)))))))))))
      (fn x => not (eqop eq_nata x Zero_nat)) i
  | create_mono_redpair_proc (A1_, A2_) (B1_, B2_) (Arctic_carrier id) =
    (fn uu => fn uua => fn uub => fn uuc =>
      error [[#"a", #"r", #"c", #"t", #"i", #"c", #" ", #"n", #"o", #"t", #" ",
               #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ", #"f",
               #"o", #"r", #" ", #"T", #"R", #"S", #"s"]])
  | create_mono_redpair_proc (A1_, A2_) (B1_, B2_) (Arc_mat_carrier (n, ic)) =
    (fn uu => fn uua => fn uub => fn uuc =>
      error [[#"a", #"r", #"c", #"t", #"i", #"c", #" ", #"n", #"o", #"t", #" ",
               #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ", #"f",
               #"o", #"r", #" ", #"T", #"R", #"S", #"s"]])
  | create_mono_redpair_proc (A1_, A2_) (B1_, B2_) (Int_mat_carrier (n, ib)) =
    (fn uu => fn uua => fn uub => fn uuc =>
      error [[#"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ", #"m",
               #"a", #"t", #"r", #"i", #"x", #"-", #"o", #"r", #"d", #"e", #"r",
               #"s", #" ", #"n", #"o", #"t", #" ", #"y", #"e", #"t", #" ", #"s",
               #"u", #"p", #"p", #"o", #"r", #"t", #"e", #"d"]]);

fun checkAllMIndex g f xs =
  snd (foldl (fn a as (n, mon) => fn x => (g n x, bind mon (fn xa => f n x)))
        (one_nat, return ()) xs);

fun checkDepGraphProc (A1_, A2_) (B1_, B2_) p dps rm rrm =
  bind (catch
         (checkSubset (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) p
           (flat_map snd dps))
         (fn x =>
           error [concat
                    [[#"D", #"e", #"p", #"e", #"n", #"d", #"e", #"n", #"c",
                       #"y", #" ", #"P", #"a", #"i", #"r", #" "],
                      showRule A2_ B2_ x,
                      [#" ", #"i", #"s", #" ", #"m", #"i", #"s", #"s", #"i",
                        #"n", #"g", #" ", #"i", #"n", #" ", #"d", #"e", #"c",
                        #"o", #"m", #"p", #"o", #"s", #"i", #"t", #"i", #"o",
                        #"n"]]]))
    (fn x =>
      catch (check_graph_decomp
              (sshowa (sshow_terma A2_ B2_) (sshow_terma A2_ B2_) o fst)
              (fn a as (aa, b) =>
                let
                  val (ab, ba) = aa;
                in
                  let
                    val (uu, t) = ab;
                  in
                    (fn ac as (uua, ct) => fn ad as (ae, bb) =>
                      let
                        val (u, uub) = ae;
                      in
                        (fn af as (cu, uuc) =>
                          gcMatchI A1_ B1_ ct u andalso gcMatchI A1_ B1_ cu t)
                      end
                        bb)
                  end
                    ba
                end
                  b)
              (map (fn a as (real, cs) =>
                     (real,
                       map (fn aa as (s, t) =>
                             ((s, t),
                               (capRM A1_ B1_ rrm s, capRM A1_ B1_ rm t)))
                         cs))
                dps))
        (fn xa =>
          error [[#"o", #"u", #"r", #" ", #"e", #"s", #"t", #"i", #"m", #"a",
                   #"t", #"i", #"o", #"n", #" ", #"(", #"E", #"D", #"G", #"*",
                   #"*", #"*", #")", #" ", #"c", #"o", #"u", #"l", #"d", #" ",
                   #"n", #"o", #"t", #" ", #"s", #"h", #"o", #"w", #" ", #"t",
                   #"h", #"a", #"t", #" ", #"y", #"o", #"u", #" ", #"g", #"a",
                   #"v", #"e", #" ", #"a", #" ", #"v", #"a", #"l", #"i", #"d",
                   #" ", #"d", #"e", #"c", #"o", #"m", #"p", #"o", #"s", #"i",
                   #"t", #"i", #"o", #"n", #" ", #"d", #"u", #"e", #" ", #"t",
                   #"o", #" ", #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l",
                   #"o", #"w", #"i", #"n", #"g", #" ", #"r", #"e", #"a", #"s",
                   #"o", #"n"],
                  xa]));

fun checkUrClosedT_af (A1_, A2_) (B1_, B2_) uu uv uw (Var x) = return ()
  | checkUrClosedT_af (A1_, A2_) (B1_, B2_) r ur pi (Fun (f, ts)) =
    let
      val n = size_list ts;
    in
      bind (catch
             (checkAllMI
               (fn i =>
                 (if pi f n i
                   then checkUrClosedT_af (A1_, A2_) (B1_, B2_) r ur pi
                          (nth ts i)
                   else return ()))
               (upt Zero_nat n))
             (fn x => error (snd x)))
        (fn x =>
          catch (checkAllMI
                  (fn lr =>
                    checkB
                      (member (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) lr ur
                        orelse
                        not (matchCapBelowI A1_ B1_ (fst lr) r (Fun (f, ts))))
                      (append
                        (append
                          (append
                            (append
                              [#"d", #"u", #"e", #" ", #"t", #"o", #" ", #"t",
                                #"h", #"e", #" ", #"s", #"u", #"b", #"t", #"e",
                                #"r", #"m", #" "]
                              (sshow_term A2_ B2_ (Fun (f, ts))))
                            [#" ", #"o", #"f", #" ", #"s", #"o", #"m", #"e",
                              #" ", #"u", #"s", #"a", #"b", #"l", #"e", #" ",
                              #"r", #"h", #"s", #",", #" ", #"r", #"u", #"l",
                              #"e", #" "])
                          (showRule A2_ B2_ lr))
                        [#" ", #"s", #"h", #"o", #"u", #"l", #"d", #" ", #"b",
                          #"e", #" ", #"u", #"s", #"a", #"b", #"l", #"e",
                          #"."]))
                  r)
            (fn xa => error (snd xa)))
    end;

fun checkUrPClosed_af (A1_, A2_) (B1_, B2_) r ur p pi =
  bind (catch
         (catch
           (checkAllMI
             (fn lr => checkUrClosedT_af (A1_, A2_) (B1_, B2_) r ur pi (snd lr))
             ur)
           (fn x => error (snd x)))
         (fn x =>
           error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e", #"n",
                    #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n", #"g", #" ",
                    #"c", #"l", #"o", #"s", #"u", #"r", #"e", #" ", #"p", #"r",
                    #"o", #"p", #"e", #"r", #"t", #"i", #"e", #"s", #" ", #"o",
                    #"f", #" ", #"r", #"h", #"s", #" ", #"o", #"f", #" ", #"u",
                    #"s", #"a", #"b", #"l", #"e", #" ", #"r", #"u", #"l", #"e",
                    #"s"],
                   x]))
    (fn x =>
      catch (catch
              (checkAllMI
                (fn st =>
                  checkUrClosedT_af (A1_, A2_) (B1_, B2_) r ur pi (snd st))
                p)
              (fn xa => error (snd xa)))
        (fn xa =>
          error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e", #"n",
                   #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n", #"g", #" ",
                   #"c", #"l", #"o", #"s", #"u", #"r", #"e", #" ", #"p", #"r",
                   #"o", #"p", #"e", #"r", #"t", #"i", #"e", #"s", #" ", #"o",
                   #"f", #" ", #"r", #"h", #"s", #" ", #"o", #"f", #" ", #"D",
                   #"P", #"s"],
                  xa]));

fun checkUrPClosedRM_af (A1_, A2_) (B1_, B2_) r rmo ur p pi =
  (case rmo of NONE => checkUrPClosed_af (A1_, A2_) (B1_, B2_) r ur p pi
     | SOME rm =>
       bind (catch
              (catch
                (checkAllMI
                  (fn lr =>
                    checkUrClosedTRM_af (A1_, A2_) (B1_, B2_) rm ur pi (snd lr))
                  ur)
                (fn x => error (snd x)))
              (fn x =>
                error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e",
                         #"n", #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n",
                         #"g", #" ", #"c", #"l", #"o", #"s", #"u", #"r", #"e",
                         #" ", #"p", #"r", #"o", #"p", #"e", #"r", #"t", #"i",
                         #"e", #"s", #" ", #"o", #"f", #" ", #"r", #"h", #"s",
                         #" ", #"o", #"f", #" ", #"u", #"s", #"a", #"b", #"l",
                         #"e", #" ", #"r", #"u", #"l", #"e", #"s"],
                        x]))
         (fn x =>
           catch (catch
                   (checkAllMI
                     (fn st =>
                       checkUrClosedTRM_af (A1_, A2_) (B1_, B2_) rm ur pi
                         (snd st))
                     p)
                   (fn xa => error (snd xa)))
             (fn xa =>
               error [[#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"e",
                        #"n", #" ", #"c", #"h", #"e", #"c", #"k", #"i", #"n",
                        #"g", #" ", #"c", #"l", #"o", #"s", #"u", #"r", #"e",
                        #" ", #"p", #"r", #"o", #"p", #"e", #"r", #"t", #"i",
                        #"e", #"s", #" ", #"o", #"f", #" ", #"r", #"h", #"s",
                        #" ", #"o", #"f", #" ", #"D", #"P", #"s"],
                       xa])));

fun checkRedPairUrAfProc (A1_, A2_) (B1_, B2_) checkSS checkNSS name
  representation p rP r rm ur pi =
  let
    val a = partition (fn foo => isValue (checkSS foo)) p;
    val (sp, nSp) = a;
  in
    catch (bind (bind (bind (catch
                              (checkUrPClosedRM_af (A1_, A2_) (B1_, B2_) r rm ur
                                p pi)
                              (fn x =>
                                error ([#"p", #"r", #"o", #"b", #"l", #"e",
 #"m", #" ", #"w", #"i", #"t", #"h", #" ", #"c", #"l", #"o", #"s", #"u", #"r",
 #"e", #" ", #"p", #"r", #"o", #"p", #"e", #"r", #"t", #"i", #"e", #"s", #" ",
 #"o", #"f", #" ", #"u", #"s", #"a", #"b", #"l", #"e", #" ", #"r", #"u", #"l",
 #"e", #"s", #" ", #"w", #".", #"r", #".", #"t", #".", #" ", #"a", #"r", #"g",
 #"u", #"m", #"e", #"n", #"t", #" ", #"f", #"i", #"l", #"t", #"e", #"r"]
:: x)))
                        (fn x =>
                          catch (catch (checkAllMI checkNSS ur)
                                  (fn xa => error (snd xa)))
                            (fn xa =>
                              error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m",
                                       #" ", #"w", #"h", #"e", #"n", #" ", #"o",
                                       #"r", #"i", #"e", #"n", #"t", #"i", #"n",
                                       #"g", #" ", #"u", #"s", #"a", #"b", #"l",
                                       #"e", #" ", #"r", #"u", #"l", #"e", #"s"]
                                      :: xa))))
                  (fn x =>
                    catch (catch (checkAllMI checkNSS nSp)
                            (fn xa => error (snd xa)))
                      (fn xa =>
                        error ([#"p", #"r", #"o", #"b", #"l", #"e", #"m", #" ",
                                 #"w", #"h", #"e", #"n", #" ", #"o", #"r", #"i",
                                 #"e", #"n", #"t", #"i", #"n", #"g", #" ", #"D",
                                 #"P", #"s"]
                                :: xa))))
            (fn x =>
              catch (checkSubset (eqb (eq_terma A1_ B1_) (eq_terma A1_ B1_)) nSp
                      rP)
                (fn xa =>
                  error ((case checkSS xa of Left err => err
                           | Right ok =>
                             [[#"i", #"n", #"t", #"e", #"r", #"n", #"a", #"l",
                                #" ", #"e", #"r", #"r", #"o", #"r"],
                               [#"p", #"l", #"e", #"a", #"s", #"e", #" ", #"c",
                                 #"o", #"n", #"t", #"a", #"c", #"t", #" ", #"o",
                                 #"n", #"e", #" ", #"o", #"f", #" ", #"t", #"h",
                                 #"e", #" ", #"C", #"e", #"T", #"A", #" ", #"m",
                                 #"e", #"m", #"b", #"e", #"r", #"s"],
                               [#"h", #"t", #"t", #"p", #":", #"/", #"/", #"c",
                                 #"l", #"-", #"i", #"n", #"f", #"o", #"r", #"m",
                                 #"a", #"t", #"i", #"k", #".", #"u", #"i", #"b",
                                 #"k", #".", #"a", #"c", #".", #"a", #"t", #"/",
                                 #"s", #"o", #"f", #"t", #"w", #"a", #"r", #"e",
                                 #"/", #"c", #"e", #"t", #"a"],
                               [#"c", #"e", #"t", #"a", #"@", #"i", #"n", #"f",
                                 #"o", #"r", #"m", #"a", #"t", #"i", #"k", #".",
                                 #"u", #"i", #"b", #"k", #".", #"a", #"c", #".",
                                 #"a", #"t"]])))))
      (fn x =>
        error ([#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                 #"a", #"p", #"p", #"l", #"y", #" ", #"t", #"h", #"e", #" ",
                 #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n", #" ",
                 #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o", #"c", #"e",
                 #"s", #"s", #"o", #"r", #" ", #"w", #"i", #"t", #"h", #" ",
                 #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                 #"i", #"n", #"g"]
                :: append representation
                     ([] ::
                       [#"a", #"n", #"d", #" ", #"t", #"h", #"e", #" ", #"f",
                         #"o", #"l", #"l", #"o", #"w", #"i", #"n", #"g", #" ",
                         #"u", #"s", #"a", #"b", #"l", #"e", #" ", #"r", #"u",
                         #"l", #"e", #"s"]
                         :: append (showRules A2_ B2_ ur) ([] :: x))))
  end;

fun create_redpair_ur_af_proc_comp (A1_, A2_) (B1_, B2_) (C1_, C2_) c i p rP r
  rm ur =
  let
    val pi = to_lpoly_inter B1_ (get_comp_1 c) i;
  in
    bind (check_lpoly_coeffs A2_ B2_ (get_comp_wf_elem c) (get_comp_apos c)
           (fn x => get_comp_ge c x (get_comp_0 c)) i)
      (fn x =>
        checkRedPairUrAfProc (B1_, B2_) (C1_, C2_)
          (check_polo_sI (A1_, A2_) B2_ (C1_, C2_) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (A1_, A2_) B2_ (C1_, C2_) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (append
            (append
              [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l", #" ",
                #"o", #"r", #"d", #"e", #"r", #" ", #"("]
              (get_comp_description c))
            [#")"])
          (create_lpoly_repr (A1_, A2_) (B1_, B2_) c i) p rP r rm ur
          (create_afI A1_ B1_ (get_comp_0 c) i))
  end;

fun create_redpair_ur_af_proc (A1_, A2_) (B1_, B2_) (Int_carrier ia) =
  create_redpair_ur_af_proc_comp (eq_int, sshow_inta) (A1_, A2_) (B1_, B2_)
    ((fn x => true),
      ((0 : IntInf.int),
        ((1 : IntInf.int),
          ((fn a => fn b => IntInf.+ (a, b)),
            ((fn a => fn b => IntInf.* (a, b)),
              (max ord_int (0 : IntInf.int),
                ((fn x => fn y => IntInf.<= (y, x)),
                  ((fn x => fn y => IntInf.< (y, x)),
                    (true, (int_name, (fn x => true)))))))))))
    ia
  | create_redpair_ur_af_proc (A1_, A2_) (B1_, B2_) (Nat_carrier i) =
    create_redpair_ur_af_proc_comp (eq_nata, sshow_nata) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (Zero_nat,
          (one_nat,
            (plus_nat,
              (times_nat,
                (id, ((fn x => fn y => less_eq_nat y x),
                       ((fn x => fn y => less_nat y x),
                         (true, (nat_name, (fn x => true)))))))))))
      i
  | create_redpair_ur_af_proc (A1_, A2_) (B1_, B2_) (Arctic_carrier ida) =
    create_redpair_ur_af_proc_comp (eq_arctica, sshow_arctica) (A1_, A2_)
      (B1_, B2_)
      ((fn x => true),
        (MinInfty,
          (Num_arc (0 : IntInf.int),
            (plus_arctic,
              (mult_arctic,
                (id, (geq_arctic,
                       (gt_arctic, (false, (arctic_name, pos_arctic))))))))))
      ida
  | create_redpair_ur_af_proc (A1_, A2_) (B1_, B2_) (Int_mat_carrier (n, ib)) =
    create_redpair_ur_af_proc_comp
      (eq_lista (eq_lista eq_int), sshow_lista (sshow_lista sshow_inta))
      (A1_, A2_) (B1_, B2_)
      (wf_mat_rcI (Suc n) (Suc n),
        (mat0I (0 : IntInf.int) (Suc n) (Suc n),
          (mat1I (0 : IntInf.int) (1 : IntInf.int) (Suc n),
            (mat_plusI (fn a => fn b => IntInf.+ (a, b)),
              (mat_mulI (0 : IntInf.int) (fn a => fn b => IntInf.+ (a, b))
                 (fn a => fn b => IntInf.* (a, b)) (Suc n),
                (mat_max0I (max ord_int (0 : IntInf.int)),
                  (mat_geI (fn x => fn y => IntInf.<= (y, x)),
                    (mat_gtI (fn x => fn y => IntInf.<= (y, x))
                       (fn x => fn y => IntInf.< (y, x)),
                      (true,
                        (int_mat_name, mat_arc_posI (fn x => true)))))))))))
      ib
  | create_redpair_ur_af_proc (A1_, A2_) (B1_, B2_) (Arc_mat_carrier (n, ic)) =
    create_redpair_ur_af_proc_comp
      (eq_lista (eq_lista eq_arctica), sshow_lista (sshow_lista sshow_arctica))
      (A1_, A2_) (B1_, B2_)
      (wf_mat_rcI (Suc n) (Suc n),
        (mat0I MinInfty (Suc n) (Suc n),
          (mat1I MinInfty (Num_arc (0 : IntInf.int)) (Suc n),
            (mat_plusI plus_arctic,
              (mat_mulI MinInfty plus_arctic mult_arctic (Suc n),
                (mat_max0I id,
                  (mat_geI geq_arctic,
                    (mat_geI gt_arctic,
                      (false, (arc_mat_name, mat_arc_posI pos_arctic))))))))))
      ic;

fun create_redpair_proc_comp (A1_, A2_) (B1_, B2_) (C1_, C2_) c i p rP r =
  let
    val pi = to_lpoly_inter B1_ (get_comp_1 c) i;
  in
    bind (check_lpoly_coeffs A2_ B2_ (get_comp_wf_elem c) (get_comp_apos c)
           (fn x => get_comp_ge c x (get_comp_0 c)) i)
      (fn x =>
        checkRedPairProc B1_ C1_
          (check_polo_sI (A1_, A2_) B2_ (C1_, C2_) (get_comp_mono c)
            (get_comp_0 c) (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) (get_comp_gt c) pi)
          (check_polo_nsI (A1_, A2_) B2_ (C1_, C2_) (get_comp_0 c)
            (get_comp_1 c) (get_comp_plus c) (get_comp_times c)
            (get_comp_max0 c) (get_comp_ge c) pi)
          (append
            (append
              [#"p", #"o", #"l", #"y", #"n", #"o", #"m", #"i", #"a", #"l", #" ",
                #"o", #"r", #"d", #"e", #"r", #" ", #"("]
              (get_comp_description c))
            [#")"])
          (create_lpoly_repr (A1_, A2_) (B1_, B2_) c i) p rP r)
  end;

fun create_redpair_proc (A1_, A2_) (B1_, B2_) (Int_carrier ia) =
  create_redpair_proc_comp (eq_int, sshow_inta) (A1_, A2_) (B1_, B2_)
    ((fn x => true),
      ((0 : IntInf.int),
        ((1 : IntInf.int),
          ((fn a => fn b => IntInf.+ (a, b)),
            ((fn a => fn b => IntInf.* (a, b)),
              (max ord_int (0 : IntInf.int),
                ((fn x => fn y => IntInf.<= (y, x)),
                  ((fn x => fn y => IntInf.< (y, x)),
                    (true, (int_name, (fn x => true)))))))))))
    ia
  | create_redpair_proc (A1_, A2_) (B1_, B2_) (Nat_carrier i) =
    create_redpair_proc_comp (eq_nata, sshow_nata) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (Zero_nat,
          (one_nat,
            (plus_nat,
              (times_nat,
                (id, ((fn x => fn y => less_eq_nat y x),
                       ((fn x => fn y => less_nat y x),
                         (true, (nat_name, (fn x => true)))))))))))
      i
  | create_redpair_proc (A1_, A2_) (B1_, B2_) (Arctic_carrier ida) =
    create_redpair_proc_comp (eq_arctica, sshow_arctica) (A1_, A2_) (B1_, B2_)
      ((fn x => true),
        (MinInfty,
          (Num_arc (0 : IntInf.int),
            (plus_arctic,
              (mult_arctic,
                (id, (geq_arctic,
                       (gt_arctic, (false, (arctic_name, pos_arctic))))))))))
      ida
  | create_redpair_proc (A1_, A2_) (B1_, B2_) (Int_mat_carrier (n, ib)) =
    create_redpair_proc_comp
      (eq_lista (eq_lista eq_int), sshow_lista (sshow_lista sshow_inta))
      (A1_, A2_) (B1_, B2_)
      (wf_mat_rcI (Suc n) (Suc n),
        (mat0I (0 : IntInf.int) (Suc n) (Suc n),
          (mat1I (0 : IntInf.int) (1 : IntInf.int) (Suc n),
            (mat_plusI (fn a => fn b => IntInf.+ (a, b)),
              (mat_mulI (0 : IntInf.int) (fn a => fn b => IntInf.+ (a, b))
                 (fn a => fn b => IntInf.* (a, b)) (Suc n),
                (mat_max0I (max ord_int (0 : IntInf.int)),
                  (mat_geI (fn x => fn y => IntInf.<= (y, x)),
                    (mat_gtI (fn x => fn y => IntInf.<= (y, x))
                       (fn x => fn y => IntInf.< (y, x)),
                      (true,
                        (int_mat_name, mat_arc_posI (fn x => true)))))))))))
      ib
  | create_redpair_proc (A1_, A2_) (B1_, B2_) (Arc_mat_carrier (n, ic)) =
    create_redpair_proc_comp
      (eq_lista (eq_lista eq_arctica), sshow_lista (sshow_lista sshow_arctica))
      (A1_, A2_) (B1_, B2_)
      (wf_mat_rcI (Suc n) (Suc n),
        (mat0I MinInfty (Suc n) (Suc n),
          (mat1I MinInfty (Num_arc (0 : IntInf.int)) (Suc n),
            (mat_plusI plus_arctic,
              (mat_mulI MinInfty plus_arctic mult_arctic (Suc n),
                (mat_max0I id,
                  (mat_geI geq_arctic,
                    (mat_geI gt_arctic,
                      (false, (arc_mat_name, mat_arc_posI pos_arctic))))))))))
      ic;

fun checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm PisEmpty =
  debug i [#"P", #" ", #"i", #"s", #" ", #"e", #"m", #"p", #"t", #"y"]
    ((case p of [] => return ()
       | x :: xa =>
         error (append i
                  [#":", #" ", #"P", #" ", #"i", #"s", #" ", #"n", #"o", #"t",
                    #" ", #"e", #"m", #"p", #"t", #"y", #" ", #"i", #"n", #" ",
                    #"t", #"h", #"e", #" ", #"f", #"o", #"l", #"l", #"o", #"w",
                    #"i", #"n", #"g", #" ", #"D", #"P", #"-", #"p", #"r", #"o",
                    #"b", #"l", #"e", #"m", #" "]
                 :: showDPP (sshow_laba A2_ B2_) (sshow_lista sshow_chara) p
                      r)))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm
    (RedPairProc (redp, rP, prf)) =
    debug i [#"R", #"e", #"d", #"P", #"a", #"i", #"r", #"P", #"r", #"o", #"c"]
      (bind (catch
              (create_redpair_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) redp p rP r)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"t", #"h", #"e", #" ", #"r",
                           #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n", #" ",
                           #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o", #"c",
                           #"e", #"s", #"s", #"o", #"r", #" ", #"t", #"o", #" ",
                           #"s", #"w", #"i", #"t", #"c", #"h", #" ", #"f", #"r",
                           #"o", #"m", #" "]
                        :: append
                             (showDPP (sshow_laba A2_ B2_)
                               (sshow_lista sshow_chara) p r)
                             ([#"t", #"o", #" ", #"t", #"h", #"e", #" ", #"D",
                                #"P", #" ", #"p", #"r", #"o", #"b", #"l", #"e",
                                #"m", #" "]
                               :: append
                                    (showDPP (sshow_laba A2_ B2_)
                                      (sshow_lista sshow_chara) rP r)
                                    x))))
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) rP r
                  m rm rrm prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n",
                         #" ", #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o",
                         #"c", #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm
    (RedPairUrProc (redp, rP, ur, prf)) =
    debug i
      [#"R", #"e", #"d", #"P", #"a", #"i", #"r", #"P", #"r", #"o", #"c", #" ",
        #"(", #"u", #"s", #"a", #"b", #"l", #"e", #" ", #"r", #"u", #"l", #"e",
        #"s", #" ", #"+", #" ", #"a", #"f", #")"]
      (bind (catch
              (create_redpair_ur_af_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) redp p rP r rm ur)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"t", #"h", #"e", #" ", #"r",
                           #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n", #" ",
                           #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o", #"c",
                           #"e", #"s", #"s", #"o", #"r", #" ", #"(", #"w", #"i",
                           #"t", #"h", #" ", #"u", #"s", #"a", #"b", #"l", #"e",
                           #" ", #"r", #"u", #"l", #"e", #"s", #" ", #"a", #"n",
                           #"d", #" ", #"a", #"f", #")", #" ", #"t", #"o", #" ",
                           #"s", #"w", #"i", #"t", #"c", #"h", #" ", #"f", #"r",
                           #"o", #"m", #" "]
                        :: append
                             (showDPP (sshow_laba A2_ B2_)
                               (sshow_lista sshow_chara) p r)
                             ([#"t", #"o", #" ", #"t", #"h", #"e", #" ", #"D",
                                #"P", #" ", #"p", #"r", #"o", #"b", #"l", #"e",
                                #"m", #" "]
                               :: append
                                    (showDPP (sshow_laba A2_ B2_)
                                      (sshow_lista sshow_chara) rP r)
                                    x))))
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) rP r
                  m rm rrm prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n",
                         #" ", #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o",
                         #"c", #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm (DepGraphProc edpts) =
    debug i
      [#"D", #"e", #"p", #"G", #"r", #"a", #"p", #"h", #"P", #"r", #"o", #"c"]
      (bind (catch
              (checkDepGraphProc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) p
                (map (fn x =>
                       ((case fst x of NONE => false | SOME xa => true), snd x))
                  edpts)
                rm rrm)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"i", #"l", #"e", #" ", #"t", #"r", #"y", #"i",
                           #"n", #"g", #" ", #"t", #"o", #" ", #"p", #"e", #"r",
                           #"f", #"o", #"r", #"m", #" ", #"S", #"C", #"C", #"-",
                           #"d", #"e", #"c", #"o", #"m", #"p", #"o", #"s", #"i",
                           #"t", #"i", #"o", #"n", #" ", #" ", #"o", #"n", #" "]
                        :: append
                             (showDPP (sshow_laba A2_ B2_)
                               (sshow_lista sshow_chara) p r)
                             x)))
        (fn x =>
          catch (checkAllMIndex
                  (fn j => fn rp =>
                    (case fst rp of NONE => j | SOME prf => Suc j))
                  (fn j => fn rp =>
                    (case fst rp of NONE => return ()
                       | SOME a =>
                         checkDPProof (A1_, A2_) (B1_, B2_)
                           (append (append i [#"."]) (sshow_nat j)) (snd rp) r m
                           rm rrm a))
                  edpts)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"d", #"e", #"p", #"e", #"n", #"d", #"e", #"n", #"c",
                         #"y", #" ", #"g", #"r", #"a", #"p", #"h", #" ", #"p",
                         #"r", #"o", #"c", #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm
    (MonoRedPairProc (redp, rP, rR, prf)) =
    debug i
      [#"M", #"o", #"n", #"o", #"R", #"e", #"d", #"P", #"a", #"i", #"r", #"P",
        #"r", #"o", #"c"]
      (bind (catch
              (create_mono_redpair_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) redp p rP r rR)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"t", #"h", #"e", #" ", #"m",
                           #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ",
                           #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n",
                           #" ", #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o",
                           #"c", #"e", #"s", #"s", #"o", #"r", #" ", #"t", #"o",
                           #" ", #"s", #"w", #"i", #"t", #"c", #"h", #" ", #"f",
                           #"r", #"o", #"m", #" "]
                        :: append
                             (showDPP (sshow_laba A2_ B2_)
                               (sshow_lista sshow_chara) p r)
                             ([#"t", #"o", #" ", #"t", #"h", #"e", #" ", #"D",
                                #"P", #" ", #"p", #"r", #"o", #"b", #"l", #"e",
                                #"m", #" "]
                               :: append
                                    (showDPP (sshow_laba A2_ B2_)
                                      (sshow_lista sshow_chara) rP rR)
                                    x))))
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) rP
                  rR m (computeRuleMap (eq_laba A1_ B1_) rR)
                  (computeRuleMap (eq_laba A1_ B1_) (reverse_trs rR)) prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c",
                         #" ", #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o",
                         #"n", #" ", #"p", #"a", #"i", #"r", #" ", #"p", #"r",
                         #"o", #"c", #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm
    (MonoRedPairUrProc (redp, rP, rR, ur, prf)) =
    debug i
      [#"M", #"o", #"n", #"o", #"R", #"e", #"d", #"P", #"a", #"i", #"r", #"P",
        #"r", #"o", #"c", #" ", #"(", #"u", #"s", #"a", #"b", #"l", #"e", #" ",
        #"r", #"u", #"l", #"e", #"s", #")"]
      (bind (catch
              (create_mono_redpair_ur_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) redp p rP r rm rR
                ur)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"t", #"h", #"e", #" ", #"m",
                           #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c", #" ",
                           #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o", #"n",
                           #" ", #"p", #"a", #"i", #"r", #" ", #"p", #"r", #"o",
                           #"c", #"e", #"s", #"s", #"o", #"r", #" ", #"(", #"w",
                           #"i", #"t", #"h", #" ", #"u", #"s", #"a", #"b", #"l",
                           #"e", #" ", #"r", #"u", #"l", #"e", #"s", #")", #" ",
                           #"t", #"o", #" ", #"s", #"w", #"i", #"t", #"c", #"h",
                           #" ", #"f", #"r", #"o", #"m", #" "]
                        :: append
                             (showDPP (sshow_laba A2_ B2_)
                               (sshow_lista sshow_chara) p r)
                             ([#"t", #"o", #" ", #"t", #"h", #"e", #" ", #"D",
                                #"P", #" ", #"p", #"r", #"o", #"b", #"l", #"e",
                                #"m", #" "]
                               :: append
                                    (showDPP (sshow_laba A2_ B2_)
                                      (sshow_lista sshow_chara) rP rR)
                                    x))))
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) rP
                  rR m (computeRuleMap (eq_laba A1_ B1_) rR)
                  (computeRuleMap (eq_laba A1_ B1_) (reverse_trs rR)) prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"m", #"o", #"n", #"o", #"t", #"o", #"n", #"i", #"c",
                         #" ", #"r", #"e", #"d", #"u", #"c", #"t", #"i", #"o",
                         #"n", #" ", #"p", #"a", #"i", #"r", #" ", #"(", #"w",
                         #"i", #"t", #"h", #" ", #"u", #"s", #"a", #"b", #"l",
                         #"e", #" ", #"r", #"u", #"l", #"e", #"s", #")", #" ",
                         #"p", #"r", #"o", #"c", #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r (Suc m) rm rrm
    (UnlabProc (ulP, ulR, prf)) =
    debug i [#"U", #"n", #"l", #"a", #"b", #"P", #"r", #"o", #"c"]
      (bind (check_unlab_proc (A1_, A2_) (B1_, B2_)
              (eq_lista eq_char, sshow_lista sshow_chara) p r ulP ulR)
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) ulP
                  ulR m (computeRuleMap (eq_laba A1_ B1_) ulR)
                  (computeRuleMap (eq_laba A1_ B1_) (reverse_trs ulR)) prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"u", #"n", #"l", #"a", #"b", #"e", #"l", #"i", #"n",
                         #"g", #" ", #"p", #"r", #"o", #"c", #"e", #"s", #"s",
                         #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r Zero_nat rm rrm
    (UnlabProc (ulP, ulR, prf)) =
    error [[#"u", #"n", #"l", #"a", #"b", #"e", #"l", #"i", #"n", #"g", #" ",
             #"n", #"o", #"t", #" ", #"a", #"l", #"l", #"o", #"w", #"e", #"d",
             #" ", #"d", #"u", #"e", #" ", #"t", #"o", #" ", #"l", #"a", #"b",
             #"e", #"l", #"-", #"d", #"e", #"p", #"t", #"h", #" ", #"o", #"f",
             #" ", #"0"]]
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm
    (SemlabProc (sli, l, lP, lR, prf)) =
    debug i [#"S", #"e", #"m", #"l", #"a", #"b", #"P", #"r", #"o", #"c"]
      (bind (catch
              (create_sem_lab_proc (A1_, A2_) (B1_, B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) sli l p r lP lR)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"t", #"h", #"e", #" ", #"s",
                           #"e", #"m", #"a", #"n", #"t", #"i", #"c", #" ", #"l",
                           #"a", #"b", #"e", #"l", #"i", #"n", #"g", #" ", #"p",
                           #"r", #"o", #"c", #"e", #"s", #"s", #"o", #"r"]
                        :: x)))
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) lP
                  lR (Suc m) (computeRuleMap (eq_laba A1_ B1_) lR)
                  (computeRuleMap (eq_laba A1_ B1_) (reverse_trs lR)) prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"s", #"e", #"m", #"a", #"n", #"t", #"i", #"c", #" ",
                         #"l", #"a", #"b", #"e", #"l", #"i", #"n", #"g", #" ",
                         #"p", #"r", #"o", #"c", #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm
    (RootlabProc (lP, lR, prf)) =
    debug i [#"R", #"o", #"o", #"t", #"l", #"a", #"b", #"P", #"r", #"o", #"c"]
      (bind (catch (rl_proc (A1_, A2_) (B1_, B2_) p r lP lR)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"t", #"h", #"e", #" ", #"r",
                           #"o", #"o", #"t", #" ", #"l", #"a", #"b", #"e", #"l",
                           #"i", #"n", #"g", #" ", #"p", #"r", #"o", #"c", #"e",
                           #"s", #"s", #"o", #"r"]
                        :: x)))
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) lP
                  lR (Suc m) (computeRuleMap (eq_laba A1_ B1_) lR)
                  (computeRuleMap (eq_laba A1_ B1_) (reverse_trs lR)) prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"r", #"o", #"o", #"t", #"-", #"l", #"a", #"b", #"e",
                         #"l", #"i", #"n", #"g", #" ", #"p", #"r", #"o", #"c",
                         #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm (SpscProc (sp, rP, prf)) =
    debug i [#"S", #"u", #"b", #"t", #"e", #"r", #"m", #"P", #"r", #"o", #"c"]
      (bind (catch
              (create_spsc_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) sp p rP r)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"t", #"h", #"e", #" ", #"s",
                           #"u", #"b", #"t", #"e", #"r", #"m", #" ", #"c", #"r",
                           #"i", #"t", #"e", #"r", #"i", #"o", #"n", #" ", #"p",
                           #"r", #"o", #"c", #"e", #"s", #"s", #"o", #"r"]
                        :: x)))
        (fn x =>
          catch (checkDPProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) rP r
                  m rm rrm prf)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"s", #"u", #"b", #"t", #"e", #"r", #"m", #" ", #"c",
                         #"r", #"i", #"t", #"e", #"r", #"i", #"o", #"n", #" ",
                         #"p", #"r", #"o", #"c", #"e", #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm (SizeChangeSubTerm graphs)
    = debug i
        [#"S", #"i", #"z", #"e", #"C", #"h", #"a", #"n", #"g", #"e", #"S", #"u",
          #"b", #"T", #"e", #"r", #"m"]
        (check_sct_subterm_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
          (eq_lista eq_char, sshow_lista sshow_chara) p graphs r rm rrm)
  | checkDPProof (A1_, A2_) (B1_, B2_) i p r m rm rrm
    (SizeChangeRedPair (redp, ur, graphs)) =
    debug i
      [#"S", #"i", #"z", #"e", #"C", #"h", #"a", #"n", #"g", #"e", #"R", #"e",
        #"d", #"P", #"a", #"i", #"r"]
      (create_sct_ur_af_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
        (eq_lista eq_char, sshow_lista sshow_chara) redp p r rm rrm ur graphs);

fun checkTRSProof (A1_, A2_) (B1_, B2_) i r m (DPTrans (p, prf)) =
  debug i [#"D", #"P", #" ", #"t", #"r", #"a", #"n", #"s"]
    (bind (catch
            (checkDPs (eq_laba A1_ B1_, sshow_laba A2_ B2_)
              (eq_lista eq_char, sshow_lista sshow_chara) r p)
            (fn x =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                         #"h", #"e", #"n", #" ", #"s", #"w", #"i", #"t", #"c",
                         #"h", #"i", #"n", #"g", #" ", #"f", #"r", #"o", #"m",
                         #" ", #"t", #"h", #"e", #" ", #"T", #"R", #"S"]
                      :: append
                           (showTRS (sshow_laba A2_ B2_)
                             (sshow_lista sshow_chara) r)
                           ([#"t", #"o", #" ", #"t", #"h", #"e", #" ", #"i",
                              #"n", #"i", #"t", #"i", #"a", #"l", #" ", #"D",
                              #"P", #" ", #"p", #"r", #"o", #"b", #"l", #"e",
                              #"m"]
                             :: append
                                  (showDPP (sshow_shpa (sshow_laba A2_ B2_))
                                    (sshow_lista sshow_chara) p (computeID r))
                                  x))))
      (fn x =>
        catch (checkDPProof
                (eq_shpa (eq_laba A1_ B1_), sshow_shpa (sshow_laba A2_ B2_))
                (B1_, B2_) (append i [#".", #"1"]) (map (map_rule init_lab) p)
                (map (map_rule init_lab) (computeID r)) Zero_nat
                (computeRuleMap (eq_laba (eq_shpa (eq_laba A1_ B1_)) B1_)
                  (map (map_rule init_lab) (computeID r)))
                (computeRuleMap (eq_laba (eq_shpa (eq_laba A1_ B1_)) B1_)
                  (reverse_trs (map (map_rule init_lab) (computeID r))))
                prf)
          (fn xa =>
            error (append i
                     [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                       #"e", #"l", #"o", #"w", #" ", #"s", #"w", #"i", #"t",
                       #"c", #"h", #" ", #"t", #"o", #" ", #"d", #"e", #"p",
                       #"e", #"n", #"d", #"e", #"n", #"c", #"y", #" ", #"p",
                       #"a", #"i", #"r", #"s"]
                    :: indent xa))))
  | checkTRSProof (A1_, A2_) (B1_, B2_) i r m (RuleRemoval (rp, rR, prfa)) =
    debug i
      [#"r", #"u", #"l", #"e", #" ", #"r", #"e", #"m", #"o", #"v", #"a", #"l"]
      (bind (catch
              (create_rule_removal_proc (eq_laba A1_ B1_, sshow_laba A2_ B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) rp r rR)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"i",
                           #"n", #" ", #"r", #"u", #"l", #"e", #" ", #"r", #"e",
                           #"m", #"o", #"v", #"a", #"l", #" ", #"w", #"h", #"e",
                           #"n", #" ", #"s", #"w", #"i", #"t", #"c", #"h", #"i",
                           #"n", #"g", #" ", #"f", #"r", #"o", #"m", #" ", #"t",
                           #"h", #"e", #" ", #"T", #"R", #"S"]
                        :: append
                             (showTRS (sshow_laba A2_ B2_)
                               (sshow_lista sshow_chara) r)
                             ([#"t", #"o", #" ", #"t", #"h", #"e", #" ", #"T",
                                #"R", #"S"]
                               :: append
                                    (showTRS (sshow_laba A2_ B2_)
                                      (sshow_lista sshow_chara) rR)
                                    x))))
        (fn x =>
          catch (checkTRSProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) rR
                  m prfa)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"r", #"u", #"l", #"e", #" ", #"r", #"e", #"m", #"o",
                         #"v", #"a", #"l", #" ", #"p", #"r", #"o", #"c", #"e",
                         #"s", #"s", #"o", #"r"]
                      :: indent xa))))
  | checkTRSProof (A1_, A2_) (B1_, B2_) i r (Suc m) (Unlab (uR, prfa)) =
    debug i [#"u", #"n", #"l", #"a", #"b", #"e", #"l"]
      (bind (catch
              (unlab_trs_tt (A1_, A2_) (B1_, B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) r uR)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"i",
                           #"n", #" ", #"u", #"n", #"l", #"a", #"b", #"e", #"l",
                           #"i", #"n", #"g", #" ", #"o", #"f", #" ", #"T", #"R",
                           #"S", #" ", #"w", #"h", #"e", #"n", #" ", #"s", #"w",
                           #"i", #"t", #"c", #"h", #"i", #"n", #"g", #" ", #"f",
                           #"r", #"o", #"m", #" ", #"t", #"h", #"e", #" ", #"T",
                           #"R", #"S"]
                        :: append
                             (showTRS (sshow_laba A2_ B2_)
                               (sshow_lista sshow_chara) r)
                             ([#"t", #"o", #" ", #"t", #"h", #"e", #" ", #"T",
                                #"R", #"S"]
                               :: append
                                    (showTRS (sshow_laba A2_ B2_)
                                      (sshow_lista sshow_chara) uR)
                                    x))))
        (fn x =>
          catch (checkTRSProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) uR
                  m prfa)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"t", #"h", #"e", #" ",
                         #"u", #"n", #"l", #"a", #"b", #"e", #"l", #"i", #"n",
                         #"g"]
                      :: indent xa))))
  | checkTRSProof (A1_, A2_) (B1_, B2_) i r Zero_nat (Unlab (uR, prfa)) =
    error [[#"u", #"n", #"l", #"a", #"b", #"e", #"l", #"i", #"n", #"g", #" ",
             #"n", #"o", #"t", #" ", #"a", #"l", #"l", #"o", #"w", #"e", #"d",
             #" ", #"f", #"o", #"r", #" ", #"l", #"a", #"b", #"e", #"l", #" ",
             #"d", #"e", #"p", #"t", #"h", #" ", #"0"]]
  | checkTRSProof (A1_, A2_) (B1_, B2_) i r m RisEmpty =
    debug i [#"R", #" ", #"i", #"s", #" ", #"e", #"m", #"p", #"t", #"y"]
      (checkB (null r)
        (append
          [append i
             [#":", #" ", #"R", #" ", #"i", #"s", #" ", #"n", #"o", #"t", #" ",
               #"e", #"m", #"p", #"t", #"y", #",", #" ", #"b", #"u", #"t", #" ",
               #"i", #"t", #" ", #"i", #"s", #" ", #"t", #"h", #"e", #" ", #"T",
               #"R", #"S"]]
          (showTRS (sshow_laba A2_ B2_) (sshow_lista sshow_chara) r)))
  | checkTRSProof (A1_, A2_) (B1_, B2_) i r m (Semlab (sli, l, lR, prfa)) =
    debug i [#"S", #"e", #"m", #"l", #"a", #"b"]
      (bind (catch
              (create_sem_lab_tt (A1_, A2_) (B1_, B2_)
                (eq_lista eq_char, sshow_lista sshow_chara) sli l r lR)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"s", #"e", #"m", #"a", #"n",
                           #"t", #"i", #"c", #" ", #"l", #"a", #"b", #"e", #"l",
                           #"i", #"n", #"g"]
                        :: x)))
        (fn x =>
          catch (checkTRSProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) lR
                  (Suc m) prfa)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"s", #"e", #"m", #"a",
                         #"n", #"t", #"i", #"c", #" ", #"l", #"a", #"b", #"e",
                         #"l", #"i", #"n", #"g"]
                      :: indent xa))))
  | checkTRSProof (A1_, A2_) (B1_, B2_) i r m (Rootlab (lR, prfa)) =
    debug i [#"R", #"o", #"o", #"t", #"l", #"a", #"b"]
      (bind (catch (rl_tt (A1_, A2_) (B1_, B2_) r lR)
              (fn x =>
                error (append i
                         [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"w",
                           #"h", #"e", #"n", #" ", #"a", #"p", #"p", #"l", #"y",
                           #"i", #"n", #"g", #" ", #"r", #"o", #"o", #"t", #"-",
                           #"l", #"a", #"b", #"e", #"l", #"i", #"n", #"g"]
                        :: x)))
        (fn x =>
          catch (checkTRSProof (A1_, A2_) (B1_, B2_) (append i [#".", #"1"]) lR
                  (Suc m) prfa)
            (fn xa =>
              error (append i
                       [#":", #" ", #"e", #"r", #"r", #"o", #"r", #" ", #"b",
                         #"e", #"l", #"o", #"w", #" ", #"r", #"o", #"o", #"t",
                         #"-", #"l", #"a", #"b", #"e", #"l", #"i", #"n", #"g"]
                      :: indent xa))));

fun certifyCertProblem (A1_, A2_) (B1_, B2_)
  (CertificationProblem (true, trs, Left prf)) =
  (case checkTRSProof (A1_, A2_) (B1_, B2_) [#"1"] (map (map_rule init_lab) trs)
          Zero_nat prf
     of Left err =>
       Error (showList
               ([#"t", #"e", #"r", #"m", #"i", #"n", #"a", #"t", #"i", #"o",
                  #"n", #" ", #"p", #"r", #"o", #"o", #"f", #" ", #"n", #"o",
                  #"t", #" ", #"a", #"c", #"c", #"e", #"p", #"t", #"e", #"d"]
                 :: err))
     | Right x => Certified true)
  | certifyCertProblem (A1_, A2_) (B1_, B2_)
    (CertificationProblem (false, trs, Right prfa)) =
    (case checkTRSDisproof (A1_, A2_)
            (eq_lista eq_char, sshow_lista sshow_chara) [#"1"] trs prfa
       of Left err =>
         Error (showList
                 ([#"n", #"o", #"n", #"-", #"t", #"e", #"r", #"m", #"i", #"n",
                    #"a", #"t", #"i", #"o", #"n", #" ", #"p", #"r", #"o", #"o",
                    #"f", #" ", #"n", #"o", #"t", #" ", #"a", #"c", #"c", #"e",
                    #"p", #"t", #"e", #"d"]
                   :: err))
       | Right x => Certified false)
  | certifyCertProblem (A1_, A2_) (B1_, B2_)
    (CertificationProblem (false, va, Left v)) =
    Error [#"N", #"o", #"t", #" ", #"a", #"l", #"l", #"o", #"w", #"e", #"d",
            #" ", #"t", #"o", #" ", #"g", #"i", #"v", #"e", #" ", #"a", #" ",
            #"t", #"e", #"r", #"m", #"i", #"n", #"a", #"t", #"i", #"o", #"n",
            #" ", #"p", #"r", #"o", #"o", #"f", #" ", #"f", #"o", #"r", #" ",
            #"a", #" ", #"n", #"o", #"n", #"-", #"t", #"e", #"r", #"m", #"i",
            #"n", #"a", #"t", #"i", #"o", #"n", #" ", #"g", #"o", #"a", #"l",
            #" ", #"o", #"r", #" ", #"v", #"i", #"c", #"e", #"-", #"v", #"e",
            #"r", #"s", #"a"]
  | certifyCertProblem (A1_, A2_) (B1_, B2_)
    (CertificationProblem (true, va, Right vc)) =
    Error [#"N", #"o", #"t", #" ", #"a", #"l", #"l", #"o", #"w", #"e", #"d",
            #" ", #"t", #"o", #" ", #"g", #"i", #"v", #"e", #" ", #"a", #" ",
            #"t", #"e", #"r", #"m", #"i", #"n", #"a", #"t", #"i", #"o", #"n",
            #" ", #"p", #"r", #"o", #"o", #"f", #" ", #"f", #"o", #"r", #" ",
            #"a", #" ", #"n", #"o", #"n", #"-", #"t", #"e", #"r", #"m", #"i",
            #"n", #"a", #"t", #"i", #"o", #"n", #" ", #"g", #"o", #"a", #"l",
            #" ", #"o", #"r", #" ", #"v", #"i", #"c", #"e", #"-", #"v", #"e",
            #"r", #"s", #"a"];

fun xml_certificationProblem A_ C_ (CertificationProblem (term, rls, prf)) =
  let
    val inp =
      Xml ([#"i", #"n", #"p", #"u", #"t"], [],
            [Xml ([#"t", #"r", #"s", #"I", #"n", #"p", #"u", #"t"], [],
                   [Xml ([#"t", #"r", #"s"], [], [xml_rules A_ C_ rls], NONE)],
                   NONE)],
            NONE);
    val goal = (if isLeft prf then terminationGoal else nonTerminationGoal);
  in
    Xml ([#"p", #"r", #"o", #"b", #"l", #"e", #"m"], [], [goal, inp], NONE)
  end;

fun eqWhiteSpace s1 s2 =
  eqop (eq_lista eq_char) (filter (fn c => not (member eq_char c wspace)) s1)
    (filter (fn c => not (member eq_char c wspace)) s2);

fun extract_goal_input s =
  (case search_len s [#"<", #"g", #"o", #"a", #"l", #">"]
          (size_list [#"<", #"g", #"o", #"a", #"l", #">"])
     of NONE => NONE
     | SOME a =>
       let
         val (x, mgoal) = a;
       in
         (case search_len mgoal [#"<", #"/", #"g", #"o", #"a", #"l", #">"]
                 (size_list [#"<", #"/", #"g", #"o", #"a", #"l", #">"])
            of NONE => NONE
            | SOME aa =>
              let
                val (goal, s') = aa;
              in
                (case search_len s' [#"<", #"i", #"n", #"p", #"u", #"t", #">"]
                        (size_list [#"<", #"i", #"n", #"p", #"u", #"t", #">"])
                   of NONE => NONE
                   | SOME ab =>
                     let
                       val (xa, minput) = ab;
                     in
                       (case search_len minput
                               [#"<", #"/", #"i", #"n", #"p", #"u", #"t", #">"]
                               (size_list
                                 [#"<", #"/", #"i", #"n", #"p", #"u", #"t",
                                   #">"])
                          of NONE => NONE
                          | SOME ac =>
                            let
                              val (input, xb) = ac;
                            in
                              SOME (concat
                                     [[#"<", #"p", #"r", #"o", #"b", #"l", #"e",
#"m", #">"],
                                       [#"<", #"g", #"o", #"a", #"l", #">"],
                                       goal,
                                       [#"<", #"/", #"g", #"o", #"a", #"l",
 #">"],
                                       [#"<", #"i", #"n", #"p", #"u", #"t",
 #">"],
                                       input,
                                       [#"<", #"/", #"i", #"n", #"p", #"u",
 #"t", #">"],
                                       [#"<", #"/", #"p", #"r", #"o", #"b",
 #"l", #"e", #"m", #">"]])
                            end)
                     end)
              end)
       end);

fun certifyProof problemString =
  (case parseCertProblem problemString
     of Left err =>
       Error (append
               (append
                 [#"e", #"r", #"r", #"o", #"r", #" ", #"w", #"h", #"i", #"l",
                   #"e", #" ", #"p", #"a", #"r", #"s", #"i", #"n", #"g"]
                 newline)
               err)
     | Right cp =>
       (case extract_goal_input problemString
          of NONE =>
            Error [#"c", #"o", #"u", #"l", #"d", #" ", #"n", #"o", #"t", #" ",
                    #"f", #"i", #"n", #"d", #" ", #"g", #"o", #"a", #"l", #" ",
                    #"a", #"n", #"d", #" ", #"i", #"n", #"p", #"u", #"t", #" ",
                    #"i", #"n", #" ", #"s", #"t", #"r", #"i", #"n", #"g", #" ",
                    #"f", #"o", #"r", #" ", #"c", #"o", #"m", #"p", #"a", #"r",
                    #"i", #"s", #"o", #"n", #" ", #"w", #"i", #"t", #"h", #" ",
                    #"p", #"a", #"r", #"s", #"e", #"d", #" ", #"r", #"e", #"s",
                    #"u", #"l", #"t"]
          | SOME inpProblem =>
            (if eqWhiteSpace inpProblem
                  (sshow_xml
                    (xml_certificationProblem (sshow_lista sshow_chara)
                      (sshow_lista sshow_chara) cp))
              then certifyCertProblem
                     (eq_lista eq_char, sshow_lista sshow_chara)
                     (eq_lista eq_nata, sshow_lista sshow_nata) cp
              else Error (concat
                           [[#"p", #"a", #"r", #"s", #"e", #"d", #" ", #"p",
                              #"r", #"o", #"b", #"l", #"e", #"m", #" ", #"d",
                              #"o", #"e", #"s", #" ", #"n", #"o", #"t", #" ",
                              #"c", #"o", #"r", #"r", #"e", #"s", #"p", #"o",
                              #"n", #"d", #" ", #"t", #"o", #" ", #"i", #"n",
                              #"p", #"u", #"t"],
                             newline, [#"i", #"n", #"p", #"u", #"t", #":"],
                             newline,
                             filter (fn c => not (member eq_char c wspace))
                               inpProblem,
                             newline, newline,
                             [#"p", #"a", #"r", #"s", #"e", #"d", #":"],
                             newline,
                             filter (fn c => not (member eq_char c wspace))
                               (sshow_xml
                                 (xml_certificationProblem
                                   (sshow_lista sshow_chara)
                                   (sshow_lista sshow_chara) cp)),
                             newline]))));

end; (*struct Ceta*)
